/*
** DO NOT EDIT THIS FILE !!!
** It was generated automatically by:
**
**   User: joergr
**   Host: thinkpad2
**   Date: 2024-11-16 10:42:05
**   Prog: /home/joergr/Source/dcmtk-full/public/dcmdata/libsrc/mkdictbi
**
**   From: ../data/dicom.dic
**         ../data/private.dic
**
*/

#include "dcmtk/dcmdata/dcdict.h"
#if DCM_DICT_DEFAULT == DCM_DICT_DEFAULT_USE_BUILTIN
#include "dcmtk/dcmdata/dcdicent.h"

struct DBI_SimpleEntry {
    Uint16 group;
    Uint16 element;
    Uint16 upperGroup;
    Uint16 upperElement;
    DcmEVR evr;
    const char* tagName;
    int vmMin;
    int vmMax;
    const char* standardVersion;
    DcmDictRangeRestriction groupRestriction;
    DcmDictRangeRestriction elementRestriction;
    const char* privateCreator;
};

static const DBI_SimpleEntry simpleBuiltinDict[] = {
    { 0x0000, 0x0000, 0x0000, 0x0000,
      EVR_UL, "CommandGroupLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0001, 0x0000, 0x0001,
      EVR_UL, "RETIRED_CommandLengthToEnd", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0002, 0x0000, 0x0002,
      EVR_UI, "AffectedSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0003, 0x0000, 0x0003,
      EVR_UI, "RequestedSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0010, 0x0000, 0x0010,
      EVR_SH, "RETIRED_CommandRecognitionCode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0100, 0x0000, 0x0100,
      EVR_US, "CommandField", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0110, 0x0000, 0x0110,
      EVR_US, "MessageID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0120, 0x0000, 0x0120,
      EVR_US, "MessageIDBeingRespondedTo", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0200, 0x0000, 0x0200,
      EVR_AE, "RETIRED_Initiator", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0300, 0x0000, 0x0300,
      EVR_AE, "RETIRED_Receiver", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0400, 0x0000, 0x0400,
      EVR_AE, "RETIRED_FindLocation", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0600, 0x0000, 0x0600,
      EVR_AE, "MoveDestination", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0700, 0x0000, 0x0700,
      EVR_US, "Priority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0800, 0x0000, 0x0800,
      EVR_US, "CommandDataSetType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0850, 0x0000, 0x0850,
      EVR_US, "RETIRED_NumberOfMatches", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0860, 0x0000, 0x0860,
      EVR_US, "RETIRED_ResponseSequenceNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0900, 0x0000, 0x0900,
      EVR_US, "Status", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0901, 0x0000, 0x0901,
      EVR_AT, "OffendingElement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0902, 0x0000, 0x0902,
      EVR_LO, "ErrorComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0903, 0x0000, 0x0903,
      EVR_US, "ErrorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1000, 0x0000, 0x1000,
      EVR_UI, "AffectedSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1001, 0x0000, 0x1001,
      EVR_UI, "RequestedSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1002, 0x0000, 0x1002,
      EVR_US, "EventTypeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1005, 0x0000, 0x1005,
      EVR_AT, "AttributeIdentifierList", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1008, 0x0000, 0x1008,
      EVR_US, "ActionTypeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1020, 0x0000, 0x1020,
      EVR_US, "NumberOfRemainingSuboperations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1021, 0x0000, 0x1021,
      EVR_US, "NumberOfCompletedSuboperations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1022, 0x0000, 0x1022,
      EVR_US, "NumberOfFailedSuboperations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1023, 0x0000, 0x1023,
      EVR_US, "NumberOfWarningSuboperations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1030, 0x0000, 0x1030,
      EVR_AE, "MoveOriginatorApplicationEntityTitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x1031, 0x0000, 0x1031,
      EVR_US, "MoveOriginatorMessageID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x4000, 0x0000, 0x4000,
      EVR_LT, "RETIRED_DialogReceiver", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x4010, 0x0000, 0x4010,
      EVR_LT, "RETIRED_TerminalType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5010, 0x0000, 0x5010,
      EVR_SH, "RETIRED_MessageSetID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5020, 0x0000, 0x5020,
      EVR_SH, "RETIRED_EndMessageID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5110, 0x0000, 0x5110,
      EVR_LT, "RETIRED_DisplayFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5120, 0x0000, 0x5120,
      EVR_LT, "RETIRED_PagePositionID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5130, 0x0000, 0x5130,
      EVR_CS, "RETIRED_TextFormatID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5140, 0x0000, 0x5140,
      EVR_CS, "RETIRED_NormalReverse", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5150, 0x0000, 0x5150,
      EVR_CS, "RETIRED_AddGrayScale", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5160, 0x0000, 0x5160,
      EVR_CS, "RETIRED_Borders", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5170, 0x0000, 0x5170,
      EVR_IS, "RETIRED_Copies", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5180, 0x0000, 0x5180,
      EVR_CS, "RETIRED_CommandMagnificationType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x5190, 0x0000, 0x5190,
      EVR_CS, "RETIRED_Erase", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x51a0, 0x0000, 0x51a0,
      EVR_CS, "RETIRED_Print", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x51b0, 0x0000, 0x51b0,
      EVR_US, "RETIRED_Overlays", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0000, 0x0002, 0x0000,
      EVR_UL, "FileMetaInformationGroupLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0001, 0x0002, 0x0001,
      EVR_OB, "FileMetaInformationVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0002, 0x0002, 0x0002,
      EVR_UI, "MediaStorageSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0003, 0x0002, 0x0003,
      EVR_UI, "MediaStorageSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0010, 0x0002, 0x0010,
      EVR_UI, "TransferSyntaxUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0012, 0x0002, 0x0012,
      EVR_UI, "ImplementationClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0013, 0x0002, 0x0013,
      EVR_SH, "ImplementationVersionName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0016, 0x0002, 0x0016,
      EVR_AE, "SourceApplicationEntityTitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0017, 0x0002, 0x0017,
      EVR_AE, "SendingApplicationEntityTitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0018, 0x0002, 0x0018,
      EVR_AE, "ReceivingApplicationEntityTitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0026, 0x0002, 0x0026,
      EVR_UR, "SourcePresentationAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0027, 0x0002, 0x0027,
      EVR_UR, "SendingPresentationAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0028, 0x0002, 0x0028,
      EVR_UR, "ReceivingPresentationAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0031, 0x0002, 0x0031,
      EVR_OB, "RTVMetaInformationVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0032, 0x0002, 0x0032,
      EVR_UI, "RTVCommunicationSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0033, 0x0002, 0x0033,
      EVR_UI, "RTVCommunicationSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0035, 0x0002, 0x0035,
      EVR_OB, "RTVSourceIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0036, 0x0002, 0x0036,
      EVR_OB, "RTVFlowIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0037, 0x0002, 0x0037,
      EVR_UL, "RTVFlowRTPSamplingRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0038, 0x0002, 0x0038,
      EVR_FD, "RTVFlowActualFrameDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0100, 0x0002, 0x0100,
      EVR_UI, "PrivateInformationCreatorUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0002, 0x0102, 0x0002, 0x0102,
      EVR_OB, "PrivateInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0003, 0x0000, 0x0003, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
  , { 0x0003, 0x0008, 0x0003, 0x0008,
      EVR_US, "ISICommandField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0011, 0x0003, 0x0011,
      EVR_US, "AttachIDApplicationCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0012, 0x0003, 0x0012,
      EVR_UL, "AttachIDMessageCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0013, 0x0003, 0x0013,
      EVR_DA, "AttachIDDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0014, 0x0003, 0x0014,
      EVR_TM, "AttachIDTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0020, 0x0003, 0x0020,
      EVR_US, "MessageType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0030, 0x0003, 0x0030,
      EVR_DA, "MaxWaitingDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0003, 0x0031, 0x0003, 0x0031,
      EVR_TM, "MaxWaitingTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x0004, 0x1130, 0x0004, 0x1130,
      EVR_CS, "FileSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1141, 0x0004, 0x1141,
      EVR_CS, "FileSetDescriptorFileID", 1, 8, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1142, 0x0004, 0x1142,
      EVR_CS, "SpecificCharacterSetOfFileSetDescriptorFile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1200, 0x0004, 0x1200,
      EVR_up, "OffsetOfTheFirstDirectoryRecordOfTheRootDirectoryEntity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1202, 0x0004, 0x1202,
      EVR_up, "OffsetOfTheLastDirectoryRecordOfTheRootDirectoryEntity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1212, 0x0004, 0x1212,
      EVR_US, "FileSetConsistencyFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1220, 0x0004, 0x1220,
      EVR_SQ, "DirectoryRecordSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1400, 0x0004, 0x1400,
      EVR_up, "OffsetOfTheNextDirectoryRecord", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1410, 0x0004, 0x1410,
      EVR_US, "RecordInUseFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1420, 0x0004, 0x1420,
      EVR_up, "OffsetOfReferencedLowerLevelDirectoryEntity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1430, 0x0004, 0x1430,
      EVR_CS, "DirectoryRecordType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1432, 0x0004, 0x1432,
      EVR_UI, "PrivateRecordUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1500, 0x0004, 0x1500,
      EVR_CS, "ReferencedFileID", 1, 8, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1504, 0x0004, 0x1504,
      EVR_up, "RETIRED_MRDRDirectoryRecordOffset", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1510, 0x0004, 0x1510,
      EVR_UI, "ReferencedSOPClassUIDInFile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1511, 0x0004, 0x1511,
      EVR_UI, "ReferencedSOPInstanceUIDInFile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1512, 0x0004, 0x1512,
      EVR_UI, "ReferencedTransferSyntaxUIDInFile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x151a, 0x0004, 0x151a,
      EVR_UI, "ReferencedRelatedGeneralSOPClassUIDInFile", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0004, 0x1600, 0x0004, 0x1600,
      EVR_UL, "RETIRED_NumberOfReferences", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0005, 0x0000, 0x0005, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
#endif
  , { 0x0006, 0x0001, 0x0006, 0x0001,
      EVR_SQ, "CurrentFrameFunctionalGroupsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0001, 0x0008, 0x0001,
      EVR_UL, "RETIRED_LengthToEnd", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0005, 0x0008, 0x0005,
      EVR_CS, "SpecificCharacterSet", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0006, 0x0008, 0x0006,
      EVR_SQ, "LanguageCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0008, 0x0008, 0x0008,
      EVR_CS, "ImageType", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0010, 0x0008, 0x0010,
      EVR_SH, "RETIRED_RecognitionCode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0012, 0x0008, 0x0012,
      EVR_DA, "InstanceCreationDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0013, 0x0008, 0x0013,
      EVR_TM, "InstanceCreationTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0014, 0x0008, 0x0014,
      EVR_UI, "InstanceCreatorUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0015, 0x0008, 0x0015,
      EVR_DT, "InstanceCoercionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0016, 0x0008, 0x0016,
      EVR_UI, "SOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0017, 0x0008, 0x0017,
      EVR_UI, "AcquisitionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0018, 0x0008, 0x0018,
      EVR_UI, "SOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0019, 0x0008, 0x0019,
      EVR_UI, "PyramidUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x001a, 0x0008, 0x001a,
      EVR_UI, "RelatedGeneralSOPClassUID", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x001b, 0x0008, 0x001b,
      EVR_UI, "OriginalSpecializedSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x001c, 0x0008, 0x001c,
      EVR_CS, "SyntheticData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0020, 0x0008, 0x0020,
      EVR_DA, "StudyDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0021, 0x0008, 0x0021,
      EVR_DA, "SeriesDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0022, 0x0008, 0x0022,
      EVR_DA, "AcquisitionDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0023, 0x0008, 0x0023,
      EVR_DA, "ContentDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0024, 0x0008, 0x0024,
      EVR_DA, "RETIRED_OverlayDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0025, 0x0008, 0x0025,
      EVR_DA, "RETIRED_CurveDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x002a, 0x0008, 0x002a,
      EVR_DT, "AcquisitionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0030, 0x0008, 0x0030,
      EVR_TM, "StudyTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0031, 0x0008, 0x0031,
      EVR_TM, "SeriesTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0032, 0x0008, 0x0032,
      EVR_TM, "AcquisitionTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0033, 0x0008, 0x0033,
      EVR_TM, "ContentTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0034, 0x0008, 0x0034,
      EVR_TM, "RETIRED_OverlayTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0035, 0x0008, 0x0035,
      EVR_TM, "RETIRED_CurveTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0040, 0x0008, 0x0040,
      EVR_US, "RETIRED_DataSetType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0041, 0x0008, 0x0041,
      EVR_LO, "RETIRED_DataSetSubtype", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0042, 0x0008, 0x0042,
      EVR_CS, "RETIRED_NuclearMedicineSeriesType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0050, 0x0008, 0x0050,
      EVR_SH, "AccessionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0051, 0x0008, 0x0051,
      EVR_SQ, "IssuerOfAccessionNumberSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0052, 0x0008, 0x0052,
      EVR_CS, "QueryRetrieveLevel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0053, 0x0008, 0x0053,
      EVR_CS, "QueryRetrieveView", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0054, 0x0008, 0x0054,
      EVR_AE, "RetrieveAETitle", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0055, 0x0008, 0x0055,
      EVR_AE, "StationAETitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0056, 0x0008, 0x0056,
      EVR_CS, "InstanceAvailability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0058, 0x0008, 0x0058,
      EVR_UI, "FailedSOPInstanceUIDList", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0060, 0x0008, 0x0060,
      EVR_CS, "Modality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0061, 0x0008, 0x0061,
      EVR_CS, "ModalitiesInStudy", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0062, 0x0008, 0x0062,
      EVR_UI, "SOPClassesInStudy", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0063, 0x0008, 0x0063,
      EVR_SQ, "AnatomicRegionsInStudyCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0064, 0x0008, 0x0064,
      EVR_CS, "ConversionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0068, 0x0008, 0x0068,
      EVR_CS, "PresentationIntentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0070, 0x0008, 0x0070,
      EVR_LO, "Manufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0080, 0x0008, 0x0080,
      EVR_LO, "InstitutionName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0081, 0x0008, 0x0081,
      EVR_ST, "InstitutionAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0082, 0x0008, 0x0082,
      EVR_SQ, "InstitutionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0090, 0x0008, 0x0090,
      EVR_PN, "ReferringPhysicianName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0092, 0x0008, 0x0092,
      EVR_ST, "ReferringPhysicianAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0094, 0x0008, 0x0094,
      EVR_SH, "ReferringPhysicianTelephoneNumbers", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0096, 0x0008, 0x0096,
      EVR_SQ, "ReferringPhysicianIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x009c, 0x0008, 0x009c,
      EVR_PN, "ConsultingPhysicianName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x009d, 0x0008, 0x009d,
      EVR_SQ, "ConsultingPhysicianIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0100, 0x0008, 0x0100,
      EVR_SH, "CodeValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0101, 0x0008, 0x0101,
      EVR_LO, "ExtendedCodeValue", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0102, 0x0008, 0x0102,
      EVR_SH, "CodingSchemeDesignator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0103, 0x0008, 0x0103,
      EVR_SH, "CodingSchemeVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0104, 0x0008, 0x0104,
      EVR_LO, "CodeMeaning", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0105, 0x0008, 0x0105,
      EVR_CS, "MappingResource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0106, 0x0008, 0x0106,
      EVR_DT, "ContextGroupVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0107, 0x0008, 0x0107,
      EVR_DT, "ContextGroupLocalVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0108, 0x0008, 0x0108,
      EVR_LT, "ExtendedCodeMeaning", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0109, 0x0008, 0x0109,
      EVR_SQ, "CodingSchemeResourcesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010a, 0x0008, 0x010a,
      EVR_CS, "CodingSchemeURLType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010b, 0x0008, 0x010b,
      EVR_CS, "ContextGroupExtensionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010c, 0x0008, 0x010c,
      EVR_UI, "CodingSchemeUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010d, 0x0008, 0x010d,
      EVR_UI, "ContextGroupExtensionCreatorUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010e, 0x0008, 0x010e,
      EVR_UR, "CodingSchemeURL", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x010f, 0x0008, 0x010f,
      EVR_CS, "ContextIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0110, 0x0008, 0x0110,
      EVR_SQ, "CodingSchemeIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0112, 0x0008, 0x0112,
      EVR_LO, "CodingSchemeRegistry", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0114, 0x0008, 0x0114,
      EVR_ST, "CodingSchemeExternalID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0115, 0x0008, 0x0115,
      EVR_ST, "CodingSchemeName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0116, 0x0008, 0x0116,
      EVR_ST, "CodingSchemeResponsibleOrganization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0117, 0x0008, 0x0117,
      EVR_UI, "ContextUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0118, 0x0008, 0x0118,
      EVR_UI, "MappingResourceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0119, 0x0008, 0x0119,
      EVR_UC, "LongCodeValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0120, 0x0008, 0x0120,
      EVR_UR, "URNCodeValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0121, 0x0008, 0x0121,
      EVR_SQ, "EquivalentCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0122, 0x0008, 0x0122,
      EVR_LO, "MappingResourceName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0123, 0x0008, 0x0123,
      EVR_SQ, "ContextGroupIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0124, 0x0008, 0x0124,
      EVR_SQ, "MappingResourceIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0201, 0x0008, 0x0201,
      EVR_SH, "TimezoneOffsetFromUTC", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0220, 0x0008, 0x0220,
      EVR_SQ, "ResponsibleGroupCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0221, 0x0008, 0x0221,
      EVR_CS, "EquipmentModality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0222, 0x0008, 0x0222,
      EVR_LO, "ManufacturerRelatedModelGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0300, 0x0008, 0x0300,
      EVR_SQ, "PrivateDataElementCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0301, 0x0008, 0x0301,
      EVR_US, "PrivateGroupReference", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0302, 0x0008, 0x0302,
      EVR_LO, "PrivateCreatorReference", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0303, 0x0008, 0x0303,
      EVR_CS, "BlockIdentifyingInformationStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0304, 0x0008, 0x0304,
      EVR_US, "NonidentifyingPrivateElements", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0305, 0x0008, 0x0305,
      EVR_SQ, "DeidentificationActionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0306, 0x0008, 0x0306,
      EVR_US, "IdentifyingPrivateElements", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0307, 0x0008, 0x0307,
      EVR_CS, "DeidentificationAction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0308, 0x0008, 0x0308,
      EVR_US, "PrivateDataElement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0309, 0x0008, 0x0309,
      EVR_UL, "PrivateDataElementValueMultiplicity", 1, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030a, 0x0008, 0x030a,
      EVR_CS, "PrivateDataElementValueRepresentation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030b, 0x0008, 0x030b,
      EVR_UL, "PrivateDataElementNumberOfItems", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030c, 0x0008, 0x030c,
      EVR_UC, "PrivateDataElementName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030d, 0x0008, 0x030d,
      EVR_UC, "PrivateDataElementKeyword", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030e, 0x0008, 0x030e,
      EVR_UT, "PrivateDataElementDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x030f, 0x0008, 0x030f,
      EVR_UT, "PrivateDataElementEncoding", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0310, 0x0008, 0x0310,
      EVR_SQ, "PrivateDataElementDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0400, 0x0008, 0x0400,
      EVR_SQ, "ScopeOfInventorySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0401, 0x0008, 0x0401,
      EVR_LT, "InventoryPurpose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0402, 0x0008, 0x0402,
      EVR_LT, "InventoryInstanceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0403, 0x0008, 0x0403,
      EVR_CS, "InventoryLevel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0404, 0x0008, 0x0404,
      EVR_DT, "ItemInventoryDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0405, 0x0008, 0x0405,
      EVR_CS, "RemovedFromOperationalUse", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0406, 0x0008, 0x0406,
      EVR_SQ, "ReasonForRemovalCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0407, 0x0008, 0x0407,
      EVR_UR, "StoredInstanceBaseURI", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0408, 0x0008, 0x0408,
      EVR_UR, "FolderAccessURI", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0409, 0x0008, 0x0409,
      EVR_UR, "FileAccessURI", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040a, 0x0008, 0x040a,
      EVR_CS, "ContainerFileType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040b, 0x0008, 0x040b,
      EVR_UR, "FilenameInContainer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040c, 0x0008, 0x040c,
      EVR_UV, "FileOffsetInContainer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040d, 0x0008, 0x040d,
      EVR_UV, "FileLengthInContainer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040e, 0x0008, 0x040e,
      EVR_UI, "StoredInstanceTransferSyntaxUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x040f, 0x0008, 0x040f,
      EVR_CS, "ExtendedMatchingMechanisms", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0410, 0x0008, 0x0410,
      EVR_SQ, "RangeMatchingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0411, 0x0008, 0x0411,
      EVR_SQ, "ListOfUIDMatchingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0412, 0x0008, 0x0412,
      EVR_SQ, "EmptyValueMatchingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0413, 0x0008, 0x0413,
      EVR_SQ, "GeneralMatchingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0414, 0x0008, 0x0414,
      EVR_US, "RequestedStatusInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0415, 0x0008, 0x0415,
      EVR_CS, "RetainInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0416, 0x0008, 0x0416,
      EVR_DT, "ExpirationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0417, 0x0008, 0x0417,
      EVR_CS, "TransactionStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0418, 0x0008, 0x0418,
      EVR_LT, "TransactionStatusComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0419, 0x0008, 0x0419,
      EVR_SQ, "FileSetAccessSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041a, 0x0008, 0x041a,
      EVR_SQ, "FileAccessSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041b, 0x0008, 0x041b,
      EVR_OB, "RecordKey", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041c, 0x0008, 0x041c,
      EVR_OB, "PriorRecordKey", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041d, 0x0008, 0x041d,
      EVR_SQ, "MetadataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041e, 0x0008, 0x041e,
      EVR_SQ, "UpdatedMetadataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x041f, 0x0008, 0x041f,
      EVR_DT, "StudyUpdateDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0420, 0x0008, 0x0420,
      EVR_SQ, "InventoryAccessEndPointsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0421, 0x0008, 0x0421,
      EVR_SQ, "StudyAccessEndPointsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0422, 0x0008, 0x0422,
      EVR_SQ, "IncorporatedInventoryInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0423, 0x0008, 0x0423,
      EVR_SQ, "InventoriedStudiesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0424, 0x0008, 0x0424,
      EVR_SQ, "InventoriedSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0425, 0x0008, 0x0425,
      EVR_SQ, "InventoriedInstancesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0426, 0x0008, 0x0426,
      EVR_CS, "InventoryCompletionStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0427, 0x0008, 0x0427,
      EVR_UL, "NumberOfStudyRecordsInInstance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0428, 0x0008, 0x0428,
      EVR_UV, "TotalNumberOfStudyRecords", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x0429, 0x0008, 0x0429,
      EVR_UV, "MaximumNumberOfRecords", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1000, 0x0008, 0x1000,
      EVR_AE, "RETIRED_NetworkID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1010, 0x0008, 0x1010,
      EVR_SH, "StationName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1030, 0x0008, 0x1030,
      EVR_LO, "StudyDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1032, 0x0008, 0x1032,
      EVR_SQ, "ProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x103e, 0x0008, 0x103e,
      EVR_LO, "SeriesDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x103f, 0x0008, 0x103f,
      EVR_SQ, "SeriesDescriptionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1040, 0x0008, 0x1040,
      EVR_LO, "InstitutionalDepartmentName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1041, 0x0008, 0x1041,
      EVR_SQ, "InstitutionalDepartmentTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1048, 0x0008, 0x1048,
      EVR_PN, "PhysiciansOfRecord", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1049, 0x0008, 0x1049,
      EVR_SQ, "PhysiciansOfRecordIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1050, 0x0008, 0x1050,
      EVR_PN, "PerformingPhysicianName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1052, 0x0008, 0x1052,
      EVR_SQ, "PerformingPhysicianIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1060, 0x0008, 0x1060,
      EVR_PN, "NameOfPhysiciansReadingStudy", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1062, 0x0008, 0x1062,
      EVR_SQ, "PhysiciansReadingStudyIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1070, 0x0008, 0x1070,
      EVR_PN, "OperatorsName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1072, 0x0008, 0x1072,
      EVR_SQ, "OperatorIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1080, 0x0008, 0x1080,
      EVR_LO, "AdmittingDiagnosesDescription", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1084, 0x0008, 0x1084,
      EVR_SQ, "AdmittingDiagnosesCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1088, 0x0008, 0x1088,
      EVR_LO, "PyramidDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1090, 0x0008, 0x1090,
      EVR_LO, "ManufacturerModelName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1100, 0x0008, 0x1100,
      EVR_SQ, "RETIRED_ReferencedResultsSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1110, 0x0008, 0x1110,
      EVR_SQ, "ReferencedStudySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1111, 0x0008, 0x1111,
      EVR_SQ, "ReferencedPerformedProcedureStepSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1112, 0x0008, 0x1112,
      EVR_SQ, "ReferencedInstancesBySOPClassSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1115, 0x0008, 0x1115,
      EVR_SQ, "ReferencedSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1120, 0x0008, 0x1120,
      EVR_SQ, "ReferencedPatientSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1125, 0x0008, 0x1125,
      EVR_SQ, "ReferencedVisitSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1130, 0x0008, 0x1130,
      EVR_SQ, "RETIRED_ReferencedOverlaySequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1134, 0x0008, 0x1134,
      EVR_SQ, "ReferencedStereometricInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x113a, 0x0008, 0x113a,
      EVR_SQ, "ReferencedWaveformSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1140, 0x0008, 0x1140,
      EVR_SQ, "ReferencedImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1145, 0x0008, 0x1145,
      EVR_SQ, "RETIRED_ReferencedCurveSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x114a, 0x0008, 0x114a,
      EVR_SQ, "ReferencedInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x114b, 0x0008, 0x114b,
      EVR_SQ, "ReferencedRealWorldValueMappingInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x114c, 0x0008, 0x114c,
      EVR_SQ, "ReferencedSegmentationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x114d, 0x0008, 0x114d,
      EVR_SQ, "ReferencedSurfaceSegmentationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1150, 0x0008, 0x1150,
      EVR_UI, "ReferencedSOPClassUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1155, 0x0008, 0x1155,
      EVR_UI, "ReferencedSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1156, 0x0008, 0x1156,
      EVR_SQ, "DefinitionSourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x115a, 0x0008, 0x115a,
      EVR_UI, "SOPClassesSupported", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1160, 0x0008, 0x1160,
      EVR_IS, "ReferencedFrameNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1161, 0x0008, 0x1161,
      EVR_UL, "SimpleFrameList", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1162, 0x0008, 0x1162,
      EVR_UL, "CalculatedFrameList", 3, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1163, 0x0008, 0x1163,
      EVR_FD, "TimeRange", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1164, 0x0008, 0x1164,
      EVR_SQ, "FrameExtractionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1167, 0x0008, 0x1167,
      EVR_UI, "MultiFrameSourceSOPInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1190, 0x0008, 0x1190,
      EVR_UR, "RetrieveURL", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1195, 0x0008, 0x1195,
      EVR_UI, "TransactionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1196, 0x0008, 0x1196,
      EVR_US, "WarningReason", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1197, 0x0008, 0x1197,
      EVR_US, "FailureReason", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1198, 0x0008, 0x1198,
      EVR_SQ, "FailedSOPSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1199, 0x0008, 0x1199,
      EVR_SQ, "ReferencedSOPSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x119a, 0x0008, 0x119a,
      EVR_SQ, "OtherFailuresSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x119b, 0x0008, 0x119b,
      EVR_SQ, "FailedStudySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1200, 0x0008, 0x1200,
      EVR_SQ, "StudiesContainingOtherReferencedInstancesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x1250, 0x0008, 0x1250,
      EVR_SQ, "RelatedSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2110, 0x0008, 0x2110,
      EVR_CS, "RETIRED_LossyImageCompressionRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2111, 0x0008, 0x2111,
      EVR_ST, "DerivationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2112, 0x0008, 0x2112,
      EVR_SQ, "SourceImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2120, 0x0008, 0x2120,
      EVR_SH, "StageName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2122, 0x0008, 0x2122,
      EVR_IS, "StageNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2124, 0x0008, 0x2124,
      EVR_IS, "NumberOfStages", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2127, 0x0008, 0x2127,
      EVR_SH, "ViewName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2128, 0x0008, 0x2128,
      EVR_IS, "ViewNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2129, 0x0008, 0x2129,
      EVR_IS, "NumberOfEventTimers", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x212a, 0x0008, 0x212a,
      EVR_IS, "NumberOfViewsInStage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2130, 0x0008, 0x2130,
      EVR_DS, "EventElapsedTimes", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2132, 0x0008, 0x2132,
      EVR_LO, "EventTimerNames", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2133, 0x0008, 0x2133,
      EVR_SQ, "EventTimerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2134, 0x0008, 0x2134,
      EVR_FD, "EventTimeOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2135, 0x0008, 0x2135,
      EVR_SQ, "EventCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2142, 0x0008, 0x2142,
      EVR_IS, "StartTrim", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2143, 0x0008, 0x2143,
      EVR_IS, "StopTrim", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2144, 0x0008, 0x2144,
      EVR_IS, "RecommendedDisplayFrameRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2200, 0x0008, 0x2200,
      EVR_CS, "RETIRED_TransducerPosition", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2204, 0x0008, 0x2204,
      EVR_CS, "RETIRED_TransducerOrientation", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2208, 0x0008, 0x2208,
      EVR_CS, "RETIRED_AnatomicStructure", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2218, 0x0008, 0x2218,
      EVR_SQ, "AnatomicRegionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2220, 0x0008, 0x2220,
      EVR_SQ, "AnatomicRegionModifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2228, 0x0008, 0x2228,
      EVR_SQ, "PrimaryAnatomicStructureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2229, 0x0008, 0x2229,
      EVR_SQ, "RETIRED_AnatomicStructureSpaceOrRegionSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2230, 0x0008, 0x2230,
      EVR_SQ, "PrimaryAnatomicStructureModifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2240, 0x0008, 0x2240,
      EVR_SQ, "RETIRED_TransducerPositionSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2242, 0x0008, 0x2242,
      EVR_SQ, "RETIRED_TransducerPositionModifierSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2244, 0x0008, 0x2244,
      EVR_SQ, "RETIRED_TransducerOrientationSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2246, 0x0008, 0x2246,
      EVR_SQ, "RETIRED_TransducerOrientationModifierSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2251, 0x0008, 0x2251,
      EVR_SQ, "RETIRED_AnatomicStructureSpaceOrRegionCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2253, 0x0008, 0x2253,
      EVR_SQ, "RETIRED_AnatomicPortalOfEntranceCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2255, 0x0008, 0x2255,
      EVR_SQ, "RETIRED_AnatomicApproachDirectionCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2256, 0x0008, 0x2256,
      EVR_ST, "RETIRED_AnatomicPerspectiveDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2257, 0x0008, 0x2257,
      EVR_SQ, "RETIRED_AnatomicPerspectiveCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2258, 0x0008, 0x2258,
      EVR_ST, "RETIRED_AnatomicLocationOfExaminingInstrumentDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x2259, 0x0008, 0x2259,
      EVR_SQ, "RETIRED_AnatomicLocationOfExaminingInstrumentCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x225a, 0x0008, 0x225a,
      EVR_SQ, "RETIRED_AnatomicStructureSpaceOrRegionModifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x225c, 0x0008, 0x225c,
      EVR_SQ, "RETIRED_OnAxisBackgroundAnatomicStructureCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x3001, 0x0008, 0x3001,
      EVR_SQ, "AlternateRepresentationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x3002, 0x0008, 0x3002,
      EVR_UI, "AvailableTransferSyntaxUID", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x3010, 0x0008, 0x3010,
      EVR_UI, "IrradiationEventUID", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x3011, 0x0008, 0x3011,
      EVR_SQ, "SourceIrradiationEventSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x3012, 0x0008, 0x3012,
      EVR_UI, "RadiopharmaceuticalAdministrationEventUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x4000, 0x0008, 0x4000,
      EVR_LT, "RETIRED_IdentifyingComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9007, 0x0008, 0x9007,
      EVR_CS, "FrameType", 4, 5, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9092, 0x0008, 0x9092,
      EVR_SQ, "ReferencedImageEvidenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9121, 0x0008, 0x9121,
      EVR_SQ, "ReferencedRawDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9123, 0x0008, 0x9123,
      EVR_UI, "CreatorVersionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9124, 0x0008, 0x9124,
      EVR_SQ, "DerivationImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9154, 0x0008, 0x9154,
      EVR_SQ, "SourceImageEvidenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9205, 0x0008, 0x9205,
      EVR_CS, "PixelPresentation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9206, 0x0008, 0x9206,
      EVR_CS, "VolumetricProperties", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9207, 0x0008, 0x9207,
      EVR_CS, "VolumeBasedCalculationTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9208, 0x0008, 0x9208,
      EVR_CS, "ComplexImageComponent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9209, 0x0008, 0x9209,
      EVR_CS, "AcquisitionContrast", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9215, 0x0008, 0x9215,
      EVR_SQ, "DerivationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9237, 0x0008, 0x9237,
      EVR_SQ, "ReferencedPresentationStateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9410, 0x0008, 0x9410,
      EVR_SQ, "ReferencedOtherPlaneSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9458, 0x0008, 0x9458,
      EVR_SQ, "FrameDisplaySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9459, 0x0008, 0x9459,
      EVR_FL, "RecommendedDisplayFrameRateInFloat", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0008, 0x9460, 0x0008, 0x9460,
      EVR_CS, "SkipFrameRangeFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_LT, "DataObjectRecognitionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_LT, "OriginalFileName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_DS, "NumberOfMeasurements", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_SQ, "AnonymizerUIDMap", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_OB, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "MERGE TECHNOLOGIES, INC." }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_UL, "FileLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0009, 0x0000, 0x0009, 0x0000,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_UN, "RISPatientInfoIMGEF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_US, "SIENETCommandField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_UN, "CRHeaderInformation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FFP DATA" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_UL, "FileSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_LO, "FullFidelity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0001, 0x0009, 0x0001,
      EVR_UN, "SIENETGeneralPurposeIMGEF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ISI" }
  , { 0x0009, 0x0002, 0x0009, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0009, 0x0002, 0x0009, 0x0002,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0002, 0x0009, 0x0002,
      EVR_SH, "SuiteId", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0003, 0x0009, 0x0003,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0009, 0x0003, 0x0009, 0x0003,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0004, 0x0009, 0x0004,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0009, 0x0004, 0x0009, 0x0004,
      EVR_LO, "ImageDataConsistence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0004, 0x0009, 0x0004,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0004, 0x0009, 0x0004,
      EVR_SH, "ImageControlUnit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0004, 0x0009, 0x0004,
      EVR_SH, "ProductId", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0005, 0x0009, 0x0005,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0005, 0x0009, 0x0005,
      EVR_OW, "ImageUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0006, 0x0009, 0x0006,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0006, 0x0009, 0x0006,
      EVR_OW, "RouteImageUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0007, 0x0009, 0x0007,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0008, 0x0009, 0x0008,
      EVR_CS, "ExposureChannel", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0009, 0x0008, 0x0009, 0x0008,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0008, 0x0009, 0x0008,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0008, 0x0009, 0x0008,
      EVR_UL, "ImageDisplayInformationVersionNo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0009, 0x0009, 0x0009,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0009, 0x0009, 0x0009,
      EVR_UL, "PatientInformationVersionNo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x000a, 0x0009, 0x000a,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x000b, 0x0009, 0x000b,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x000c, 0x0009, 0x000c,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x000c, 0x0009, 0x000c,
      EVR_OW, "FilmUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x000d, 0x0009, 0x000d,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x000e, 0x0009, 0x000e,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x000f, 0x0009, 0x000f,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "Comments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_CS, "ExposureUnitTypeCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "OriginalFileLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "StorageMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "GeneratorIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_UI, "AnonymizerUIDKey", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DICOM" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "Comments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LT, "Comments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "RecognitionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0010, 0x0009, 0x0010,
      EVR_LO, "SPIRelease", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR" }
  , { 0x0009, 0x0011, 0x0009, 0x0011,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA" }
  , { 0x0009, 0x0011, 0x0009, 0x0011,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0011, 0x0009, 0x0011,
      EVR_LT, "GantryIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0011, 0x0009, 0x0011,
      EVR_SL, "StudyFlags", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_UL, "EvaluationMaskImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_LT, "X-RayTubeIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DICOM" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_SL, "StudyType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0012, 0x0009, 0x0012,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR" }
  , { 0x0009, 0x0013, 0x0009, 0x0013,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA" }
  , { 0x0009, 0x0013, 0x0009, 0x0013,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0013, 0x0009, 0x0013,
      EVR_LT, "DetectorIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0014, 0x0009, 0x0014,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA" }
  , { 0x0009, 0x0014, 0x0009, 0x0014,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0014, 0x0009, 0x0014,
      EVR_LT, "ReceiverPLA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0009, 0x0014, 0x0009, 0x0014,
      EVR_LT, "DASIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LO, "UniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LO, "SPIImageUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ACUSON" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LT, "SMIIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LO, "SPIImageUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0009, 0x0015, 0x0009, 0x0015,
      EVR_LO, "SPIImageUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0009, 0x0016, 0x0009, 0x0016,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0016, 0x0009, 0x0016,
      EVR_US, "TransferPriority", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0009, 0x0016, 0x0009, 0x0016,
      EVR_LT, "CPUIdentificationLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0017, 0x0009, 0x0017,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0018, 0x0009, 0x0018,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0018, 0x0009, 0x0018,
      EVR_LT, "DataSetIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0009, 0x001a, 0x0009, 0x001a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x001e, 0x0009, 0x001e,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_LT, "HeaderVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  LAB" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_UI, "AnonymizerUIDValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  ORI" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0020, 0x0009, 0x0020,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0021, 0x0009, 0x0021,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0021, 0x0009, 0x0021,
      EVR_SL, "SeriesFlags", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0022, 0x0009, 0x0022,
      EVR_SH, "UserOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0023, 0x0009, 0x0023,
      EVR_SL, "InitiationType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0024, 0x0009, 0x0024,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x0024, 0x0009, 0x0024,
      EVR_SL, "InitiationDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0025, 0x0009, 0x0025,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x0025, 0x0009, 0x0025,
      EVR_SL, "InitiationCountRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0026, 0x0009, 0x0026,
      EVR_DA, "LastMoveDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0009, 0x0026, 0x0009, 0x0026,
      EVR_SL, "NumberEnergySets", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0027, 0x0009, 0x0027,
      EVR_TM, "LastMoveTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0009, 0x0027, 0x0009, 0x0027,
      EVR_SL, "ImageActualDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0027, 0x0009, 0x0027,
      EVR_SL, "NumberDetectors", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0029, 0x0009, 0x0029,
      EVR_LT, "ActualUser", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0009, 0x0029, 0x0009, 0x0029,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x002a, 0x0009, 0x002a,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x002c, 0x0009, 0x002c,
      EVR_LO, "SeriesComments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x002d, 0x0009, 0x002d,
      EVR_SL, "TrackBeatAverage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x002e, 0x0009, 0x002e,
      EVR_FD, "DistancePrescribed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x002f, 0x0009, 0x002f,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_SQ, "AnonymizerPatientIDMap", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  ORI" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_SH, "ServiceId", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_UL, "ByteOffsetOfOriginalHeader", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0030, 0x0009, 0x0030,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0031, 0x0009, 0x0031,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0031, 0x0009, 0x0031,
      EVR_LT, "PACSUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0031, 0x0009, 0x0031,
      EVR_SH, "MobileLocationNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x0031, 0x0009, 0x0031,
      EVR_UL, "LengthOfOriginalHeader", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0032, 0x0009, 0x0032,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0032, 0x0009, 0x0032,
      EVR_TM, "ExposureStartTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0009, 0x0034, 0x0009, 0x0034,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0034, 0x0009, 0x0034,
      EVR_LT, "ClusterUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0035, 0x0009, 0x0035,
      EVR_SL, "GantryLocusType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0037, 0x0009, 0x0037,
      EVR_SL, "StartingHeartRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0038, 0x0009, 0x0038,
      EVR_LT, "SystemUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0038, 0x0009, 0x0038,
      EVR_SL, "RRWindowWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0039, 0x0009, 0x0039,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0039, 0x0009, 0x0039,
      EVR_SL, "RRWindowOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x003a, 0x0009, 0x003a,
      EVR_SL, "PercentCycleImaged", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x003e, 0x0009, 0x003e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x003f, 0x0009, 0x003f,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_US, "DataObjectType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_LO, "AnonymizerPatientIDKey", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_SQ, "AlternateImageSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_US, "DataObjectType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_US, "DataObjectType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0040, 0x0009, 0x0040,
      EVR_UL, "ByteOffsetOfPixelmatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0041, 0x0009, 0x0041,
      EVR_LT, "DataObjectSubtype", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0009, 0x0041, 0x0009, 0x0041,
      EVR_SL, "PatientFlags", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0041, 0x0009, 0x0041,
      EVR_LO, "DataObjectSubtype", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0009, 0x0041, 0x0009, 0x0041,
      EVR_LO, "DataObjectSubtype", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0009, 0x0041, 0x0009, 0x0041,
      EVR_UL, "LengthOfPixelmatrixInBytes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0042, 0x0009, 0x0042,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0042, 0x0009, 0x0042,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x0042, 0x0009, 0x0042,
      EVR_DA, "PatientCreationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0043, 0x0009, 0x0043,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x0043, 0x0009, 0x0043,
      EVR_TM, "PatientCreationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0009, 0x0050, 0x0009, 0x0050,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0050, 0x0009, 0x0050,
      EVR_LO, "AnonymizerPatientIDValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DCMTK_ANONYMIZER" }
  , { 0x0009, 0x0050, 0x0009, 0x0050,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0050, 0x0009, 0x0050,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0051, 0x0009, 0x0051,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  IDE" }
  , { 0x0009, 0x0051, 0x0009, 0x0051,
      EVR_LT, "StudyUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0051, 0x0009, 0x0051,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x0060, 0x0009, 0x0060,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0061, 0x0009, 0x0061,
      EVR_LT, "SeriesUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0070, 0x0009, 0x0070,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0075, 0x0009, 0x0075,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0080, 0x0009, 0x0080,
      EVR_LO, "KanjiHospitalName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0080, 0x0009, 0x0080,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0090, 0x0009, 0x0090,
      EVR_ST, "DistributionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x0090, 0x0009, 0x0090,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0009, 0x0091, 0x0009, 0x0091,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x0092, 0x0009, 0x0092,
      EVR_SH, "KanjiDepartmentName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x00c0, 0x0009, 0x00c0,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0009, 0x00c1, 0x0009, 0x00c1,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0009, 0x00e2, 0x0009, 0x00e2,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00e3, 0x0009, 0x00e3,
      EVR_UI, "EquipmentUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00e6, 0x0009, 0x00e6,
      EVR_SH, "GenesisVersionNow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00e7, 0x0009, 0x00e7,
      EVR_UL, "ExamRecordChecksum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00e8, 0x0009, 0x00e8,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00e9, 0x0009, 0x00e9,
      EVR_SL, "ActualSeriesDataTimeStamp", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IDEN_01" }
  , { 0x0009, 0x00f0, 0x0009, 0x00f0,
      EVR_CS, "BlackeningProcessFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0009, 0x00f2, 0x0009, 0x00f2,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x00f3, 0x0009, 0x00f3,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x00f4, 0x0009, 0x00f4,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x00f5, 0x0009, 0x00f5,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x00f5, 0x0009, 0x00f5,
      EVR_LT, "PDMEFIDPlaceholder", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x00f6, 0x0009, 0x00f6,
      EVR_LT, "PDMDataObjectTypeExtension", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0009, 0x00f7, 0x0009, 0x00f7,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0009, 0x00f8, 0x0009, 0x00f8,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0009, 0x00fb, 0x0009, 0x00fb,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
#endif
  , { 0x0010, 0x0010, 0x0010, 0x0010,
      EVR_PN, "PatientName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0020, 0x0010, 0x0020,
      EVR_LO, "PatientID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0021, 0x0010, 0x0021,
      EVR_LO, "IssuerOfPatientID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0022, 0x0010, 0x0022,
      EVR_CS, "TypeOfPatientID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0024, 0x0010, 0x0024,
      EVR_SQ, "IssuerOfPatientIDQualifiersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0026, 0x0010, 0x0026,
      EVR_SQ, "SourcePatientGroupIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0027, 0x0010, 0x0027,
      EVR_SQ, "GroupOfPatientsIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0028, 0x0010, 0x0028,
      EVR_US, "SubjectRelativePositionInImage", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0030, 0x0010, 0x0030,
      EVR_DA, "PatientBirthDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0032, 0x0010, 0x0032,
      EVR_TM, "PatientBirthTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0033, 0x0010, 0x0033,
      EVR_LO, "PatientBirthDateInAlternativeCalendar", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0034, 0x0010, 0x0034,
      EVR_LO, "PatientDeathDateInAlternativeCalendar", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0035, 0x0010, 0x0035,
      EVR_CS, "PatientAlternativeCalendar", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0040, 0x0010, 0x0040,
      EVR_CS, "PatientSex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0050, 0x0010, 0x0050,
      EVR_SQ, "PatientInsurancePlanCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0101, 0x0010, 0x0101,
      EVR_SQ, "PatientPrimaryLanguageCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0102, 0x0010, 0x0102,
      EVR_SQ, "PatientPrimaryLanguageModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0200, 0x0010, 0x0200,
      EVR_CS, "QualityControlSubject", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0201, 0x0010, 0x0201,
      EVR_SQ, "QualityControlSubjectTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0212, 0x0010, 0x0212,
      EVR_UC, "StrainDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0213, 0x0010, 0x0213,
      EVR_LO, "StrainNomenclature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0214, 0x0010, 0x0214,
      EVR_LO, "StrainStockNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0215, 0x0010, 0x0215,
      EVR_SQ, "StrainSourceRegistryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0216, 0x0010, 0x0216,
      EVR_SQ, "StrainStockSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0217, 0x0010, 0x0217,
      EVR_LO, "StrainSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0218, 0x0010, 0x0218,
      EVR_UT, "StrainAdditionalInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0219, 0x0010, 0x0219,
      EVR_SQ, "StrainCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0221, 0x0010, 0x0221,
      EVR_SQ, "GeneticModificationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0222, 0x0010, 0x0222,
      EVR_UC, "GeneticModificationsDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0223, 0x0010, 0x0223,
      EVR_LO, "GeneticModificationsNomenclature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x0229, 0x0010, 0x0229,
      EVR_SQ, "GeneticModificationsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1000, 0x0010, 0x1000,
      EVR_LO, "RETIRED_OtherPatientIDs", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1001, 0x0010, 0x1001,
      EVR_PN, "OtherPatientNames", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1002, 0x0010, 0x1002,
      EVR_SQ, "OtherPatientIDsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1005, 0x0010, 0x1005,
      EVR_PN, "PatientBirthName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1010, 0x0010, 0x1010,
      EVR_AS, "PatientAge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1020, 0x0010, 0x1020,
      EVR_DS, "PatientSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1021, 0x0010, 0x1021,
      EVR_SQ, "PatientSizeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1022, 0x0010, 0x1022,
      EVR_DS, "PatientBodyMassIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1023, 0x0010, 0x1023,
      EVR_DS, "MeasuredAPDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1024, 0x0010, 0x1024,
      EVR_DS, "MeasuredLateralDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1030, 0x0010, 0x1030,
      EVR_DS, "PatientWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1040, 0x0010, 0x1040,
      EVR_LO, "PatientAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1050, 0x0010, 0x1050,
      EVR_LO, "RETIRED_InsurancePlanIdentification", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1060, 0x0010, 0x1060,
      EVR_PN, "PatientMotherBirthName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1080, 0x0010, 0x1080,
      EVR_LO, "MilitaryRank", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1081, 0x0010, 0x1081,
      EVR_LO, "BranchOfService", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1090, 0x0010, 0x1090,
      EVR_LO, "RETIRED_MedicalRecordLocator", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x1100, 0x0010, 0x1100,
      EVR_SQ, "ReferencedPatientPhotoSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2000, 0x0010, 0x2000,
      EVR_LO, "MedicalAlerts", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2110, 0x0010, 0x2110,
      EVR_LO, "Allergies", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2150, 0x0010, 0x2150,
      EVR_LO, "CountryOfResidence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2152, 0x0010, 0x2152,
      EVR_LO, "RegionOfResidence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2154, 0x0010, 0x2154,
      EVR_SH, "PatientTelephoneNumbers", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2155, 0x0010, 0x2155,
      EVR_LT, "PatientTelecomInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2160, 0x0010, 0x2160,
      EVR_SH, "EthnicGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2161, 0x0010, 0x2161,
      EVR_SQ, "EthnicGroupCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2180, 0x0010, 0x2180,
      EVR_SH, "Occupation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x21a0, 0x0010, 0x21a0,
      EVR_CS, "SmokingStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x21b0, 0x0010, 0x21b0,
      EVR_LT, "AdditionalPatientHistory", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x21c0, 0x0010, 0x21c0,
      EVR_US, "PregnancyStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x21d0, 0x0010, 0x21d0,
      EVR_DA, "LastMenstrualDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x21f0, 0x0010, 0x21f0,
      EVR_LO, "PatientReligiousPreference", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2201, 0x0010, 0x2201,
      EVR_LO, "PatientSpeciesDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2202, 0x0010, 0x2202,
      EVR_SQ, "PatientSpeciesCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2203, 0x0010, 0x2203,
      EVR_CS, "PatientSexNeutered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2210, 0x0010, 0x2210,
      EVR_CS, "AnatomicalOrientationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2292, 0x0010, 0x2292,
      EVR_LO, "PatientBreedDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2293, 0x0010, 0x2293,
      EVR_SQ, "PatientBreedCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2294, 0x0010, 0x2294,
      EVR_SQ, "BreedRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2295, 0x0010, 0x2295,
      EVR_LO, "BreedRegistrationNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2296, 0x0010, 0x2296,
      EVR_SQ, "BreedRegistryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2297, 0x0010, 0x2297,
      EVR_PN, "ResponsiblePerson", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2298, 0x0010, 0x2298,
      EVR_CS, "ResponsiblePersonRole", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x2299, 0x0010, 0x2299,
      EVR_LO, "ResponsibleOrganization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x4000, 0x0010, 0x4000,
      EVR_LT, "PatientComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0010, 0x9431, 0x0010, 0x9431,
      EVR_FL, "ExaminedBodyThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0011, 0x0001, 0x0011, 0x0001,
      EVR_LT, "PatientDOB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_PATNT_01" }
  , { 0x0011, 0x0001, 0x0011, 0x0001,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0011, 0x0002, 0x0011, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0011, 0x0003, 0x0011, 0x0003,
      EVR_LT, "PatientUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x0004, 0x0011, 0x0004,
      EVR_LT, "PatientID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x000a, 0x0011, 0x000a,
      EVR_SL, "SeriesType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x000a, 0x0011, 0x000a,
      EVR_LT, "CaseID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x000a, 0x0011, 0x000a,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x000b, 0x0011, 0x000b,
      EVR_SL, "EffectiveSeriesDuration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x000c, 0x0011, 0x000c,
      EVR_SL, "NumBeats", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x000d, 0x0011, 0x000d,
      EVR_LO, "RadioNuclideName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LO, "Organ", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LO, "Organ", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LT, "PatientEntryID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LT, "Organ", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_LT, "PatientUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_DA, "RegistrationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x0010, 0x0011, 0x0010,
      EVR_SS, "PatientStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PATI_01" }
  , { 0x0011, 0x0011, 0x0011, 0x0011,
      EVR_LT, "PatientID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0011, 0x0011, 0x0011,
      EVR_TM, "RegistrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x0012, 0x0011, 0x0012,
      EVR_LO, "DatasetName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0013, 0x0011, 0x0013,
      EVR_SL, "DatasetType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0015, 0x0011, 0x0015,
      EVR_LO, "AllergyIndication", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0011, 0x0015, 0x0011, 0x0015,
      EVR_SL, "DetectorNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0015, 0x0011, 0x0015,
      EVR_LO, "AllergyIndication", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0011, 0x0015, 0x0011, 0x0015,
      EVR_LT, "AllergyIndication", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0011, 0x0016, 0x0011, 0x0016,
      EVR_SL, "EnergyNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0017, 0x0011, 0x0017,
      EVR_SL, "RRIntervalWindowNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0018, 0x0011, 0x0018,
      EVR_SL, "MGBinNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0018, 0x0011, 0x0018,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 2;1" }
  , { 0x0011, 0x0019, 0x0011, 0x0019,
      EVR_FD, "RadiusOfRotation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x001a, 0x0011, 0x001a,
      EVR_SL, "DetectorCountZone", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x001b, 0x0011, 0x001b,
      EVR_SL, "NumEnergyWindows", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x001c, 0x0011, 0x001c,
      EVR_SL, "EnergyOffset", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x001d, 0x0011, 0x001d,
      EVR_SL, "EnergyRange", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x001f, 0x0011, 0x001f,
      EVR_SL, "ImageOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0020, 0x0011, 0x0020,
      EVR_LO, "Pregnancy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0011, 0x0020, 0x0011, 0x0020,
      EVR_SL, "FluoroTimer", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0020, 0x0011, 0x0020,
      EVR_LO, "Pregnancy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0011, 0x0020, 0x0011, 0x0020,
      EVR_LT, "Pregnancy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0011, 0x0020, 0x0011, 0x0020,
      EVR_DA, "PatientRegistrationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0021, 0x0011, 0x0021,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0011, 0x0021, 0x0011, 0x0021,
      EVR_TM, "PatientRegistrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0022, 0x0011, 0x0022,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0011, 0x0022, 0x0011, 0x0022,
      EVR_LT, "RequestID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x0022, 0x0011, 0x0022,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x0023, 0x0011, 0x0023,
      EVR_SL, "UseFOVMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0023, 0x0011, 0x0023,
      EVR_LT, "ExaminationUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x0023, 0x0011, 0x0023,
      EVR_DS, "UsedPatientWeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x0024, 0x0011, 0x0024,
      EVR_SL, "FOVMaskYCutoffAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0025, 0x0011, 0x0025,
      EVR_SL, "PtopDoseAreaProduct", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0025, 0x0011, 0x0025,
      EVR_SL, "FOVMaskCutoffAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0026, 0x0011, 0x0026,
      EVR_SL, "PtopTotalSkinDose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0026, 0x0011, 0x0026,
      EVR_SL, "TableOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0027, 0x0011, 0x0027,
      EVR_SL, "ROITopLeft", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0028, 0x0011, 0x0028,
      EVR_SL, "ROIBottomRight", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0028, 0x0011, 0x0028,
      EVR_SL, "FluoroTimerB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0011, 0x0028, 0x0011, 0x0028,
      EVR_SL, "FluoroTimerA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0011, 0x0029, 0x0011, 0x0029,
      EVR_SL, "FluoroSkinDoseB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0011, 0x0029, 0x0011, 0x0029,
      EVR_SL, "FluoroSkinDoseA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0011, 0x002a, 0x0011, 0x002a,
      EVR_SL, "TotalSkinDoseB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0011, 0x002a, 0x0011, 0x002a,
      EVR_SL, "TotalSkinDoseA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0011, 0x002b, 0x0011, 0x002b,
      EVR_SL, "FluoroDoseAreaProductB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0011, 0x002b, 0x0011, 0x002b,
      EVR_SL, "FluoroDoseAreaProductA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0011, 0x002c, 0x0011, 0x002c,
      EVR_SL, "TotalDoseAreaProductB", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0011, 0x002c, 0x0011, 0x002c,
      EVR_SL, "TotalDoseAreaProductA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0011, 0x0030, 0x0011, 0x0030,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0030, 0x0011, 0x0030,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0030, 0x0011, 0x0030,
      EVR_LT, "PatientnameRIS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0031, 0x0011, 0x0031,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0011, 0x0031, 0x0011, 0x0031,
      EVR_LT, "PatientprenameRIS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0032, 0x0011, 0x0032,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0011, 0x0033, 0x0011, 0x0033,
      EVR_LO, "EnergyCorrectName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0034, 0x0011, 0x0034,
      EVR_LO, "SpatialCorrectName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0035, 0x0011, 0x0035,
      EVR_LO, "PatientInitialPuckCounter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0035, 0x0011, 0x0035,
      EVR_LO, "TuningCalibName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0036, 0x0011, 0x0036,
      EVR_LO, "UniformityCorrectName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0037, 0x0011, 0x0037,
      EVR_LO, "AcquisitionSpecificCorrectName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0038, 0x0011, 0x0038,
      EVR_SL, "ByteOrder", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x003a, 0x0011, 0x003a,
      EVR_SL, "PictureFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x003b, 0x0011, 0x003b,
      EVR_FD, "PixelScale", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x003c, 0x0011, 0x003c,
      EVR_FD, "PixelOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x003e, 0x0011, 0x003e,
      EVR_SL, "FOVShape", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x003f, 0x0011, 0x003f,
      EVR_SL, "DatasetFlags", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0040, 0x0011, 0x0040,
      EVR_SS, "SPIDataObjectType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0011, 0x0040, 0x0011, 0x0040,
      EVR_LT, "PatientHospitalStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0040, 0x0011, 0x0040,
      EVR_IS, "OrganCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0011, 0x0041, 0x0011, 0x0041,
      EVR_LT, "MedicalAlerts", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0042, 0x0011, 0x0042,
      EVR_LT, "ContrastAllergies", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0011, 0x0044, 0x0011, 0x0044,
      EVR_FD, "ThresholdCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0045, 0x0011, 0x0045,
      EVR_FD, "ThresholdWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0046, 0x0011, 0x0046,
      EVR_SL, "InterpolationType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0055, 0x0011, 0x0055,
      EVR_FD, "Period", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x0056, 0x0011, 0x0056,
      EVR_FD, "ElapsedTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0011, 0x00a1, 0x0011, 0x00a1,
      EVR_DA, "PatientRegistrationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00a2, 0x0011, 0x00a2,
      EVR_TM, "PatientRegistrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00b0, 0x0011, 0x00b0,
      EVR_LT, "PatientLastName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00b2, 0x0011, 0x00b2,
      EVR_LT, "PatientFirstName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00b4, 0x0011, 0x00b4,
      EVR_LT, "PatientHospitalStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00bc, 0x0011, 0x00bc,
      EVR_TM, "CurrentLocationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00c0, 0x0011, 0x00c0,
      EVR_LT, "PatientInsuranceStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00d0, 0x0011, 0x00d0,
      EVR_LT, "PatientBillingType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0011, 0x00d2, 0x0011, 0x00d2,
      EVR_LT, "PatientBillingAddress", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x0012, 0x0010, 0x0012, 0x0010,
      EVR_LO, "ClinicalTrialSponsorName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0020, 0x0012, 0x0020,
      EVR_LO, "ClinicalTrialProtocolID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0021, 0x0012, 0x0021,
      EVR_LO, "ClinicalTrialProtocolName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0022, 0x0012, 0x0022,
      EVR_LO, "IssuerOfClinicalTrialProtocolID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0023, 0x0012, 0x0023,
      EVR_SQ, "OtherClinicalTrialProtocolIDsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0030, 0x0012, 0x0030,
      EVR_LO, "ClinicalTrialSiteID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0031, 0x0012, 0x0031,
      EVR_LO, "ClinicalTrialSiteName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0032, 0x0012, 0x0032,
      EVR_LO, "IssuerOfClinicalTrialSiteID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0040, 0x0012, 0x0040,
      EVR_LO, "ClinicalTrialSubjectID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0041, 0x0012, 0x0041,
      EVR_LO, "IssuerOfClinicalTrialSubjectID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0042, 0x0012, 0x0042,
      EVR_LO, "ClinicalTrialSubjectReadingID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0043, 0x0012, 0x0043,
      EVR_LO, "IssuerOfClinicalTrialSubjectReadingID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0050, 0x0012, 0x0050,
      EVR_LO, "ClinicalTrialTimePointID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0051, 0x0012, 0x0051,
      EVR_ST, "ClinicalTrialTimePointDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0052, 0x0012, 0x0052,
      EVR_FD, "LongitudinalTemporalOffsetFromEvent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0053, 0x0012, 0x0053,
      EVR_CS, "LongitudinalTemporalEventType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0054, 0x0012, 0x0054,
      EVR_SQ, "ClinicalTrialTimePointTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0055, 0x0012, 0x0055,
      EVR_LO, "IssuerOfClinicalTrialTimePointID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0060, 0x0012, 0x0060,
      EVR_LO, "ClinicalTrialCoordinatingCenterName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0062, 0x0012, 0x0062,
      EVR_CS, "PatientIdentityRemoved", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0063, 0x0012, 0x0063,
      EVR_LO, "DeidentificationMethod", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0064, 0x0012, 0x0064,
      EVR_SQ, "DeidentificationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0071, 0x0012, 0x0071,
      EVR_LO, "ClinicalTrialSeriesID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0072, 0x0012, 0x0072,
      EVR_LO, "ClinicalTrialSeriesDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0073, 0x0012, 0x0073,
      EVR_LO, "IssuerOfClinicalTrialSeriesID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0081, 0x0012, 0x0081,
      EVR_LO, "ClinicalTrialProtocolEthicsCommitteeName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0082, 0x0012, 0x0082,
      EVR_LO, "ClinicalTrialProtocolEthicsCommitteeApprovalNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0083, 0x0012, 0x0083,
      EVR_SQ, "ConsentForClinicalTrialUseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0084, 0x0012, 0x0084,
      EVR_CS, "DistributionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0085, 0x0012, 0x0085,
      EVR_CS, "ConsentForDistributionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0086, 0x0012, 0x0086,
      EVR_DA, "EthicsCommitteeApprovalEffectivenessStartDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0012, 0x0087, 0x0012, 0x0087,
      EVR_DA, "EthicsCommitteeApprovalEffectivenessEndDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0013, 0x0000, 0x0013, 0x0000,
      EVR_LT, "ModifyingPhysician", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0010, 0x0013, 0x0010,
      EVR_FD, "DigitalFOV", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0010, 0x0013, 0x0010,
      EVR_DA, "ModificationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0011, 0x0013, 0x0011,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0012, 0x0013, 0x0012,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0012, 0x0013, 0x0012,
      EVR_TM, "ModificationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0016, 0x0013, 0x0016,
      EVR_SL, "AutoTrackPeak", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0017, 0x0013, 0x0017,
      EVR_SL, "AutoTrackWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0018, 0x0013, 0x0018,
      EVR_FD, "TransmissionScanTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0019, 0x0013, 0x0019,
      EVR_FD, "TransmissionMaskWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x001a, 0x0013, 0x001a,
      EVR_FD, "CopperAttenuatorThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x001c, 0x0013, 0x001c,
      EVR_FD, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x001d, 0x0013, 0x001d,
      EVR_FD, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x001e, 0x0013, 0x001e,
      EVR_FD, "TomoViewOffset", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0020, 0x0013, 0x0020,
      EVR_LO, "PatientName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0022, 0x0013, 0x0022,
      EVR_LO, "PatientId", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0026, 0x0013, 0x0026,
      EVR_LT, "StudyComments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GENIE_1" }
  , { 0x0013, 0x0030, 0x0013, 0x0030,
      EVR_DA, "PatientBirthdate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0031, 0x0013, 0x0031,
      EVR_DS, "PatientWeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0032, 0x0013, 0x0032,
      EVR_LT, "PatientsMaidenName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0033, 0x0013, 0x0033,
      EVR_LT, "ReferringPhysician", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0034, 0x0013, 0x0034,
      EVR_LT, "AdmittingDiagnosis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0035, 0x0013, 0x0035,
      EVR_LO, "PatientSex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0040, 0x0013, 0x0040,
      EVR_LO, "ProcedureDescription", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0042, 0x0013, 0x0042,
      EVR_LO, "RestDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0044, 0x0013, 0x0044,
      EVR_LO, "PatientPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0046, 0x0013, 0x0046,
      EVR_LT, "ViewDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0050, 0x0013, 0x0050,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0051, 0x0013, 0x0051,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0052, 0x0013, 0x0052,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0053, 0x0013, 0x0053,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0054, 0x0013, 0x0054,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0055, 0x0013, 0x0055,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0013, 0x0056, 0x0013, 0x0056,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
#endif
  , { 0x0014, 0x0023, 0x0014, 0x0023,
      EVR_ST, "RETIRED_CADFileFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0024, 0x0014, 0x0024,
      EVR_ST, "RETIRED_ComponentReferenceSystem", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0025, 0x0014, 0x0025,
      EVR_ST, "ComponentManufacturingProcedure", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0028, 0x0014, 0x0028,
      EVR_ST, "ComponentManufacturer", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0030, 0x0014, 0x0030,
      EVR_DS, "MaterialThickness", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0032, 0x0014, 0x0032,
      EVR_DS, "MaterialPipeDiameter", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0034, 0x0014, 0x0034,
      EVR_DS, "MaterialIsolationDiameter", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0042, 0x0014, 0x0042,
      EVR_ST, "MaterialGrade", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0044, 0x0014, 0x0044,
      EVR_ST, "MaterialPropertiesDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0045, 0x0014, 0x0045,
      EVR_ST, "RETIRED_MaterialPropertiesFileFormatRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0046, 0x0014, 0x0046,
      EVR_LT, "MaterialNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0050, 0x0014, 0x0050,
      EVR_CS, "ComponentShape", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0052, 0x0014, 0x0052,
      EVR_CS, "CurvatureType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0054, 0x0014, 0x0054,
      EVR_DS, "OuterDiameter", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0056, 0x0014, 0x0056,
      EVR_DS, "InnerDiameter", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0100, 0x0014, 0x0100,
      EVR_LO, "ComponentWelderIDs", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0101, 0x0014, 0x0101,
      EVR_CS, "SecondaryApprovalStatus", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0102, 0x0014, 0x0102,
      EVR_DA, "SecondaryReviewDate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0103, 0x0014, 0x0103,
      EVR_TM, "SecondaryReviewTime", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0104, 0x0014, 0x0104,
      EVR_PN, "SecondaryReviewerName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0105, 0x0014, 0x0105,
      EVR_ST, "RepairID", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0106, 0x0014, 0x0106,
      EVR_SQ, "MultipleComponentApprovalSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0107, 0x0014, 0x0107,
      EVR_CS, "OtherApprovalStatus", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0108, 0x0014, 0x0108,
      EVR_CS, "OtherSecondaryApprovalStatus", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0200, 0x0014, 0x0200,
      EVR_SQ, "DataElementLabelSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0201, 0x0014, 0x0201,
      EVR_SQ, "DataElementLabelItemSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0202, 0x0014, 0x0202,
      EVR_AT, "DataElement", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0203, 0x0014, 0x0203,
      EVR_LO, "DataElementName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0204, 0x0014, 0x0204,
      EVR_LO, "DataElementDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0205, 0x0014, 0x0205,
      EVR_CS, "DataElementConditionality", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0206, 0x0014, 0x0206,
      EVR_IS, "DataElementMinimumCharacters", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x0207, 0x0014, 0x0207,
      EVR_IS, "DataElementMaximumCharacters", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x1010, 0x0014, 0x1010,
      EVR_ST, "ActualEnvironmentalConditions", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x1020, 0x0014, 0x1020,
      EVR_DA, "ExpiryDate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x1040, 0x0014, 0x1040,
      EVR_ST, "EnvironmentalConditions", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2002, 0x0014, 0x2002,
      EVR_SQ, "EvaluatorSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2004, 0x0014, 0x2004,
      EVR_IS, "EvaluatorNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2006, 0x0014, 0x2006,
      EVR_PN, "EvaluatorName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2008, 0x0014, 0x2008,
      EVR_IS, "EvaluationAttempt", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2012, 0x0014, 0x2012,
      EVR_SQ, "IndicationSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2014, 0x0014, 0x2014,
      EVR_IS, "IndicationNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2016, 0x0014, 0x2016,
      EVR_SH, "IndicationLabel", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2018, 0x0014, 0x2018,
      EVR_ST, "IndicationDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x201a, 0x0014, 0x201a,
      EVR_CS, "IndicationType", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x201c, 0x0014, 0x201c,
      EVR_CS, "IndicationDisposition", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x201e, 0x0014, 0x201e,
      EVR_SQ, "IndicationROISequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2030, 0x0014, 0x2030,
      EVR_SQ, "IndicationPhysicalPropertySequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2032, 0x0014, 0x2032,
      EVR_SH, "PropertyLabel", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2202, 0x0014, 0x2202,
      EVR_IS, "CoordinateSystemNumberOfAxes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2204, 0x0014, 0x2204,
      EVR_SQ, "CoordinateSystemAxesSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2206, 0x0014, 0x2206,
      EVR_ST, "CoordinateSystemAxisDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2208, 0x0014, 0x2208,
      EVR_CS, "CoordinateSystemDataSetMapping", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x220a, 0x0014, 0x220a,
      EVR_IS, "CoordinateSystemAxisNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x220c, 0x0014, 0x220c,
      EVR_CS, "CoordinateSystemAxisType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x220e, 0x0014, 0x220e,
      EVR_CS, "CoordinateSystemAxisUnits", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2210, 0x0014, 0x2210,
      EVR_OB, "CoordinateSystemAxisValues", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2220, 0x0014, 0x2220,
      EVR_SQ, "CoordinateSystemTransformSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2222, 0x0014, 0x2222,
      EVR_ST, "TransformDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2224, 0x0014, 0x2224,
      EVR_IS, "TransformNumberOfAxes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2226, 0x0014, 0x2226,
      EVR_IS, "TransformOrderOfAxes", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x2228, 0x0014, 0x2228,
      EVR_CS, "TransformedAxisUnits", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x222a, 0x0014, 0x222a,
      EVR_DS, "CoordinateSystemTransformRotationAndScaleMatrix", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x222c, 0x0014, 0x222c,
      EVR_DS, "CoordinateSystemTransformTranslationMatrix", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3011, 0x0014, 0x3011,
      EVR_DS, "InternalDetectorFrameTime", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3012, 0x0014, 0x3012,
      EVR_DS, "NumberOfFramesIntegrated", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3020, 0x0014, 0x3020,
      EVR_SQ, "DetectorTemperatureSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3022, 0x0014, 0x3022,
      EVR_ST, "SensorName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3024, 0x0014, 0x3024,
      EVR_DS, "HorizontalOffsetOfSensor", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3026, 0x0014, 0x3026,
      EVR_DS, "VerticalOffsetOfSensor", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3028, 0x0014, 0x3028,
      EVR_DS, "SensorTemperature", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3040, 0x0014, 0x3040,
      EVR_SQ, "DarkCurrentSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3050, 0x0014, 0x3050,
      EVR_ox, "DarkCurrentCounts", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3060, 0x0014, 0x3060,
      EVR_SQ, "GainCorrectionReferenceSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3070, 0x0014, 0x3070,
      EVR_ox, "AirCounts", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3071, 0x0014, 0x3071,
      EVR_DS, "KVUsedInGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3072, 0x0014, 0x3072,
      EVR_DS, "MAUsedInGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3073, 0x0014, 0x3073,
      EVR_DS, "NumberOfFramesUsedForIntegration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3074, 0x0014, 0x3074,
      EVR_LO, "FilterMaterialUsedInGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3075, 0x0014, 0x3075,
      EVR_DS, "FilterThicknessUsedInGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3076, 0x0014, 0x3076,
      EVR_DA, "DateOfGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3077, 0x0014, 0x3077,
      EVR_TM, "TimeOfGainCalibration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3080, 0x0014, 0x3080,
      EVR_OB, "BadPixelImage", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3099, 0x0014, 0x3099,
      EVR_LT, "CalibrationNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3100, 0x0014, 0x3100,
      EVR_LT, "LinearityCorrectionTechnique", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x3101, 0x0014, 0x3101,
      EVR_LT, "BeamHardeningCorrectionTechnique", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4002, 0x0014, 0x4002,
      EVR_SQ, "PulserEquipmentSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4004, 0x0014, 0x4004,
      EVR_CS, "PulserType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4006, 0x0014, 0x4006,
      EVR_LT, "PulserNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4008, 0x0014, 0x4008,
      EVR_SQ, "ReceiverEquipmentSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x400a, 0x0014, 0x400a,
      EVR_CS, "AmplifierType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x400c, 0x0014, 0x400c,
      EVR_LT, "ReceiverNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x400e, 0x0014, 0x400e,
      EVR_SQ, "PreAmplifierEquipmentSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x400f, 0x0014, 0x400f,
      EVR_LT, "PreAmplifierNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4010, 0x0014, 0x4010,
      EVR_SQ, "TransmitTransducerSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4011, 0x0014, 0x4011,
      EVR_SQ, "ReceiveTransducerSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4012, 0x0014, 0x4012,
      EVR_US, "NumberOfElements", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4013, 0x0014, 0x4013,
      EVR_CS, "ElementShape", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4014, 0x0014, 0x4014,
      EVR_DS, "ElementDimensionA", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4015, 0x0014, 0x4015,
      EVR_DS, "ElementDimensionB", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4016, 0x0014, 0x4016,
      EVR_DS, "ElementPitchA", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4017, 0x0014, 0x4017,
      EVR_DS, "MeasuredBeamDimensionA", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4018, 0x0014, 0x4018,
      EVR_DS, "MeasuredBeamDimensionB", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4019, 0x0014, 0x4019,
      EVR_DS, "LocationOfMeasuredBeamDiameter", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x401a, 0x0014, 0x401a,
      EVR_DS, "NominalFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x401b, 0x0014, 0x401b,
      EVR_DS, "MeasuredCenterFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x401c, 0x0014, 0x401c,
      EVR_DS, "MeasuredBandwidth", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x401d, 0x0014, 0x401d,
      EVR_DS, "ElementPitchB", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4020, 0x0014, 0x4020,
      EVR_SQ, "PulserSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4022, 0x0014, 0x4022,
      EVR_DS, "PulseWidth", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4024, 0x0014, 0x4024,
      EVR_DS, "ExcitationFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4026, 0x0014, 0x4026,
      EVR_CS, "ModulationType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4028, 0x0014, 0x4028,
      EVR_DS, "Damping", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4030, 0x0014, 0x4030,
      EVR_SQ, "ReceiverSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4031, 0x0014, 0x4031,
      EVR_DS, "AcquiredSoundpathLength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4032, 0x0014, 0x4032,
      EVR_CS, "AcquisitionCompressionType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4033, 0x0014, 0x4033,
      EVR_IS, "AcquisitionSampleSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4034, 0x0014, 0x4034,
      EVR_DS, "RectifierSmoothing", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4035, 0x0014, 0x4035,
      EVR_SQ, "DACSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4036, 0x0014, 0x4036,
      EVR_CS, "DACType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4038, 0x0014, 0x4038,
      EVR_DS, "DACGainPoints", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x403a, 0x0014, 0x403a,
      EVR_DS, "DACTimePoints", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x403c, 0x0014, 0x403c,
      EVR_DS, "DACAmplitude", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4040, 0x0014, 0x4040,
      EVR_SQ, "PreAmplifierSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4050, 0x0014, 0x4050,
      EVR_SQ, "TransmitTransducerSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4051, 0x0014, 0x4051,
      EVR_SQ, "ReceiveTransducerSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4052, 0x0014, 0x4052,
      EVR_DS, "IncidentAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4054, 0x0014, 0x4054,
      EVR_ST, "CouplingTechnique", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4056, 0x0014, 0x4056,
      EVR_ST, "CouplingMedium", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4057, 0x0014, 0x4057,
      EVR_DS, "CouplingVelocity", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4058, 0x0014, 0x4058,
      EVR_DS, "ProbeCenterLocationX", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4059, 0x0014, 0x4059,
      EVR_DS, "ProbeCenterLocationZ", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x405a, 0x0014, 0x405a,
      EVR_DS, "SoundPathLength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x405c, 0x0014, 0x405c,
      EVR_ST, "DelayLawIdentifier", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4060, 0x0014, 0x4060,
      EVR_SQ, "GateSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4062, 0x0014, 0x4062,
      EVR_DS, "GateThreshold", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4064, 0x0014, 0x4064,
      EVR_DS, "VelocityOfSound", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4070, 0x0014, 0x4070,
      EVR_SQ, "CalibrationSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4072, 0x0014, 0x4072,
      EVR_ST, "CalibrationProcedure", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4074, 0x0014, 0x4074,
      EVR_SH, "ProcedureVersion", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4076, 0x0014, 0x4076,
      EVR_DA, "ProcedureCreationDate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4078, 0x0014, 0x4078,
      EVR_DA, "ProcedureExpirationDate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x407a, 0x0014, 0x407a,
      EVR_DA, "ProcedureLastModifiedDate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x407c, 0x0014, 0x407c,
      EVR_TM, "CalibrationTime", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x407e, 0x0014, 0x407e,
      EVR_DA, "CalibrationDate", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4080, 0x0014, 0x4080,
      EVR_SQ, "ProbeDriveEquipmentSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4081, 0x0014, 0x4081,
      EVR_CS, "DriveType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4082, 0x0014, 0x4082,
      EVR_LT, "ProbeDriveNotes", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4083, 0x0014, 0x4083,
      EVR_SQ, "DriveProbeSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4084, 0x0014, 0x4084,
      EVR_DS, "ProbeInductance", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4085, 0x0014, 0x4085,
      EVR_DS, "ProbeResistance", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4086, 0x0014, 0x4086,
      EVR_SQ, "ReceiveProbeSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4087, 0x0014, 0x4087,
      EVR_SQ, "ProbeDriveSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4088, 0x0014, 0x4088,
      EVR_DS, "BridgeResistors", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4089, 0x0014, 0x4089,
      EVR_DS, "ProbeOrientationAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x408b, 0x0014, 0x408b,
      EVR_DS, "UserSelectedGainY", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x408c, 0x0014, 0x408c,
      EVR_DS, "UserSelectedPhase", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x408d, 0x0014, 0x408d,
      EVR_DS, "UserSelectedOffsetX", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x408e, 0x0014, 0x408e,
      EVR_DS, "UserSelectedOffsetY", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4091, 0x0014, 0x4091,
      EVR_SQ, "ChannelSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x4092, 0x0014, 0x4092,
      EVR_DS, "ChannelThreshold", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x409a, 0x0014, 0x409a,
      EVR_SQ, "ScannerSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x409b, 0x0014, 0x409b,
      EVR_ST, "ScanProcedure", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x409c, 0x0014, 0x409c,
      EVR_DS, "TranslationRateX", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x409d, 0x0014, 0x409d,
      EVR_DS, "TranslationRateY", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x409f, 0x0014, 0x409f,
      EVR_DS, "ChannelOverlap", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x40a0, 0x0014, 0x40a0,
      EVR_LO, "ImageQualityIndicatorType", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x40a1, 0x0014, 0x40a1,
      EVR_LO, "ImageQualityIndicatorMaterial", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x40a2, 0x0014, 0x40a2,
      EVR_LO, "ImageQualityIndicatorSize", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5002, 0x0014, 0x5002,
      EVR_IS, "LINACEnergy", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5004, 0x0014, 0x5004,
      EVR_IS, "LINACOutput", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5100, 0x0014, 0x5100,
      EVR_US, "ActiveAperture", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5101, 0x0014, 0x5101,
      EVR_DS, "TotalAperture", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5102, 0x0014, 0x5102,
      EVR_DS, "ApertureElevation", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5103, 0x0014, 0x5103,
      EVR_DS, "MainLobeAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5104, 0x0014, 0x5104,
      EVR_DS, "MainRoofAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5105, 0x0014, 0x5105,
      EVR_CS, "ConnectorType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5106, 0x0014, 0x5106,
      EVR_SH, "WedgeModelNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5107, 0x0014, 0x5107,
      EVR_DS, "WedgeAngleFloat", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5108, 0x0014, 0x5108,
      EVR_DS, "WedgeRoofAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5109, 0x0014, 0x5109,
      EVR_CS, "WedgeElement1Position", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510a, 0x0014, 0x510a,
      EVR_DS, "WedgeMaterialVelocity", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510b, 0x0014, 0x510b,
      EVR_SH, "WedgeMaterial", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510c, 0x0014, 0x510c,
      EVR_DS, "WedgeOffsetZ", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510d, 0x0014, 0x510d,
      EVR_DS, "WedgeOriginOffsetX", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510e, 0x0014, 0x510e,
      EVR_DS, "WedgeTimeDelay", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x510f, 0x0014, 0x510f,
      EVR_SH, "WedgeName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5110, 0x0014, 0x5110,
      EVR_SH, "WedgeManufacturerName", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5111, 0x0014, 0x5111,
      EVR_LO, "WedgeDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5112, 0x0014, 0x5112,
      EVR_DS, "NominalBeamAngle", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5113, 0x0014, 0x5113,
      EVR_DS, "WedgeOffsetX", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5114, 0x0014, 0x5114,
      EVR_DS, "WedgeOffsetY", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5115, 0x0014, 0x5115,
      EVR_DS, "WedgeTotalLength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5116, 0x0014, 0x5116,
      EVR_DS, "WedgeInContactLength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5117, 0x0014, 0x5117,
      EVR_DS, "WedgeFrontGap", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5118, 0x0014, 0x5118,
      EVR_DS, "WedgeTotalHeight", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x5119, 0x0014, 0x5119,
      EVR_DS, "WedgeFrontHeight", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511a, 0x0014, 0x511a,
      EVR_DS, "WedgeRearHeight", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511b, 0x0014, 0x511b,
      EVR_DS, "WedgeTotalWidth", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511c, 0x0014, 0x511c,
      EVR_DS, "WedgeInContactWidth", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511d, 0x0014, 0x511d,
      EVR_DS, "WedgeChamferHeight", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511e, 0x0014, 0x511e,
      EVR_CS, "WedgeCurve", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x511f, 0x0014, 0x511f,
      EVR_DS, "RadiusAlongWedge", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6001, 0x0014, 0x6001,
      EVR_SQ, "ThermalCameraSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6002, 0x0014, 0x6002,
      EVR_DS, "AcquisitionFrameRate", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6003, 0x0014, 0x6003,
      EVR_DS, "IntegrationTime", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6004, 0x0014, 0x6004,
      EVR_DS, "NumberOfCalibrationFrames", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6005, 0x0014, 0x6005,
      EVR_DS, "NumberOfRowsInFullAcquisitionImage", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6006, 0x0014, 0x6006,
      EVR_DS, "NumberOfColumnsInFullAcquisitionImage", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6007, 0x0014, 0x6007,
      EVR_SQ, "ThermalSourceSettingsSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6008, 0x0014, 0x6008,
      EVR_DS, "SourceHorizontalPitch", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6009, 0x0014, 0x6009,
      EVR_DS, "SourceVerticalPitch", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600a, 0x0014, 0x600a,
      EVR_DS, "SourceHorizontalScanSpeed", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600b, 0x0014, 0x600b,
      EVR_DS, "ThermalSourceModulationFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600c, 0x0014, 0x600c,
      EVR_SQ, "InductionSourceSettingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600d, 0x0014, 0x600d,
      EVR_DS, "CoilFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600e, 0x0014, 0x600e,
      EVR_DS, "CurrentAmplitudeAcrossCoil", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x600f, 0x0014, 0x600f,
      EVR_SQ, "FlashSourceSettingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6010, 0x0014, 0x6010,
      EVR_DS, "FlashDuration", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6011, 0x0014, 0x6011,
      EVR_DS, "FlashFrameNumber", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6012, 0x0014, 0x6012,
      EVR_SQ, "LaserSourceSettingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6013, 0x0014, 0x6013,
      EVR_DS, "HorizontalLaserSpotDimension", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6014, 0x0014, 0x6014,
      EVR_DS, "VerticalLaserSpotDimension", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6015, 0x0014, 0x6015,
      EVR_DS, "LaserWavelength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6016, 0x0014, 0x6016,
      EVR_DS, "LaserPower", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6017, 0x0014, 0x6017,
      EVR_SQ, "ForcedGasSettingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6018, 0x0014, 0x6018,
      EVR_SQ, "VibrationSourceSettingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6019, 0x0014, 0x6019,
      EVR_DS, "VibrationExcitationFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601a, 0x0014, 0x601a,
      EVR_DS, "VibrationExcitationVoltage", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601b, 0x0014, 0x601b,
      EVR_CS, "ThermographyDataCaptureMethod", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601c, 0x0014, 0x601c,
      EVR_CS, "ThermalTechnique", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601d, 0x0014, 0x601d,
      EVR_SQ, "ThermalCameraCoreSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601e, 0x0014, 0x601e,
      EVR_CS, "DetectorWavelengthRange", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x601f, 0x0014, 0x601f,
      EVR_CS, "ThermalCameraCalibrationType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6020, 0x0014, 0x6020,
      EVR_UV, "AcquisitionImageCounter", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6021, 0x0014, 0x6021,
      EVR_DS, "FrontPanelTemperature", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6022, 0x0014, 0x6022,
      EVR_DS, "AirGapTemperature", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6023, 0x0014, 0x6023,
      EVR_DS, "VerticalPixelSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6024, 0x0014, 0x6024,
      EVR_DS, "HorizontalPixelSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6025, 0x0014, 0x6025,
      EVR_ST, "DataStreamingProtocol", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6026, 0x0014, 0x6026,
      EVR_SQ, "LensSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6027, 0x0014, 0x6027,
      EVR_DS, "FieldOfView", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6028, 0x0014, 0x6028,
      EVR_LO, "LensFilterManufacturer", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6029, 0x0014, 0x6029,
      EVR_CS, "CutoffFilterType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602a, 0x0014, 0x602a,
      EVR_DS, "LensFilterCutOffWavelength", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602b, 0x0014, 0x602b,
      EVR_SQ, "ThermalSourceSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602c, 0x0014, 0x602c,
      EVR_CS, "ThermalSourceMotionState", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602d, 0x0014, 0x602d,
      EVR_CS, "ThermalSourceMotionType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602e, 0x0014, 0x602e,
      EVR_SQ, "InductionHeatingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x602f, 0x0014, 0x602f,
      EVR_ST, "CoilConfigurationID", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6030, 0x0014, 0x6030,
      EVR_DS, "NumberOfTurnsInCoil", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6031, 0x0014, 0x6031,
      EVR_CS, "ShapeOfIndividualTurn", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6032, 0x0014, 0x6032,
      EVR_DS, "SizeOfIndividualTurn", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6033, 0x0014, 0x6033,
      EVR_DS, "DistanceBetweenTurns", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6034, 0x0014, 0x6034,
      EVR_SQ, "FlashHeatingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6035, 0x0014, 0x6035,
      EVR_DS, "NumberOfLamps", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6036, 0x0014, 0x6036,
      EVR_ST, "FlashSynchronizationProtocol", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6037, 0x0014, 0x6037,
      EVR_CS, "FlashModificationStatus", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6038, 0x0014, 0x6038,
      EVR_SQ, "LaserHeatingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6039, 0x0014, 0x6039,
      EVR_LO, "LaserManufacturer", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603a, 0x0014, 0x603a,
      EVR_LO, "LaserModelNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603b, 0x0014, 0x603b,
      EVR_ST, "LaserTypeDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603c, 0x0014, 0x603c,
      EVR_SQ, "ForcedGasHeatingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603d, 0x0014, 0x603d,
      EVR_LO, "GasUsedForHeatingCoolingPart", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603e, 0x0014, 0x603e,
      EVR_SQ, "VibrationSonicHeatingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x603f, 0x0014, 0x603f,
      EVR_LO, "ProbeManufacturer", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6040, 0x0014, 0x6040,
      EVR_LO, "ProbeModelNumber", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6041, 0x0014, 0x6041,
      EVR_DS, "ApertureSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6042, 0x0014, 0x6042,
      EVR_DS, "ProbeResonantFrequency", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6043, 0x0014, 0x6043,
      EVR_UT, "HeatSourceDescription", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6044, 0x0014, 0x6044,
      EVR_CS, "SurfacePreparationWithOpticalCoating", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6045, 0x0014, 0x6045,
      EVR_ST, "OpticalCoatingType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6046, 0x0014, 0x6046,
      EVR_DS, "ThermalConductivityOfExposedSurface", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6047, 0x0014, 0x6047,
      EVR_DS, "MaterialDensity", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6048, 0x0014, 0x6048,
      EVR_DS, "SpecificHeatOfInspectionSurface", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6049, 0x0014, 0x6049,
      EVR_DS, "EmissivityOfInspectionSurface", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x604a, 0x0014, 0x604a,
      EVR_CS, "ElectromagneticClassificationOfInspectionSurface", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x604c, 0x0014, 0x604c,
      EVR_DS, "MovingWindowSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x604d, 0x0014, 0x604d,
      EVR_CS, "MovingWindowType", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x604e, 0x0014, 0x604e,
      EVR_DS, "MovingWindowWeights", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x604f, 0x0014, 0x604f,
      EVR_DS, "MovingWindowPitch", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6050, 0x0014, 0x6050,
      EVR_CS, "MovingWindowPaddingScheme", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6051, 0x0014, 0x6051,
      EVR_DS, "MovingWindowPaddingLength", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6052, 0x0014, 0x6052,
      EVR_SQ, "SpatialFilteringParametersSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6053, 0x0014, 0x6053,
      EVR_CS, "SpatialFilteringScheme", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6056, 0x0014, 0x6056,
      EVR_DS, "HorizontalMovingWindowSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6057, 0x0014, 0x6057,
      EVR_DS, "VerticalMovingWindowSize", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6059, 0x0014, 0x6059,
      EVR_SQ, "PolynomialFittingSequence", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605a, 0x0014, 0x605a,
      EVR_CS, "FittingDataType", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605b, 0x0014, 0x605b,
      EVR_CS, "OperationOnTimeAxisBeforeFitting", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605c, 0x0014, 0x605c,
      EVR_CS, "OperationOnPixelIntensityBeforeFitting", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605d, 0x0014, 0x605d,
      EVR_DS, "OrderOfPolynomial", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605e, 0x0014, 0x605e,
      EVR_CS, "IndependentVariableForPolynomialFit", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x605f, 0x0014, 0x605f,
      EVR_DS, "PolynomialCoefficients", 1, -1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0014, 0x6060, 0x0014, 0x6060,
      EVR_CS, "ThermographyPixelDataUnit", 1, 1, "DICOM/DICONDE",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0015, 0x0001, 0x0015, 0x0001,
      EVR_DS, "StenosisCalibrationRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_EXAMS_01" }
  , { 0x0015, 0x0002, 0x0015, 0x0002,
      EVR_DS, "StenosisMagnification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_EXAMS_01" }
  , { 0x0015, 0x0003, 0x0015, 0x0003,
      EVR_DS, "CardiacCalibrationRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_EXAMS_01" }
#endif
  , { 0x0016, 0x0001, 0x0016, 0x0001,
      EVR_DS, "WhitePoint", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0002, 0x0016, 0x0002,
      EVR_DS, "PrimaryChromaticities", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0003, 0x0016, 0x0003,
      EVR_UT, "BatteryLevel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0004, 0x0016, 0x0004,
      EVR_DS, "ExposureTimeInSeconds", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0005, 0x0016, 0x0005,
      EVR_DS, "FNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0006, 0x0016, 0x0006,
      EVR_IS, "OECFRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0007, 0x0016, 0x0007,
      EVR_IS, "OECFColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0008, 0x0016, 0x0008,
      EVR_UC, "OECFColumnNames", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0009, 0x0016, 0x0009,
      EVR_DS, "OECFValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000a, 0x0016, 0x000a,
      EVR_IS, "SpatialFrequencyResponseRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000b, 0x0016, 0x000b,
      EVR_IS, "SpatialFrequencyResponseColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000c, 0x0016, 0x000c,
      EVR_UC, "SpatialFrequencyResponseColumnNames", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000d, 0x0016, 0x000d,
      EVR_DS, "SpatialFrequencyResponseValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000e, 0x0016, 0x000e,
      EVR_IS, "ColorFilterArrayPatternRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x000f, 0x0016, 0x000f,
      EVR_IS, "ColorFilterArrayPatternColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0010, 0x0016, 0x0010,
      EVR_DS, "ColorFilterArrayPatternValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0011, 0x0016, 0x0011,
      EVR_US, "FlashFiringStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0012, 0x0016, 0x0012,
      EVR_US, "FlashReturnStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0013, 0x0016, 0x0013,
      EVR_US, "FlashMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0014, 0x0016, 0x0014,
      EVR_US, "FlashFunctionPresent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0015, 0x0016, 0x0015,
      EVR_US, "FlashRedEyeMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0016, 0x0016, 0x0016,
      EVR_US, "ExposureProgram", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0017, 0x0016, 0x0017,
      EVR_UT, "SpectralSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0018, 0x0016, 0x0018,
      EVR_IS, "PhotographicSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0019, 0x0016, 0x0019,
      EVR_IS, "SelfTimerMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001a, 0x0016, 0x001a,
      EVR_US, "SensitivityType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001b, 0x0016, 0x001b,
      EVR_IS, "StandardOutputSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001c, 0x0016, 0x001c,
      EVR_IS, "RecommendedExposureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001d, 0x0016, 0x001d,
      EVR_IS, "ISOSpeed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001e, 0x0016, 0x001e,
      EVR_IS, "ISOSpeedLatitudeyyy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x001f, 0x0016, 0x001f,
      EVR_IS, "ISOSpeedLatitudezzz", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0020, 0x0016, 0x0020,
      EVR_UT, "EXIFVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0021, 0x0016, 0x0021,
      EVR_DS, "ShutterSpeedValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0022, 0x0016, 0x0022,
      EVR_DS, "ApertureValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0023, 0x0016, 0x0023,
      EVR_DS, "BrightnessValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0024, 0x0016, 0x0024,
      EVR_DS, "ExposureBiasValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0025, 0x0016, 0x0025,
      EVR_DS, "MaxApertureValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0026, 0x0016, 0x0026,
      EVR_DS, "SubjectDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0027, 0x0016, 0x0027,
      EVR_US, "MeteringMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0028, 0x0016, 0x0028,
      EVR_US, "LightSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0029, 0x0016, 0x0029,
      EVR_DS, "FocalLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x002a, 0x0016, 0x002a,
      EVR_IS, "SubjectArea", 2, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x002b, 0x0016, 0x002b,
      EVR_OB, "MakerNote", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0030, 0x0016, 0x0030,
      EVR_DS, "Temperature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0031, 0x0016, 0x0031,
      EVR_DS, "Humidity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0032, 0x0016, 0x0032,
      EVR_DS, "Pressure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0033, 0x0016, 0x0033,
      EVR_DS, "WaterDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0034, 0x0016, 0x0034,
      EVR_DS, "Acceleration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0035, 0x0016, 0x0035,
      EVR_DS, "CameraElevationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0036, 0x0016, 0x0036,
      EVR_DS, "FlashEnergy", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0037, 0x0016, 0x0037,
      EVR_IS, "SubjectLocation", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0038, 0x0016, 0x0038,
      EVR_DS, "PhotographicExposureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0039, 0x0016, 0x0039,
      EVR_US, "SensingMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x003a, 0x0016, 0x003a,
      EVR_US, "FileSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x003b, 0x0016, 0x003b,
      EVR_US, "SceneType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0041, 0x0016, 0x0041,
      EVR_US, "CustomRendered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0042, 0x0016, 0x0042,
      EVR_US, "ExposureMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0043, 0x0016, 0x0043,
      EVR_US, "WhiteBalance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0044, 0x0016, 0x0044,
      EVR_DS, "DigitalZoomRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0045, 0x0016, 0x0045,
      EVR_IS, "FocalLengthIn35mmFilm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0046, 0x0016, 0x0046,
      EVR_US, "SceneCaptureType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0047, 0x0016, 0x0047,
      EVR_US, "GainControl", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0048, 0x0016, 0x0048,
      EVR_US, "Contrast", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0049, 0x0016, 0x0049,
      EVR_US, "Saturation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004a, 0x0016, 0x004a,
      EVR_US, "Sharpness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004b, 0x0016, 0x004b,
      EVR_OB, "DeviceSettingDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004c, 0x0016, 0x004c,
      EVR_US, "SubjectDistanceRange", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004d, 0x0016, 0x004d,
      EVR_UT, "CameraOwnerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004e, 0x0016, 0x004e,
      EVR_DS, "LensSpecification", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x004f, 0x0016, 0x004f,
      EVR_UT, "LensMake", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0050, 0x0016, 0x0050,
      EVR_UT, "LensModel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0051, 0x0016, 0x0051,
      EVR_UT, "LensSerialNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0061, 0x0016, 0x0061,
      EVR_CS, "InteroperabilityIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0062, 0x0016, 0x0062,
      EVR_OB, "InteroperabilityVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0070, 0x0016, 0x0070,
      EVR_OB, "GPSVersionID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0071, 0x0016, 0x0071,
      EVR_CS, "GPSLatitudeRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0072, 0x0016, 0x0072,
      EVR_DS, "GPSLatitude", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0073, 0x0016, 0x0073,
      EVR_CS, "GPSLongitudeRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0074, 0x0016, 0x0074,
      EVR_DS, "GPSLongitude", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0075, 0x0016, 0x0075,
      EVR_US, "GPSAltitudeRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0076, 0x0016, 0x0076,
      EVR_DS, "GPSAltitude", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0077, 0x0016, 0x0077,
      EVR_DT, "GPSTimeStamp", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0078, 0x0016, 0x0078,
      EVR_UT, "GPSSatellites", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0079, 0x0016, 0x0079,
      EVR_CS, "GPSStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007a, 0x0016, 0x007a,
      EVR_CS, "GPSMeasureMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007b, 0x0016, 0x007b,
      EVR_DS, "GPSDOP", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007c, 0x0016, 0x007c,
      EVR_CS, "GPSSpeedRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007d, 0x0016, 0x007d,
      EVR_DS, "GPSSpeed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007e, 0x0016, 0x007e,
      EVR_CS, "GPSTrackRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x007f, 0x0016, 0x007f,
      EVR_DS, "GPSTrack", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0080, 0x0016, 0x0080,
      EVR_CS, "GPSImgDirectionRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0081, 0x0016, 0x0081,
      EVR_DS, "GPSImgDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0082, 0x0016, 0x0082,
      EVR_UT, "GPSMapDatum", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0083, 0x0016, 0x0083,
      EVR_CS, "GPSDestLatitudeRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0084, 0x0016, 0x0084,
      EVR_DS, "GPSDestLatitude", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0085, 0x0016, 0x0085,
      EVR_CS, "GPSDestLongitudeRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0086, 0x0016, 0x0086,
      EVR_DS, "GPSDestLongitude", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0087, 0x0016, 0x0087,
      EVR_CS, "GPSDestBearingRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0088, 0x0016, 0x0088,
      EVR_DS, "GPSDestBearing", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x0089, 0x0016, 0x0089,
      EVR_CS, "GPSDestDistanceRef", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x008a, 0x0016, 0x008a,
      EVR_DS, "GPSDestDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x008b, 0x0016, 0x008b,
      EVR_OB, "GPSProcessingMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x008c, 0x0016, 0x008c,
      EVR_OB, "GPSAreaInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x008d, 0x0016, 0x008d,
      EVR_DT, "GPSDateStamp", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x008e, 0x0016, 0x008e,
      EVR_IS, "GPSDifferential", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x1001, 0x0016, 0x1001,
      EVR_CS, "LightSourcePolarization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x1002, 0x0016, 0x1002,
      EVR_DS, "EmitterColorTemperature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x1003, 0x0016, 0x1003,
      EVR_CS, "ContactMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x1004, 0x0016, 0x1004,
      EVR_CS, "ImmersionMedia", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0016, 0x1005, 0x0016, 0x1005,
      EVR_DS, "OpticalMagnificationFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0017, 0x0000, 0x0017, 0x0000,
      EVR_LO, "ExtendedBodyPart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0017, 0x0010, 0x0017, 0x0010,
      EVR_LO, "ExtendedViewPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0017, 0x00f0, 0x0017, 0x00f0,
      EVR_IS, "ImagesSOPClass", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
#endif
  , { 0x0018, 0x0010, 0x0018, 0x0010,
      EVR_LO, "ContrastBolusAgent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0012, 0x0018, 0x0012,
      EVR_SQ, "ContrastBolusAgentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0013, 0x0018, 0x0013,
      EVR_FL, "ContrastBolusT1Relaxivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0014, 0x0018, 0x0014,
      EVR_SQ, "ContrastBolusAdministrationRouteSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0015, 0x0018, 0x0015,
      EVR_CS, "BodyPartExamined", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0020, 0x0018, 0x0020,
      EVR_CS, "ScanningSequence", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0021, 0x0018, 0x0021,
      EVR_CS, "SequenceVariant", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0022, 0x0018, 0x0022,
      EVR_CS, "ScanOptions", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0023, 0x0018, 0x0023,
      EVR_CS, "MRAcquisitionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0024, 0x0018, 0x0024,
      EVR_SH, "SequenceName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0025, 0x0018, 0x0025,
      EVR_CS, "AngioFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0026, 0x0018, 0x0026,
      EVR_SQ, "InterventionDrugInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0027, 0x0018, 0x0027,
      EVR_TM, "InterventionDrugStopTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0028, 0x0018, 0x0028,
      EVR_DS, "InterventionDrugDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0029, 0x0018, 0x0029,
      EVR_SQ, "InterventionDrugCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x002a, 0x0018, 0x002a,
      EVR_SQ, "AdditionalDrugSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0030, 0x0018, 0x0030,
      EVR_LO, "RETIRED_Radionuclide", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0031, 0x0018, 0x0031,
      EVR_LO, "Radiopharmaceutical", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0032, 0x0018, 0x0032,
      EVR_DS, "RETIRED_EnergyWindowCenterline", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0033, 0x0018, 0x0033,
      EVR_DS, "RETIRED_EnergyWindowTotalWidth", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0034, 0x0018, 0x0034,
      EVR_LO, "InterventionDrugName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0035, 0x0018, 0x0035,
      EVR_TM, "InterventionDrugStartTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0036, 0x0018, 0x0036,
      EVR_SQ, "InterventionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0037, 0x0018, 0x0037,
      EVR_CS, "RETIRED_TherapyType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0038, 0x0018, 0x0038,
      EVR_CS, "InterventionStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0039, 0x0018, 0x0039,
      EVR_CS, "RETIRED_TherapyDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x003a, 0x0018, 0x003a,
      EVR_ST, "InterventionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0040, 0x0018, 0x0040,
      EVR_IS, "CineRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0042, 0x0018, 0x0042,
      EVR_CS, "InitialCineRunState", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0050, 0x0018, 0x0050,
      EVR_DS, "SliceThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0060, 0x0018, 0x0060,
      EVR_DS, "KVP", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0070, 0x0018, 0x0070,
      EVR_IS, "CountsAccumulated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0071, 0x0018, 0x0071,
      EVR_CS, "AcquisitionTerminationCondition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0072, 0x0018, 0x0072,
      EVR_DS, "EffectiveDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0073, 0x0018, 0x0073,
      EVR_CS, "AcquisitionStartCondition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0074, 0x0018, 0x0074,
      EVR_IS, "AcquisitionStartConditionData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0075, 0x0018, 0x0075,
      EVR_IS, "AcquisitionTerminationConditionData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0080, 0x0018, 0x0080,
      EVR_DS, "RepetitionTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0081, 0x0018, 0x0081,
      EVR_DS, "EchoTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0082, 0x0018, 0x0082,
      EVR_DS, "InversionTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0083, 0x0018, 0x0083,
      EVR_DS, "NumberOfAverages", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0084, 0x0018, 0x0084,
      EVR_DS, "ImagingFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0085, 0x0018, 0x0085,
      EVR_SH, "ImagedNucleus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0086, 0x0018, 0x0086,
      EVR_IS, "EchoNumbers", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0087, 0x0018, 0x0087,
      EVR_DS, "MagneticFieldStrength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0088, 0x0018, 0x0088,
      EVR_DS, "SpacingBetweenSlices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0089, 0x0018, 0x0089,
      EVR_IS, "NumberOfPhaseEncodingSteps", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0090, 0x0018, 0x0090,
      EVR_DS, "DataCollectionDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0091, 0x0018, 0x0091,
      EVR_IS, "EchoTrainLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0093, 0x0018, 0x0093,
      EVR_DS, "PercentSampling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0094, 0x0018, 0x0094,
      EVR_DS, "PercentPhaseFieldOfView", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x0095, 0x0018, 0x0095,
      EVR_DS, "PixelBandwidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1000, 0x0018, 0x1000,
      EVR_LO, "DeviceSerialNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1002, 0x0018, 0x1002,
      EVR_UI, "DeviceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1003, 0x0018, 0x1003,
      EVR_LO, "DeviceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1004, 0x0018, 0x1004,
      EVR_LO, "PlateID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1005, 0x0018, 0x1005,
      EVR_LO, "GeneratorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1006, 0x0018, 0x1006,
      EVR_LO, "GridID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1007, 0x0018, 0x1007,
      EVR_LO, "CassetteID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1008, 0x0018, 0x1008,
      EVR_LO, "GantryID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1009, 0x0018, 0x1009,
      EVR_UT, "UniqueDeviceIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x100a, 0x0018, 0x100a,
      EVR_SQ, "UDISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x100b, 0x0018, 0x100b,
      EVR_UI, "ManufacturerDeviceClassUID", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1010, 0x0018, 0x1010,
      EVR_LO, "SecondaryCaptureDeviceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1011, 0x0018, 0x1011,
      EVR_LO, "RETIRED_HardcopyCreationDeviceID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1012, 0x0018, 0x1012,
      EVR_DA, "DateOfSecondaryCapture", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1014, 0x0018, 0x1014,
      EVR_TM, "TimeOfSecondaryCapture", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1016, 0x0018, 0x1016,
      EVR_LO, "SecondaryCaptureDeviceManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1017, 0x0018, 0x1017,
      EVR_LO, "RETIRED_HardcopyDeviceManufacturer", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1018, 0x0018, 0x1018,
      EVR_LO, "SecondaryCaptureDeviceManufacturerModelName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1019, 0x0018, 0x1019,
      EVR_LO, "SecondaryCaptureDeviceSoftwareVersions", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x101a, 0x0018, 0x101a,
      EVR_LO, "RETIRED_HardcopyDeviceSoftwareVersion", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x101b, 0x0018, 0x101b,
      EVR_LO, "RETIRED_HardcopyDeviceManufacturerModelName", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1020, 0x0018, 0x1020,
      EVR_LO, "SoftwareVersions", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1022, 0x0018, 0x1022,
      EVR_SH, "VideoImageFormatAcquired", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1023, 0x0018, 0x1023,
      EVR_LO, "DigitalImageFormatAcquired", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1030, 0x0018, 0x1030,
      EVR_LO, "ProtocolName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1040, 0x0018, 0x1040,
      EVR_LO, "ContrastBolusRoute", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1041, 0x0018, 0x1041,
      EVR_DS, "ContrastBolusVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1042, 0x0018, 0x1042,
      EVR_TM, "ContrastBolusStartTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1043, 0x0018, 0x1043,
      EVR_TM, "ContrastBolusStopTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1044, 0x0018, 0x1044,
      EVR_DS, "ContrastBolusTotalDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1045, 0x0018, 0x1045,
      EVR_IS, "SyringeCounts", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1046, 0x0018, 0x1046,
      EVR_DS, "ContrastFlowRate", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1047, 0x0018, 0x1047,
      EVR_DS, "ContrastFlowDuration", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1048, 0x0018, 0x1048,
      EVR_CS, "ContrastBolusIngredient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1049, 0x0018, 0x1049,
      EVR_DS, "ContrastBolusIngredientConcentration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1050, 0x0018, 0x1050,
      EVR_DS, "SpatialResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1060, 0x0018, 0x1060,
      EVR_DS, "TriggerTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1061, 0x0018, 0x1061,
      EVR_LO, "TriggerSourceOrType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1062, 0x0018, 0x1062,
      EVR_IS, "NominalInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1063, 0x0018, 0x1063,
      EVR_DS, "FrameTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1064, 0x0018, 0x1064,
      EVR_LO, "CardiacFramingType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1065, 0x0018, 0x1065,
      EVR_DS, "FrameTimeVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1066, 0x0018, 0x1066,
      EVR_DS, "FrameDelay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1067, 0x0018, 0x1067,
      EVR_DS, "ImageTriggerDelay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1068, 0x0018, 0x1068,
      EVR_DS, "MultiplexGroupTimeOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1069, 0x0018, 0x1069,
      EVR_DS, "TriggerTimeOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x106a, 0x0018, 0x106a,
      EVR_CS, "SynchronizationTrigger", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x106c, 0x0018, 0x106c,
      EVR_US, "SynchronizationChannel", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x106e, 0x0018, 0x106e,
      EVR_UL, "TriggerSamplePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1070, 0x0018, 0x1070,
      EVR_LO, "RadiopharmaceuticalRoute", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1071, 0x0018, 0x1071,
      EVR_DS, "RadiopharmaceuticalVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1072, 0x0018, 0x1072,
      EVR_TM, "RadiopharmaceuticalStartTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1073, 0x0018, 0x1073,
      EVR_TM, "RadiopharmaceuticalStopTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1074, 0x0018, 0x1074,
      EVR_DS, "RadionuclideTotalDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1075, 0x0018, 0x1075,
      EVR_DS, "RadionuclideHalfLife", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1076, 0x0018, 0x1076,
      EVR_DS, "RadionuclidePositronFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1077, 0x0018, 0x1077,
      EVR_DS, "RadiopharmaceuticalSpecificActivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1078, 0x0018, 0x1078,
      EVR_DT, "RadiopharmaceuticalStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1079, 0x0018, 0x1079,
      EVR_DT, "RadiopharmaceuticalStopDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1080, 0x0018, 0x1080,
      EVR_CS, "BeatRejectionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1081, 0x0018, 0x1081,
      EVR_IS, "LowRRValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1082, 0x0018, 0x1082,
      EVR_IS, "HighRRValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1083, 0x0018, 0x1083,
      EVR_IS, "IntervalsAcquired", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1084, 0x0018, 0x1084,
      EVR_IS, "IntervalsRejected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1085, 0x0018, 0x1085,
      EVR_LO, "PVCRejection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1086, 0x0018, 0x1086,
      EVR_IS, "SkipBeats", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1088, 0x0018, 0x1088,
      EVR_IS, "HeartRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1090, 0x0018, 0x1090,
      EVR_IS, "CardiacNumberOfImages", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1094, 0x0018, 0x1094,
      EVR_IS, "TriggerWindow", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1100, 0x0018, 0x1100,
      EVR_DS, "ReconstructionDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1110, 0x0018, 0x1110,
      EVR_DS, "DistanceSourceToDetector", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1111, 0x0018, 0x1111,
      EVR_DS, "DistanceSourceToPatient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1114, 0x0018, 0x1114,
      EVR_DS, "EstimatedRadiographicMagnificationFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1120, 0x0018, 0x1120,
      EVR_DS, "GantryDetectorTilt", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1121, 0x0018, 0x1121,
      EVR_DS, "GantryDetectorSlew", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1130, 0x0018, 0x1130,
      EVR_DS, "TableHeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1131, 0x0018, 0x1131,
      EVR_DS, "TableTraverse", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1134, 0x0018, 0x1134,
      EVR_CS, "TableMotion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1135, 0x0018, 0x1135,
      EVR_DS, "TableVerticalIncrement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1136, 0x0018, 0x1136,
      EVR_DS, "TableLateralIncrement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1137, 0x0018, 0x1137,
      EVR_DS, "TableLongitudinalIncrement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1138, 0x0018, 0x1138,
      EVR_DS, "TableAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x113a, 0x0018, 0x113a,
      EVR_CS, "TableType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1140, 0x0018, 0x1140,
      EVR_CS, "RotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1141, 0x0018, 0x1141,
      EVR_DS, "RETIRED_AngularPosition", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1142, 0x0018, 0x1142,
      EVR_DS, "RadialPosition", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1143, 0x0018, 0x1143,
      EVR_DS, "ScanArc", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1144, 0x0018, 0x1144,
      EVR_DS, "AngularStep", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1145, 0x0018, 0x1145,
      EVR_DS, "CenterOfRotationOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1146, 0x0018, 0x1146,
      EVR_DS, "RETIRED_RotationOffset", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1147, 0x0018, 0x1147,
      EVR_CS, "FieldOfViewShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1149, 0x0018, 0x1149,
      EVR_IS, "FieldOfViewDimensions", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1150, 0x0018, 0x1150,
      EVR_IS, "ExposureTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1151, 0x0018, 0x1151,
      EVR_IS, "XRayTubeCurrent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1152, 0x0018, 0x1152,
      EVR_IS, "Exposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1153, 0x0018, 0x1153,
      EVR_IS, "ExposureInuAs", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1154, 0x0018, 0x1154,
      EVR_DS, "AveragePulseWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1155, 0x0018, 0x1155,
      EVR_CS, "RadiationSetting", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1156, 0x0018, 0x1156,
      EVR_CS, "RectificationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x115a, 0x0018, 0x115a,
      EVR_CS, "RadiationMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x115e, 0x0018, 0x115e,
      EVR_DS, "ImageAndFluoroscopyAreaDoseProduct", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1160, 0x0018, 0x1160,
      EVR_SH, "FilterType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1161, 0x0018, 0x1161,
      EVR_LO, "TypeOfFilters", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1162, 0x0018, 0x1162,
      EVR_DS, "IntensifierSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1164, 0x0018, 0x1164,
      EVR_DS, "ImagerPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1166, 0x0018, 0x1166,
      EVR_CS, "Grid", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1170, 0x0018, 0x1170,
      EVR_IS, "GeneratorPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1180, 0x0018, 0x1180,
      EVR_SH, "CollimatorGridName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1181, 0x0018, 0x1181,
      EVR_CS, "CollimatorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1182, 0x0018, 0x1182,
      EVR_IS, "FocalDistance", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1183, 0x0018, 0x1183,
      EVR_DS, "XFocusCenter", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1184, 0x0018, 0x1184,
      EVR_DS, "YFocusCenter", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1190, 0x0018, 0x1190,
      EVR_DS, "FocalSpots", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1191, 0x0018, 0x1191,
      EVR_CS, "AnodeTargetMaterial", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11a0, 0x0018, 0x11a0,
      EVR_DS, "BodyPartThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11a2, 0x0018, 0x11a2,
      EVR_DS, "CompressionForce", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11a3, 0x0018, 0x11a3,
      EVR_DS, "CompressionPressure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11a4, 0x0018, 0x11a4,
      EVR_LO, "PaddleDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11a5, 0x0018, 0x11a5,
      EVR_DS, "CompressionContactArea", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b0, 0x0018, 0x11b0,
      EVR_LO, "AcquisitionMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b1, 0x0018, 0x11b1,
      EVR_LO, "DoseModeName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b2, 0x0018, 0x11b2,
      EVR_CS, "AcquiredSubtractionMaskFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b3, 0x0018, 0x11b3,
      EVR_CS, "FluoroscopyPersistenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b4, 0x0018, 0x11b4,
      EVR_CS, "FluoroscopyLastImageHoldPersistenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b5, 0x0018, 0x11b5,
      EVR_IS, "UpperLimitNumberOfPersistentFluoroscopyFrames", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b6, 0x0018, 0x11b6,
      EVR_CS, "ContrastBolusAutoInjectionTriggerFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b7, 0x0018, 0x11b7,
      EVR_FD, "ContrastBolusInjectionDelay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b8, 0x0018, 0x11b8,
      EVR_SQ, "XAAcquisitionPhaseDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11b9, 0x0018, 0x11b9,
      EVR_FD, "XAAcquisitionFrameRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11ba, 0x0018, 0x11ba,
      EVR_SQ, "XAPlaneDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11bb, 0x0018, 0x11bb,
      EVR_LO, "AcquisitionFieldOfViewLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11bc, 0x0018, 0x11bc,
      EVR_SQ, "XRayFilterDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11bd, 0x0018, 0x11bd,
      EVR_FD, "XAAcquisitionDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11be, 0x0018, 0x11be,
      EVR_CS, "ReconstructionPipelineType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11bf, 0x0018, 0x11bf,
      EVR_SQ, "ImageFilterDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11c0, 0x0018, 0x11c0,
      EVR_CS, "AppliedMaskSubtractionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x11c1, 0x0018, 0x11c1,
      EVR_SQ, "RequestedSeriesDescriptionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1200, 0x0018, 0x1200,
      EVR_DA, "DateOfLastCalibration", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1201, 0x0018, 0x1201,
      EVR_TM, "TimeOfLastCalibration", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1202, 0x0018, 0x1202,
      EVR_DT, "DateTimeOfLastCalibration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1203, 0x0018, 0x1203,
      EVR_DT, "CalibrationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1204, 0x0018, 0x1204,
      EVR_DA, "DateOfManufacture", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1205, 0x0018, 0x1205,
      EVR_DA, "DateOfInstallation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1210, 0x0018, 0x1210,
      EVR_SH, "ConvolutionKernel", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1240, 0x0018, 0x1240,
      EVR_IS, "RETIRED_UpperLowerPixelValues", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1242, 0x0018, 0x1242,
      EVR_IS, "ActualFrameDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1243, 0x0018, 0x1243,
      EVR_IS, "CountRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1244, 0x0018, 0x1244,
      EVR_US, "PreferredPlaybackSequencing", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1250, 0x0018, 0x1250,
      EVR_SH, "ReceiveCoilName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1251, 0x0018, 0x1251,
      EVR_SH, "TransmitCoilName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1260, 0x0018, 0x1260,
      EVR_SH, "PlateType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1261, 0x0018, 0x1261,
      EVR_LO, "PhosphorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1271, 0x0018, 0x1271,
      EVR_FD, "WaterEquivalentDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1272, 0x0018, 0x1272,
      EVR_SQ, "WaterEquivalentDiameterCalculationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1300, 0x0018, 0x1300,
      EVR_DS, "ScanVelocity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1301, 0x0018, 0x1301,
      EVR_CS, "WholeBodyTechnique", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1302, 0x0018, 0x1302,
      EVR_IS, "ScanLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1310, 0x0018, 0x1310,
      EVR_US, "AcquisitionMatrix", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1312, 0x0018, 0x1312,
      EVR_CS, "InPlanePhaseEncodingDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1314, 0x0018, 0x1314,
      EVR_DS, "FlipAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1315, 0x0018, 0x1315,
      EVR_CS, "VariableFlipAngleFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1316, 0x0018, 0x1316,
      EVR_DS, "SAR", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1318, 0x0018, 0x1318,
      EVR_DS, "dBdt", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1320, 0x0018, 0x1320,
      EVR_FL, "B1rms", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1400, 0x0018, 0x1400,
      EVR_LO, "AcquisitionDeviceProcessingDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1401, 0x0018, 0x1401,
      EVR_LO, "AcquisitionDeviceProcessingCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1402, 0x0018, 0x1402,
      EVR_CS, "CassetteOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1403, 0x0018, 0x1403,
      EVR_CS, "CassetteSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1404, 0x0018, 0x1404,
      EVR_US, "ExposuresOnPlate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1405, 0x0018, 0x1405,
      EVR_IS, "RelativeXRayExposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1411, 0x0018, 0x1411,
      EVR_DS, "ExposureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1412, 0x0018, 0x1412,
      EVR_DS, "TargetExposureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1413, 0x0018, 0x1413,
      EVR_DS, "DeviationIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1450, 0x0018, 0x1450,
      EVR_DS, "ColumnAngulation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1460, 0x0018, 0x1460,
      EVR_DS, "TomoLayerHeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1470, 0x0018, 0x1470,
      EVR_DS, "TomoAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1480, 0x0018, 0x1480,
      EVR_DS, "TomoTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1490, 0x0018, 0x1490,
      EVR_CS, "TomoType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1491, 0x0018, 0x1491,
      EVR_CS, "TomoClass", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1495, 0x0018, 0x1495,
      EVR_IS, "NumberOfTomosynthesisSourceImages", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1500, 0x0018, 0x1500,
      EVR_CS, "PositionerMotion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1508, 0x0018, 0x1508,
      EVR_CS, "PositionerType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1510, 0x0018, 0x1510,
      EVR_DS, "PositionerPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1511, 0x0018, 0x1511,
      EVR_DS, "PositionerSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1520, 0x0018, 0x1520,
      EVR_DS, "PositionerPrimaryAngleIncrement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1521, 0x0018, 0x1521,
      EVR_DS, "PositionerSecondaryAngleIncrement", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1530, 0x0018, 0x1530,
      EVR_DS, "DetectorPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1531, 0x0018, 0x1531,
      EVR_DS, "DetectorSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1600, 0x0018, 0x1600,
      EVR_CS, "ShutterShape", 1, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1602, 0x0018, 0x1602,
      EVR_IS, "ShutterLeftVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1604, 0x0018, 0x1604,
      EVR_IS, "ShutterRightVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1606, 0x0018, 0x1606,
      EVR_IS, "ShutterUpperHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1608, 0x0018, 0x1608,
      EVR_IS, "ShutterLowerHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1610, 0x0018, 0x1610,
      EVR_IS, "CenterOfCircularShutter", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1612, 0x0018, 0x1612,
      EVR_IS, "RadiusOfCircularShutter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1620, 0x0018, 0x1620,
      EVR_IS, "VerticesOfThePolygonalShutter", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1622, 0x0018, 0x1622,
      EVR_US, "ShutterPresentationValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1623, 0x0018, 0x1623,
      EVR_US, "ShutterOverlayGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1624, 0x0018, 0x1624,
      EVR_US, "ShutterPresentationColorCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1630, 0x0018, 0x1630,
      EVR_CS, "OutlineShapeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1631, 0x0018, 0x1631,
      EVR_FD, "OutlineLeftVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1632, 0x0018, 0x1632,
      EVR_FD, "OutlineRightVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1633, 0x0018, 0x1633,
      EVR_FD, "OutlineUpperHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1634, 0x0018, 0x1634,
      EVR_FD, "OutlineLowerHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1635, 0x0018, 0x1635,
      EVR_FD, "CenterOfCircularOutline", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1636, 0x0018, 0x1636,
      EVR_FD, "DiameterOfCircularOutline", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1637, 0x0018, 0x1637,
      EVR_UL, "NumberOfPolygonalVertices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1638, 0x0018, 0x1638,
      EVR_OF, "VerticesOfThePolygonalOutline", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1700, 0x0018, 0x1700,
      EVR_CS, "CollimatorShape", 1, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1702, 0x0018, 0x1702,
      EVR_IS, "CollimatorLeftVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1704, 0x0018, 0x1704,
      EVR_IS, "CollimatorRightVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1706, 0x0018, 0x1706,
      EVR_IS, "CollimatorUpperHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1708, 0x0018, 0x1708,
      EVR_IS, "CollimatorLowerHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1710, 0x0018, 0x1710,
      EVR_IS, "CenterOfCircularCollimator", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1712, 0x0018, 0x1712,
      EVR_IS, "RadiusOfCircularCollimator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1720, 0x0018, 0x1720,
      EVR_IS, "VerticesOfThePolygonalCollimator", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1800, 0x0018, 0x1800,
      EVR_CS, "AcquisitionTimeSynchronized", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1801, 0x0018, 0x1801,
      EVR_SH, "TimeSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1802, 0x0018, 0x1802,
      EVR_CS, "TimeDistributionProtocol", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x1803, 0x0018, 0x1803,
      EVR_LO, "NTPSourceAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2001, 0x0018, 0x2001,
      EVR_IS, "PageNumberVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2002, 0x0018, 0x2002,
      EVR_SH, "FrameLabelVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2003, 0x0018, 0x2003,
      EVR_DS, "FramePrimaryAngleVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2004, 0x0018, 0x2004,
      EVR_DS, "FrameSecondaryAngleVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2005, 0x0018, 0x2005,
      EVR_DS, "SliceLocationVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2006, 0x0018, 0x2006,
      EVR_SH, "DisplayWindowLabelVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2010, 0x0018, 0x2010,
      EVR_DS, "NominalScannedPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2020, 0x0018, 0x2020,
      EVR_CS, "DigitizingDeviceTransportDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2030, 0x0018, 0x2030,
      EVR_DS, "RotationOfScannedFilm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2041, 0x0018, 0x2041,
      EVR_SQ, "BiopsyTargetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2042, 0x0018, 0x2042,
      EVR_UI, "TargetUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2043, 0x0018, 0x2043,
      EVR_FL, "LocalizingCursorPosition", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2044, 0x0018, 0x2044,
      EVR_FL, "CalculatedTargetPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2045, 0x0018, 0x2045,
      EVR_SH, "TargetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x2046, 0x0018, 0x2046,
      EVR_FL, "DisplayedZValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3100, 0x0018, 0x3100,
      EVR_CS, "IVUSAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3101, 0x0018, 0x3101,
      EVR_DS, "IVUSPullbackRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3102, 0x0018, 0x3102,
      EVR_DS, "IVUSGatedRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3103, 0x0018, 0x3103,
      EVR_IS, "IVUSPullbackStartFrameNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3104, 0x0018, 0x3104,
      EVR_IS, "IVUSPullbackStopFrameNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x3105, 0x0018, 0x3105,
      EVR_IS, "LesionNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x4000, 0x0018, 0x4000,
      EVR_LT, "RETIRED_AcquisitionComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5000, 0x0018, 0x5000,
      EVR_SH, "OutputPower", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5010, 0x0018, 0x5010,
      EVR_LO, "TransducerData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5011, 0x0018, 0x5011,
      EVR_SQ, "TransducerIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5012, 0x0018, 0x5012,
      EVR_DS, "FocusDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5020, 0x0018, 0x5020,
      EVR_LO, "ProcessingFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5021, 0x0018, 0x5021,
      EVR_LO, "RETIRED_PostprocessingFunction", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5022, 0x0018, 0x5022,
      EVR_DS, "MechanicalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5024, 0x0018, 0x5024,
      EVR_DS, "BoneThermalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5026, 0x0018, 0x5026,
      EVR_DS, "CranialThermalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5027, 0x0018, 0x5027,
      EVR_DS, "SoftTissueThermalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5028, 0x0018, 0x5028,
      EVR_DS, "SoftTissueFocusThermalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5029, 0x0018, 0x5029,
      EVR_DS, "SoftTissueSurfaceThermalIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5030, 0x0018, 0x5030,
      EVR_DS, "RETIRED_DynamicRange", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5040, 0x0018, 0x5040,
      EVR_DS, "RETIRED_TotalGain", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5050, 0x0018, 0x5050,
      EVR_IS, "DepthOfScanField", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5100, 0x0018, 0x5100,
      EVR_CS, "PatientPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5101, 0x0018, 0x5101,
      EVR_CS, "ViewPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5104, 0x0018, 0x5104,
      EVR_SQ, "ProjectionEponymousNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5210, 0x0018, 0x5210,
      EVR_DS, "RETIRED_ImageTransformationMatrix", 6, 6, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x5212, 0x0018, 0x5212,
      EVR_DS, "RETIRED_ImageTranslationVector", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6000, 0x0018, 0x6000,
      EVR_DS, "Sensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6011, 0x0018, 0x6011,
      EVR_SQ, "SequenceOfUltrasoundRegions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6012, 0x0018, 0x6012,
      EVR_US, "RegionSpatialFormat", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6014, 0x0018, 0x6014,
      EVR_US, "RegionDataType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6016, 0x0018, 0x6016,
      EVR_UL, "RegionFlags", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6018, 0x0018, 0x6018,
      EVR_UL, "RegionLocationMinX0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x601a, 0x0018, 0x601a,
      EVR_UL, "RegionLocationMinY0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x601c, 0x0018, 0x601c,
      EVR_UL, "RegionLocationMaxX1", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x601e, 0x0018, 0x601e,
      EVR_UL, "RegionLocationMaxY1", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6020, 0x0018, 0x6020,
      EVR_SL, "ReferencePixelX0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6022, 0x0018, 0x6022,
      EVR_SL, "ReferencePixelY0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6024, 0x0018, 0x6024,
      EVR_US, "PhysicalUnitsXDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6026, 0x0018, 0x6026,
      EVR_US, "PhysicalUnitsYDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6028, 0x0018, 0x6028,
      EVR_FD, "ReferencePixelPhysicalValueX", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x602a, 0x0018, 0x602a,
      EVR_FD, "ReferencePixelPhysicalValueY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x602c, 0x0018, 0x602c,
      EVR_FD, "PhysicalDeltaX", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x602e, 0x0018, 0x602e,
      EVR_FD, "PhysicalDeltaY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6030, 0x0018, 0x6030,
      EVR_UL, "TransducerFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6031, 0x0018, 0x6031,
      EVR_CS, "TransducerType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6032, 0x0018, 0x6032,
      EVR_UL, "PulseRepetitionFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6034, 0x0018, 0x6034,
      EVR_FD, "DopplerCorrectionAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6036, 0x0018, 0x6036,
      EVR_FD, "SteeringAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6038, 0x0018, 0x6038,
      EVR_UL, "RETIRED_DopplerSampleVolumeXPositionRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6039, 0x0018, 0x6039,
      EVR_SL, "DopplerSampleVolumeXPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603a, 0x0018, 0x603a,
      EVR_UL, "RETIRED_DopplerSampleVolumeYPositionRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603b, 0x0018, 0x603b,
      EVR_SL, "DopplerSampleVolumeYPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603c, 0x0018, 0x603c,
      EVR_UL, "RETIRED_TMLinePositionX0Retired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603d, 0x0018, 0x603d,
      EVR_SL, "TMLinePositionX0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603e, 0x0018, 0x603e,
      EVR_UL, "RETIRED_TMLinePositionY0Retired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x603f, 0x0018, 0x603f,
      EVR_SL, "TMLinePositionY0", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6040, 0x0018, 0x6040,
      EVR_UL, "RETIRED_TMLinePositionX1Retired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6041, 0x0018, 0x6041,
      EVR_SL, "TMLinePositionX1", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6042, 0x0018, 0x6042,
      EVR_UL, "RETIRED_TMLinePositionY1Retired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6043, 0x0018, 0x6043,
      EVR_SL, "TMLinePositionY1", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6044, 0x0018, 0x6044,
      EVR_US, "PixelComponentOrganization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6046, 0x0018, 0x6046,
      EVR_UL, "PixelComponentMask", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6048, 0x0018, 0x6048,
      EVR_UL, "PixelComponentRangeStart", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x604a, 0x0018, 0x604a,
      EVR_UL, "PixelComponentRangeStop", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x604c, 0x0018, 0x604c,
      EVR_US, "PixelComponentPhysicalUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x604e, 0x0018, 0x604e,
      EVR_US, "PixelComponentDataType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6050, 0x0018, 0x6050,
      EVR_UL, "NumberOfTableBreakPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6052, 0x0018, 0x6052,
      EVR_UL, "TableOfXBreakPoints", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6054, 0x0018, 0x6054,
      EVR_FD, "TableOfYBreakPoints", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6056, 0x0018, 0x6056,
      EVR_UL, "NumberOfTableEntries", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6058, 0x0018, 0x6058,
      EVR_UL, "TableOfPixelValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x605a, 0x0018, 0x605a,
      EVR_FL, "TableOfParameterValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6060, 0x0018, 0x6060,
      EVR_FL, "RWaveTimeVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x6070, 0x0018, 0x6070,
      EVR_US, "ActiveImageAreaOverlayGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7000, 0x0018, 0x7000,
      EVR_CS, "DetectorConditionsNominalFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7001, 0x0018, 0x7001,
      EVR_DS, "DetectorTemperature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7004, 0x0018, 0x7004,
      EVR_CS, "DetectorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7005, 0x0018, 0x7005,
      EVR_CS, "DetectorConfiguration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7006, 0x0018, 0x7006,
      EVR_LT, "DetectorDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7008, 0x0018, 0x7008,
      EVR_LT, "DetectorMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x700a, 0x0018, 0x700a,
      EVR_SH, "DetectorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x700c, 0x0018, 0x700c,
      EVR_DA, "DateOfLastDetectorCalibration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x700e, 0x0018, 0x700e,
      EVR_TM, "TimeOfLastDetectorCalibration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7010, 0x0018, 0x7010,
      EVR_IS, "ExposuresOnDetectorSinceLastCalibration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7011, 0x0018, 0x7011,
      EVR_IS, "ExposuresOnDetectorSinceManufactured", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7012, 0x0018, 0x7012,
      EVR_DS, "DetectorTimeSinceLastExposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7014, 0x0018, 0x7014,
      EVR_DS, "DetectorActiveTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7016, 0x0018, 0x7016,
      EVR_DS, "DetectorActivationOffsetFromExposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x701a, 0x0018, 0x701a,
      EVR_DS, "DetectorBinning", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7020, 0x0018, 0x7020,
      EVR_DS, "DetectorElementPhysicalSize", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7022, 0x0018, 0x7022,
      EVR_DS, "DetectorElementSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7024, 0x0018, 0x7024,
      EVR_CS, "DetectorActiveShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7026, 0x0018, 0x7026,
      EVR_DS, "DetectorActiveDimensions", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7028, 0x0018, 0x7028,
      EVR_DS, "DetectorActiveOrigin", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x702a, 0x0018, 0x702a,
      EVR_LO, "DetectorManufacturerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x702b, 0x0018, 0x702b,
      EVR_LO, "DetectorManufacturerModelName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7030, 0x0018, 0x7030,
      EVR_DS, "FieldOfViewOrigin", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7032, 0x0018, 0x7032,
      EVR_DS, "FieldOfViewRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7034, 0x0018, 0x7034,
      EVR_CS, "FieldOfViewHorizontalFlip", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7036, 0x0018, 0x7036,
      EVR_FL, "PixelDataAreaOriginRelativeToFOV", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7038, 0x0018, 0x7038,
      EVR_FL, "PixelDataAreaRotationAngleRelativeToFOV", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7040, 0x0018, 0x7040,
      EVR_LT, "GridAbsorbingMaterial", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7041, 0x0018, 0x7041,
      EVR_LT, "GridSpacingMaterial", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7042, 0x0018, 0x7042,
      EVR_DS, "GridThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7044, 0x0018, 0x7044,
      EVR_DS, "GridPitch", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7046, 0x0018, 0x7046,
      EVR_IS, "GridAspectRatio", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7048, 0x0018, 0x7048,
      EVR_DS, "GridPeriod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x704c, 0x0018, 0x704c,
      EVR_DS, "GridFocalDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7050, 0x0018, 0x7050,
      EVR_CS, "FilterMaterial", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7052, 0x0018, 0x7052,
      EVR_DS, "FilterThicknessMinimum", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7054, 0x0018, 0x7054,
      EVR_DS, "FilterThicknessMaximum", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7056, 0x0018, 0x7056,
      EVR_FL, "FilterBeamPathLengthMinimum", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7058, 0x0018, 0x7058,
      EVR_FL, "FilterBeamPathLengthMaximum", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7060, 0x0018, 0x7060,
      EVR_CS, "ExposureControlMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7062, 0x0018, 0x7062,
      EVR_LT, "ExposureControlModeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7064, 0x0018, 0x7064,
      EVR_CS, "ExposureStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x7065, 0x0018, 0x7065,
      EVR_DS, "PhototimerSetting", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x8150, 0x0018, 0x8150,
      EVR_DS, "ExposureTimeInuS", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x8151, 0x0018, 0x8151,
      EVR_DS, "XRayTubeCurrentInuA", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9004, 0x0018, 0x9004,
      EVR_CS, "ContentQualification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9005, 0x0018, 0x9005,
      EVR_SH, "PulseSequenceName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9006, 0x0018, 0x9006,
      EVR_SQ, "MRImagingModifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9008, 0x0018, 0x9008,
      EVR_CS, "EchoPulseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9009, 0x0018, 0x9009,
      EVR_CS, "InversionRecovery", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9010, 0x0018, 0x9010,
      EVR_CS, "FlowCompensation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9011, 0x0018, 0x9011,
      EVR_CS, "MultipleSpinEcho", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9012, 0x0018, 0x9012,
      EVR_CS, "MultiPlanarExcitation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9014, 0x0018, 0x9014,
      EVR_CS, "PhaseContrast", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9015, 0x0018, 0x9015,
      EVR_CS, "TimeOfFlightContrast", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9016, 0x0018, 0x9016,
      EVR_CS, "Spoiling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9017, 0x0018, 0x9017,
      EVR_CS, "SteadyStatePulseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9018, 0x0018, 0x9018,
      EVR_CS, "EchoPlanarPulseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9019, 0x0018, 0x9019,
      EVR_FD, "TagAngleFirstAxis", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9020, 0x0018, 0x9020,
      EVR_CS, "MagnetizationTransfer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9021, 0x0018, 0x9021,
      EVR_CS, "T2Preparation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9022, 0x0018, 0x9022,
      EVR_CS, "BloodSignalNulling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9024, 0x0018, 0x9024,
      EVR_CS, "SaturationRecovery", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9025, 0x0018, 0x9025,
      EVR_CS, "SpectrallySelectedSuppression", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9026, 0x0018, 0x9026,
      EVR_CS, "SpectrallySelectedExcitation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9027, 0x0018, 0x9027,
      EVR_CS, "SpatialPresaturation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9028, 0x0018, 0x9028,
      EVR_CS, "Tagging", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9029, 0x0018, 0x9029,
      EVR_CS, "OversamplingPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9030, 0x0018, 0x9030,
      EVR_FD, "TagSpacingFirstDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9032, 0x0018, 0x9032,
      EVR_CS, "GeometryOfKSpaceTraversal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9033, 0x0018, 0x9033,
      EVR_CS, "SegmentedKSpaceTraversal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9034, 0x0018, 0x9034,
      EVR_CS, "RectilinearPhaseEncodeReordering", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9035, 0x0018, 0x9035,
      EVR_FD, "TagThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9036, 0x0018, 0x9036,
      EVR_CS, "PartialFourierDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9037, 0x0018, 0x9037,
      EVR_CS, "CardiacSynchronizationTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9041, 0x0018, 0x9041,
      EVR_LO, "ReceiveCoilManufacturerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9042, 0x0018, 0x9042,
      EVR_SQ, "MRReceiveCoilSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9043, 0x0018, 0x9043,
      EVR_CS, "ReceiveCoilType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9044, 0x0018, 0x9044,
      EVR_CS, "QuadratureReceiveCoil", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9045, 0x0018, 0x9045,
      EVR_SQ, "MultiCoilDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9046, 0x0018, 0x9046,
      EVR_LO, "MultiCoilConfiguration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9047, 0x0018, 0x9047,
      EVR_SH, "MultiCoilElementName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9048, 0x0018, 0x9048,
      EVR_CS, "MultiCoilElementUsed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9049, 0x0018, 0x9049,
      EVR_SQ, "MRTransmitCoilSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9050, 0x0018, 0x9050,
      EVR_LO, "TransmitCoilManufacturerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9051, 0x0018, 0x9051,
      EVR_CS, "TransmitCoilType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9052, 0x0018, 0x9052,
      EVR_FD, "SpectralWidth", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9053, 0x0018, 0x9053,
      EVR_FD, "ChemicalShiftReference", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9054, 0x0018, 0x9054,
      EVR_CS, "VolumeLocalizationTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9058, 0x0018, 0x9058,
      EVR_US, "MRAcquisitionFrequencyEncodingSteps", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9059, 0x0018, 0x9059,
      EVR_CS, "Decoupling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9060, 0x0018, 0x9060,
      EVR_CS, "DecoupledNucleus", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9061, 0x0018, 0x9061,
      EVR_FD, "DecouplingFrequency", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9062, 0x0018, 0x9062,
      EVR_CS, "DecouplingMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9063, 0x0018, 0x9063,
      EVR_FD, "DecouplingChemicalShiftReference", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9064, 0x0018, 0x9064,
      EVR_CS, "KSpaceFiltering", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9065, 0x0018, 0x9065,
      EVR_CS, "TimeDomainFiltering", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9066, 0x0018, 0x9066,
      EVR_US, "NumberOfZeroFills", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9067, 0x0018, 0x9067,
      EVR_CS, "BaselineCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9069, 0x0018, 0x9069,
      EVR_FD, "ParallelReductionFactorInPlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9070, 0x0018, 0x9070,
      EVR_FD, "CardiacRRIntervalSpecified", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9073, 0x0018, 0x9073,
      EVR_FD, "AcquisitionDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9074, 0x0018, 0x9074,
      EVR_DT, "FrameAcquisitionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9075, 0x0018, 0x9075,
      EVR_CS, "DiffusionDirectionality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9076, 0x0018, 0x9076,
      EVR_SQ, "DiffusionGradientDirectionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9077, 0x0018, 0x9077,
      EVR_CS, "ParallelAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9078, 0x0018, 0x9078,
      EVR_CS, "ParallelAcquisitionTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9079, 0x0018, 0x9079,
      EVR_FD, "InversionTimes", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9080, 0x0018, 0x9080,
      EVR_ST, "MetaboliteMapDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9081, 0x0018, 0x9081,
      EVR_CS, "PartialFourier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9082, 0x0018, 0x9082,
      EVR_FD, "EffectiveEchoTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9083, 0x0018, 0x9083,
      EVR_SQ, "MetaboliteMapCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9084, 0x0018, 0x9084,
      EVR_SQ, "ChemicalShiftSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9085, 0x0018, 0x9085,
      EVR_CS, "CardiacSignalSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9087, 0x0018, 0x9087,
      EVR_FD, "DiffusionBValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9089, 0x0018, 0x9089,
      EVR_FD, "DiffusionGradientOrientation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9090, 0x0018, 0x9090,
      EVR_FD, "VelocityEncodingDirection", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9091, 0x0018, 0x9091,
      EVR_FD, "VelocityEncodingMinimumValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9092, 0x0018, 0x9092,
      EVR_SQ, "VelocityEncodingAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9093, 0x0018, 0x9093,
      EVR_US, "NumberOfKSpaceTrajectories", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9094, 0x0018, 0x9094,
      EVR_CS, "CoverageOfKSpace", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9095, 0x0018, 0x9095,
      EVR_UL, "SpectroscopyAcquisitionPhaseRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9096, 0x0018, 0x9096,
      EVR_FD, "RETIRED_ParallelReductionFactorInPlaneRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9098, 0x0018, 0x9098,
      EVR_FD, "TransmitterFrequency", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9100, 0x0018, 0x9100,
      EVR_CS, "ResonantNucleus", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9101, 0x0018, 0x9101,
      EVR_CS, "FrequencyCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9103, 0x0018, 0x9103,
      EVR_SQ, "MRSpectroscopyFOVGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9104, 0x0018, 0x9104,
      EVR_FD, "SlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9105, 0x0018, 0x9105,
      EVR_FD, "SlabOrientation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9106, 0x0018, 0x9106,
      EVR_FD, "MidSlabPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9107, 0x0018, 0x9107,
      EVR_SQ, "MRSpatialSaturationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9112, 0x0018, 0x9112,
      EVR_SQ, "MRTimingAndRelatedParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9114, 0x0018, 0x9114,
      EVR_SQ, "MREchoSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9115, 0x0018, 0x9115,
      EVR_SQ, "MRModifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9117, 0x0018, 0x9117,
      EVR_SQ, "MRDiffusionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9118, 0x0018, 0x9118,
      EVR_SQ, "CardiacSynchronizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9119, 0x0018, 0x9119,
      EVR_SQ, "MRAveragesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9125, 0x0018, 0x9125,
      EVR_SQ, "MRFOVGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9126, 0x0018, 0x9126,
      EVR_SQ, "VolumeLocalizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9127, 0x0018, 0x9127,
      EVR_UL, "SpectroscopyAcquisitionDataColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9147, 0x0018, 0x9147,
      EVR_CS, "DiffusionAnisotropyType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9151, 0x0018, 0x9151,
      EVR_DT, "FrameReferenceDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9152, 0x0018, 0x9152,
      EVR_SQ, "MRMetaboliteMapSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9155, 0x0018, 0x9155,
      EVR_FD, "ParallelReductionFactorOutOfPlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9159, 0x0018, 0x9159,
      EVR_UL, "SpectroscopyAcquisitionOutOfPlanePhaseSteps", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9166, 0x0018, 0x9166,
      EVR_CS, "RETIRED_BulkMotionStatus", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9168, 0x0018, 0x9168,
      EVR_FD, "ParallelReductionFactorSecondInPlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9169, 0x0018, 0x9169,
      EVR_CS, "CardiacBeatRejectionTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9170, 0x0018, 0x9170,
      EVR_CS, "RespiratoryMotionCompensationTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9171, 0x0018, 0x9171,
      EVR_CS, "RespiratorySignalSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9172, 0x0018, 0x9172,
      EVR_CS, "BulkMotionCompensationTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9173, 0x0018, 0x9173,
      EVR_CS, "BulkMotionSignalSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9174, 0x0018, 0x9174,
      EVR_CS, "ApplicableSafetyStandardAgency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9175, 0x0018, 0x9175,
      EVR_LO, "ApplicableSafetyStandardDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9176, 0x0018, 0x9176,
      EVR_SQ, "OperatingModeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9177, 0x0018, 0x9177,
      EVR_CS, "OperatingModeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9178, 0x0018, 0x9178,
      EVR_CS, "OperatingMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9179, 0x0018, 0x9179,
      EVR_CS, "SpecificAbsorptionRateDefinition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9180, 0x0018, 0x9180,
      EVR_CS, "GradientOutputType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9181, 0x0018, 0x9181,
      EVR_FD, "SpecificAbsorptionRateValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9182, 0x0018, 0x9182,
      EVR_FD, "GradientOutput", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9183, 0x0018, 0x9183,
      EVR_CS, "FlowCompensationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9184, 0x0018, 0x9184,
      EVR_FD, "TaggingDelay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9185, 0x0018, 0x9185,
      EVR_ST, "RespiratoryMotionCompensationTechniqueDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9186, 0x0018, 0x9186,
      EVR_SH, "RespiratorySignalSourceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9195, 0x0018, 0x9195,
      EVR_FD, "RETIRED_ChemicalShiftMinimumIntegrationLimitInHz", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9196, 0x0018, 0x9196,
      EVR_FD, "RETIRED_ChemicalShiftMaximumIntegrationLimitInHz", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9197, 0x0018, 0x9197,
      EVR_SQ, "MRVelocityEncodingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9198, 0x0018, 0x9198,
      EVR_CS, "FirstOrderPhaseCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9199, 0x0018, 0x9199,
      EVR_CS, "WaterReferencedPhaseCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9200, 0x0018, 0x9200,
      EVR_CS, "MRSpectroscopyAcquisitionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9214, 0x0018, 0x9214,
      EVR_CS, "RespiratoryCyclePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9217, 0x0018, 0x9217,
      EVR_FD, "VelocityEncodingMaximumValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9218, 0x0018, 0x9218,
      EVR_FD, "TagSpacingSecondDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9219, 0x0018, 0x9219,
      EVR_SS, "TagAngleSecondAxis", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9220, 0x0018, 0x9220,
      EVR_FD, "FrameAcquisitionDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9226, 0x0018, 0x9226,
      EVR_SQ, "MRImageFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9227, 0x0018, 0x9227,
      EVR_SQ, "MRSpectroscopyFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9231, 0x0018, 0x9231,
      EVR_US, "MRAcquisitionPhaseEncodingStepsInPlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9232, 0x0018, 0x9232,
      EVR_US, "MRAcquisitionPhaseEncodingStepsOutOfPlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9234, 0x0018, 0x9234,
      EVR_UL, "SpectroscopyAcquisitionPhaseColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9236, 0x0018, 0x9236,
      EVR_CS, "CardiacCyclePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9239, 0x0018, 0x9239,
      EVR_SQ, "SpecificAbsorptionRateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9240, 0x0018, 0x9240,
      EVR_US, "RFEchoTrainLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9241, 0x0018, 0x9241,
      EVR_US, "GradientEchoTrainLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9250, 0x0018, 0x9250,
      EVR_CS, "ArterialSpinLabelingContrast", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9251, 0x0018, 0x9251,
      EVR_SQ, "MRArterialSpinLabelingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9252, 0x0018, 0x9252,
      EVR_LO, "ASLTechniqueDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9253, 0x0018, 0x9253,
      EVR_US, "ASLSlabNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9254, 0x0018, 0x9254,
      EVR_FD, "ASLSlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9255, 0x0018, 0x9255,
      EVR_FD, "ASLSlabOrientation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9256, 0x0018, 0x9256,
      EVR_FD, "ASLMidSlabPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9257, 0x0018, 0x9257,
      EVR_CS, "ASLContext", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9258, 0x0018, 0x9258,
      EVR_UL, "ASLPulseTrainDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9259, 0x0018, 0x9259,
      EVR_CS, "ASLCrusherFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925a, 0x0018, 0x925a,
      EVR_FD, "ASLCrusherFlowLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925b, 0x0018, 0x925b,
      EVR_LO, "ASLCrusherDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925c, 0x0018, 0x925c,
      EVR_CS, "ASLBolusCutoffFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925d, 0x0018, 0x925d,
      EVR_SQ, "ASLBolusCutoffTimingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925e, 0x0018, 0x925e,
      EVR_LO, "ASLBolusCutoffTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x925f, 0x0018, 0x925f,
      EVR_UL, "ASLBolusCutoffDelayTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9260, 0x0018, 0x9260,
      EVR_SQ, "ASLSlabSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9295, 0x0018, 0x9295,
      EVR_FD, "ChemicalShiftMinimumIntegrationLimitInppm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9296, 0x0018, 0x9296,
      EVR_FD, "ChemicalShiftMaximumIntegrationLimitInppm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9297, 0x0018, 0x9297,
      EVR_CS, "WaterReferenceAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9298, 0x0018, 0x9298,
      EVR_IS, "EchoPeakPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9301, 0x0018, 0x9301,
      EVR_SQ, "CTAcquisitionTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9302, 0x0018, 0x9302,
      EVR_CS, "AcquisitionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9303, 0x0018, 0x9303,
      EVR_FD, "TubeAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9304, 0x0018, 0x9304,
      EVR_SQ, "CTAcquisitionDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9305, 0x0018, 0x9305,
      EVR_FD, "RevolutionTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9306, 0x0018, 0x9306,
      EVR_FD, "SingleCollimationWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9307, 0x0018, 0x9307,
      EVR_FD, "TotalCollimationWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9308, 0x0018, 0x9308,
      EVR_SQ, "CTTableDynamicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9309, 0x0018, 0x9309,
      EVR_FD, "TableSpeed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9310, 0x0018, 0x9310,
      EVR_FD, "TableFeedPerRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9311, 0x0018, 0x9311,
      EVR_FD, "SpiralPitchFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9312, 0x0018, 0x9312,
      EVR_SQ, "CTGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9313, 0x0018, 0x9313,
      EVR_FD, "DataCollectionCenterPatient", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9314, 0x0018, 0x9314,
      EVR_SQ, "CTReconstructionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9315, 0x0018, 0x9315,
      EVR_CS, "ReconstructionAlgorithm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9316, 0x0018, 0x9316,
      EVR_CS, "ConvolutionKernelGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9317, 0x0018, 0x9317,
      EVR_FD, "ReconstructionFieldOfView", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9318, 0x0018, 0x9318,
      EVR_FD, "ReconstructionTargetCenterPatient", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9319, 0x0018, 0x9319,
      EVR_FD, "ReconstructionAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9320, 0x0018, 0x9320,
      EVR_SH, "ImageFilter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9321, 0x0018, 0x9321,
      EVR_SQ, "CTExposureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9322, 0x0018, 0x9322,
      EVR_FD, "ReconstructionPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9323, 0x0018, 0x9323,
      EVR_CS, "ExposureModulationType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9324, 0x0018, 0x9324,
      EVR_FD, "RETIRED_EstimatedDoseSaving", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9325, 0x0018, 0x9325,
      EVR_SQ, "CTXRayDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9326, 0x0018, 0x9326,
      EVR_SQ, "CTPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9327, 0x0018, 0x9327,
      EVR_FD, "TablePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9328, 0x0018, 0x9328,
      EVR_FD, "ExposureTimeInms", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9329, 0x0018, 0x9329,
      EVR_SQ, "CTImageFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9330, 0x0018, 0x9330,
      EVR_FD, "XRayTubeCurrentInmA", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9332, 0x0018, 0x9332,
      EVR_FD, "ExposureInmAs", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9333, 0x0018, 0x9333,
      EVR_CS, "ConstantVolumeFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9334, 0x0018, 0x9334,
      EVR_CS, "FluoroscopyFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9335, 0x0018, 0x9335,
      EVR_FD, "DistanceSourceToDataCollectionCenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9337, 0x0018, 0x9337,
      EVR_US, "ContrastBolusAgentNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9338, 0x0018, 0x9338,
      EVR_SQ, "ContrastBolusIngredientCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9340, 0x0018, 0x9340,
      EVR_SQ, "ContrastAdministrationProfileSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9341, 0x0018, 0x9341,
      EVR_SQ, "ContrastBolusUsageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9342, 0x0018, 0x9342,
      EVR_CS, "ContrastBolusAgentAdministered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9343, 0x0018, 0x9343,
      EVR_CS, "ContrastBolusAgentDetected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9344, 0x0018, 0x9344,
      EVR_CS, "ContrastBolusAgentPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9345, 0x0018, 0x9345,
      EVR_FD, "CTDIvol", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9346, 0x0018, 0x9346,
      EVR_SQ, "CTDIPhantomTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9351, 0x0018, 0x9351,
      EVR_FL, "CalciumScoringMassFactorPatient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9352, 0x0018, 0x9352,
      EVR_FL, "CalciumScoringMassFactorDevice", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9353, 0x0018, 0x9353,
      EVR_FL, "EnergyWeightingFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9360, 0x0018, 0x9360,
      EVR_SQ, "CTAdditionalXRaySourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9361, 0x0018, 0x9361,
      EVR_CS, "MultienergyCTAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9362, 0x0018, 0x9362,
      EVR_SQ, "MultienergyCTAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9363, 0x0018, 0x9363,
      EVR_SQ, "MultienergyCTProcessingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9364, 0x0018, 0x9364,
      EVR_SQ, "MultienergyCTCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9365, 0x0018, 0x9365,
      EVR_SQ, "MultienergyCTXRaySourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9366, 0x0018, 0x9366,
      EVR_US, "XRaySourceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9367, 0x0018, 0x9367,
      EVR_UC, "XRaySourceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9368, 0x0018, 0x9368,
      EVR_CS, "MultienergySourceTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9369, 0x0018, 0x9369,
      EVR_DT, "SourceStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936a, 0x0018, 0x936a,
      EVR_DT, "SourceEndDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936b, 0x0018, 0x936b,
      EVR_US, "SwitchingPhaseNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936c, 0x0018, 0x936c,
      EVR_DS, "SwitchingPhaseNominalDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936d, 0x0018, 0x936d,
      EVR_DS, "SwitchingPhaseTransitionDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936e, 0x0018, 0x936e,
      EVR_DS, "EffectiveBinEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x936f, 0x0018, 0x936f,
      EVR_SQ, "MultienergyCTXRayDetectorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9370, 0x0018, 0x9370,
      EVR_US, "XRayDetectorIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9371, 0x0018, 0x9371,
      EVR_UC, "XRayDetectorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9372, 0x0018, 0x9372,
      EVR_CS, "MultienergyDetectorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9373, 0x0018, 0x9373,
      EVR_ST, "XRayDetectorLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9374, 0x0018, 0x9374,
      EVR_DS, "NominalMaxEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9375, 0x0018, 0x9375,
      EVR_DS, "NominalMinEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9376, 0x0018, 0x9376,
      EVR_US, "ReferencedXRayDetectorIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9377, 0x0018, 0x9377,
      EVR_US, "ReferencedXRaySourceIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9378, 0x0018, 0x9378,
      EVR_US, "ReferencedPathIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9379, 0x0018, 0x9379,
      EVR_SQ, "MultienergyCTPathSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937a, 0x0018, 0x937a,
      EVR_US, "MultienergyCTPathIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937b, 0x0018, 0x937b,
      EVR_UT, "MultienergyAcquisitionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937c, 0x0018, 0x937c,
      EVR_FD, "MonoenergeticEnergyEquivalent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937d, 0x0018, 0x937d,
      EVR_SQ, "MaterialCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937e, 0x0018, 0x937e,
      EVR_CS, "DecompositionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x937f, 0x0018, 0x937f,
      EVR_UT, "DecompositionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9380, 0x0018, 0x9380,
      EVR_SQ, "DecompositionAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9381, 0x0018, 0x9381,
      EVR_SQ, "DecompositionMaterialSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9382, 0x0018, 0x9382,
      EVR_SQ, "MaterialAttenuationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9383, 0x0018, 0x9383,
      EVR_DS, "PhotonEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9384, 0x0018, 0x9384,
      EVR_DS, "XRayMassAttenuationCoefficient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9401, 0x0018, 0x9401,
      EVR_SQ, "ProjectionPixelCalibrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9402, 0x0018, 0x9402,
      EVR_FL, "DistanceSourceToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9403, 0x0018, 0x9403,
      EVR_FL, "DistanceObjectToTableTop", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9404, 0x0018, 0x9404,
      EVR_FL, "ObjectPixelSpacingInCenterOfBeam", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9405, 0x0018, 0x9405,
      EVR_SQ, "PositionerPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9406, 0x0018, 0x9406,
      EVR_SQ, "TablePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9407, 0x0018, 0x9407,
      EVR_SQ, "CollimatorShapeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9410, 0x0018, 0x9410,
      EVR_CS, "PlanesInAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9412, 0x0018, 0x9412,
      EVR_SQ, "XAXRFFrameCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9417, 0x0018, 0x9417,
      EVR_SQ, "FrameAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9420, 0x0018, 0x9420,
      EVR_CS, "XRayReceptorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9423, 0x0018, 0x9423,
      EVR_LO, "AcquisitionProtocolName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9424, 0x0018, 0x9424,
      EVR_LT, "AcquisitionProtocolDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9425, 0x0018, 0x9425,
      EVR_CS, "ContrastBolusIngredientOpaque", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9426, 0x0018, 0x9426,
      EVR_FL, "DistanceReceptorPlaneToDetectorHousing", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9427, 0x0018, 0x9427,
      EVR_CS, "IntensifierActiveShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9428, 0x0018, 0x9428,
      EVR_FL, "IntensifierActiveDimensions", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9429, 0x0018, 0x9429,
      EVR_FL, "PhysicalDetectorSize", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9430, 0x0018, 0x9430,
      EVR_FL, "PositionOfIsocenterProjection", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9432, 0x0018, 0x9432,
      EVR_SQ, "FieldOfViewSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9433, 0x0018, 0x9433,
      EVR_LO, "FieldOfViewDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9434, 0x0018, 0x9434,
      EVR_SQ, "ExposureControlSensingRegionsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9435, 0x0018, 0x9435,
      EVR_CS, "ExposureControlSensingRegionShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9436, 0x0018, 0x9436,
      EVR_SS, "ExposureControlSensingRegionLeftVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9437, 0x0018, 0x9437,
      EVR_SS, "ExposureControlSensingRegionRightVerticalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9438, 0x0018, 0x9438,
      EVR_SS, "ExposureControlSensingRegionUpperHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9439, 0x0018, 0x9439,
      EVR_SS, "ExposureControlSensingRegionLowerHorizontalEdge", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9440, 0x0018, 0x9440,
      EVR_SS, "CenterOfCircularExposureControlSensingRegion", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9441, 0x0018, 0x9441,
      EVR_US, "RadiusOfCircularExposureControlSensingRegion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9442, 0x0018, 0x9442,
      EVR_SS, "VerticesOfThePolygonalExposureControlSensingRegion", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9447, 0x0018, 0x9447,
      EVR_FL, "ColumnAngulationPatient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9449, 0x0018, 0x9449,
      EVR_FL, "BeamAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9451, 0x0018, 0x9451,
      EVR_SQ, "FrameDetectorParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9452, 0x0018, 0x9452,
      EVR_FL, "CalculatedAnatomyThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9455, 0x0018, 0x9455,
      EVR_SQ, "CalibrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9456, 0x0018, 0x9456,
      EVR_SQ, "ObjectThicknessSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9457, 0x0018, 0x9457,
      EVR_CS, "PlaneIdentification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9461, 0x0018, 0x9461,
      EVR_FL, "FieldOfViewDimensionsInFloat", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9462, 0x0018, 0x9462,
      EVR_SQ, "IsocenterReferenceSystemSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9463, 0x0018, 0x9463,
      EVR_FL, "PositionerIsocenterPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9464, 0x0018, 0x9464,
      EVR_FL, "PositionerIsocenterSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9465, 0x0018, 0x9465,
      EVR_FL, "PositionerIsocenterDetectorRotationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9466, 0x0018, 0x9466,
      EVR_FL, "TableXPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9467, 0x0018, 0x9467,
      EVR_FL, "TableYPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9468, 0x0018, 0x9468,
      EVR_FL, "TableZPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9469, 0x0018, 0x9469,
      EVR_FL, "TableHorizontalRotationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9470, 0x0018, 0x9470,
      EVR_FL, "TableHeadTiltAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9471, 0x0018, 0x9471,
      EVR_FL, "TableCradleTiltAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9472, 0x0018, 0x9472,
      EVR_SQ, "FrameDisplayShutterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9473, 0x0018, 0x9473,
      EVR_FL, "AcquiredImageAreaDoseProduct", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9474, 0x0018, 0x9474,
      EVR_CS, "CArmPositionerTabletopRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9476, 0x0018, 0x9476,
      EVR_SQ, "XRayGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9477, 0x0018, 0x9477,
      EVR_SQ, "IrradiationEventIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9504, 0x0018, 0x9504,
      EVR_SQ, "XRay3DFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9506, 0x0018, 0x9506,
      EVR_SQ, "ContributingSourcesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9507, 0x0018, 0x9507,
      EVR_SQ, "XRay3DAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9508, 0x0018, 0x9508,
      EVR_FL, "PrimaryPositionerScanArc", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9509, 0x0018, 0x9509,
      EVR_FL, "SecondaryPositionerScanArc", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9510, 0x0018, 0x9510,
      EVR_FL, "PrimaryPositionerScanStartAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9511, 0x0018, 0x9511,
      EVR_FL, "SecondaryPositionerScanStartAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9514, 0x0018, 0x9514,
      EVR_FL, "PrimaryPositionerIncrement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9515, 0x0018, 0x9515,
      EVR_FL, "SecondaryPositionerIncrement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9516, 0x0018, 0x9516,
      EVR_DT, "StartAcquisitionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9517, 0x0018, 0x9517,
      EVR_DT, "EndAcquisitionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9518, 0x0018, 0x9518,
      EVR_SS, "PrimaryPositionerIncrementSign", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9519, 0x0018, 0x9519,
      EVR_SS, "SecondaryPositionerIncrementSign", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9524, 0x0018, 0x9524,
      EVR_LO, "ApplicationName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9525, 0x0018, 0x9525,
      EVR_LO, "ApplicationVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9526, 0x0018, 0x9526,
      EVR_LO, "ApplicationManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9527, 0x0018, 0x9527,
      EVR_CS, "AlgorithmType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9528, 0x0018, 0x9528,
      EVR_LO, "AlgorithmDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9530, 0x0018, 0x9530,
      EVR_SQ, "XRay3DReconstructionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9531, 0x0018, 0x9531,
      EVR_LO, "ReconstructionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9538, 0x0018, 0x9538,
      EVR_SQ, "PerProjectionAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9541, 0x0018, 0x9541,
      EVR_SQ, "DetectorPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9542, 0x0018, 0x9542,
      EVR_SQ, "XRayAcquisitionDoseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9543, 0x0018, 0x9543,
      EVR_FD, "XRaySourceIsocenterPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9544, 0x0018, 0x9544,
      EVR_FD, "XRaySourceIsocenterSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9545, 0x0018, 0x9545,
      EVR_FD, "BreastSupportIsocenterPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9546, 0x0018, 0x9546,
      EVR_FD, "BreastSupportIsocenterSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9547, 0x0018, 0x9547,
      EVR_FD, "BreastSupportXPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9548, 0x0018, 0x9548,
      EVR_FD, "BreastSupportYPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9549, 0x0018, 0x9549,
      EVR_FD, "BreastSupportZPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9550, 0x0018, 0x9550,
      EVR_FD, "DetectorIsocenterPrimaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9551, 0x0018, 0x9551,
      EVR_FD, "DetectorIsocenterSecondaryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9552, 0x0018, 0x9552,
      EVR_FD, "DetectorXPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9553, 0x0018, 0x9553,
      EVR_FD, "DetectorYPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9554, 0x0018, 0x9554,
      EVR_FD, "DetectorZPositionToIsocenter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9555, 0x0018, 0x9555,
      EVR_SQ, "XRayGridSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9556, 0x0018, 0x9556,
      EVR_SQ, "XRayFilterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9557, 0x0018, 0x9557,
      EVR_FD, "DetectorActiveAreaTLHCPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9558, 0x0018, 0x9558,
      EVR_FD, "DetectorActiveAreaOrientation", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9559, 0x0018, 0x9559,
      EVR_CS, "PositionerPrimaryAngleDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9601, 0x0018, 0x9601,
      EVR_SQ, "DiffusionBMatrixSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9602, 0x0018, 0x9602,
      EVR_FD, "DiffusionBValueXX", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9603, 0x0018, 0x9603,
      EVR_FD, "DiffusionBValueXY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9604, 0x0018, 0x9604,
      EVR_FD, "DiffusionBValueXZ", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9605, 0x0018, 0x9605,
      EVR_FD, "DiffusionBValueYY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9606, 0x0018, 0x9606,
      EVR_FD, "DiffusionBValueYZ", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9607, 0x0018, 0x9607,
      EVR_FD, "DiffusionBValueZZ", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9621, 0x0018, 0x9621,
      EVR_SQ, "FunctionalMRSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9622, 0x0018, 0x9622,
      EVR_CS, "FunctionalSettlingPhaseFramesPresent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9623, 0x0018, 0x9623,
      EVR_DT, "FunctionalSyncPulse", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9624, 0x0018, 0x9624,
      EVR_CS, "SettlingPhaseFrame", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9701, 0x0018, 0x9701,
      EVR_DT, "DecayCorrectionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9715, 0x0018, 0x9715,
      EVR_FD, "StartDensityThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9716, 0x0018, 0x9716,
      EVR_FD, "StartRelativeDensityDifferenceThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9717, 0x0018, 0x9717,
      EVR_FD, "StartCardiacTriggerCountThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9718, 0x0018, 0x9718,
      EVR_FD, "StartRespiratoryTriggerCountThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9719, 0x0018, 0x9719,
      EVR_FD, "TerminationCountsThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9720, 0x0018, 0x9720,
      EVR_FD, "TerminationDensityThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9721, 0x0018, 0x9721,
      EVR_FD, "TerminationRelativeDensityThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9722, 0x0018, 0x9722,
      EVR_FD, "TerminationTimeThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9723, 0x0018, 0x9723,
      EVR_FD, "TerminationCardiacTriggerCountThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9724, 0x0018, 0x9724,
      EVR_FD, "TerminationRespiratoryTriggerCountThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9725, 0x0018, 0x9725,
      EVR_CS, "DetectorGeometry", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9726, 0x0018, 0x9726,
      EVR_FD, "TransverseDetectorSeparation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9727, 0x0018, 0x9727,
      EVR_FD, "AxialDetectorDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9729, 0x0018, 0x9729,
      EVR_US, "RadiopharmaceuticalAgentNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9732, 0x0018, 0x9732,
      EVR_SQ, "PETFrameAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9733, 0x0018, 0x9733,
      EVR_SQ, "PETDetectorMotionDetailsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9734, 0x0018, 0x9734,
      EVR_SQ, "PETTableDynamicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9735, 0x0018, 0x9735,
      EVR_SQ, "PETPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9736, 0x0018, 0x9736,
      EVR_SQ, "PETFrameCorrectionFactorsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9737, 0x0018, 0x9737,
      EVR_SQ, "RadiopharmaceuticalUsageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9738, 0x0018, 0x9738,
      EVR_CS, "AttenuationCorrectionSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9739, 0x0018, 0x9739,
      EVR_US, "NumberOfIterations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9740, 0x0018, 0x9740,
      EVR_US, "NumberOfSubsets", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9749, 0x0018, 0x9749,
      EVR_SQ, "PETReconstructionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9751, 0x0018, 0x9751,
      EVR_SQ, "PETFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9755, 0x0018, 0x9755,
      EVR_CS, "TimeOfFlightInformationUsed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9756, 0x0018, 0x9756,
      EVR_CS, "ReconstructionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9758, 0x0018, 0x9758,
      EVR_CS, "DecayCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9759, 0x0018, 0x9759,
      EVR_CS, "AttenuationCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9760, 0x0018, 0x9760,
      EVR_CS, "ScatterCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9761, 0x0018, 0x9761,
      EVR_CS, "DeadTimeCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9762, 0x0018, 0x9762,
      EVR_CS, "GantryMotionCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9763, 0x0018, 0x9763,
      EVR_CS, "PatientMotionCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9764, 0x0018, 0x9764,
      EVR_CS, "CountLossNormalizationCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9765, 0x0018, 0x9765,
      EVR_CS, "RandomsCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9766, 0x0018, 0x9766,
      EVR_CS, "NonUniformRadialSamplingCorrected", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9767, 0x0018, 0x9767,
      EVR_CS, "SensitivityCalibrated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9768, 0x0018, 0x9768,
      EVR_CS, "DetectorNormalizationCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9769, 0x0018, 0x9769,
      EVR_CS, "IterativeReconstructionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9770, 0x0018, 0x9770,
      EVR_CS, "AttenuationCorrectionTemporalRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9771, 0x0018, 0x9771,
      EVR_SQ, "PatientPhysiologicalStateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9772, 0x0018, 0x9772,
      EVR_SQ, "PatientPhysiologicalStateCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9801, 0x0018, 0x9801,
      EVR_FD, "DepthsOfFocus", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9803, 0x0018, 0x9803,
      EVR_SQ, "ExcludedIntervalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9804, 0x0018, 0x9804,
      EVR_DT, "ExclusionStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9805, 0x0018, 0x9805,
      EVR_FD, "ExclusionDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9806, 0x0018, 0x9806,
      EVR_SQ, "USImageDescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9807, 0x0018, 0x9807,
      EVR_SQ, "ImageDataTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9808, 0x0018, 0x9808,
      EVR_CS, "DataType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9809, 0x0018, 0x9809,
      EVR_SQ, "TransducerScanPatternCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x980b, 0x0018, 0x980b,
      EVR_CS, "AliasedDataType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x980c, 0x0018, 0x980c,
      EVR_CS, "PositionMeasuringDeviceUsed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x980d, 0x0018, 0x980d,
      EVR_SQ, "TransducerGeometryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x980e, 0x0018, 0x980e,
      EVR_SQ, "TransducerBeamSteeringCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x980f, 0x0018, 0x980f,
      EVR_SQ, "TransducerApplicationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9810, 0x0018, 0x9810,
      EVR_xs, "ZeroVelocityPixelValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9821, 0x0018, 0x9821,
      EVR_SQ, "PhotoacousticExcitationCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9822, 0x0018, 0x9822,
      EVR_FD, "ExcitationSpectralWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9823, 0x0018, 0x9823,
      EVR_FD, "ExcitationEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9824, 0x0018, 0x9824,
      EVR_FD, "ExcitationPulseDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9825, 0x0018, 0x9825,
      EVR_SQ, "ExcitationWavelengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9826, 0x0018, 0x9826,
      EVR_FD, "ExcitationWavelength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9828, 0x0018, 0x9828,
      EVR_CS, "IlluminationTranslationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9829, 0x0018, 0x9829,
      EVR_CS, "AcousticCouplingMediumFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982a, 0x0018, 0x982a,
      EVR_SQ, "AcousticCouplingMediumCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982b, 0x0018, 0x982b,
      EVR_FD, "AcousticCouplingMediumTemperature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982c, 0x0018, 0x982c,
      EVR_SQ, "TransducerResponseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982d, 0x0018, 0x982d,
      EVR_FD, "CenterFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982e, 0x0018, 0x982e,
      EVR_FD, "FractionalBandwidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x982f, 0x0018, 0x982f,
      EVR_FD, "LowerCutoffFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9830, 0x0018, 0x9830,
      EVR_FD, "UpperCutoffFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9831, 0x0018, 0x9831,
      EVR_SQ, "TransducerTechnologySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9832, 0x0018, 0x9832,
      EVR_SQ, "SoundSpeedCorrectionMechanismCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9833, 0x0018, 0x9833,
      EVR_FD, "ObjectSoundSpeed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9834, 0x0018, 0x9834,
      EVR_FD, "AcousticCouplingMediumSoundSpeed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9835, 0x0018, 0x9835,
      EVR_SQ, "PhotoacousticImageFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9836, 0x0018, 0x9836,
      EVR_SQ, "ImageDataTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9900, 0x0018, 0x9900,
      EVR_LO, "ReferenceLocationLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9901, 0x0018, 0x9901,
      EVR_UT, "ReferenceLocationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9902, 0x0018, 0x9902,
      EVR_SQ, "ReferenceBasisCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9903, 0x0018, 0x9903,
      EVR_SQ, "ReferenceGeometryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9904, 0x0018, 0x9904,
      EVR_DS, "OffsetDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9905, 0x0018, 0x9905,
      EVR_CS, "OffsetDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9906, 0x0018, 0x9906,
      EVR_SQ, "PotentialScheduledProtocolCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9907, 0x0018, 0x9907,
      EVR_SQ, "PotentialRequestedProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9908, 0x0018, 0x9908,
      EVR_UC, "PotentialReasonsForProcedure", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9909, 0x0018, 0x9909,
      EVR_SQ, "PotentialReasonsForProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990a, 0x0018, 0x990a,
      EVR_UC, "PotentialDiagnosticTasks", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990b, 0x0018, 0x990b,
      EVR_SQ, "ContraindicationsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990c, 0x0018, 0x990c,
      EVR_SQ, "ReferencedDefinedProtocolSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990d, 0x0018, 0x990d,
      EVR_SQ, "ReferencedPerformedProtocolSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990e, 0x0018, 0x990e,
      EVR_SQ, "PredecessorProtocolSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x990f, 0x0018, 0x990f,
      EVR_UT, "ProtocolPlanningInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9910, 0x0018, 0x9910,
      EVR_UT, "ProtocolDesignRationale", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9911, 0x0018, 0x9911,
      EVR_SQ, "PatientSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9912, 0x0018, 0x9912,
      EVR_SQ, "ModelSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9913, 0x0018, 0x9913,
      EVR_SQ, "ParametersSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9914, 0x0018, 0x9914,
      EVR_SQ, "InstructionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9915, 0x0018, 0x9915,
      EVR_US, "InstructionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9916, 0x0018, 0x9916,
      EVR_LO, "InstructionText", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9917, 0x0018, 0x9917,
      EVR_UT, "InstructionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9918, 0x0018, 0x9918,
      EVR_CS, "InstructionPerformedFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9919, 0x0018, 0x9919,
      EVR_DT, "InstructionPerformedDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991a, 0x0018, 0x991a,
      EVR_UT, "InstructionPerformanceComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991b, 0x0018, 0x991b,
      EVR_SQ, "PatientPositioningInstructionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991c, 0x0018, 0x991c,
      EVR_SQ, "PositioningMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991d, 0x0018, 0x991d,
      EVR_SQ, "PositioningLandmarkSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991e, 0x0018, 0x991e,
      EVR_UI, "TargetFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x991f, 0x0018, 0x991f,
      EVR_SQ, "AcquisitionProtocolElementSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9920, 0x0018, 0x9920,
      EVR_SQ, "AcquisitionProtocolElementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9921, 0x0018, 0x9921,
      EVR_US, "ProtocolElementNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9922, 0x0018, 0x9922,
      EVR_LO, "ProtocolElementName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9923, 0x0018, 0x9923,
      EVR_UT, "ProtocolElementCharacteristicsSummary", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9924, 0x0018, 0x9924,
      EVR_UT, "ProtocolElementPurpose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9930, 0x0018, 0x9930,
      EVR_CS, "AcquisitionMotion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9931, 0x0018, 0x9931,
      EVR_SQ, "AcquisitionStartLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9932, 0x0018, 0x9932,
      EVR_SQ, "AcquisitionEndLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9933, 0x0018, 0x9933,
      EVR_SQ, "ReconstructionProtocolElementSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9934, 0x0018, 0x9934,
      EVR_SQ, "ReconstructionProtocolElementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9935, 0x0018, 0x9935,
      EVR_SQ, "StorageProtocolElementSpecificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9936, 0x0018, 0x9936,
      EVR_SQ, "StorageProtocolElementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9937, 0x0018, 0x9937,
      EVR_LO, "RequestedSeriesDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9938, 0x0018, 0x9938,
      EVR_US, "SourceAcquisitionProtocolElementNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9939, 0x0018, 0x9939,
      EVR_US, "SourceAcquisitionBeamNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x993a, 0x0018, 0x993a,
      EVR_US, "SourceReconstructionProtocolElementNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x993b, 0x0018, 0x993b,
      EVR_SQ, "ReconstructionStartLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x993c, 0x0018, 0x993c,
      EVR_SQ, "ReconstructionEndLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x993d, 0x0018, 0x993d,
      EVR_SQ, "ReconstructionAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x993e, 0x0018, 0x993e,
      EVR_SQ, "ReconstructionTargetCenterLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9941, 0x0018, 0x9941,
      EVR_UT, "ImageFilterDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9942, 0x0018, 0x9942,
      EVR_FD, "CTDIvolNotificationTrigger", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9943, 0x0018, 0x9943,
      EVR_FD, "DLPNotificationTrigger", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9944, 0x0018, 0x9944,
      EVR_CS, "AutoKVPSelectionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9945, 0x0018, 0x9945,
      EVR_FD, "AutoKVPUpperBound", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9946, 0x0018, 0x9946,
      EVR_FD, "AutoKVPLowerBound", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0x9947, 0x0018, 0x9947,
      EVR_CS, "ProtocolDefinedPatientPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0xa001, 0x0018, 0xa001,
      EVR_SQ, "ContributingEquipmentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0xa002, 0x0018, 0xa002,
      EVR_DT, "ContributionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0018, 0xa003, 0x0018, 0xa003,
      EVR_ST, "ContributionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_CS, "ImageBlankingShape", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_UN, "PhysiologicalDataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_IS, "AECField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_LO, "APRName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0019, 0x0000, 0x0019, 0x0000,
      EVR_US, "ReviewMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_DS, "AngleValueLArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_UN, "PhysiologicalDataChannelAndKind", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_IS, "AECFilmScreen", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_US, "AnatomicalBackgroundPercent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0001, 0x0019, 0x0001,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_DS, "AngleValuePArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_IS, "ImageBlankingLeftVerticalEdge", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_US, "SampleBitsAllocated", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_SL, "NumberOfCellsInDetector", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_IS, "AECDensity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_US, "NumberOfPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0002, 0x0019, 0x0002,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_DS, "AngleValueCArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_US, "SampleBitsStored", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_DS, "CellNumberAtTheta", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_US, "ApplyAnatomicalBackground", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0003, 0x0019, 0x0003,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_CS, "AngleLabelLArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_IS, "ImageBlankingRightVerticalEdge", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_US, "SampleHighBit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_DS, "CellSpacing", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_SS, "PixelShiftArray", 4, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0004, 0x0019, 0x0004,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_CS, "AngleLabelPArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_US, "SampleRepresentation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_US, "Brightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0005, 0x0019, 0x0005,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_CS, "AngleLabelCArm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_IS, "ImageBlankingUpperHorizontalEdge", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_UN, "SmallestSampleValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_US, "Contrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0006, 0x0019, 0x0006,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0007, 0x0019, 0x0007,
      EVR_ST, "ProcedureName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0007, 0x0019, 0x0007,
      EVR_UN, "LargestSampleValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0007, 0x0019, 0x0007,
      EVR_US, "Enabled", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0007, 0x0019, 0x0007,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0007, 0x0019, 0x0007,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_ST, "ExamName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_IS, "ImageBlankingLowerHorizontalEdge", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_UN, "NumberOfSamples", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_US, "NativeEdgeEnhancementPercentGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_LT, "OrientationHeadFeet", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0008, 0x0019, 0x0008,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_SH, "PatientSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_DS, "MainMagneticField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_UN, "SampleData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_SS, "NativeEdgeEnhancementLUTIndex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_LT, "ViewDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x0009, 0x0019, 0x0009,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x000a, 0x0019, 0x000a,
      EVR_IS, "RecordView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x000a, 0x0019, 0x000a,
      EVR_UN, "SampleRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x000a, 0x0019, 0x000a,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x000a, 0x0019, 0x000a,
      EVR_SS, "NativeEdgeEnhancementKernelSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x000a, 0x0019, 0x000a,
      EVR_LT, "OrientationSupineProne", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x000b, 0x0019, 0x000b,
      EVR_US, "SubtrEdgeEnhancementPercentGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x000b, 0x0019, 0x000b,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x000c, 0x0019, 0x000c,
      EVR_SS, "SubtrEdgeEnhancementLUTIndex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x000c, 0x0019, 0x000c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x000d, 0x0019, 0x000d,
      EVR_SS, "SubtrEdgeEnhancementKernelSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x000d, 0x0019, 0x000d,
      EVR_TM, "Time", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x000e, 0x0019, 0x000e,
      EVR_IS, "FlowCompensation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x000e, 0x0019, 0x000e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x000e, 0x0019, 0x000e,
      EVR_US, "FadePercent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x000e, 0x0019, 0x000e,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x0019, 0x000f, 0x0019, 0x000f,
      EVR_DS, "HorizontalFrameOfReference", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x000f, 0x0019, 0x000f,
      EVR_US, "FlippedBeforeLateralityApplied", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_LT, "MeasurementMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_LT, "VideoBeamBoost", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-XSB-DCI Release 1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_UN, "ECGTimeMapDataBitsAllocated", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_DS, "InjectorDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_IS, "CenterOfCircularImageBlanking", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_UN, "PhysiologicalDataType2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_US, "MainsFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_IS, "PatientThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_US, "ApplyFade", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_DS, "SourceSideCollimatorAperture", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_DS, "DistanceSourceToSourceSideCollimator", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_ST, "CRImageParamsCommon", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113681" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_DS, "NetFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_LT, "ParameterFileName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  ACQU" }
  , { 0x0019, 0x0010, 0x0019, 0x0010,
      EVR_DS, "TotalMeasurementTimeNominal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_LT, "ImageType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_US, "ChannelGeneratingVideoSync", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-XSB-DCI Release 1" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_UN, "ECGTimeMapDataBitsStored", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_CS, "AutoInject", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_UN, "PhysiologicalDataChannelAndKind2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_SS, "SeriesContrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_IS, "ChemicalShiftNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS-MR-1" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_DS, "DetectorSideCollimatorAperture", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_DS, "DistanceSourceToDetectorSideCollimator", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_ST, "CRImageIPParamsSingle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113681" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_LO, "SequenceFileName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  ACQU" }
  , { 0x0019, 0x0011, 0x0019, 0x0011,
      EVR_DS, "TotalMeasurementTimeCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_US, "VideoGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-XSB-DCI Release 1" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_UN, "ECGTimeMapDataHighBit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_IS, "RadiusOfCircularImageBlanking", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_US, "SampleBitsAllocated2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_SS, "LastPseq", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_IS, "PhaseNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS-MR-1" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_US, "Zoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_DS, "MagneticFieldStrength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_ST, "CRImageIPParamsLeft", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113681" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_LT, "SequenceFileOwner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  ACQU" }
  , { 0x0019, 0x0012, 0x0019, 0x0012,
      EVR_DS, "StartDelayTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_US, "VideoOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-XSB-DCI Release 1" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_UN, "ECGTimeMapDataRepresentation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_US, "SampleBitsStored2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_SS, "StartNumberForBaseline", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_SS, "PanX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_ST, "CRImageIPParamsRight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113681" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_LT, "SequenceDescription", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  ACQU" }
  , { 0x0019, 0x0013, 0x0019, 0x0013,
      EVR_DS, "DwellTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_UN, "ECGTimeMapDataSmallestDataValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_IS, "AcquisitionMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_US, "SampleHighBit2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_SS, "EndNumberForBaseline", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_SS, "PanY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_DS, "ADCVoltage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_LT, "EPIFileName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  ACQU" }
  , { 0x0019, 0x0014, 0x0019, 0x0014,
      EVR_IS, "NumberOfPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_LT, "SoftwareVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_UN, "ECGTimeMapDataLargestDataValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_CS, "CameraRotationEnabled", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_LT, "OfflineUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_US, "SampleRepresentation2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_SS, "StartNumberForEnhancedScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_SS, "NativeEdgeEnhancementAdvPercGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_LO, "AcquiredPlane", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_LO, "KanjiBodyPartForExposure", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0015, 0x0019, 0x0015,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_UN, "ECGTimeMapDataNumberOfDataValues", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_CS, "ReverseSweep", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_UN, "SmallestSampleValue2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_SS, "EndNumberForEnhancedScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_SS, "SubtrEdgeEnhancementAdvPercGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_DS, "ADCOffset", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0016, 0x0019, 0x0016,
      EVR_UL, "SequenceControlMask", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_UN, "ECGTimeMapData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-DCI Release 1" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_IS, "SpatialFilterStrength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_UN, "LargestSampleValue2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_SS, "SeriesPlane", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_US, "InvertFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x0017, 0x0019, 0x0017,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_IS, "ZoomFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_UN, "NumberOfSamples2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_LO, "FirstScanRAS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_IS, "BeamDistance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0018, 0x0019, 0x0018,
      EVR_UL, "MeasurementStatusMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0019, 0x0019, 0x0019,
      EVR_IS, "XZoomCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x0019, 0x0019, 0x0019,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0019, 0x0019, 0x0019,
      EVR_UN, "SampleData2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x0019, 0x0019, 0x0019,
      EVR_DS, "FirstScanLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0019, 0x0019, 0x0019,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_IS, "YZoomCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_UN, "SampleRate2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_LO, "LastScanRAS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_OB, "Quant1KOverlay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x001a, 0x0019, 0x001a,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001b, 0x0019, 0x001b,
      EVR_DS, "Focus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001b, 0x0019, 0x001b,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001b, 0x0019, 0x001b,
      EVR_DS, "LastScanLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x001b, 0x0019, 0x001b,
      EVR_US, "OriginalResolution", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x001b, 0x0019, 0x001b,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001c, 0x0019, 0x001c,
      EVR_CS, "Dose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001c, 0x0019, 0x001c,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001c, 0x0019, 0x001c,
      EVR_DS, "AutoWindowCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x001c, 0x0019, 0x001c,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001d, 0x0019, 0x001d,
      EVR_IS, "SideMark", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001d, 0x0019, 0x001d,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001d, 0x0019, 0x001d,
      EVR_DS, "AutoWindowWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  VIEW 1.0" }
  , { 0x0019, 0x001d, 0x0019, 0x001d,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001e, 0x0019, 0x001e,
      EVR_IS, "PercentageLandscape", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001e, 0x0019, 0x001e,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001e, 0x0019, 0x001e,
      EVR_DS, "DisplayFieldOfView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x001e, 0x0019, 0x001e,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x001f, 0x0019, 0x001f,
      EVR_DS, "ExposureDuration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DLX_SERIE_01" }
  , { 0x0019, 0x001f, 0x0019, 0x001f,
      EVR_SS, "Internal", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x001f, 0x0019, 0x001f,
      EVR_SS, "DefaultTableIsoCenterHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x001f, 0x0019, 0x001f,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_LT, "MPMCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_DS, "RTDDataCompressionFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-XSB-DCI Release 1" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_SS, "SystemCalibFactorPlaneA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_TM, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_IS, "WorkstationNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_DS, "TransmitterAmplitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_DS, "ExposureTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_IS, "NumberOfPossibleChannels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_SL, "SceneFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_LT, "MeasurementMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_SL, "SystemCalibFactorPlaneB", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0020, 0x0019, 0x0020,
      EVR_IS, "NumberOfFourierLinesNominal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_LT, "Latitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_IS, "NumberOfTransmitterAmplitudes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_DS, "ExposureCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_IS, "MeanChannelNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0021, 0x0019, 0x0021,
      EVR_IS, "NumberOfFourierLinesCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_LT, "Sensitivity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_SS, "XRayParameterSetNo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_UN, "RouteAET", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_DS, "TransmitterAttenuator", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_DS, "DetectorSpacing", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_SL, "RefPhotofileFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0022, 0x0019, 0x0022,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0023, 0x0019, 0x0023,
      EVR_LT, "EDR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0023, 0x0019, 0x0023,
      EVR_DS, "PCRPrintScale", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0023, 0x0019, 0x0023,
      EVR_DS, "TableSpeed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0023, 0x0019, 0x0023,
      EVR_DS, "DetectorCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0023, 0x0019, 0x0023,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_LT, "LFix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_SS, "XRaySystem", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_UN, "PCRPrintJobEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_DS, "MidScanTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_DS, "TransmitterCalibration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_DS, "ReadingIntegrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_LO, "SceneName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0024, 0x0019, 0x0024,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_LT, "SFix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_IS, "PCRNoFilmCopies", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_SS, "MidScanFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_LT, "OriginalPixelDataQuality", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_DS, "KVPGeneratorPowerCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0025, 0x0019, 0x0025,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_LT, "PresetMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_IS, "PCRFilmLayoutPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_SL, "DegreesOfAzimuth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_DS, "TransmitterReference", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_DS, "GeneratorVoltage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_SS, "AcquisitionIndex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0026, 0x0019, 0x0026,
      EVR_IS, "NumberOfFourierLinesAfterZero", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0027, 0x0019, 0x0027,
      EVR_LT, "Region", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0027, 0x0019, 0x0027,
      EVR_UN, "PCRPrintReportName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0027, 0x0019, 0x0027,
      EVR_DS, "GantryPeriod", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0027, 0x0019, 0x0027,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_LT, "Subregion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_SS, "AcquiredDisplayMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_IS, "TubeNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_SS, "MixedPulseMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0028, 0x0019, 0x0028,
      EVR_IS, "FirstMeasuredFourierLine", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0029, 0x0019, 0x0029,
      EVR_IS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x002a, 0x0019, 0x002a,
      EVR_SS, "AcquisitionDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x002a, 0x0019, 0x002a,
      EVR_DS, "XrayOnPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x002a, 0x0019, 0x002a,
      EVR_SS, "NoOfPositions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x002a, 0x0019, 0x002a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x002b, 0x0019, 0x002b,
      EVR_DS, "XrayOffPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x002c, 0x0019, 0x002c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x002c, 0x0019, 0x002c,
      EVR_SL, "NumberOfTriggers", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x002c, 0x0019, 0x002c,
      EVR_SS, "NoOfPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x002c, 0x0019, 0x002c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x002d, 0x0019, 0x002d,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x002e, 0x0019, 0x002e,
      EVR_SS, "MaxFramesLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x002e, 0x0019, 0x002e,
      EVR_DS, "AngleOfFirstView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x002e, 0x0019, 0x002e,
      EVR_SS, "FrameRateForPositions", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x002e, 0x0019, 0x002e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x002f, 0x0019, 0x002f,
      EVR_DS, "TriggerFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_LT, "Orientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_US, "MaximumFrameSizeNIU", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_UL, "MaximumImageFrameSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_LO, "CREDRMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_IS, "BuckyGrid", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_US, "ECGTriggering", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_SS, "NoOfFramesForPositions", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-PCR Release 2" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_LT, "CalculationMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0030, 0x0019, 0x0030,
      EVR_IS, "AcquisitionColumns", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0031, 0x0019, 0x0031,
      EVR_LT, "MarkOnFilm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0031, 0x0019, 0x0031,
      EVR_UN, "ECG1Offset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0031, 0x0019, 0x0031,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0019, 0x0031, 0x0019, 0x0031,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0031, 0x0019, 0x0031,
      EVR_IS, "ReconstructionColumns", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_LT, "RotationOnDRC", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_SS, "SubtractedFilterType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_UN, "ECG2Offset1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_SS, "SteppingDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_LO, "KanjiMenuName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0032, 0x0019, 0x0032,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0033, 0x0019, 0x0033,
      EVR_UN, "ECG2Offset2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0034, 0x0019, 0x0034,
      EVR_SS, "FilterFactorNative", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0034, 0x0019, 0x0034,
      EVR_IS, "Focus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0034, 0x0019, 0x0034,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0034, 0x0019, 0x0034,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0036, 0x0019, 0x0036,
      EVR_SS, "AnatomicBackgroundFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0036, 0x0019, 0x0036,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0036, 0x0019, 0x0036,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0038, 0x0019, 0x0038,
      EVR_SS, "WindowUpperLimitNative", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0038, 0x0019, 0x0038,
      EVR_IS, "Child", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0038, 0x0019, 0x0038,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0038, 0x0019, 0x0038,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0039, 0x0019, 0x0039,
      EVR_SS, "AxialType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x0039, 0x0019, 0x0039,
      EVR_SS, "ScanFOVType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x003a, 0x0019, 0x003a,
      EVR_SS, "WindowLowerLimitNative", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x003a, 0x0019, 0x003a,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x003a, 0x0019, 0x003a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x003a, 0x0019, 0x003a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x003b, 0x0019, 0x003b,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x003c, 0x0019, 0x003c,
      EVR_SS, "WindowBrightnessPhase1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x003c, 0x0019, 0x003c,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x003c, 0x0019, 0x003c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x003c, 0x0019, 0x003c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x003e, 0x0019, 0x003e,
      EVR_SS, "WindowBrightnessPhase2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x003e, 0x0019, 0x003e,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x003e, 0x0019, 0x003e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x003e, 0x0019, 0x003e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x003f, 0x0019, 0x003f,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_LT, "ReaderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_SS, "WindowContrastPhase1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_SS, "StatReconFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_LO, "CRLatitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_DS, "FrameRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_IS, "CollimatorDistanceX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_UL, "MasterControlMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_CS, "ImageProcessingType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_IS, "ArrayCoilElementNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0040, 0x0019, 0x0040,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0041, 0x0019, 0x0041,
      EVR_LT, "SubModality", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0041, 0x0019, 0x0041,
      EVR_SS, "ComputeType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0041, 0x0019, 0x0041,
      EVR_IS, "CollimatorDistanceY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0041, 0x0019, 0x0041,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0041, 0x0019, 0x0041,
      EVR_UL, "ArrayCoilElementSelectMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_LT, "ReaderSerialNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_SS, "WindowContrastPhase2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_SS, "SegmentNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_US, "ProcessingMask", 5, 5, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_IS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_UL, "ArrayCoilElementDataMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0042, 0x0019, 0x0042,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0043, 0x0019, 0x0043,
      EVR_SS, "TotalSegmentsRequested", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0043, 0x0019, 0x0043,
      EVR_IS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0043, 0x0019, 0x0043,
      EVR_IS, "ArrayCoilElementToADCConnect", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_SS, "FilterFactorSub", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_DS, "InterscanDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_SS, "ImageTransferDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_DS, "ArrayCoilElementNoiseLevel", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0044, 0x0019, 0x0044,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0045, 0x0019, 0x0045,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0045, 0x0019, 0x0045,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0045, 0x0019, 0x0045,
      EVR_IS, "ArrayCoilADCPairNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0046, 0x0019, 0x0046,
      EVR_SS, "PeakOpacified", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0046, 0x0019, 0x0046,
      EVR_SL, "InversFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0046, 0x0019, 0x0046,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0046, 0x0019, 0x0046,
      EVR_UL, "ArrayCoilCombinationMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0046, 0x0019, 0x0046,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0047, 0x0019, 0x0047,
      EVR_SS, "ViewCompressionFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0047, 0x0019, 0x0047,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0048, 0x0019, 0x0048,
      EVR_SL, "MaskFrame", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0048, 0x0019, 0x0048,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0048, 0x0019, 0x0048,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0048, 0x0019, 0x0048,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0048, 0x0019, 0x0048,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0049, 0x0019, 0x0049,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0049, 0x0019, 0x0049,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x004a, 0x0019, 0x004a,
      EVR_SL, "BIHFrame", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x004a, 0x0019, 0x004a,
      EVR_SS, "TotalNumberOfRefChannels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x004a, 0x0019, 0x004a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x004a, 0x0019, 0x004a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x004b, 0x0019, 0x004b,
      EVR_SL, "DataSizeForScanData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x004c, 0x0019, 0x004c,
      EVR_SS, "CentBeamAngulationCaudCran", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x004c, 0x0019, 0x004c,
      EVR_SS, "BlankingCircleDiameter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x004c, 0x0019, 0x004c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x004e, 0x0019, 0x004e,
      EVR_SS, "CentBeamAngulationLRAnterior", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x004e, 0x0019, 0x004e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_LT, "CassetteScale", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_SS, "LongitudinalPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_US, "VideoScanMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_LO, "CRGroupNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_IS, "CentralBeamHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_DS, "ReceiverTotalGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_SL, "StandDataValid", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_DS, "DetectorAlignment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_CS, "EDRMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_IS, "NoiseLevel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_IS, "NumberOfAverages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0050, 0x0019, 0x0050,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0051, 0x0019, 0x0051,
      EVR_LT, "CassetteMatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0051, 0x0019, 0x0051,
      EVR_US, "VideoLineRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0051, 0x0019, 0x0051,
      EVR_DS, "ReceiverAmplifierGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_LT, "CassetteSubmatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_SS, "SideOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_SS, "ReconPostProcessingFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_DS, "ReceiverPreamplifierGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_SS, "TableTilt", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0052, 0x0019, 0x0052,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0053, 0x0019, 0x0053,
      EVR_LT, "Barcode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_SS, "IsoCenterHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_DS, "ReceiverCableAttenuation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_SS, "TableAxisRotation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0054, 0x0019, 0x0054,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0055, 0x0019, 0x0055,
      EVR_DS, "ReceiverReferenceGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0056, 0x0019, 0x0056,
      EVR_SS, "ImageTwist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0056, 0x0019, 0x0056,
      EVR_DS, "ReceiverFilterFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0056, 0x0019, 0x0056,
      EVR_SS, "TableLongitudalPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0056, 0x0019, 0x0056,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0057, 0x0019, 0x0057,
      EVR_SS, "CTWaterNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0058, 0x0019, 0x0058,
      EVR_SS, "SourceImageDistance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0058, 0x0019, 0x0058,
      EVR_SS, "CTBoneNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0058, 0x0019, 0x0058,
      EVR_SS, "TableSideOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0058, 0x0019, 0x0058,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x005a, 0x0019, 0x005a,
      EVR_SS, "MechanicalMagnificationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x005a, 0x0019, 0x005a,
      EVR_FL, "AcquisitionDuration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x005a, 0x0019, 0x005a,
      EVR_SS, "TableIsoCenterHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x005a, 0x0019, 0x005a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x005c, 0x0019, 0x005c,
      EVR_SL, "CalibrationFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x005c, 0x0019, 0x005c,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x005c, 0x0019, 0x005c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x005d, 0x0019, 0x005d,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x005e, 0x0019, 0x005e,
      EVR_SL, "CalibrationAngleCranCaud", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x005e, 0x0019, 0x005e,
      EVR_SL, "NumberOfChannels1To512", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x005e, 0x0019, 0x005e,
      EVR_SL, "CollimationDataValid", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x005e, 0x0019, 0x005e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x005f, 0x0019, 0x005f,
      EVR_SL, "IncrementBetweenChannels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_LT, "ContrastType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_SL, "CalibrationAngleRAOLAO", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_SL, "StartingView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_US, "XrayTechnique", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_IS, "BuckyAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_DS, "ReconstructionScaleFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_SL, "PeriSequenceNo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_DS, "FocusAlignment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_SH, "RadiographersCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_IS, "NumberOfDataBytes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_DS, "FlipAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0060, 0x0019, 0x0060,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0061, 0x0019, 0x0061,
      EVR_LT, "RotationAmount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0061, 0x0019, 0x0061,
      EVR_SL, "NumberOfViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0061, 0x0019, 0x0061,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0061, 0x0019, 0x0061,
      EVR_DS, "ImageIdentifierFromat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0061, 0x0019, 0x0061,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_LT, "RotationCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_SL, "CalibrationTableToFloorDist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_SL, "IncrementBetweenViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_US, "IrisDiaphragm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_IS, "NumberOfVirtuellChannels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_DS, "ReferenceScaleFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_SL, "PeriTotalScenes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0062, 0x0019, 0x0062,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0063, 0x0019, 0x0063,
      EVR_LT, "DensityShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0063, 0x0019, 0x0063,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0063, 0x0019, 0x0063,
      EVR_CS, "Filter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_US, "FrequencyRank", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_SL, "CalibrationIsocenterToFloorDist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_CS, "CineParallel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_SL, "PeriOverlapTop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0064, 0x0019, 0x0064,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0065, 0x0019, 0x0065,
      EVR_LT, "FrequencyEnhancement", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0065, 0x0019, 0x0065,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0065, 0x0019, 0x0065,
      EVR_CS, "CineMaster", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0065, 0x0019, 0x0065,
      EVR_UL, "FocalSpotDeflectionAmplitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0066, 0x0019, 0x0066,
      EVR_LT, "FrequencyType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0066, 0x0019, 0x0066,
      EVR_SL, "CalibrationIsocenterToSourceDist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0066, 0x0019, 0x0066,
      EVR_SL, "PeriOverlapBottom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0066, 0x0019, 0x0066,
      EVR_UL, "FocalSpotDeflectionPhase", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0066, 0x0019, 0x0066,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0067, 0x0019, 0x0067,
      EVR_LT, "KernelLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0067, 0x0019, 0x0067,
      EVR_UL, "FocalSpotDeflectionOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0068, 0x0019, 0x0068,
      EVR_UL, "KernelMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0068, 0x0019, 0x0068,
      EVR_SL, "CalibrationSourceToII", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0068, 0x0019, 0x0068,
      EVR_IS, "CArmAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0068, 0x0019, 0x0068,
      EVR_SL, "RawImageNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0068, 0x0019, 0x0068,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0069, 0x0019, 0x0069,
      EVR_UL, "ConvolutionMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0069, 0x0019, 0x0069,
      EVR_IS, "CollimatorAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x006a, 0x0019, 0x006a,
      EVR_SL, "CalibrationIIZoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x006a, 0x0019, 0x006a,
      EVR_SS, "DependantOnNumberOfViewsProcessed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x006a, 0x0019, 0x006a,
      EVR_SL, "XRayDataValid", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x006a, 0x0019, 0x006a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x006b, 0x0019, 0x006b,
      EVR_SS, "FieldOfViewInDetectorCells", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x006c, 0x0019, 0x006c,
      EVR_SL, "CalibrationIIField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x006c, 0x0019, 0x006c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x006e, 0x0019, 0x006e,
      EVR_SL, "CalibrationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x006e, 0x0019, 0x006e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_LT, "PLASource", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_SL, "CalibrationObjectToImageDistance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_UN, "RADProtocolPrinter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_SS, "ValueOfBackProjectionButton", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_LO, "CRImageSerialNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_US, "ExposureChannel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_IS, "FilterNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_IS, "NumberOfReadings", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_DS, "PhaseGradientAmplitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_DS, "WaterScalingFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_IS, "SplitExposureFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_IS, "NumberOfPrescans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0070, 0x0019, 0x0070,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_LT, "PLADestination", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_UN, "RADProtocolMedium", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_SS, "SetIfFatqEstimatesWereUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_UN, "ExposureChannelFirstImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_LT, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_DS, "ReadoutGradientAmplitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_DS, "InterpolationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_IS, "NoOfSplitExposureFrames", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0071, 0x0019, 0x0071,
      EVR_IS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_SL, "CalibrationSystemFactor", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_DS, "ZChannelAvgOverViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_US, "ProcessingChannel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_DS, "SelectionGradientAmplitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0072, 0x0019, 0x0072,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0073, 0x0019, 0x0073,
      EVR_DS, "AvgOfLeftRefChannelsOverViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0073, 0x0019, 0x0073,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_SL, "CalibrationSystemCorrection", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_DS, "MaxLeftChannelOverViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_LO, "FilterMaterial1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_IS, "NumberOfProjections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0074, 0x0019, 0x0074,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0075, 0x0019, 0x0075,
      EVR_LT, "UIDOriginalImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0075, 0x0019, 0x0075,
      EVR_DS, "AvgOfRightRefChannelsOverViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0075, 0x0019, 0x0075,
      EVR_LO, "FilterMaterial2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0075, 0x0019, 0x0075,
      EVR_IS, "NumberOfBytes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_SL, "CalibrationSystemIIFormats", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_DS, "MaxRightChannelOverViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_SL, "FillingAverageFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0076, 0x0019, 0x0076,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0077, 0x0019, 0x0077,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0078, 0x0019, 0x0078,
      EVR_SL, "CalibrationGantryDataValid", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0078, 0x0019, 0x0078,
      EVR_DS, "FilterThickness1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0078, 0x0019, 0x0078,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0078, 0x0019, 0x0078,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0078, 0x0019, 0x0078,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0079, 0x0019, 0x0079,
      EVR_DS, "FilterThickness2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0079, 0x0019, 0x0079,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x007a, 0x0019, 0x007a,
      EVR_SS, "CollimatorSquareBreadth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x007a, 0x0019, 0x007a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x007a, 0x0019, 0x007a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x007c, 0x0019, 0x007c,
      EVR_SS, "CollimatorSquareHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x007c, 0x0019, 0x007c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x007c, 0x0019, 0x007c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x007d, 0x0019, 0x007d,
      EVR_DS, "SecondEcho", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x007e, 0x0019, 0x007e,
      EVR_SS, "CollimatorSquareDiameter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x007e, 0x0019, 0x007e,
      EVR_SS, "NumberOfEchos", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x007e, 0x0019, 0x007e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x007e, 0x0019, 0x007e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x007f, 0x0019, 0x007f,
      EVR_DS, "TableDelta", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_SS, "CollimaterFingerTurnAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_DS, "AcquisitionDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_IS, "BuckyFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_LO, "CRBarCodeNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_DS, "GradientDelayTime", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_LT, "ReconstructionAlgorithmSet", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_LT, "PatientRegion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_IS, "ReadingPositionSpecification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR SPECTRO;1" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0080, 0x0019, 0x0080,
      EVR_LT, "ReaderHeader", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_SS, "Contiguous", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_UN, "RelativeImageTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_IS, "ObjectPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_LT, "ReconstructionAlgorithmIndex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_IS, "ReadingSensitivityCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0081, 0x0019, 0x0081,
      EVR_LT, "FilterTypeForRawData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_SS, "CollimaterFingerPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_DS, "TotalGradientDelayTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_LT, "RegenerationSoftwareVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_LT, "PatientPhaseOfLife", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0082, 0x0019, 0x0082,
      EVR_DS, "FilterParameterForRawData", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0083, 0x0019, 0x0083,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0083, 0x0019, 0x0083,
      EVR_LT, "FilterTypeForImageData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0084, 0x0019, 0x0084,
      EVR_SS, "CollimaterDiaphragmTurnAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0084, 0x0019, 0x0084,
      EVR_DS, "PeakSAR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0084, 0x0019, 0x0084,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0084, 0x0019, 0x0084,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0084, 0x0019, 0x0084,
      EVR_DS, "FilterParameterForImageData", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0085, 0x0019, 0x0085,
      EVR_SS, "MonitorSAR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0085, 0x0019, 0x0085,
      EVR_LT, "FilterTypeForPhaseCorrection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0086, 0x0019, 0x0086,
      EVR_SS, "CollimaterDiaphragmPosition1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0086, 0x0019, 0x0086,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0086, 0x0019, 0x0086,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0086, 0x0019, 0x0086,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0086, 0x0019, 0x0086,
      EVR_DS, "FilterParameterForPhaseCorrection", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0087, 0x0019, 0x0087,
      EVR_DS, "CardiacRepetitionTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0087, 0x0019, 0x0087,
      EVR_LT, "NormalizationFilterTypeForImageData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_SS, "CollimaterDiaphragmPosition2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_SS, "ImagesPerCardiacCycle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0088, 0x0019, 0x0088,
      EVR_DS, "NormalizationFilterParameterForImageData", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x008a, 0x0019, 0x008a,
      EVR_SS, "CollimaterDiaphragmMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x008a, 0x0019, 0x008a,
      EVR_SS, "ActualReceiveGainAnalog", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x008a, 0x0019, 0x008a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x008a, 0x0019, 0x008a,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x008b, 0x0019, 0x008b,
      EVR_SS, "ActualReceiveGainDigital", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x008c, 0x0019, 0x008c,
      EVR_SS, "CollimaterBeamLimitBreadth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x008c, 0x0019, 0x008c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x008c, 0x0019, 0x008c,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x008d, 0x0019, 0x008d,
      EVR_DS, "DelayAfterTrigger", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x008e, 0x0019, 0x008e,
      EVR_SS, "CollimaterBeamLimitHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x008e, 0x0019, 0x008e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x008e, 0x0019, 0x008e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x008f, 0x0019, 0x008f,
      EVR_SS, "SwapPhaseFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x008f, 0x0019, 0x008f,
      EVR_SS, "SwapPhaseFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_SS, "CollimaterBeamLimitDiameter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LO, "UnprocessedFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_SS, "PauseInterval", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_CS, "VideoWhiteCompression", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LO, "CRFilmOutputExposures", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE ??? From Adantage Review CS" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LO, "DeskCommand", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LT, "SensitivityCorrectionLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_DS, "OsteoOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_SH, "FilmAnnotationCharacterString1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_IS, "NumberOfSaturationRegions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0090, 0x0019, 0x0090,
      EVR_LT, "PLAOfSecondaryDestination", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x0091, 0x0019, 0x0091,
      EVR_UN, "KeyValues", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0091, 0x0019, 0x0091,
      EVR_DS, "PulseTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0091, 0x0019, 0x0091,
      EVR_DS, "SaturationPhaseEncodingVectorCoronalComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0091, 0x0019, 0x0091,
      EVR_SH, "FilmAnnotationCharacterString2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0019, 0x0091, 0x0019, 0x0091,
      EVR_DS, "SaturationPhaseEncodingVectorSagittalComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_SS, "X-RayControlMOde", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_UN, "DestinationPostprocessingFunction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_SL, "SliceOffsetOnFrequencyAxis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_DS, "SaturationReadoutVectorCoronalComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_DS, "OsteoRegressionLineSlope", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0092, 0x0019, 0x0092,
      EVR_DS, "SaturationReadoutVectorSagittalComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0093, 0x0019, 0x0093,
      EVR_DS, "CenterFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0093, 0x0019, 0x0093,
      EVR_DS, "OsteoRegressionLineIntercept", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0093, 0x0019, 0x0093,
      EVR_DS, "EPIStimulationMonitorMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_SS, "X-RaySystem", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_SS, "TransmitGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_DS, "OsteoStandardizationCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0094, 0x0019, 0x0094,
      EVR_DS, "ImageRotationAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0095, 0x0019, 0x0095,
      EVR_SS, "AnalogReceiverGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_SS, "FocalSpot", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_SS, "DigitalReceiverGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_IS, "OsteoPhantomNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0096, 0x0019, 0x0096,
      EVR_UL, "CoilIDMask", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0097, 0x0019, 0x0097,
      EVR_SL, "BitmapDefiningCVs", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0097, 0x0019, 0x0097,
      EVR_UL, "CoilClassMask", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0098, 0x0019, 0x0098,
      EVR_SS, "ExposureControl", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x0098, 0x0019, 0x0098,
      EVR_SS, "CenterFrequencyMethod", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x0098, 0x0019, 0x0098,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x0098, 0x0019, 0x0098,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x0098, 0x0019, 0x0098,
      EVR_DS, "CoilPosition", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x0099, 0x0019, 0x0099,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x009a, 0x0019, 0x009a,
      EVR_SL, "XRayVoltage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x009a, 0x0019, 0x009a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x009a, 0x0019, 0x009a,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x009b, 0x0019, 0x009b,
      EVR_SS, "PulseSequenceMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x009c, 0x0019, 0x009c,
      EVR_SL, "XRayCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x009c, 0x0019, 0x009c,
      EVR_SS, "PulseSequenceName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x009c, 0x0019, 0x009c,
      EVR_LO, "PulseSequenceName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x009c, 0x0019, 0x009c,
      EVR_SL, "IntensifierLevelCalibrationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x009c, 0x0019, 0x009c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x009d, 0x0019, 0x009d,
      EVR_DT, "PulseSequenceDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x009e, 0x0019, 0x009e,
      EVR_SL, "XRayCurrentTimeProduct", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x009e, 0x0019, 0x009e,
      EVR_LO, "InternalPulseSequenceName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x009e, 0x0019, 0x009e,
      EVR_SL, "NativeReviewFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x009e, 0x0019, 0x009e,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x009f, 0x0019, 0x009f,
      EVR_SS, "CoilType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x009f, 0x0019, 0x009f,
      EVR_SS, "TransmittingCoil", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_SL, "XRayPulseTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_UN, "Version", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_SS, "SurfaceCoilType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_DS, "ExtendedExposureTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_US, "Angulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_US, "RFWatchdogMask", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_DS, "EPIReconstructionPhase", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x00a0, 0x0019, 0x00a0,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_UN, "RangingMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_SS, "ExtremityCoilFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_DS, "ActualExposureTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_US, "Rotation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_DS, "EPIReconstructionSlope", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_DS, "EPIReconstructionSlope", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0019, 0x00a1, 0x0019, 0x00a1,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_SL, "XRaySceneTimeFluoroClock", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_UN, "AbdomenBrightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_SL, "RawDataRunNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_DS, "RFPowerErrorIndicator", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_SL, "SceneNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a2, 0x0019, 0x00a2,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00a3, 0x0019, 0x00a3,
      EVR_UN, "FixedBrightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a3, 0x0019, 0x00a3,
      EVR_UL, "CalibratedFieldStrength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a3, 0x0019, 0x00a3,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00a3, 0x0019, 0x00a3,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_SS, "MaximumPulseRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_SS, "SATFatWaterBone", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_UN, "DetailContrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_SS, "SATFatWaterBone", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_SS, "AcquisitionMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a4, 0x0019, 0x00a4,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00a5, 0x0019, 0x00a5,
      EVR_UN, "ContrastBalance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a5, 0x0019, 0x00a5,
      EVR_DS, "ReceiveBandwidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a5, 0x0019, 0x00a5,
      EVR_DS, "SpecificAbsorptionRateWholeBody", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00a5, 0x0019, 0x00a5,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a5, 0x0019, 0x00a5,
      EVR_SS, "AcquisitonFrameRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_SS, "PulsesPerScene", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_UN, "StructureBoost", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_DS, "SpecificEnergyDose", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_SL, "ECGFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a6, 0x0019, 0x00a6,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00a7, 0x0019, 0x00a7,
      EVR_UN, "StructurePreference", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a7, 0x0019, 0x00a7,
      EVR_DS, "UserData0", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a7, 0x0019, 0x00a7,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a7, 0x0019, 0x00a7,
      EVR_SL, "AdditionalSceneData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_SL, "DoseAreaProductOfScene", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_UN, "NoiseRobustness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_DS, "UserData1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_DS, "ExtendedXRayTubeCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_SL, "FileCopyFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00a8, 0x0019, 0x00a8,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00a9, 0x0019, 0x00a9,
      EVR_UN, "NoiseDoseLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00a9, 0x0019, 0x00a9,
      EVR_DS, "UserData2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00a9, 0x0019, 0x00a9,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00a9, 0x0019, 0x00a9,
      EVR_SL, "PhlebovisionFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_SS, "Dose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_UN, "NoiseDoseStep", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_DS, "UserData3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_SL, "Co2Flag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00aa, 0x0019, 0x00aa,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00ab, 0x0019, 0x00ab,
      EVR_UN, "NoiseFrequencyLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00ab, 0x0019, 0x00ab,
      EVR_DS, "UserData4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ab, 0x0019, 0x00ab,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00ab, 0x0019, 0x00ab,
      EVR_SS, "MaxSpeed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_SS, "DoseRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_UN, "WeakContrastLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_DS, "UserData5", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_SS, "StepWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00ac, 0x0019, 0x00ac,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE B" }
  , { 0x0019, 0x00ad, 0x0019, 0x00ad,
      EVR_UN, "StrongContrastLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00ad, 0x0019, 0x00ad,
      EVR_DS, "UserData6", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ad, 0x0019, 0x00ad,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00ad, 0x0019, 0x00ad,
      EVR_SL, "DigitalAcquisitionZoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x00ae, 0x0019, 0x00ae,
      EVR_SL, "IIToCoverDistance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00ae, 0x0019, 0x00ae,
      EVR_UN, "StructureBoostOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00ae, 0x0019, 0x00ae,
      EVR_DS, "UserData7", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ae, 0x0019, 0x00ae,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00af, 0x0019, 0x00af,
      EVR_UN, "SmoothGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00af, 0x0019, 0x00af,
      EVR_DS, "UserData8", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00af, 0x0019, 0x00af,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_SS, "LastFramePhase1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_UN, "MeasureField1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_DS, "UserData9", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_UL, "AdjustmentStatusMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00b0, 0x0019, 0x00b0,
      EVR_DS, "FeedPerRotation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00b1, 0x0019, 0x00b1,
      EVR_SS, "FrameRatePhase1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b1, 0x0019, 0x00b1,
      EVR_IS, "MinimumRRInterval", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00b1, 0x0019, 0x00b1,
      EVR_UN, "MeasureField2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b1, 0x0019, 0x00b1,
      EVR_DS, "UserData10", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b1, 0x0019, 0x00b1,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0019, 0x00b2, 0x0019, 0x00b2,
      EVR_SS, "LastFramePhase2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b2, 0x0019, 0x00b2,
      EVR_IS, "MaximumRRInterval", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00b2, 0x0019, 0x00b2,
      EVR_UN, "KeyPercentile1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b2, 0x0019, 0x00b2,
      EVR_DS, "UserData11", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b3, 0x0019, 0x00b3,
      EVR_SS, "FrameRatePhase2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b3, 0x0019, 0x00b3,
      EVR_IS, "NumberOfRejections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00b3, 0x0019, 0x00b3,
      EVR_UN, "KeyPercentile2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b3, 0x0019, 0x00b3,
      EVR_DS, "UserData12", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b4, 0x0019, 0x00b4,
      EVR_SS, "LastFramePhase3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b4, 0x0019, 0x00b4,
      EVR_IS, "NumberOfRRIntervals", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00b4, 0x0019, 0x00b4,
      EVR_UN, "DensityLUT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b4, 0x0019, 0x00b4,
      EVR_DS, "UserData13", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b5, 0x0019, 0x00b5,
      EVR_SS, "FrameRatePhase3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b5, 0x0019, 0x00b5,
      EVR_IS, "ArrhythmiaRejection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00b5, 0x0019, 0x00b5,
      EVR_UN, "Brightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b5, 0x0019, 0x00b5,
      EVR_DS, "UserData14", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b6, 0x0019, 0x00b6,
      EVR_SS, "LastFramePhase4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b6, 0x0019, 0x00b6,
      EVR_UN, "Gamma", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0019, 0x00b6, 0x0019, 0x00b6,
      EVR_DS, "UserData15", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b7, 0x0019, 0x00b7,
      EVR_SS, "FrameRatePhase4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b7, 0x0019, 0x00b7,
      EVR_DS, "UserData16", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b8, 0x0019, 0x00b8,
      EVR_SS, "GammaOfNativeImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b8, 0x0019, 0x00b8,
      EVR_DS, "UserData17", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00b9, 0x0019, 0x00b9,
      EVR_SS, "GammaOfTVSystem", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00b9, 0x0019, 0x00b9,
      EVR_DS, "UserData18", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ba, 0x0019, 0x00ba,
      EVR_DS, "UserData19", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00bb, 0x0019, 0x00bb,
      EVR_SL, "PixelshiftX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00bb, 0x0019, 0x00bb,
      EVR_DS, "UserData20", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00bc, 0x0019, 0x00bc,
      EVR_SL, "PixelshiftY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00bc, 0x0019, 0x00bc,
      EVR_DS, "UserData21", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00bd, 0x0019, 0x00bd,
      EVR_SL, "MaskAverageFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00bd, 0x0019, 0x00bd,
      EVR_DS, "UserData22", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00bd, 0x0019, 0x00bd,
      EVR_IS, "PulmoTriggerLevel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00be, 0x0019, 0x00be,
      EVR_SL, "BlankingCircleFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00be, 0x0019, 0x00be,
      EVR_DS, "ProjectionAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00be, 0x0019, 0x00be,
      EVR_DS, "ExpiratoricReserveVolume", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00bf, 0x0019, 0x00bf,
      EVR_SL, "CircleRowStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00bf, 0x0019, 0x00bf,
      EVR_DS, "VitalCapacity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00c0, 0x0019, 0x00c0,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00c0, 0x0019, 0x00c0,
      EVR_SL, "CircleRowEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c0, 0x0019, 0x00c0,
      EVR_SS, "BitmapOfSATSelections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00c0, 0x0019, 0x00c0,
      EVR_SS, "SaturationPlanes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c0, 0x0019, 0x00c0,
      EVR_DS, "PulmoWater", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00c1, 0x0019, 0x00c1,
      EVR_SL, "CircleColumnStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c1, 0x0019, 0x00c1,
      EVR_SS, "SurfaceCoilIntensityCorrectionFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00c1, 0x0019, 0x00c1,
      EVR_SS, "SurfaceCoilIntensityCorrectionFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c1, 0x0019, 0x00c1,
      EVR_DS, "EPICapacity", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00c1, 0x0019, 0x00c1,
      EVR_DS, "PulmoAir", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00c2, 0x0019, 0x00c2,
      EVR_SL, "CircleColumnEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c2, 0x0019, 0x00c2,
      EVR_SS, "SATLocationR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c2, 0x0019, 0x00c2,
      EVR_DS, "EPIInductance", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00c2, 0x0019, 0x00c2,
      EVR_DA, "PulmoDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00c3, 0x0019, 0x00c3,
      EVR_SL, "CircleDiameter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c3, 0x0019, 0x00c3,
      EVR_SS, "SATLocationL", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c3, 0x0019, 0x00c3,
      EVR_IS, "EPISwitchConfigurationCode", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00c3, 0x0019, 0x00c3,
      EVR_TM, "PulmoTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  COAD" }
  , { 0x0019, 0x00c4, 0x0019, 0x00c4,
      EVR_SL, "RectangularCollimaterFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c4, 0x0019, 0x00c4,
      EVR_SS, "SATLocationA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c4, 0x0019, 0x00c4,
      EVR_IS, "EPISwitchHardwareCode", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00c5, 0x0019, 0x00c5,
      EVR_SL, "RectangleRowStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c5, 0x0019, 0x00c5,
      EVR_SS, "SATLocationP", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c5, 0x0019, 0x00c5,
      EVR_DS, "EPISwitchDelayTime", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00c6, 0x0019, 0x00c6,
      EVR_IS, "CycledMultipleSlice", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00c6, 0x0019, 0x00c6,
      EVR_SL, "RectangleRowEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c6, 0x0019, 0x00c6,
      EVR_SS, "SATLocationH", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c7, 0x0019, 0x00c7,
      EVR_SL, "RectangleColumnStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c7, 0x0019, 0x00c7,
      EVR_SS, "SATLocationF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c8, 0x0019, 0x00c8,
      EVR_SL, "RectangleColumnEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c8, 0x0019, 0x00c8,
      EVR_SS, "SATThicknessRL", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00c9, 0x0019, 0x00c9,
      EVR_SL, "RectangleAngulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00c9, 0x0019, 0x00c9,
      EVR_SS, "SATThicknessAP", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ca, 0x0019, 0x00ca,
      EVR_SL, "IrisCollimatorFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00ca, 0x0019, 0x00ca,
      EVR_SS, "SATThicknessHF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00cb, 0x0019, 0x00cb,
      EVR_SL, "IrisRowStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00cb, 0x0019, 0x00cb,
      EVR_SS, "PhaseContrastFlowAxis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00cb, 0x0019, 0x00cb,
      EVR_SS, "PrescribedFlowAxis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00cc, 0x0019, 0x00cc,
      EVR_SL, "IrisRowEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00cc, 0x0019, 0x00cc,
      EVR_SS, "PhaseContrastVelocityEncoding", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00cc, 0x0019, 0x00cc,
      EVR_SS, "VelocityEncoding", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00cd, 0x0019, 0x00cd,
      EVR_SL, "IrisColumnStart", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00cd, 0x0019, 0x00cd,
      EVR_SS, "ThicknessDisclaimer", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ce, 0x0019, 0x00ce,
      EVR_IS, "REST", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00ce, 0x0019, 0x00ce,
      EVR_SL, "IrisColumnEnd", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00ce, 0x0019, 0x00ce,
      EVR_SS, "PrescanType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00cf, 0x0019, 0x00cf,
      EVR_SL, "IrisAngulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00cf, 0x0019, 0x00cf,
      EVR_SS, "PrescanStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d0, 0x0019, 0x00d0,
      EVR_SH, "RawDataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d1, 0x0019, 0x00d1,
      EVR_SS, "NumberOfFramesPlane", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d1, 0x0019, 0x00d1,
      EVR_DS, "FlowSensitivity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d2, 0x0019, 0x00d2,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d2, 0x0019, 0x00d2,
      EVR_SS, "ProjectionAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d2, 0x0019, 0x00d2,
      EVR_LT, "CalculationSubmode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d3, 0x0019, 0x00d3,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d3, 0x0019, 0x00d3,
      EVR_SH, "ProjectionAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d3, 0x0019, 0x00d3,
      EVR_DS, "FieldOfViewRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d4, 0x0019, 0x00d4,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d4, 0x0019, 0x00d4,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d4, 0x0019, 0x00d4,
      EVR_IS, "BaseRawMatrixSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d5, 0x0019, 0x00d5,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00d5, 0x0019, 0x00d5,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d5, 0x0019, 0x00d5,
      EVR_SS, "FractionalEcho", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00d5, 0x0019, 0x00d5,
      EVR_SS, "FractionalEcho", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d5, 0x0019, 0x00d5,
      EVR_IS, "2DOversamplingLines", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d6, 0x0019, 0x00d6,
      EVR_IS, "FourierInterpolation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00d6, 0x0019, 0x00d6,
      EVR_SS, "Internal", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d6, 0x0019, 0x00d6,
      EVR_SS, "PrepPulse", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d6, 0x0019, 0x00d6,
      EVR_IS, "3DPhaseOversamplingPartitions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d7, 0x0019, 0x00d7,
      EVR_SS, "Internal", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d7, 0x0019, 0x00d7,
      EVR_SS, "CardiacPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d7, 0x0019, 0x00d7,
      EVR_IS, "EchoLinePosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d8, 0x0019, 0x00d8,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d8, 0x0019, 0x00d8,
      EVR_SS, "VariableEchoFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00d8, 0x0019, 0x00d8,
      EVR_SS, "VariableEchoFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d8, 0x0019, 0x00d8,
      EVR_IS, "EchoColumnPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00d9, 0x0019, 0x00d9,
      EVR_IS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00d9, 0x0019, 0x00d9,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00d9, 0x0019, 0x00d9,
      EVR_DS, "ConcatenatedSat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00d9, 0x0019, 0x00d9,
      EVR_DS, "ConcatenatedSAT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00d9, 0x0019, 0x00d9,
      EVR_IS, "LinesPerSegment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00da, 0x0019, 0x00da,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00da, 0x0019, 0x00da,
      EVR_SS, "ReferenceChannelUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00da, 0x0019, 0x00da,
      EVR_LT, "PhaseCodingDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  COAD" }
  , { 0x0019, 0x00db, 0x0019, 0x00db,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00db, 0x0019, 0x00db,
      EVR_DS, "BackProjectorCoefficient", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00dc, 0x0019, 0x00dc,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00dc, 0x0019, 0x00dc,
      EVR_SS, "PrimarySpeedCorrectionUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00dd, 0x0019, 0x00dd,
      EVR_SL, "AnatomicBackground", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00dd, 0x0019, 0x00dd,
      EVR_SS, "OverrangeCorrectionUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00de, 0x0019, 0x00de,
      EVR_SL, "AutoWindowBase", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00de, 0x0019, 0x00de,
      EVR_DS, "DynamicZAlphaValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00df, 0x0019, 0x00df,
      EVR_SS, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00df, 0x0019, 0x00df,
      EVR_DS, "UserData23", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e0, 0x0019, 0x00e0,
      EVR_IS, "Prepulse", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00e0, 0x0019, 0x00e0,
      EVR_SL, "Internal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA PLANE A" }
  , { 0x0019, 0x00e0, 0x0019, 0x00e0,
      EVR_DS, "UserData24", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e1, 0x0019, 0x00e1,
      EVR_DS, "PrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00e1, 0x0019, 0x00e1,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e2, 0x0019, 0x00e2,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00e2, 0x0019, 0x00e2,
      EVR_DS, "VelocityEncodeScale", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e3, 0x0019, 0x00e3,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00e3, 0x0019, 0x00e3,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e4, 0x0019, 0x00e4,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e5, 0x0019, 0x00e5,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e6, 0x0019, 0x00e6,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e8, 0x0019, 0x00e8,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00e9, 0x0019, 0x00e9,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00eb, 0x0019, 0x00eb,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ec, 0x0019, 0x00ec,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f0, 0x0019, 0x00f0,
      EVR_LT, "WSProtocolString1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00f0, 0x0019, 0x00f0,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f1, 0x0019, 0x00f1,
      EVR_LT, "WSProtocolString2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00f1, 0x0019, 0x00f1,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f2, 0x0019, 0x00f2,
      EVR_LT, "WSProtocolString3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00f2, 0x0019, 0x00f2,
      EVR_SS, "NumberOfPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0019, 0x00f2, 0x0019, 0x00f2,
      EVR_SS, "FastPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f3, 0x0019, 0x00f3,
      EVR_LT, "WSProtocolString4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0019, 0x00f3, 0x0019, 0x00f3,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f4, 0x0019, 0x00f4,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00f9, 0x0019, 0x00f9,
      EVR_DS, "TransmitGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACQU_01" }
  , { 0x0019, 0x00ff, 0x0019, 0x00ff,
      EVR_SS, "Internal", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0019, 0x1000, 0x0019, 0x1000,
      EVR_DS, "FieldOfView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR R5.5/PART" }
  , { 0x0019, 0x1000, 0x0019, 0x1000,
      EVR_DS, "FieldOfView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR R5.6/PART" }
  , { 0x0019, 0x1000, 0x0019, 0x1000,
      EVR_DS, "FieldOfView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1005, 0x0019, 0x1005,
      EVR_DS, "CCAngulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1006, 0x0019, 0x1006,
      EVR_DS, "APAngulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1007, 0x0019, 0x1007,
      EVR_DS, "LRAngulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1008, 0x0019, 0x1008,
      EVR_IS, "PatientPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1009, 0x0019, 0x1009,
      EVR_IS, "PatientOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100a, 0x0019, 0x100a,
      EVR_IS, "SliceOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100b, 0x0019, 0x100b,
      EVR_DS, "LROffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100c, 0x0019, 0x100c,
      EVR_DS, "CCOffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100d, 0x0019, 0x100d,
      EVR_DS, "APOffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100e, 0x0019, 0x100e,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x100f, 0x0019, 0x100f,
      EVR_IS, "NumberOfSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1010, 0x0019, 0x1010,
      EVR_DS, "SliceFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1011, 0x0019, 0x1011,
      EVR_DS, "EchoTimes", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1015, 0x0019, 0x1015,
      EVR_IS, "DynamicStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1018, 0x0019, 0x1018,
      EVR_DS, "HeartbeatInterval", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1019, 0x0019, 0x1019,
      EVR_DS, "RepetitionTimeFFE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x101a, 0x0019, 0x101a,
      EVR_DS, "FFEFlipAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x101b, 0x0019, 0x101b,
      EVR_IS, "NumberOfScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1021, 0x0019, 0x1021,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1022, 0x0019, 0x1022,
      EVR_DS, "DynamicScanTimeBegin", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1024, 0x0019, 0x1024,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1064, 0x0019, 0x1064,
      EVR_DS, "RepetitionTimeSE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1065, 0x0019, 0x1065,
      EVR_DS, "RepetitionTimeIR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1069, 0x0019, 0x1069,
      EVR_IS, "NumberOfPhases", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x106a, 0x0019, 0x106a,
      EVR_IS, "CardiacFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x106b, 0x0019, 0x106b,
      EVR_DS, "InversionDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x106c, 0x0019, 0x106c,
      EVR_DS, "GateDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x106d, 0x0019, 0x106d,
      EVR_DS, "GateWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x106e, 0x0019, 0x106e,
      EVR_DS, "TriggerDelayTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1080, 0x0019, 0x1080,
      EVR_IS, "NumberOfChemicalShifts", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1081, 0x0019, 0x1081,
      EVR_DS, "ChemicalShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1084, 0x0019, 0x1084,
      EVR_IS, "NumberOfRows", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1085, 0x0019, 0x1085,
      EVR_IS, "NumberOfSamples", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1094, 0x0019, 0x1094,
      EVR_LO, "MagnetizationTransferContrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1095, 0x0019, 0x1095,
      EVR_LO, "SpectralPresaturationWithInversionRecovery", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1096, 0x0019, 0x1096,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1097, 0x0019, 0x1097,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10a0, 0x0019, 0x10a0,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10a1, 0x0019, 0x10a1,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10a3, 0x0019, 0x10a3,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10a4, 0x0019, 0x10a4,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10c8, 0x0019, 0x10c8,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10c9, 0x0019, 0x10c9,
      EVR_IS, "FoldoverDirectionTransverse", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10ca, 0x0019, 0x10ca,
      EVR_IS, "FoldoverDirectionSagittal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10cb, 0x0019, 0x10cb,
      EVR_IS, "FoldoverDirectionCoronal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10cc, 0x0019, 0x10cc,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10cd, 0x0019, 0x10cd,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10ce, 0x0019, 0x10ce,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10cf, 0x0019, 0x10cf,
      EVR_IS, "NumberOfEchoes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d0, 0x0019, 0x10d0,
      EVR_IS, "ScanResolution", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d2, 0x0019, 0x10d2,
      EVR_LO, "WaterFatShift", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d4, 0x0019, 0x10d4,
      EVR_IS, "ArtifactReduction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d5, 0x0019, 0x10d5,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d6, 0x0019, 0x10d6,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d7, 0x0019, 0x10d7,
      EVR_DS, "ScanPercentage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d8, 0x0019, 0x10d8,
      EVR_IS, "Halfscan", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10d9, 0x0019, 0x10d9,
      EVR_IS, "EPIFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10da, 0x0019, 0x10da,
      EVR_IS, "TurboFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10db, 0x0019, 0x10db,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10e0, 0x0019, 0x10e0,
      EVR_IS, "PercentageOfScanCompleted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x10e1, 0x0019, 0x10e1,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1100, 0x0019, 0x1100,
      EVR_IS, "NumberOfStacks", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1101, 0x0019, 0x1101,
      EVR_IS, "StackType", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1102, 0x0019, 0x1102,
      EVR_IS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x110b, 0x0019, 0x110b,
      EVR_DS, "LROffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x110c, 0x0019, 0x110c,
      EVR_DS, "CCOffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x110d, 0x0019, 0x110d,
      EVR_DS, "APOffcenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x1145, 0x0019, 0x1145,
      EVR_IS, "ReconstructionResolution", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x11fc, 0x0019, 0x11fc,
      EVR_IS, "ResonanceFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x12c0, 0x0019, 0x12c0,
      EVR_DS, "TriggerDelayTimes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x12e0, 0x0019, 0x12e0,
      EVR_IS, "PrepulseType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x12e1, 0x0019, 0x12e1,
      EVR_DS, "PrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0019, 0x12e3, 0x0019, 0x12e3,
      EVR_DS, "PhaseContrastVelocity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
#endif
  , { 0x0020, 0x000d, 0x0020, 0x000d,
      EVR_UI, "StudyInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x000e, 0x0020, 0x000e,
      EVR_UI, "SeriesInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0010, 0x0020, 0x0010,
      EVR_SH, "StudyID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0011, 0x0020, 0x0011,
      EVR_IS, "SeriesNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0012, 0x0020, 0x0012,
      EVR_IS, "AcquisitionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0013, 0x0020, 0x0013,
      EVR_IS, "InstanceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0014, 0x0020, 0x0014,
      EVR_IS, "RETIRED_IsotopeNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0015, 0x0020, 0x0015,
      EVR_IS, "RETIRED_PhaseNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0016, 0x0020, 0x0016,
      EVR_IS, "RETIRED_IntervalNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0017, 0x0020, 0x0017,
      EVR_IS, "RETIRED_TimeSlotNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0018, 0x0020, 0x0018,
      EVR_IS, "RETIRED_AngleNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0019, 0x0020, 0x0019,
      EVR_IS, "ItemNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0020, 0x0020, 0x0020,
      EVR_CS, "PatientOrientation", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0022, 0x0020, 0x0022,
      EVR_IS, "RETIRED_OverlayNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0024, 0x0020, 0x0024,
      EVR_IS, "RETIRED_CurveNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0026, 0x0020, 0x0026,
      EVR_IS, "RETIRED_LUTNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0027, 0x0020, 0x0027,
      EVR_LO, "PyramidLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0030, 0x0020, 0x0030,
      EVR_DS, "RETIRED_ImagePosition", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0032, 0x0020, 0x0032,
      EVR_DS, "ImagePositionPatient", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0035, 0x0020, 0x0035,
      EVR_DS, "RETIRED_ImageOrientation", 6, 6, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0037, 0x0020, 0x0037,
      EVR_DS, "ImageOrientationPatient", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0050, 0x0020, 0x0050,
      EVR_DS, "RETIRED_Location", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0052, 0x0020, 0x0052,
      EVR_UI, "FrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0060, 0x0020, 0x0060,
      EVR_CS, "Laterality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0062, 0x0020, 0x0062,
      EVR_CS, "ImageLaterality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0070, 0x0020, 0x0070,
      EVR_LO, "RETIRED_ImageGeometryType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0080, 0x0020, 0x0080,
      EVR_CS, "RETIRED_MaskingImage", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x00aa, 0x0020, 0x00aa,
      EVR_IS, "RETIRED_ReportNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0100, 0x0020, 0x0100,
      EVR_IS, "TemporalPositionIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0105, 0x0020, 0x0105,
      EVR_IS, "NumberOfTemporalPositions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0110, 0x0020, 0x0110,
      EVR_DS, "TemporalResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0200, 0x0020, 0x0200,
      EVR_UI, "SynchronizationFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x0242, 0x0020, 0x0242,
      EVR_UI, "SOPInstanceUIDOfConcatenationSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1000, 0x0020, 0x1000,
      EVR_IS, "RETIRED_SeriesInStudy", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1001, 0x0020, 0x1001,
      EVR_IS, "RETIRED_AcquisitionsInSeries", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1002, 0x0020, 0x1002,
      EVR_IS, "ImagesInAcquisition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1003, 0x0020, 0x1003,
      EVR_IS, "RETIRED_ImagesInSeries", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1004, 0x0020, 0x1004,
      EVR_IS, "RETIRED_AcquisitionsInStudy", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1005, 0x0020, 0x1005,
      EVR_IS, "RETIRED_ImagesInStudy", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1020, 0x0020, 0x1020,
      EVR_LO, "RETIRED_Reference", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x103f, 0x0020, 0x103f,
      EVR_LO, "TargetPositionReferenceIndicator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1040, 0x0020, 0x1040,
      EVR_LO, "PositionReferenceIndicator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1041, 0x0020, 0x1041,
      EVR_DS, "SliceLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1070, 0x0020, 0x1070,
      EVR_IS, "RETIRED_OtherStudyNumbers", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1200, 0x0020, 0x1200,
      EVR_IS, "NumberOfPatientRelatedStudies", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1202, 0x0020, 0x1202,
      EVR_IS, "NumberOfPatientRelatedSeries", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1204, 0x0020, 0x1204,
      EVR_IS, "NumberOfPatientRelatedInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1206, 0x0020, 0x1206,
      EVR_IS, "NumberOfStudyRelatedSeries", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1208, 0x0020, 0x1208,
      EVR_IS, "NumberOfStudyRelatedInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x1209, 0x0020, 0x1209,
      EVR_IS, "NumberOfSeriesRelatedInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3401, 0x0020, 0x3401,
      EVR_CS, "RETIRED_ModifyingDeviceID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3402, 0x0020, 0x3402,
      EVR_CS, "RETIRED_ModifiedImageID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3403, 0x0020, 0x3403,
      EVR_DA, "RETIRED_ModifiedImageDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3404, 0x0020, 0x3404,
      EVR_LO, "RETIRED_ModifyingDeviceManufacturer", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3405, 0x0020, 0x3405,
      EVR_TM, "RETIRED_ModifiedImageTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3406, 0x0020, 0x3406,
      EVR_LO, "RETIRED_ModifiedImageDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x4000, 0x0020, 0x4000,
      EVR_LT, "ImageComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x5000, 0x0020, 0x5000,
      EVR_AT, "RETIRED_OriginalImageIdentification", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x5002, 0x0020, 0x5002,
      EVR_LO, "RETIRED_OriginalImageIdentificationNomenclature", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9056, 0x0020, 0x9056,
      EVR_SH, "StackID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9057, 0x0020, 0x9057,
      EVR_UL, "InStackPositionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9071, 0x0020, 0x9071,
      EVR_SQ, "FrameAnatomySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9072, 0x0020, 0x9072,
      EVR_CS, "FrameLaterality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9111, 0x0020, 0x9111,
      EVR_SQ, "FrameContentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9113, 0x0020, 0x9113,
      EVR_SQ, "PlanePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9116, 0x0020, 0x9116,
      EVR_SQ, "PlaneOrientationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9128, 0x0020, 0x9128,
      EVR_UL, "TemporalPositionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9153, 0x0020, 0x9153,
      EVR_FD, "NominalCardiacTriggerDelayTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9154, 0x0020, 0x9154,
      EVR_FL, "NominalCardiacTriggerTimePriorToRPeak", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9155, 0x0020, 0x9155,
      EVR_FL, "ActualCardiacTriggerTimePriorToRPeak", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9156, 0x0020, 0x9156,
      EVR_US, "FrameAcquisitionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9157, 0x0020, 0x9157,
      EVR_UL, "DimensionIndexValues", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9158, 0x0020, 0x9158,
      EVR_LT, "FrameComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9161, 0x0020, 0x9161,
      EVR_UI, "ConcatenationUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9162, 0x0020, 0x9162,
      EVR_US, "InConcatenationNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9163, 0x0020, 0x9163,
      EVR_US, "InConcatenationTotalNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9164, 0x0020, 0x9164,
      EVR_UI, "DimensionOrganizationUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9165, 0x0020, 0x9165,
      EVR_AT, "DimensionIndexPointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9167, 0x0020, 0x9167,
      EVR_AT, "FunctionalGroupPointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9170, 0x0020, 0x9170,
      EVR_SQ, "UnassignedSharedConvertedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9171, 0x0020, 0x9171,
      EVR_SQ, "UnassignedPerFrameConvertedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9172, 0x0020, 0x9172,
      EVR_SQ, "ConversionSourceAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9213, 0x0020, 0x9213,
      EVR_LO, "DimensionIndexPrivateCreator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9221, 0x0020, 0x9221,
      EVR_SQ, "DimensionOrganizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9222, 0x0020, 0x9222,
      EVR_SQ, "DimensionIndexSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9228, 0x0020, 0x9228,
      EVR_UL, "ConcatenationFrameOffsetNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9238, 0x0020, 0x9238,
      EVR_LO, "FunctionalGroupPrivateCreator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9241, 0x0020, 0x9241,
      EVR_FL, "NominalPercentageOfCardiacPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9245, 0x0020, 0x9245,
      EVR_FL, "NominalPercentageOfRespiratoryPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9246, 0x0020, 0x9246,
      EVR_FL, "StartingRespiratoryAmplitude", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9247, 0x0020, 0x9247,
      EVR_CS, "StartingRespiratoryPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9248, 0x0020, 0x9248,
      EVR_FL, "EndingRespiratoryAmplitude", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9249, 0x0020, 0x9249,
      EVR_CS, "EndingRespiratoryPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9250, 0x0020, 0x9250,
      EVR_CS, "RespiratoryTriggerType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9251, 0x0020, 0x9251,
      EVR_FD, "RRIntervalTimeNominal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9252, 0x0020, 0x9252,
      EVR_FD, "ActualCardiacTriggerDelayTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9253, 0x0020, 0x9253,
      EVR_SQ, "RespiratorySynchronizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9254, 0x0020, 0x9254,
      EVR_FD, "RespiratoryIntervalTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9255, 0x0020, 0x9255,
      EVR_FD, "NominalRespiratoryTriggerDelayTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9256, 0x0020, 0x9256,
      EVR_FD, "RespiratoryTriggerDelayThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9257, 0x0020, 0x9257,
      EVR_FD, "ActualRespiratoryTriggerDelayTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9301, 0x0020, 0x9301,
      EVR_FD, "ImagePositionVolume", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9302, 0x0020, 0x9302,
      EVR_FD, "ImageOrientationVolume", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9307, 0x0020, 0x9307,
      EVR_CS, "UltrasoundAcquisitionGeometry", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9308, 0x0020, 0x9308,
      EVR_FD, "ApexPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9309, 0x0020, 0x9309,
      EVR_FD, "VolumeToTransducerMappingMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930a, 0x0020, 0x930a,
      EVR_FD, "VolumeToTableMappingMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930b, 0x0020, 0x930b,
      EVR_CS, "VolumeToTransducerRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930c, 0x0020, 0x930c,
      EVR_CS, "PatientFrameOfReferenceSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930d, 0x0020, 0x930d,
      EVR_FD, "TemporalPositionTimeOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930e, 0x0020, 0x930e,
      EVR_SQ, "PlanePositionVolumeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x930f, 0x0020, 0x930f,
      EVR_SQ, "PlaneOrientationVolumeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9310, 0x0020, 0x9310,
      EVR_SQ, "TemporalPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9311, 0x0020, 0x9311,
      EVR_CS, "DimensionOrganizationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9312, 0x0020, 0x9312,
      EVR_UI, "VolumeFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9313, 0x0020, 0x9313,
      EVR_UI, "TableFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9421, 0x0020, 0x9421,
      EVR_LO, "DimensionDescriptionLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9450, 0x0020, 0x9450,
      EVR_SQ, "PatientOrientationInFrameSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9453, 0x0020, 0x9453,
      EVR_LO, "FrameLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9518, 0x0020, 0x9518,
      EVR_US, "AcquisitionIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9529, 0x0020, 0x9529,
      EVR_SQ, "ContributingSOPInstancesReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x9536, 0x0020, 0x9536,
      EVR_US, "ReconstructionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0021, 0x0000, 0x0021, 0x0000,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0021, 0x0000, 0x0021, 0x0000,
      EVR_LT, "WindowOfImagesID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-CWS Release 1" }
  , { 0x0021, 0x0000, 0x0021, 0x0000,
      EVR_DS, "NoiseReduction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0000, 0x0021, 0x0000,
      EVR_LT, "SequenceType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0000, 0x0021, 0x0000,
      EVR_US, "AcquisitionType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_CS, "WindowOfImagesType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-CWS Release 1" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_IS, "ReconstructionNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS-MR-1" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_DS, "ContrastAmplification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_IS, "VectorSizeOriginal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0001, 0x0021, 0x0001,
      EVR_US, "AcquisitionMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_IS, "WindowOfImagesScope", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private-CWS Release 1" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_IS, "SliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS-MR-1" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_DS, "EdgeContrastBoosting", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_IS, "VectorSizeExtended", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0002, 0x0021, 0x0002,
      EVR_US, "FootswitchIndex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0003, 0x0021, 0x0003,
      EVR_SS, "SeriesFromWhichPrescribed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0003, 0x0021, 0x0003,
      EVR_DS, "LatitudeReduction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0003, 0x0021, 0x0003,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x0021, 0x0003, 0x0021, 0x0003,
      EVR_DS, "AcquiredSpectralRange", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0003, 0x0021, 0x0003,
      EVR_US, "AcquisitionRoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0004, 0x0021, 0x0004,
      EVR_DS, "VOIPosition", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0004, 0x0021, 0x0004,
      EVR_SL, "CurrentTimeProduct", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0005, 0x0021, 0x0005,
      EVR_SH, "GenesisVersionNow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0005, 0x0021, 0x0005,
      EVR_DS, "VOISize", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0005, 0x0021, 0x0005,
      EVR_SL, "Dose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0006, 0x0021, 0x0006,
      EVR_IS, "CSIMatrixSizeOriginal", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0006, 0x0021, 0x0006,
      EVR_SL, "SkinDosePercent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0007, 0x0021, 0x0007,
      EVR_UL, "SeriesRecordChecksum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0007, 0x0021, 0x0007,
      EVR_IS, "CSIMatrixSizeExtended", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0007, 0x0021, 0x0007,
      EVR_SL, "SkinDoseAccumulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0008, 0x0021, 0x0008,
      EVR_DS, "SpatialGridShift", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0008, 0x0021, 0x0008,
      EVR_SL, "SkinDoseRate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0009, 0x0021, 0x0009,
      EVR_DS, "SignalLimitsMinimum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x000a, 0x0021, 0x000a,
      EVR_UL, "CopperFilter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x000b, 0x0021, 0x000b,
      EVR_US, "MeasuringField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x000c, 0x0021, 0x000c,
      EVR_SS, "PostBlankingCircle", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x000d, 0x0021, 0x000d,
      EVR_SS, "DynaAngles", 2, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x000e, 0x0021, 0x000e,
      EVR_SS, "TotalSteps", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x000f, 0x0021, 0x000f,
      EVR_SL, "DynaXRayInfo", 3, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_UL, "CreationMask", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_LO, "FindRangeAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_IS, "RotationAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_DS, "Zoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_DS, "SignalLimitsMaximum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_CS, "FCRImageID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0021, 0x0010, 0x0021, 0x0010,
      EVR_US, "ModalityLUTInputGamma", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0011, 0x0021, 0x0011,
      EVR_DS, "ThresholdCAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0011, 0x0021, 0x0011,
      EVR_IS, "StartAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0011, 0x0021, 0x0011,
      EVR_DS, "Target", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0021, 0x0011, 0x0021, 0x0011,
      EVR_DS, "SpecInfoMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0011, 0x0021, 0x0011,
      EVR_US, "ModalityLUTOutputGamma", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0012, 0x0021, 0x0012,
      EVR_IS, "ExposureNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0021, 0x0012, 0x0021, 0x0012,
      EVR_LT, "SeriesUniqueIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0021, 0x0012, 0x0021, 0x0012,
      EVR_IS, "TubeAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0021, 0x0012, 0x0021, 0x0012,
      EVR_DS, "EPITimeRateOfChangeOfMagnitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0012, 0x0021, 0x0012,
      EVR_OB, "SH_STPAR", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0013, 0x0021, 0x0013,
      EVR_IS, "ImageSequenceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0021, 0x0013, 0x0021, 0x0013,
      EVR_DS, "EPITimeRateOfChangeOfXComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0013, 0x0021, 0x0013,
      EVR_US, "AcquisitionZoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0014, 0x0021, 0x0014,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0021, 0x0014, 0x0021, 0x0014,
      EVR_DS, "EPITimeRateOfChangeOfYComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0014, 0x0021, 0x0014,
      EVR_SS, "DynaAngulationStepWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0015, 0x0021, 0x0015,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0015, 0x0021, 0x0015,
      EVR_SS, "ImagesInStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0015, 0x0021, 0x0015,
      EVR_DS, "EPITimeRateOfChangeOfZComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0015, 0x0021, 0x0015,
      EVR_US, "Harmonization", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0016, 0x0021, 0x0016,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0016, 0x0021, 0x0016,
      EVR_DS, "EPITimeRateOfChangeLegalLimit1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0016, 0x0021, 0x0016,
      EVR_US, "DRSingleFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0017, 0x0021, 0x0017,
      EVR_DS, "EPIOperationModeFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0017, 0x0021, 0x0017,
      EVR_SL, "SourceToIsocenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0018, 0x0021, 0x0018,
      EVR_SH, "GenesisVersionNow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0018, 0x0021, 0x0018,
      EVR_DS, "EPIFieldCalculationSafetyFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0018, 0x0021, 0x0018,
      EVR_US, "PressureData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0019, 0x0021, 0x0019,
      EVR_UL, "AcqReconRecordChecksum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0019, 0x0021, 0x0019,
      EVR_DS, "EPILegalLimit1OfChangeValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0019, 0x0021, 0x0019,
      EVR_SL, "ECGIndexArray", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001a, 0x0021, 0x001a,
      EVR_US, "FDFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001b, 0x0021, 0x001b,
      EVR_OB, "SH_ZOOM", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001c, 0x0021, 0x001c,
      EVR_OB, "SH_COLPAR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001d, 0x0021, 0x001d,
      EVR_US, "K_Factor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001e, 0x0021, 0x001e,
      EVR_US, "EVE", 8, 8, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x001f, 0x0021, 0x001f,
      EVR_SL, "TotalSceneTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_DS, "FoV", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_UL, "EvaluationMask", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_DS, "TableStartLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_LO, "SensometricCurve", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_SS, "ScenesInStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_US, "ROIMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_DS, "EPILegalLimit2OfChangeValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_IS, "PhaseCorrectionRowsSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0020, 0x0021, 0x0020,
      EVR_US, "RestoreFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0021, 0x0021, 0x0021,
      EVR_DS, "SliceGap", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0021, 0x0021, 0x0021, 0x0021,
      EVR_DS, "EPIRiseTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0021, 0x0021, 0x0021,
      EVR_IS, "PhaseCorrectionColumnsSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0021, 0x0021, 0x0021,
      EVR_US, "StandMovementFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0022, 0x0021, 0x0022,
      EVR_DS, "ImageMagnificationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0022, 0x0021, 0x0022,
      EVR_DS, "StackRadialAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0021, 0x0022, 0x0021, 0x0022,
      EVR_IS, "PhaseCorrectionRowsReconstruction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0022, 0x0021, 0x0022,
      EVR_US, "FDRows", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0023, 0x0021, 0x0023,
      EVR_US, "FDColumns", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0024, 0x0021, 0x0024,
      EVR_DS, "ImageScrollOffset", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0024, 0x0021, 0x0024,
      EVR_IS, "PhaseCorrectionColumnsReconstruction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0024, 0x0021, 0x0024,
      EVR_US, "TableMovementFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0025, 0x0021, 0x0025,
      EVR_SS, "ImagesInPhotofile", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0025, 0x0021, 0x0025,
      EVR_LO, "OriginalOrganProgramName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0026, 0x0021, 0x0026,
      EVR_IS, "ImagePixelOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0026, 0x0021, 0x0026,
      EVR_DS, "CrispyXPIFilter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ACQ 1.0" }
  , { 0x0021, 0x0027, 0x0021, 0x0027,
      EVR_SS, "PlaneBImagesExist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0028, 0x0021, 0x0028,
      EVR_SS, "NoOf2MBChunks", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_LT, "ViewDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_US, "ExtendedProcessingMask", 7, 7, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_DS, "LowerWindowOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_IS, "TopogramTubePosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_SS, "ImagesInAllScenes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_DS, "ArrayCoilADCOffset", 16, 16, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_IS, "NumberOf3DRawPartitionsNominal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0030, 0x0021, 0x0030,
      EVR_CS, "SetNo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0021, 0x0031, 0x0021, 0x0031,
      EVR_DS, "UpperWindowOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0031, 0x0021, 0x0031,
      EVR_DS, "ArrayCoilPreamplifierGain", 16, 16, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0031, 0x0021, 0x0031,
      EVR_IS, "NumberOf3DRawPartitionsCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0032, 0x0021, 0x0032,
      EVR_CS, "PatientRestDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0032, 0x0021, 0x0032,
      EVR_DS, "LengthOfTopogram", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0034, 0x0021, 0x0034,
      EVR_DS, "TopogramCorrectionFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0034, 0x0021, 0x0034,
      EVR_IS, "NumberOf3DImagePartitions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0035, 0x0021, 0x0035,
      EVR_SS, "SeriesFromWhichPrescribed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0036, 0x0021, 0x0036,
      EVR_SS, "ImageFromWhichPrescribed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0036, 0x0021, 0x0036,
      EVR_DS, "MaximumTablePosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0036, 0x0021, 0x0036,
      EVR_IS, "Actual3DImagePartitionNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0037, 0x0021, 0x0037,
      EVR_SS, "ScreenFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0039, 0x0021, 0x0039,
      EVR_DS, "SlabThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_CDS Release 1" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_SS, "ArchiveSWInternalVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RA GEN" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_IS, "TableMoveDirectionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_DS, "MinPrintableDensity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_IS, "NumberOfSlicesNominal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0040, 0x0021, 0x0040,
      EVR_IS, "ImageNoInTheSet", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0021, 0x0041, 0x0021, 0x0041,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0041, 0x0021, 0x0041,
      EVR_DS, "MaxPrintableDensity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0041, 0x0021, 0x0041,
      EVR_IS, "NumberOfSlicesCurrent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0042, 0x0021, 0x0042,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0042, 0x0021, 0x0042,
      EVR_IS, "CurrentSliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0043, 0x0021, 0x0043,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0043, 0x0021, 0x0043,
      EVR_IS, "CurrentGroupNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0044, 0x0021, 0x0044,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0044, 0x0021, 0x0044,
      EVR_DS, "CurrentSliceDistanceFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0045, 0x0021, 0x0045,
      EVR_IS, "VOIStartRow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0045, 0x0021, 0x0045,
      EVR_IS, "MIPStartRow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0046, 0x0021, 0x0046,
      EVR_IS, "VOIStopRow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0046, 0x0021, 0x0046,
      EVR_IS, "MIPStopRow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0047, 0x0021, 0x0047,
      EVR_IS, "VOIStartColumn", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0047, 0x0021, 0x0047,
      EVR_IS, "MIPStartColumn", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0048, 0x0021, 0x0048,
      EVR_IS, "VOIStopColumn", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0048, 0x0021, 0x0048,
      EVR_IS, "MIPStartColumn", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0049, 0x0021, 0x0049,
      EVR_IS, "VOIStartSlice", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0049, 0x0021, 0x0049,
      EVR_IS, "MIPStartSliceName=", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x004a, 0x0021, 0x004a,
      EVR_LO, "AnatomicalReferenceForScout", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x004a, 0x0021, 0x004a,
      EVR_IS, "VOIStopSlice", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x004a, 0x0021, 0x004a,
      EVR_IS, "MIPStartSlice", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x004e, 0x0021, 0x004e,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x004f, 0x0021, 0x004f,
      EVR_SS, "LocationsInAcquisition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x004f, 0x0021, 0x004f,
      EVR_LT, "OrderofSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  RAW" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_SS, "GraphicallyPrescribed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_IS, "VectorStartRow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_LT, "SaturationType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_US, "SignalMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0050, 0x0021, 0x0050,
      EVR_CS, "PairProcessingInformation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0021, 0x0051, 0x0021, 0x0051,
      EVR_DS, "RotationFromSourceXRot", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0051, 0x0021, 0x0051,
      EVR_IS, "VectorRowStep", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0051, 0x0021, 0x0051,
      EVR_DS, "SaturationNormalVector", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0052, 0x0021, 0x0052,
      EVR_DS, "RotationFromSourceYRot", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0052, 0x0021, 0x0052,
      EVR_IS, "VectorStartColumn", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0052, 0x0021, 0x0052,
      EVR_DS, "SaturationPositionVector", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0052, 0x0021, 0x0052,
      EVR_DS, "DelayAfterTrigger", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0053, 0x0021, 0x0053,
      EVR_DS, "RotationFromSourceZRot", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0053, 0x0021, 0x0053,
      EVR_IS, "VectorColumnStep", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0053, 0x0021, 0x0053,
      EVR_DS, "SaturationThickness", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0053, 0x0021, 0x0053,
      EVR_IS, "RRInterval", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0054, 0x0021, 0x0054,
      EVR_SH, "ImagePosition", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0054, 0x0021, 0x0054,
      EVR_DS, "SaturationWidth", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0054, 0x0021, 0x0054,
      EVR_DS, "NumberOfTriggerPulses", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0055, 0x0021, 0x0055,
      EVR_SH, "ImageOrientation", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0055, 0x0021, 0x0055,
      EVR_DS, "SaturationDistance", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  RAW" }
  , { 0x0021, 0x0056, 0x0021, 0x0056,
      EVR_SL, "IntegerSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0056, 0x0021, 0x0056,
      EVR_DS, "RepetitionTimeEffective", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0057, 0x0021, 0x0057,
      EVR_SL, "IntegerSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0057, 0x0021, 0x0057,
      EVR_LT, "GatePhase", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0058, 0x0021, 0x0058,
      EVR_SL, "IntegerSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0058, 0x0021, 0x0058,
      EVR_DS, "GateThreshold", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0059, 0x0021, 0x0059,
      EVR_SL, "IntegerSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0059, 0x0021, 0x0059,
      EVR_DS, "GatedRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x005a, 0x0021, 0x005a,
      EVR_SL, "IntegerSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x005b, 0x0021, 0x005b,
      EVR_DS, "FloatSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x005c, 0x0021, 0x005c,
      EVR_DS, "FloatSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x005d, 0x0021, 0x005d,
      EVR_DS, "FloatSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x005e, 0x0021, 0x005e,
      EVR_DS, "FloatSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x005f, 0x0021, 0x005f,
      EVR_DS, "FloatSlop", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0060, 0x0021, 0x0060,
      EVR_DS, "ImagePosition", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0060, 0x0021, 0x0060,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0060, 0x0021, 0x0060,
      EVR_IS, "RangeTypeCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0060, 0x0021, 0x0060,
      EVR_IS, "NumberOfInterpolatedImages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0061, 0x0021, 0x0061,
      EVR_DS, "ImageNormal", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0062, 0x0021, 0x0062,
      EVR_IS, "ReferenceTypeCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0063, 0x0021, 0x0063,
      EVR_DS, "ImageDistance", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0065, 0x0021, 0x0065,
      EVR_US, "ImagePositioningHistoryMask", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x006a, 0x0021, 0x006a,
      EVR_DS, "ImageRow", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x006b, 0x0021, 0x006b,
      EVR_DS, "ImageColumn", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0070, 0x0021, 0x0070,
      EVR_LT, "PatientOrientationSet1", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0070, 0x0021, 0x0070,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0070, 0x0021, 0x0070,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0070, 0x0021, 0x0070,
      EVR_DS, "ObjectOrientation", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0070, 0x0021, 0x0070,
      EVR_IS, "NumberOfEchoes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0071, 0x0021, 0x0071,
      EVR_LT, "PatientOrientationSet2", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0071, 0x0021, 0x0071,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0072, 0x0021, 0x0072,
      EVR_DS, "LightOrientation", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0072, 0x0021, 0x0072,
      EVR_DS, "SecondEchoTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0073, 0x0021, 0x0073,
      EVR_DS, "SecondRepetitionTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0075, 0x0021, 0x0075,
      EVR_DS, "LightBrightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0076, 0x0021, 0x0076,
      EVR_DS, "LightContrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x007a, 0x0021, 0x007a,
      EVR_IS, "OverlayThreshold", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x007b, 0x0021, 0x007b,
      EVR_IS, "SurfaceThreshold", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x007c, 0x0021, 0x007c,
      EVR_IS, "GreyScaleThreshold", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x0080, 0x0021, 0x0080,
      EVR_LT, "StudyName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0080, 0x0021, 0x0080,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0080, 0x0021, 0x0080,
      EVR_OB, "EquipmentTypeSpecificInformation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0021, 0x0080, 0x0021, 0x0080,
      EVR_IS, "CardiacCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0081, 0x0021, 0x0081,
      EVR_DS, "AutoWindowLevelAlpha", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0082, 0x0021, 0x0082,
      EVR_LT, "StudyType", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0021, 0x0082, 0x0021, 0x0082,
      EVR_DS, "AutoWindowLevelBeta", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0083, 0x0021, 0x0083,
      EVR_DS, "AutoWindowLevelWindow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0084, 0x0021, 0x0084,
      EVR_DS, "AutoWindowLevelLevel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0090, 0x0021, 0x0090,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x0090, 0x0021, 0x0090,
      EVR_SS, "TubeFocalSpotPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0090, 0x0021, 0x0090,
      EVR_DS, "Brightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0091, 0x0021, 0x0091,
      EVR_SS, "BiopsyPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0091, 0x0021, 0x0091,
      EVR_DS, "Contrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0091, 0x0021, 0x0091,
      EVR_DS, "SaturationPhaseEncodingVectorTransverseComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0092, 0x0021, 0x0092,
      EVR_FL, "BiopsyTLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0092, 0x0021, 0x0092,
      EVR_DS, "ShapeFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0021, 0x0092, 0x0021, 0x0092,
      EVR_DS, "SaturationReadoutVectorTransverseComponent", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0093, 0x0021, 0x0093,
      EVR_FL, "BiopsyRefLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_RELA_01" }
  , { 0x0021, 0x0093, 0x0021, 0x0093,
      EVR_DS, "EPIChangeValueOfMagnitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0094, 0x0021, 0x0094,
      EVR_DS, "EPIChangeValueOfXComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0095, 0x0021, 0x0095,
      EVR_DS, "EPIChangeValueOfYComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x0096, 0x0021, 0x0096,
      EVR_DS, "EPIChangeValueOfZComponent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MR VA0  GEN" }
  , { 0x0021, 0x00a0, 0x0021, 0x00a0,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x00a0, 0x0021, 0x00a0,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00a1, 0x0021, 0x00a1,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00a2, 0x0021, 0x00a2,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x00a2, 0x0021, 0x00a2,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00a3, 0x0021, 0x00a3,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00a4, 0x0021, 0x00a4,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00a7, 0x0021, 0x00a7,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  GEN" }
  , { 0x0021, 0x00b0, 0x0021, 0x00b0,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x00c0, 0x0021, 0x00c0,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0021, 0x1000, 0x0021, 0x1000,
      EVR_IS, "ReconstructionNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1010, 0x0021, 0x1010,
      EVR_IS, "ImageType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1020, 0x0021, 0x1020,
      EVR_IS, "SliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1030, 0x0021, 0x1030,
      EVR_IS, "EchoNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1031, 0x0021, 0x1031,
      EVR_DS, "PatientReferenceID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1035, 0x0021, 0x1035,
      EVR_IS, "ChemicalShiftNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1040, 0x0021, 0x1040,
      EVR_IS, "PhaseNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1050, 0x0021, 0x1050,
      EVR_IS, "DynamicScanNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1060, 0x0021, 0x1060,
      EVR_IS, "NumberOfRowsInObject", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1061, 0x0021, 0x1061,
      EVR_IS, "RowNumber", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1062, 0x0021, 0x1062,
      EVR_IS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1100, 0x0021, 0x1100,
      EVR_DA, "ScanDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1110, 0x0021, 0x1110,
      EVR_TM, "ScanTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0021, 0x1221, 0x0021, 0x1221,
      EVR_IS, "SliceGap", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
#endif
  , { 0x0022, 0x0001, 0x0022, 0x0001,
      EVR_US, "LightPathFilterPassThroughWavelength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0002, 0x0022, 0x0002,
      EVR_US, "LightPathFilterPassBand", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0003, 0x0022, 0x0003,
      EVR_US, "ImagePathFilterPassThroughWavelength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0004, 0x0022, 0x0004,
      EVR_US, "ImagePathFilterPassBand", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0005, 0x0022, 0x0005,
      EVR_CS, "PatientEyeMovementCommanded", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0006, 0x0022, 0x0006,
      EVR_SQ, "PatientEyeMovementCommandCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0007, 0x0022, 0x0007,
      EVR_FL, "SphericalLensPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0008, 0x0022, 0x0008,
      EVR_FL, "CylinderLensPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0009, 0x0022, 0x0009,
      EVR_FL, "CylinderAxis", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000a, 0x0022, 0x000a,
      EVR_FL, "EmmetropicMagnification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000b, 0x0022, 0x000b,
      EVR_FL, "IntraOcularPressure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000c, 0x0022, 0x000c,
      EVR_FL, "HorizontalFieldOfView", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000d, 0x0022, 0x000d,
      EVR_CS, "PupilDilated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000e, 0x0022, 0x000e,
      EVR_FL, "DegreeOfDilation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x000f, 0x0022, 0x000f,
      EVR_FD, "VertexDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0010, 0x0022, 0x0010,
      EVR_FL, "StereoBaselineAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0011, 0x0022, 0x0011,
      EVR_FL, "StereoBaselineDisplacement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0012, 0x0022, 0x0012,
      EVR_FL, "StereoHorizontalPixelOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0013, 0x0022, 0x0013,
      EVR_FL, "StereoVerticalPixelOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0014, 0x0022, 0x0014,
      EVR_FL, "StereoRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0015, 0x0022, 0x0015,
      EVR_SQ, "AcquisitionDeviceTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0016, 0x0022, 0x0016,
      EVR_SQ, "IlluminationTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0017, 0x0022, 0x0017,
      EVR_SQ, "LightPathFilterTypeStackCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0018, 0x0022, 0x0018,
      EVR_SQ, "ImagePathFilterTypeStackCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0019, 0x0022, 0x0019,
      EVR_SQ, "LensesCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x001a, 0x0022, 0x001a,
      EVR_SQ, "ChannelDescriptionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x001b, 0x0022, 0x001b,
      EVR_SQ, "RefractiveStateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x001c, 0x0022, 0x001c,
      EVR_SQ, "MydriaticAgentCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x001d, 0x0022, 0x001d,
      EVR_SQ, "RelativeImagePositionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x001e, 0x0022, 0x001e,
      EVR_FL, "CameraAngleOfView", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0020, 0x0022, 0x0020,
      EVR_SQ, "StereoPairsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0021, 0x0022, 0x0021,
      EVR_SQ, "LeftImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0022, 0x0022, 0x0022,
      EVR_SQ, "RightImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0028, 0x0022, 0x0028,
      EVR_CS, "StereoPairsPresent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0030, 0x0022, 0x0030,
      EVR_FL, "AxialLengthOfTheEye", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0031, 0x0022, 0x0031,
      EVR_SQ, "OphthalmicFrameLocationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0032, 0x0022, 0x0032,
      EVR_FL, "ReferenceCoordinates", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0035, 0x0022, 0x0035,
      EVR_FL, "DepthSpatialResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0036, 0x0022, 0x0036,
      EVR_FL, "MaximumDepthDistortion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0037, 0x0022, 0x0037,
      EVR_FL, "AlongScanSpatialResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0038, 0x0022, 0x0038,
      EVR_FL, "MaximumAlongScanDistortion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0039, 0x0022, 0x0039,
      EVR_CS, "OphthalmicImageOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0041, 0x0022, 0x0041,
      EVR_FL, "DepthOfTransverseImage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0042, 0x0022, 0x0042,
      EVR_SQ, "MydriaticAgentConcentrationUnitsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0048, 0x0022, 0x0048,
      EVR_FL, "AcrossScanSpatialResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0049, 0x0022, 0x0049,
      EVR_FL, "MaximumAcrossScanDistortion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x004e, 0x0022, 0x004e,
      EVR_DS, "MydriaticAgentConcentration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0055, 0x0022, 0x0055,
      EVR_FL, "IlluminationWaveLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0056, 0x0022, 0x0056,
      EVR_FL, "IlluminationPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0057, 0x0022, 0x0057,
      EVR_FL, "IlluminationBandwidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x0058, 0x0022, 0x0058,
      EVR_SQ, "MydriaticAgentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1007, 0x0022, 0x1007,
      EVR_SQ, "OphthalmicAxialMeasurementsRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1008, 0x0022, 0x1008,
      EVR_SQ, "OphthalmicAxialMeasurementsLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1009, 0x0022, 0x1009,
      EVR_CS, "OphthalmicAxialMeasurementsDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1010, 0x0022, 0x1010,
      EVR_CS, "OphthalmicAxialLengthMeasurementsType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1012, 0x0022, 0x1012,
      EVR_SQ, "OphthalmicAxialLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1019, 0x0022, 0x1019,
      EVR_FL, "OphthalmicAxialLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1024, 0x0022, 0x1024,
      EVR_SQ, "LensStatusCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1025, 0x0022, 0x1025,
      EVR_SQ, "VitreousStatusCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1028, 0x0022, 0x1028,
      EVR_SQ, "IOLFormulaCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1029, 0x0022, 0x1029,
      EVR_LO, "IOLFormulaDetail", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1033, 0x0022, 0x1033,
      EVR_FL, "KeratometerIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1035, 0x0022, 0x1035,
      EVR_SQ, "SourceOfOphthalmicAxialLengthCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1036, 0x0022, 0x1036,
      EVR_SQ, "SourceOfCornealSizeDataCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1037, 0x0022, 0x1037,
      EVR_FL, "TargetRefraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1039, 0x0022, 0x1039,
      EVR_CS, "RefractiveProcedureOccurred", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1040, 0x0022, 0x1040,
      EVR_SQ, "RefractiveSurgeryTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1044, 0x0022, 0x1044,
      EVR_SQ, "OphthalmicUltrasoundMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1045, 0x0022, 0x1045,
      EVR_SQ, "SurgicallyInducedAstigmatismSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1046, 0x0022, 0x1046,
      EVR_CS, "TypeOfOpticalCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1047, 0x0022, 0x1047,
      EVR_SQ, "ToricIOLPowerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1048, 0x0022, 0x1048,
      EVR_SQ, "PredictedToricErrorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1049, 0x0022, 0x1049,
      EVR_CS, "PreSelectedForImplantation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x104a, 0x0022, 0x104a,
      EVR_SQ, "ToricIOLPowerForExactEmmetropiaSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x104b, 0x0022, 0x104b,
      EVR_SQ, "ToricIOLPowerForExactTargetRefractionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1050, 0x0022, 0x1050,
      EVR_SQ, "OphthalmicAxialLengthMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1053, 0x0022, 0x1053,
      EVR_FL, "IOLPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1054, 0x0022, 0x1054,
      EVR_FL, "PredictedRefractiveError", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1059, 0x0022, 0x1059,
      EVR_FL, "OphthalmicAxialLengthVelocity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1065, 0x0022, 0x1065,
      EVR_LO, "LensStatusDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1066, 0x0022, 0x1066,
      EVR_LO, "VitreousStatusDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1090, 0x0022, 0x1090,
      EVR_SQ, "IOLPowerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1092, 0x0022, 0x1092,
      EVR_SQ, "LensConstantSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1093, 0x0022, 0x1093,
      EVR_LO, "IOLManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1094, 0x0022, 0x1094,
      EVR_LO, "RETIRED_LensConstantDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1095, 0x0022, 0x1095,
      EVR_LO, "ImplantName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1096, 0x0022, 0x1096,
      EVR_SQ, "KeratometryMeasurementTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1097, 0x0022, 0x1097,
      EVR_LO, "ImplantPartNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1100, 0x0022, 0x1100,
      EVR_SQ, "ReferencedOphthalmicAxialMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1101, 0x0022, 0x1101,
      EVR_SQ, "OphthalmicAxialLengthMeasurementsSegmentNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1103, 0x0022, 0x1103,
      EVR_SQ, "RefractiveErrorBeforeRefractiveSurgeryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1121, 0x0022, 0x1121,
      EVR_FL, "IOLPowerForExactEmmetropia", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1122, 0x0022, 0x1122,
      EVR_FL, "IOLPowerForExactTargetRefraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1125, 0x0022, 0x1125,
      EVR_SQ, "AnteriorChamberDepthDefinitionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1127, 0x0022, 0x1127,
      EVR_SQ, "LensThicknessSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1128, 0x0022, 0x1128,
      EVR_SQ, "AnteriorChamberDepthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x112a, 0x0022, 0x112a,
      EVR_SQ, "CalculationCommentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x112b, 0x0022, 0x112b,
      EVR_CS, "CalculationCommentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x112c, 0x0022, 0x112c,
      EVR_LT, "CalculationComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1130, 0x0022, 0x1130,
      EVR_FL, "LensThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1131, 0x0022, 0x1131,
      EVR_FL, "AnteriorChamberDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1132, 0x0022, 0x1132,
      EVR_SQ, "SourceOfLensThicknessDataCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1133, 0x0022, 0x1133,
      EVR_SQ, "SourceOfAnteriorChamberDepthDataCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1134, 0x0022, 0x1134,
      EVR_SQ, "SourceOfRefractiveMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1135, 0x0022, 0x1135,
      EVR_SQ, "SourceOfRefractiveMeasurementsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1140, 0x0022, 0x1140,
      EVR_CS, "OphthalmicAxialLengthMeasurementModified", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1150, 0x0022, 0x1150,
      EVR_SQ, "OphthalmicAxialLengthDataSourceCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1153, 0x0022, 0x1153,
      EVR_SQ, "RETIRED_OphthalmicAxialLengthAcquisitionMethodCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1155, 0x0022, 0x1155,
      EVR_FL, "SignalToNoiseRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1159, 0x0022, 0x1159,
      EVR_LO, "OphthalmicAxialLengthDataSourceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1210, 0x0022, 0x1210,
      EVR_SQ, "OphthalmicAxialLengthMeasurementsTotalLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1211, 0x0022, 0x1211,
      EVR_SQ, "OphthalmicAxialLengthMeasurementsSegmentalLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1212, 0x0022, 0x1212,
      EVR_SQ, "OphthalmicAxialLengthMeasurementsLengthSummationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1220, 0x0022, 0x1220,
      EVR_SQ, "UltrasoundOphthalmicAxialLengthMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1225, 0x0022, 0x1225,
      EVR_SQ, "OpticalOphthalmicAxialLengthMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1230, 0x0022, 0x1230,
      EVR_SQ, "UltrasoundSelectedOphthalmicAxialLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1250, 0x0022, 0x1250,
      EVR_SQ, "OphthalmicAxialLengthSelectionMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1255, 0x0022, 0x1255,
      EVR_SQ, "OpticalSelectedOphthalmicAxialLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1257, 0x0022, 0x1257,
      EVR_SQ, "SelectedSegmentalOphthalmicAxialLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1260, 0x0022, 0x1260,
      EVR_SQ, "SelectedTotalOphthalmicAxialLengthSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1262, 0x0022, 0x1262,
      EVR_SQ, "OphthalmicAxialLengthQualityMetricSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1265, 0x0022, 0x1265,
      EVR_SQ, "RETIRED_OphthalmicAxialLengthQualityMetricTypeCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1273, 0x0022, 0x1273,
      EVR_LO, "RETIRED_OphthalmicAxialLengthQualityMetricTypeDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1300, 0x0022, 0x1300,
      EVR_SQ, "IntraocularLensCalculationsRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1310, 0x0022, 0x1310,
      EVR_SQ, "IntraocularLensCalculationsLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1330, 0x0022, 0x1330,
      EVR_SQ, "ReferencedOphthalmicAxialLengthMeasurementQCImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1415, 0x0022, 0x1415,
      EVR_CS, "OphthalmicMappingDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1420, 0x0022, 0x1420,
      EVR_SQ, "AcquisitionMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1423, 0x0022, 0x1423,
      EVR_SQ, "AcquisitionMethodAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1436, 0x0022, 0x1436,
      EVR_SQ, "OphthalmicThicknessMapTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1443, 0x0022, 0x1443,
      EVR_SQ, "OphthalmicThicknessMappingNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1445, 0x0022, 0x1445,
      EVR_SQ, "RetinalThicknessDefinitionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1450, 0x0022, 0x1450,
      EVR_SQ, "PixelValueMappingToCodedConceptSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1452, 0x0022, 0x1452,
      EVR_xs, "MappedPixelValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1454, 0x0022, 0x1454,
      EVR_LO, "PixelValueMappingExplanation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1458, 0x0022, 0x1458,
      EVR_SQ, "OphthalmicThicknessMapQualityThresholdSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1460, 0x0022, 0x1460,
      EVR_FL, "OphthalmicThicknessMapThresholdQualityRating", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1463, 0x0022, 0x1463,
      EVR_FL, "AnatomicStructureReferencePoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1465, 0x0022, 0x1465,
      EVR_SQ, "RegistrationToLocalizerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1466, 0x0022, 0x1466,
      EVR_CS, "RegisteredLocalizerUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1467, 0x0022, 0x1467,
      EVR_FL, "RegisteredLocalizerTopLeftHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1468, 0x0022, 0x1468,
      EVR_FL, "RegisteredLocalizerBottomRightHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1470, 0x0022, 0x1470,
      EVR_SQ, "OphthalmicThicknessMapQualityRatingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1472, 0x0022, 0x1472,
      EVR_SQ, "RelevantOPTAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1512, 0x0022, 0x1512,
      EVR_SQ, "TransformationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1513, 0x0022, 0x1513,
      EVR_SQ, "TransformationAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1515, 0x0022, 0x1515,
      EVR_CS, "OphthalmicAxialLengthMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1517, 0x0022, 0x1517,
      EVR_FL, "OphthalmicFOV", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1518, 0x0022, 0x1518,
      EVR_SQ, "TwoDimensionalToThreeDimensionalMapSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1525, 0x0022, 0x1525,
      EVR_SQ, "WideFieldOphthalmicPhotographyQualityRatingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1526, 0x0022, 0x1526,
      EVR_SQ, "WideFieldOphthalmicPhotographyQualityThresholdSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1527, 0x0022, 0x1527,
      EVR_FL, "WideFieldOphthalmicPhotographyThresholdQualityRating", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1528, 0x0022, 0x1528,
      EVR_FL, "XCoordinatesCenterPixelViewAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1529, 0x0022, 0x1529,
      EVR_FL, "YCoordinatesCenterPixelViewAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1530, 0x0022, 0x1530,
      EVR_UL, "NumberOfMapPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1531, 0x0022, 0x1531,
      EVR_OF, "TwoDimensionalToThreeDimensionalMapData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1612, 0x0022, 0x1612,
      EVR_SQ, "DerivationAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1615, 0x0022, 0x1615,
      EVR_SQ, "OphthalmicImageTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1616, 0x0022, 0x1616,
      EVR_LO, "OphthalmicImageTypeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1618, 0x0022, 0x1618,
      EVR_SQ, "ScanPatternTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1620, 0x0022, 0x1620,
      EVR_SQ, "ReferencedSurfaceMeshIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1622, 0x0022, 0x1622,
      EVR_CS, "OphthalmicVolumetricPropertiesFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1623, 0x0022, 0x1623,
      EVR_FL, "OphthalmicAnatomicReferencePointFrameCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1624, 0x0022, 0x1624,
      EVR_FL, "OphthalmicAnatomicReferencePointXCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1626, 0x0022, 0x1626,
      EVR_FL, "OphthalmicAnatomicReferencePointYCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1627, 0x0022, 0x1627,
      EVR_SQ, "OphthalmicEnFaceVolumeDescriptorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1628, 0x0022, 0x1628,
      EVR_SQ, "OphthalmicEnFaceImageQualityRatingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1629, 0x0022, 0x1629,
      EVR_CS, "OphthalmicEnFaceVolumeDescriptorScope", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1630, 0x0022, 0x1630,
      EVR_DS, "QualityThreshold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1632, 0x0022, 0x1632,
      EVR_SQ, "OphthalmicAnatomicReferencePointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1633, 0x0022, 0x1633,
      EVR_CS, "OphthalmicAnatomicReferencePointLocalizationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1634, 0x0022, 0x1634,
      EVR_IS, "PrimaryAnatomicStructureItemIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1640, 0x0022, 0x1640,
      EVR_SQ, "OCTBscanAnalysisAcquisitionParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1642, 0x0022, 0x1642,
      EVR_UL, "NumberOfBscansPerFrame", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1643, 0x0022, 0x1643,
      EVR_FL, "BscanSlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1644, 0x0022, 0x1644,
      EVR_FL, "DistanceBetweenBscanSlabs", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1645, 0x0022, 0x1645,
      EVR_FL, "BscanCycleTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1646, 0x0022, 0x1646,
      EVR_FL, "BscanCycleTimeVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1649, 0x0022, 0x1649,
      EVR_FL, "AscanRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1650, 0x0022, 0x1650,
      EVR_FL, "BscanRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0022, 0x1658, 0x0022, 0x1658,
      EVR_UL, "SurfaceMeshZPixelOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_DS, "CRDRE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_DS, "CRDRE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_LO, "ImageLaterality", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_DS, "HorizontalCalibrationPixelSize", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_LO, "CRExposureMenuCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_LO, "CRExposureMenuCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_US, "CRSShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0000, 0x0023, 0x0000,
      EVR_US, "CRSShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0001, 0x0023, 0x0001,
      EVR_IS, "LetterPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0023, 0x0001, 0x0023, 0x0001,
      EVR_DS, "VerticalCalibrationPixelSize", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0001, 0x0023, 0x0001,
      EVR_SL, "NumberOfSeriesInStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0002, 0x0023, 0x0002,
      EVR_IS, "BurnedInAnnotation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0023, 0x0002, 0x0023, 0x0002,
      EVR_LO, "CalibrationObject", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0002, 0x0023, 0x0002,
      EVR_SL, "NumberOfUnarchivedSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0003, 0x0023, 0x0003,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0023, 0x0003, 0x0023, 0x0003,
      EVR_DS, "CalibrationObjectSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0004, 0x0023, 0x0004,
      EVR_LO, "CalibrationMethod", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0005, 0x0023, 0x0005,
      EVR_ST, "Filename", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0006, 0x0023, 0x0006,
      EVR_IS, "FrameNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0007, 0x0023, 0x0007,
      EVR_IS, "CalibrationFactorMultiplicity", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x0008, 0x0023, 0x0008,
      EVR_IS, "CalibrationTODValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  QUANT 1.0" }
  , { 0x0023, 0x000d, 0x0023, 0x000d,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 2;1" }
  , { 0x0023, 0x000e, 0x0023, 0x000e,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 2;1" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_US, "CRDRN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_US, "CRDRN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_SS, "ReferenceImageField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_LO, "CRExposureMenuString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_LO, "CRExposureMenuString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_LO, "CRCShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0010, 0x0023, 0x0010,
      EVR_DS, "CRCShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_DS, "CRORE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_DS, "CRORE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_LO, "CREDRMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_LO, "CREDRMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_LO, "CRGT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0020, 0x0023, 0x0020,
      EVR_DS, "CRGT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_US, "CRORN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_US, "CRORN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_LO, "CRLatitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_LO, "CRLatitude", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_DS, "CRGA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0030, 0x0023, 0x0030,
      EVR_DS, "CRGA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_US, "CRORD", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_US, "CRORD", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_LO, "CRGroupNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_LO, "CRGroupNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_DS, "CRGC", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0040, 0x0023, 0x0040,
      EVR_DS, "CRGC", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_LO, "CRCassetteSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_LO, "CRCassetteSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_SS, "SummaryImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_US, "CRImageSerialNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_US, "CRImageSerialNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_DS, "CRGS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0050, 0x0023, 0x0050,
      EVR_DS, "CRGS", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_LO, "CRMachineID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_LO, "CRMachineID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_LO, "CRBarCodeNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_LO, "CRBarCodeNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_LO, "CRRT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0060, 0x0023, 0x0060,
      EVR_DS, "CRRT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_LO, "CRMachineType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_LO, "CRMachineType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_FD, "StartTimeSecsInFirstAxial", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_LO, "CRFilmOutputExposure", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_LO, "CRFilmOutputExposure", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_DS, "CRRE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0070, 0x0023, 0x0070,
      EVR_DS, "CRRE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0074, 0x0023, 0x0074,
      EVR_SL, "NumberOfUpdatesToHeader", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x007d, 0x0023, 0x007d,
      EVR_SS, "IndicatesIfStudyHasCompleteInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_STDY_01" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_LO, "CRFilmFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_LO, "CRFilmFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_US, "CRRN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_US, "CRRN", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_LO, "CRTechnicianCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0080, 0x0023, 0x0080,
      EVR_LO, "CRTechnicianCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_LO, "CRSShiftString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK1" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_LO, "CRSShiftString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK1" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_DS, "CRDRT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK2" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_DS, "CRDRT", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK2" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_LO, "CREnergySubtractionParameters", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x0090, 0x0023, 0x0090,
      EVR_LO, "CREnergySubtractionParameters", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_1.0 BLOCK3" }
  , { 0x0023, 0x00f0, 0x0023, 0x00f0,
      EVR_IS, "ImageSOPClass", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0023, 0x00f0, 0x0023, 0x00f0,
      EVR_LO, "CRDistributionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
  , { 0x0023, 0x00ff, 0x0023, 0x00ff,
      EVR_US, "CRShuttersApplied", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ACRQA_2.0 BLOCK3" }
#endif
  , { 0x0024, 0x0010, 0x0024, 0x0010,
      EVR_FL, "VisualFieldHorizontalExtent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0011, 0x0024, 0x0011,
      EVR_FL, "VisualFieldVerticalExtent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0012, 0x0024, 0x0012,
      EVR_CS, "VisualFieldShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0016, 0x0024, 0x0016,
      EVR_SQ, "ScreeningTestModeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0018, 0x0024, 0x0018,
      EVR_FL, "MaximumStimulusLuminance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0020, 0x0024, 0x0020,
      EVR_FL, "BackgroundLuminance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0021, 0x0024, 0x0021,
      EVR_SQ, "StimulusColorCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0024, 0x0024, 0x0024,
      EVR_SQ, "BackgroundIlluminationColorCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0025, 0x0024, 0x0025,
      EVR_FL, "StimulusArea", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0028, 0x0024, 0x0028,
      EVR_FL, "StimulusPresentationTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0032, 0x0024, 0x0032,
      EVR_SQ, "FixationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0033, 0x0024, 0x0033,
      EVR_SQ, "FixationMonitoringCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0034, 0x0024, 0x0034,
      EVR_SQ, "VisualFieldCatchTrialSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0035, 0x0024, 0x0035,
      EVR_US, "FixationCheckedQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0036, 0x0024, 0x0036,
      EVR_US, "PatientNotProperlyFixatedQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0037, 0x0024, 0x0037,
      EVR_CS, "PresentedVisualStimuliDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0038, 0x0024, 0x0038,
      EVR_US, "NumberOfVisualStimuli", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0039, 0x0024, 0x0039,
      EVR_CS, "ExcessiveFixationLossesDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0040, 0x0024, 0x0040,
      EVR_CS, "ExcessiveFixationLosses", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0042, 0x0024, 0x0042,
      EVR_US, "StimuliRetestingQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0044, 0x0024, 0x0044,
      EVR_LT, "CommentsOnPatientPerformanceOfVisualField", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0045, 0x0024, 0x0045,
      EVR_CS, "FalseNegativesEstimateFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0046, 0x0024, 0x0046,
      EVR_FL, "FalseNegativesEstimate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0048, 0x0024, 0x0048,
      EVR_US, "NegativeCatchTrialsQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0050, 0x0024, 0x0050,
      EVR_US, "FalseNegativesQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0051, 0x0024, 0x0051,
      EVR_CS, "ExcessiveFalseNegativesDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0052, 0x0024, 0x0052,
      EVR_CS, "ExcessiveFalseNegatives", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0053, 0x0024, 0x0053,
      EVR_CS, "FalsePositivesEstimateFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0054, 0x0024, 0x0054,
      EVR_FL, "FalsePositivesEstimate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0055, 0x0024, 0x0055,
      EVR_CS, "CatchTrialsDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0056, 0x0024, 0x0056,
      EVR_US, "PositiveCatchTrialsQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0057, 0x0024, 0x0057,
      EVR_CS, "TestPointNormalsDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0058, 0x0024, 0x0058,
      EVR_SQ, "TestPointNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0059, 0x0024, 0x0059,
      EVR_CS, "GlobalDeviationProbabilityNormalsFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0060, 0x0024, 0x0060,
      EVR_US, "FalsePositivesQuantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0061, 0x0024, 0x0061,
      EVR_CS, "ExcessiveFalsePositivesDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0062, 0x0024, 0x0062,
      EVR_CS, "ExcessiveFalsePositives", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0063, 0x0024, 0x0063,
      EVR_CS, "VisualFieldTestNormalsFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0064, 0x0024, 0x0064,
      EVR_SQ, "ResultsNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0065, 0x0024, 0x0065,
      EVR_SQ, "AgeCorrectedSensitivityDeviationAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0066, 0x0024, 0x0066,
      EVR_FL, "GlobalDeviationFromNormal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0067, 0x0024, 0x0067,
      EVR_SQ, "GeneralizedDefectSensitivityDeviationAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0068, 0x0024, 0x0068,
      EVR_FL, "LocalizedDeviationFromNormal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0069, 0x0024, 0x0069,
      EVR_LO, "PatientReliabilityIndicator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0070, 0x0024, 0x0070,
      EVR_FL, "VisualFieldMeanSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0071, 0x0024, 0x0071,
      EVR_FL, "GlobalDeviationProbability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0072, 0x0024, 0x0072,
      EVR_CS, "LocalDeviationProbabilityNormalsFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0073, 0x0024, 0x0073,
      EVR_FL, "LocalizedDeviationProbability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0074, 0x0024, 0x0074,
      EVR_CS, "ShortTermFluctuationCalculated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0075, 0x0024, 0x0075,
      EVR_FL, "ShortTermFluctuation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0076, 0x0024, 0x0076,
      EVR_CS, "ShortTermFluctuationProbabilityCalculated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0077, 0x0024, 0x0077,
      EVR_FL, "ShortTermFluctuationProbability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0078, 0x0024, 0x0078,
      EVR_CS, "CorrectedLocalizedDeviationFromNormalCalculated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0079, 0x0024, 0x0079,
      EVR_FL, "CorrectedLocalizedDeviationFromNormal", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0080, 0x0024, 0x0080,
      EVR_CS, "CorrectedLocalizedDeviationFromNormalProbabilityCalculated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0081, 0x0024, 0x0081,
      EVR_FL, "CorrectedLocalizedDeviationFromNormalProbability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0083, 0x0024, 0x0083,
      EVR_SQ, "GlobalDeviationProbabilitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0085, 0x0024, 0x0085,
      EVR_SQ, "LocalizedDeviationProbabilitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0086, 0x0024, 0x0086,
      EVR_CS, "FovealSensitivityMeasured", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0087, 0x0024, 0x0087,
      EVR_FL, "FovealSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0088, 0x0024, 0x0088,
      EVR_FL, "VisualFieldTestDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0089, 0x0024, 0x0089,
      EVR_SQ, "VisualFieldTestPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0090, 0x0024, 0x0090,
      EVR_FL, "VisualFieldTestPointXCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0091, 0x0024, 0x0091,
      EVR_FL, "VisualFieldTestPointYCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0092, 0x0024, 0x0092,
      EVR_FL, "AgeCorrectedSensitivityDeviationValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0093, 0x0024, 0x0093,
      EVR_CS, "StimulusResults", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0094, 0x0024, 0x0094,
      EVR_FL, "SensitivityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0095, 0x0024, 0x0095,
      EVR_CS, "RetestStimulusSeen", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0096, 0x0024, 0x0096,
      EVR_FL, "RetestSensitivityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0097, 0x0024, 0x0097,
      EVR_SQ, "VisualFieldTestPointNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0098, 0x0024, 0x0098,
      EVR_FL, "QuantifiedDefect", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0100, 0x0024, 0x0100,
      EVR_FL, "AgeCorrectedSensitivityDeviationProbabilityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0102, 0x0024, 0x0102,
      EVR_CS, "GeneralizedDefectCorrectedSensitivityDeviationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0103, 0x0024, 0x0103,
      EVR_FL, "GeneralizedDefectCorrectedSensitivityDeviationValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0104, 0x0024, 0x0104,
      EVR_FL, "GeneralizedDefectCorrectedSensitivityDeviationProbabilityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0105, 0x0024, 0x0105,
      EVR_FL, "MinimumSensitivityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0106, 0x0024, 0x0106,
      EVR_CS, "BlindSpotLocalized", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0107, 0x0024, 0x0107,
      EVR_FL, "BlindSpotXCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0108, 0x0024, 0x0108,
      EVR_FL, "BlindSpotYCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0110, 0x0024, 0x0110,
      EVR_SQ, "VisualAcuityMeasurementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0112, 0x0024, 0x0112,
      EVR_SQ, "RefractiveParametersUsedOnPatientSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0113, 0x0024, 0x0113,
      EVR_CS, "MeasurementLaterality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0114, 0x0024, 0x0114,
      EVR_SQ, "OphthalmicPatientClinicalInformationLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0115, 0x0024, 0x0115,
      EVR_SQ, "OphthalmicPatientClinicalInformationRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0117, 0x0024, 0x0117,
      EVR_CS, "FovealPointNormativeDataFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0118, 0x0024, 0x0118,
      EVR_FL, "FovealPointProbabilityValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0120, 0x0024, 0x0120,
      EVR_CS, "ScreeningBaselineMeasured", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0122, 0x0024, 0x0122,
      EVR_SQ, "ScreeningBaselineMeasuredSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0124, 0x0024, 0x0124,
      EVR_CS, "ScreeningBaselineType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0126, 0x0024, 0x0126,
      EVR_FL, "ScreeningBaselineValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0202, 0x0024, 0x0202,
      EVR_LO, "AlgorithmSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0306, 0x0024, 0x0306,
      EVR_LO, "DataSetName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0307, 0x0024, 0x0307,
      EVR_LO, "DataSetVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0308, 0x0024, 0x0308,
      EVR_LO, "DataSetSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0309, 0x0024, 0x0309,
      EVR_LO, "DataSetDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0317, 0x0024, 0x0317,
      EVR_SQ, "VisualFieldTestReliabilityGlobalIndexSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0320, 0x0024, 0x0320,
      EVR_SQ, "VisualFieldGlobalResultsIndexSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0325, 0x0024, 0x0325,
      EVR_SQ, "DataObservationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0338, 0x0024, 0x0338,
      EVR_CS, "IndexNormalsFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0341, 0x0024, 0x0341,
      EVR_FL, "IndexProbability", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0024, 0x0344, 0x0024, 0x0344,
      EVR_SQ, "IndexProbabilitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0025, 0x0000, 0x0025, 0x0000,
      EVR_IS, "OriginalImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0000, 0x0025, 0x0000,
      EVR_US, "ViewNative", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0001, 0x0025, 0x0001,
      EVR_IS, "NotProcessedImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0001, 0x0025, 0x0001,
      EVR_US, "OriginalSeriesNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0002, 0x0025, 0x0002,
      EVR_IS, "CutOutImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0002, 0x0025, 0x0002,
      EVR_US, "OriginalImageNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0003, 0x0025, 0x0003,
      EVR_IS, "DuplicatedImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0003, 0x0025, 0x0003,
      EVR_US, "WinCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0004, 0x0025, 0x0004,
      EVR_IS, "StoredImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0004, 0x0025, 0x0004,
      EVR_US, "WinWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0005, 0x0025, 0x0005,
      EVR_IS, "RetrievedImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0005, 0x0025, 0x0005,
      EVR_US, "WinBrightness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0006, 0x0025, 0x0006,
      EVR_IS, "RemoteImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0006, 0x0025, 0x0006,
      EVR_US, "WinContrast", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0006, 0x0025, 0x0006,
      EVR_SS, "LastPulseSequenceUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0007, 0x0025, 0x0007,
      EVR_IS, "MediaStoredImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0007, 0x0025, 0x0007,
      EVR_US, "OriginalFrameNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0007, 0x0025, 0x0007,
      EVR_SL, "ImagesInSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0008, 0x0025, 0x0008,
      EVR_IS, "ImageState", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0008, 0x0025, 0x0008,
      EVR_US, "OriginalMaskFrameNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0009, 0x0025, 0x0009,
      EVR_US, "Opac", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000a, 0x0025, 0x000a,
      EVR_US, "OriginalNumberOfFrames", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000b, 0x0025, 0x000b,
      EVR_DS, "OriginalSceneDuration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000c, 0x0025, 0x000c,
      EVR_LO, "IdentifierLOID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000d, 0x0025, 0x000d,
      EVR_SS, "OriginalSceneVFRInfo", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000e, 0x0025, 0x000e,
      EVR_SS, "OriginalFrameECGPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x000f, 0x0025, 0x000f,
      EVR_SS, "OriginalECG1stFrameOffset_retired", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0010, 0x0025, 0x0010,
      EVR_US, "RelativeLightEmissionAmountSk", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0010, 0x0025, 0x0010,
      EVR_SS, "ZoomFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0010, 0x0025, 0x0010,
      EVR_SL, "LandmarkCounter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0011, 0x0025, 0x0011,
      EVR_US, "TermOfCorrectionForEachIPTypeSt", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0011, 0x0025, 0x0011,
      EVR_US, "Flex", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0011, 0x0025, 0x0011,
      EVR_SS, "NumberOfAcquisitions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0012, 0x0025, 0x0012,
      EVR_US, "ReadingGainGp", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0012, 0x0025, 0x0012,
      EVR_US, "NumberOfMaskFrames", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0013, 0x0025, 0x0013,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0013, 0x0025, 0x0013,
      EVR_US, "NumberOfFillFrames", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0014, 0x0025, 0x0014,
      EVR_US, "SeriesNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0014, 0x0025, 0x0014,
      EVR_SL, "IndicatesNumberOfUpdatesToHeader", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0015, 0x0025, 0x0015,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0015, 0x0025, 0x0015,
      EVR_IS, "ImageNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0" }
  , { 0x0025, 0x0017, 0x0025, 0x0017,
      EVR_SL, "SeriesCompleteFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0018, 0x0025, 0x0018,
      EVR_SL, "NumberOfImagesArchived", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0019, 0x0025, 0x0019,
      EVR_SL, "LastImageNumberUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x001a, 0x0025, 0x001a,
      EVR_SH, "PrimaryReceiverSuiteAndHost", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SERS_01" }
  , { 0x0025, 0x0020, 0x0025, 0x0020,
      EVR_LO, "SourceImageFile", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0020, 0x0025, 0x0020,
      EVR_US, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0021, 0x0025, 0x0021,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0025, 0x0021, 0x0025, 0x0021,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0030, 0x0025, 0x0030,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0031, 0x0025, 0x0031,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0032, 0x0025, 0x0032,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0033, 0x0025, 0x0033,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0034, 0x0025, 0x0034,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0040, 0x0025, 0x0040,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0041, 0x0025, 0x0041,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0042, 0x0025, 0x0042,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0043, 0x0025, 0x0043,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0050, 0x0025, 0x0050,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0051, 0x0025, 0x0051,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0052, 0x0025, 0x0052,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0053, 0x0025, 0x0053,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0060, 0x0025, 0x0060,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0061, 0x0025, 0x0061,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0062, 0x0025, 0x0062,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0063, 0x0025, 0x0063,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0070, 0x0025, 0x0070,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0071, 0x0025, 0x0071,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0072, 0x0025, 0x0072,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0073, 0x0025, 0x0073,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0074, 0x0025, 0x0074,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0080, 0x0025, 0x0080,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0081, 0x0025, 0x0081,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0082, 0x0025, 0x0082,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0083, 0x0025, 0x0083,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0084, 0x0025, 0x0084,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0090, 0x0025, 0x0090,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0091, 0x0025, 0x0091,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0092, 0x0025, 0x0092,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0093, 0x0025, 0x0093,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0094, 0x0025, 0x0094,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0095, 0x0025, 0x0095,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x0096, 0x0025, 0x0096,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x00a0, 0x0025, 0x00a0,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x00a1, 0x0025, 0x00a1,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x00a2, 0x0025, 0x00a2,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x00a3, 0x0025, 0x00a3,
      EVR_SS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0025, 0x1010, 0x0025, 0x1010,
      EVR_LO, "WorkspaceID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CMR42 CIRCLECVI" }
  , { 0x0025, 0x1020, 0x0025, 0x1020,
      EVR_LO, "WorkspaceTimeString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CMR42 CIRCLECVI" }
  , { 0x0025, 0x1030, 0x0025, 0x1030,
      EVR_OB, "WorkspaceStream", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CMR42 CIRCLECVI" }
  , { 0x0027, 0x0000, 0x0027, 0x0000,
      EVR_IS, "NumberOfSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0000, 0x0027, 0x0000,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0001, 0x0027, 0x0001,
      EVR_IS, "NumberOfStudies", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0006, 0x0027, 0x0006,
      EVR_SL, "ImageArchiveFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0010, 0x0027, 0x0010,
      EVR_DT, "OldestSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0010, 0x0027, 0x0010,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0010, 0x0027, 0x0010,
      EVR_SS, "ScoutType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0011, 0x0027, 0x0011,
      EVR_DT, "NewestSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0011, 0x0027, 0x0011,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0012, 0x0027, 0x0012,
      EVR_DT, "OldestStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0012, 0x0027, 0x0012,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0013, 0x0027, 0x0013,
      EVR_DT, "NewestStudy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SVISION" }
  , { 0x0027, 0x0013, 0x0027, 0x0013,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0014, 0x0027, 0x0014,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0015, 0x0027, 0x0015,
      EVR_DS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x0016, 0x0027, 0x0016,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0027, 0x001c, 0x0027, 0x001c,
      EVR_SL, "VmaMamp", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x001d, 0x0027, 0x001d,
      EVR_SS, "VmaPhase", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x001e, 0x0027, 0x001e,
      EVR_SL, "VmaMod", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x001f, 0x0027, 0x001f,
      EVR_SL, "VmaClip", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0020, 0x0027, 0x0020,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0020, 0x0027, 0x0020,
      EVR_SS, "SmartScanOnOffFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0030, 0x0027, 0x0030,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0030, 0x0027, 0x0030,
      EVR_SH, "ForeignImageRevision", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0031, 0x0027, 0x0031,
      EVR_SS, "ImagingMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0032, 0x0027, 0x0032,
      EVR_SS, "PulseSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0033, 0x0027, 0x0033,
      EVR_SL, "ImagingOptions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0035, 0x0027, 0x0035,
      EVR_SS, "PlaneType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0036, 0x0027, 0x0036,
      EVR_SL, "ObliquePlane", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0040, 0x0027, 0x0040,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0040, 0x0027, 0x0040,
      EVR_SH, "RASLetterOfImageLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0041, 0x0027, 0x0041,
      EVR_FL, "ImageLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0042, 0x0027, 0x0042,
      EVR_FL, "CenterRCoordOfPlaneImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0043, 0x0027, 0x0043,
      EVR_FL, "CenterACoordOfPlaneImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0044, 0x0027, 0x0044,
      EVR_FL, "CenterSCoordOfPlaneImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0045, 0x0027, 0x0045,
      EVR_FL, "NormalRCoord", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0046, 0x0027, 0x0046,
      EVR_FL, "NormalACoord", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0047, 0x0027, 0x0047,
      EVR_FL, "NormalSCoord", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0048, 0x0027, 0x0048,
      EVR_FL, "RCoordOfTopRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0049, 0x0027, 0x0049,
      EVR_FL, "ACoordOfTopRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x004a, 0x0027, 0x004a,
      EVR_FL, "SCoordOfTopRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x004b, 0x0027, 0x004b,
      EVR_FL, "RCoordOfBottomRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x004c, 0x0027, 0x004c,
      EVR_FL, "ACoordOfBottomRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x004d, 0x0027, 0x004d,
      EVR_FL, "SCoordOfBottomRightCorner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0050, 0x0027, 0x0050,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0050, 0x0027, 0x0050,
      EVR_FL, "TableStartLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0051, 0x0027, 0x0051,
      EVR_FL, "TableEndLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0052, 0x0027, 0x0052,
      EVR_SH, "RASLetterForSideOfImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0053, 0x0027, 0x0053,
      EVR_SH, "RASLetterForAnteriorPosterior", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0054, 0x0027, 0x0054,
      EVR_SH, "RASLetterForScoutStartLoc", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0055, 0x0027, 0x0055,
      EVR_SH, "RASLetterForScoutEndLoc", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0060, 0x0027, 0x0060,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0060, 0x0027, 0x0060,
      EVR_FL, "ImageDimensionX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0061, 0x0027, 0x0061,
      EVR_FL, "ImageDimensionY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0062, 0x0027, 0x0062,
      EVR_FL, "NumberOfExcitations", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMAG_01" }
  , { 0x0027, 0x0070, 0x0027, 0x0070,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x0080, 0x0027, 0x0080,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x00a0, 0x0027, 0x00a0,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x00a1, 0x0027, 0x00a1,
      EVR_CS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x00a2, 0x0027, 0x00a2,
      EVR_CS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0027, 0x00a3, 0x0027, 0x00a3,
      EVR_SS, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
#endif
  , { 0x0028, 0x0002, 0x0028, 0x0002,
      EVR_US, "SamplesPerPixel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0003, 0x0028, 0x0003,
      EVR_US, "SamplesPerPixelUsed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0004, 0x0028, 0x0004,
      EVR_CS, "PhotometricInterpretation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0005, 0x0028, 0x0005,
      EVR_US, "RETIRED_ImageDimensions", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0006, 0x0028, 0x0006,
      EVR_US, "PlanarConfiguration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0008, 0x0028, 0x0008,
      EVR_IS, "NumberOfFrames", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0009, 0x0028, 0x0009,
      EVR_AT, "FrameIncrementPointer", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x000a, 0x0028, 0x000a,
      EVR_AT, "FrameDimensionPointer", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0010, 0x0028, 0x0010,
      EVR_US, "Rows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0011, 0x0028, 0x0011,
      EVR_US, "Columns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0012, 0x0028, 0x0012,
      EVR_US, "RETIRED_Planes", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0014, 0x0028, 0x0014,
      EVR_US, "UltrasoundColorDataPresent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0030, 0x0028, 0x0030,
      EVR_DS, "PixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0031, 0x0028, 0x0031,
      EVR_DS, "ZoomFactor", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0032, 0x0028, 0x0032,
      EVR_DS, "ZoomCenter", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0034, 0x0028, 0x0034,
      EVR_IS, "PixelAspectRatio", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0040, 0x0028, 0x0040,
      EVR_CS, "RETIRED_ImageFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0050, 0x0028, 0x0050,
      EVR_LO, "RETIRED_ManipulatedImage", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0051, 0x0028, 0x0051,
      EVR_CS, "CorrectedImage", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x005f, 0x0028, 0x005f,
      EVR_LO, "RETIRED_CompressionRecognitionCode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0060, 0x0028, 0x0060,
      EVR_CS, "RETIRED_CompressionCode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0061, 0x0028, 0x0061,
      EVR_SH, "RETIRED_CompressionOriginator", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0062, 0x0028, 0x0062,
      EVR_LO, "RETIRED_CompressionLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0063, 0x0028, 0x0063,
      EVR_SH, "RETIRED_CompressionDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0065, 0x0028, 0x0065,
      EVR_CS, "RETIRED_CompressionSequence", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0066, 0x0028, 0x0066,
      EVR_AT, "RETIRED_CompressionStepPointers", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0068, 0x0028, 0x0068,
      EVR_US, "RETIRED_RepeatInterval", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0069, 0x0028, 0x0069,
      EVR_US, "RETIRED_BitsGrouped", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0070, 0x0028, 0x0070,
      EVR_US, "RETIRED_PerimeterTable", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0071, 0x0028, 0x0071,
      EVR_xs, "RETIRED_PerimeterValue", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0080, 0x0028, 0x0080,
      EVR_US, "RETIRED_PredictorRows", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0081, 0x0028, 0x0081,
      EVR_US, "RETIRED_PredictorColumns", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0082, 0x0028, 0x0082,
      EVR_US, "RETIRED_PredictorConstants", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0090, 0x0028, 0x0090,
      EVR_CS, "RETIRED_BlockedPixels", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0091, 0x0028, 0x0091,
      EVR_US, "RETIRED_BlockRows", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0092, 0x0028, 0x0092,
      EVR_US, "RETIRED_BlockColumns", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0093, 0x0028, 0x0093,
      EVR_US, "RETIRED_RowOverlap", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0094, 0x0028, 0x0094,
      EVR_US, "RETIRED_ColumnOverlap", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0100, 0x0028, 0x0100,
      EVR_US, "BitsAllocated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0101, 0x0028, 0x0101,
      EVR_US, "BitsStored", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0102, 0x0028, 0x0102,
      EVR_US, "HighBit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0103, 0x0028, 0x0103,
      EVR_US, "PixelRepresentation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0104, 0x0028, 0x0104,
      EVR_xs, "RETIRED_SmallestValidPixelValue", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0105, 0x0028, 0x0105,
      EVR_xs, "RETIRED_LargestValidPixelValue", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0106, 0x0028, 0x0106,
      EVR_xs, "SmallestImagePixelValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0107, 0x0028, 0x0107,
      EVR_xs, "LargestImagePixelValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0108, 0x0028, 0x0108,
      EVR_xs, "SmallestPixelValueInSeries", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0109, 0x0028, 0x0109,
      EVR_xs, "LargestPixelValueInSeries", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0110, 0x0028, 0x0110,
      EVR_xs, "RETIRED_SmallestImagePixelValueInPlane", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0111, 0x0028, 0x0111,
      EVR_xs, "RETIRED_LargestImagePixelValueInPlane", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0120, 0x0028, 0x0120,
      EVR_xs, "PixelPaddingValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0121, 0x0028, 0x0121,
      EVR_xs, "PixelPaddingRangeLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0122, 0x0028, 0x0122,
      EVR_FL, "FloatPixelPaddingValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0123, 0x0028, 0x0123,
      EVR_FD, "DoubleFloatPixelPaddingValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0124, 0x0028, 0x0124,
      EVR_FL, "FloatPixelPaddingRangeLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0125, 0x0028, 0x0125,
      EVR_FD, "DoubleFloatPixelPaddingRangeLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0200, 0x0028, 0x0200,
      EVR_US, "RETIRED_ImageLocation", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0300, 0x0028, 0x0300,
      EVR_CS, "QualityControlImage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0301, 0x0028, 0x0301,
      EVR_CS, "BurnedInAnnotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0302, 0x0028, 0x0302,
      EVR_CS, "RecognizableVisualFeatures", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0303, 0x0028, 0x0303,
      EVR_CS, "LongitudinalTemporalInformationModified", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0304, 0x0028, 0x0304,
      EVR_UI, "ReferencedColorPaletteInstanceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0400, 0x0028, 0x0400,
      EVR_LO, "RETIRED_TransformLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0401, 0x0028, 0x0401,
      EVR_LO, "RETIRED_TransformVersionNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0402, 0x0028, 0x0402,
      EVR_US, "RETIRED_NumberOfTransformSteps", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0403, 0x0028, 0x0403,
      EVR_LO, "RETIRED_SequenceOfCompressedData", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0404, 0x0028, 0x0404,
      EVR_AT, "RETIRED_DetailsOfCoefficients", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0410, 0x0028, 0x0410,
      EVR_US, "RETIRED_RowsForNthOrderCoefficients", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0411, 0x0028, 0x0411,
      EVR_US, "RETIRED_ColumnsForNthOrderCoefficients", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0412, 0x0028, 0x0412,
      EVR_LO, "RETIRED_CoefficientCoding", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0413, 0x0028, 0x0413,
      EVR_AT, "RETIRED_CoefficientCodingPointers", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0700, 0x0028, 0x0700,
      EVR_LO, "RETIRED_DCTLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0701, 0x0028, 0x0701,
      EVR_CS, "RETIRED_DataBlockDescription", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0702, 0x0028, 0x0702,
      EVR_AT, "RETIRED_DataBlock", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0710, 0x0028, 0x0710,
      EVR_US, "RETIRED_NormalizationFactorFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0720, 0x0028, 0x0720,
      EVR_US, "RETIRED_ZonalMapNumberFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0721, 0x0028, 0x0721,
      EVR_AT, "RETIRED_ZonalMapLocation", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0722, 0x0028, 0x0722,
      EVR_US, "RETIRED_ZonalMapFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0730, 0x0028, 0x0730,
      EVR_US, "RETIRED_AdaptiveMapFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0740, 0x0028, 0x0740,
      EVR_US, "RETIRED_CodeNumberFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0800, 0x0028, 0x0800,
      EVR_CS, "RETIRED_CodeLabel", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0802, 0x0028, 0x0802,
      EVR_US, "RETIRED_NumberOfTables", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0803, 0x0028, 0x0803,
      EVR_AT, "RETIRED_CodeTableLocation", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0804, 0x0028, 0x0804,
      EVR_US, "RETIRED_BitsForCodeWord", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0808, 0x0028, 0x0808,
      EVR_AT, "RETIRED_ImageDataLocation", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0a02, 0x0028, 0x0a02,
      EVR_CS, "PixelSpacingCalibrationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x0a04, 0x0028, 0x0a04,
      EVR_LO, "PixelSpacingCalibrationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1040, 0x0028, 0x1040,
      EVR_CS, "PixelIntensityRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1041, 0x0028, 0x1041,
      EVR_SS, "PixelIntensityRelationshipSign", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1050, 0x0028, 0x1050,
      EVR_DS, "WindowCenter", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1051, 0x0028, 0x1051,
      EVR_DS, "WindowWidth", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1052, 0x0028, 0x1052,
      EVR_DS, "RescaleIntercept", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1053, 0x0028, 0x1053,
      EVR_DS, "RescaleSlope", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1054, 0x0028, 0x1054,
      EVR_LO, "RescaleType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1055, 0x0028, 0x1055,
      EVR_LO, "WindowCenterWidthExplanation", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1056, 0x0028, 0x1056,
      EVR_CS, "VOILUTFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1080, 0x0028, 0x1080,
      EVR_CS, "RETIRED_GrayScale", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1090, 0x0028, 0x1090,
      EVR_CS, "RecommendedViewingMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1100, 0x0028, 0x1100,
      EVR_xs, "RETIRED_GrayLookupTableDescriptor", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1101, 0x0028, 0x1101,
      EVR_xs, "RedPaletteColorLookupTableDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1102, 0x0028, 0x1102,
      EVR_xs, "GreenPaletteColorLookupTableDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1103, 0x0028, 0x1103,
      EVR_xs, "BluePaletteColorLookupTableDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1104, 0x0028, 0x1104,
      EVR_US, "AlphaPaletteColorLookupTableDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1111, 0x0028, 0x1111,
      EVR_xs, "RETIRED_LargeRedPaletteColorLookupTableDescriptor", 4, 4, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1112, 0x0028, 0x1112,
      EVR_xs, "RETIRED_LargeGreenPaletteColorLookupTableDescriptor", 4, 4, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1113, 0x0028, 0x1113,
      EVR_xs, "RETIRED_LargeBluePaletteColorLookupTableDescriptor", 4, 4, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1199, 0x0028, 0x1199,
      EVR_UI, "PaletteColorLookupTableUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1200, 0x0028, 0x1200,
      EVR_lt, "RETIRED_GrayLookupTableData", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1201, 0x0028, 0x1201,
      EVR_OW, "RedPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1202, 0x0028, 0x1202,
      EVR_OW, "GreenPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1203, 0x0028, 0x1203,
      EVR_OW, "BluePaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1204, 0x0028, 0x1204,
      EVR_OW, "AlphaPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1211, 0x0028, 0x1211,
      EVR_OW, "RETIRED_LargeRedPaletteColorLookupTableData", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1212, 0x0028, 0x1212,
      EVR_OW, "RETIRED_LargeGreenPaletteColorLookupTableData", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1213, 0x0028, 0x1213,
      EVR_OW, "RETIRED_LargeBluePaletteColorLookupTableData", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1214, 0x0028, 0x1214,
      EVR_UI, "RETIRED_LargePaletteColorLookupTableUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1221, 0x0028, 0x1221,
      EVR_OW, "SegmentedRedPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1222, 0x0028, 0x1222,
      EVR_OW, "SegmentedGreenPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1223, 0x0028, 0x1223,
      EVR_OW, "SegmentedBluePaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1224, 0x0028, 0x1224,
      EVR_OW, "SegmentedAlphaPaletteColorLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1230, 0x0028, 0x1230,
      EVR_SQ, "StoredValueColorRangeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1231, 0x0028, 0x1231,
      EVR_FD, "MinimumStoredValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1232, 0x0028, 0x1232,
      EVR_FD, "MaximumStoredValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1300, 0x0028, 0x1300,
      EVR_CS, "BreastImplantPresent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1350, 0x0028, 0x1350,
      EVR_CS, "PartialView", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1351, 0x0028, 0x1351,
      EVR_ST, "PartialViewDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1352, 0x0028, 0x1352,
      EVR_SQ, "PartialViewCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x135a, 0x0028, 0x135a,
      EVR_CS, "SpatialLocationsPreserved", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1401, 0x0028, 0x1401,
      EVR_SQ, "DataFrameAssignmentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1402, 0x0028, 0x1402,
      EVR_CS, "DataPathAssignment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1403, 0x0028, 0x1403,
      EVR_US, "BitsMappedToColorLookupTable", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1404, 0x0028, 0x1404,
      EVR_SQ, "BlendingLUT1Sequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1405, 0x0028, 0x1405,
      EVR_CS, "BlendingLUT1TransferFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1406, 0x0028, 0x1406,
      EVR_FD, "BlendingWeightConstant", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1407, 0x0028, 0x1407,
      EVR_US, "BlendingLookupTableDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1408, 0x0028, 0x1408,
      EVR_OW, "BlendingLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x140b, 0x0028, 0x140b,
      EVR_SQ, "EnhancedPaletteColorLookupTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x140c, 0x0028, 0x140c,
      EVR_SQ, "BlendingLUT2Sequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x140d, 0x0028, 0x140d,
      EVR_CS, "BlendingLUT2TransferFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x140e, 0x0028, 0x140e,
      EVR_CS, "DataPathID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x140f, 0x0028, 0x140f,
      EVR_CS, "RGBLUTTransferFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x1410, 0x0028, 0x1410,
      EVR_CS, "AlphaLUTTransferFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x2000, 0x0028, 0x2000,
      EVR_OB, "ICCProfile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x2002, 0x0028, 0x2002,
      EVR_CS, "ColorSpace", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x2110, 0x0028, 0x2110,
      EVR_CS, "LossyImageCompression", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x2112, 0x0028, 0x2112,
      EVR_DS, "LossyImageCompressionRatio", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x2114, 0x0028, 0x2114,
      EVR_CS, "LossyImageCompressionMethod", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3000, 0x0028, 0x3000,
      EVR_SQ, "ModalityLUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3001, 0x0028, 0x3001,
      EVR_SQ, "VariableModalityLUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3002, 0x0028, 0x3002,
      EVR_xs, "LUTDescriptor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3003, 0x0028, 0x3003,
      EVR_LO, "LUTExplanation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3004, 0x0028, 0x3004,
      EVR_LO, "ModalityLUTType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3006, 0x0028, 0x3006,
      EVR_lt, "LUTData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3010, 0x0028, 0x3010,
      EVR_SQ, "VOILUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x3110, 0x0028, 0x3110,
      EVR_SQ, "SoftcopyVOILUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x4000, 0x0028, 0x4000,
      EVR_LT, "RETIRED_ImagePresentationComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x5000, 0x0028, 0x5000,
      EVR_SQ, "RETIRED_BiPlaneAcquisitionSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6010, 0x0028, 0x6010,
      EVR_US, "RepresentativeFrameNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6020, 0x0028, 0x6020,
      EVR_US, "FrameNumbersOfInterest", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6022, 0x0028, 0x6022,
      EVR_LO, "FrameOfInterestDescription", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6023, 0x0028, 0x6023,
      EVR_CS, "FrameOfInterestType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6030, 0x0028, 0x6030,
      EVR_US, "RETIRED_MaskPointers", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6040, 0x0028, 0x6040,
      EVR_US, "RWavePointer", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6100, 0x0028, 0x6100,
      EVR_SQ, "MaskSubtractionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6101, 0x0028, 0x6101,
      EVR_CS, "MaskOperation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6102, 0x0028, 0x6102,
      EVR_US, "ApplicableFrameRange", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6110, 0x0028, 0x6110,
      EVR_US, "MaskFrameNumbers", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6112, 0x0028, 0x6112,
      EVR_US, "ContrastFrameAveraging", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6114, 0x0028, 0x6114,
      EVR_FL, "MaskSubPixelShift", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6120, 0x0028, 0x6120,
      EVR_SS, "TIDOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x6190, 0x0028, 0x6190,
      EVR_ST, "MaskOperationExplanation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7000, 0x0028, 0x7000,
      EVR_SQ, "EquipmentAdministratorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7001, 0x0028, 0x7001,
      EVR_US, "NumberOfDisplaySubsystems", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7002, 0x0028, 0x7002,
      EVR_US, "CurrentConfigurationID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7003, 0x0028, 0x7003,
      EVR_US, "DisplaySubsystemID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7004, 0x0028, 0x7004,
      EVR_SH, "DisplaySubsystemName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7005, 0x0028, 0x7005,
      EVR_LO, "DisplaySubsystemDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7006, 0x0028, 0x7006,
      EVR_CS, "SystemStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7007, 0x0028, 0x7007,
      EVR_LO, "SystemStatusComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7008, 0x0028, 0x7008,
      EVR_SQ, "TargetLuminanceCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7009, 0x0028, 0x7009,
      EVR_US, "LuminanceCharacteristicsID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700a, 0x0028, 0x700a,
      EVR_SQ, "DisplaySubsystemConfigurationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700b, 0x0028, 0x700b,
      EVR_US, "ConfigurationID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700c, 0x0028, 0x700c,
      EVR_SH, "ConfigurationName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700d, 0x0028, 0x700d,
      EVR_LO, "ConfigurationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700e, 0x0028, 0x700e,
      EVR_US, "ReferencedTargetLuminanceCharacteristicsID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x700f, 0x0028, 0x700f,
      EVR_SQ, "QAResultsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7010, 0x0028, 0x7010,
      EVR_SQ, "DisplaySubsystemQAResultsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7011, 0x0028, 0x7011,
      EVR_SQ, "ConfigurationQAResultsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7012, 0x0028, 0x7012,
      EVR_SQ, "MeasurementEquipmentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7013, 0x0028, 0x7013,
      EVR_CS, "MeasurementFunctions", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7014, 0x0028, 0x7014,
      EVR_CS, "MeasurementEquipmentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7015, 0x0028, 0x7015,
      EVR_SQ, "VisualEvaluationResultSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7016, 0x0028, 0x7016,
      EVR_SQ, "DisplayCalibrationResultSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7017, 0x0028, 0x7017,
      EVR_US, "DDLValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7018, 0x0028, 0x7018,
      EVR_FL, "CIExyWhitePoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7019, 0x0028, 0x7019,
      EVR_CS, "DisplayFunctionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701a, 0x0028, 0x701a,
      EVR_FL, "GammaValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701b, 0x0028, 0x701b,
      EVR_US, "NumberOfLuminancePoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701c, 0x0028, 0x701c,
      EVR_SQ, "LuminanceResponseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701d, 0x0028, 0x701d,
      EVR_FL, "TargetMinimumLuminance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701e, 0x0028, 0x701e,
      EVR_FL, "TargetMaximumLuminance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x701f, 0x0028, 0x701f,
      EVR_FL, "LuminanceValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7020, 0x0028, 0x7020,
      EVR_LO, "LuminanceResponseDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7021, 0x0028, 0x7021,
      EVR_CS, "WhitePointFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7022, 0x0028, 0x7022,
      EVR_SQ, "DisplayDeviceTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7023, 0x0028, 0x7023,
      EVR_SQ, "DisplaySubsystemSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7024, 0x0028, 0x7024,
      EVR_SQ, "LuminanceResultSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7025, 0x0028, 0x7025,
      EVR_CS, "AmbientLightValueSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7026, 0x0028, 0x7026,
      EVR_CS, "MeasuredCharacteristics", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7027, 0x0028, 0x7027,
      EVR_SQ, "LuminanceUniformityResultSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7028, 0x0028, 0x7028,
      EVR_SQ, "VisualEvaluationTestSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7029, 0x0028, 0x7029,
      EVR_CS, "TestResult", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x702a, 0x0028, 0x702a,
      EVR_LO, "TestResultComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x702b, 0x0028, 0x702b,
      EVR_CS, "TestImageValidation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x702c, 0x0028, 0x702c,
      EVR_SQ, "TestPatternCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x702d, 0x0028, 0x702d,
      EVR_SQ, "MeasurementPatternCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x702e, 0x0028, 0x702e,
      EVR_SQ, "VisualEvaluationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x7fe0, 0x0028, 0x7fe0,
      EVR_UR, "PixelDataProviderURL", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9001, 0x0028, 0x9001,
      EVR_UL, "DataPointRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9002, 0x0028, 0x9002,
      EVR_UL, "DataPointColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9003, 0x0028, 0x9003,
      EVR_CS, "SignalDomainColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9099, 0x0028, 0x9099,
      EVR_US, "RETIRED_LargestMonochromePixelValue", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9108, 0x0028, 0x9108,
      EVR_CS, "DataRepresentation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9110, 0x0028, 0x9110,
      EVR_SQ, "PixelMeasuresSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9132, 0x0028, 0x9132,
      EVR_SQ, "FrameVOILUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9145, 0x0028, 0x9145,
      EVR_SQ, "PixelValueTransformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9235, 0x0028, 0x9235,
      EVR_CS, "SignalDomainRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9411, 0x0028, 0x9411,
      EVR_FL, "DisplayFilterPercentage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9415, 0x0028, 0x9415,
      EVR_SQ, "FramePixelShiftSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9416, 0x0028, 0x9416,
      EVR_US, "SubtractionItemID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9422, 0x0028, 0x9422,
      EVR_SQ, "PixelIntensityRelationshipLUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9443, 0x0028, 0x9443,
      EVR_SQ, "FramePixelDataPropertiesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9444, 0x0028, 0x9444,
      EVR_CS, "GeometricalProperties", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9445, 0x0028, 0x9445,
      EVR_FL, "GeometricMaximumDistortion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9446, 0x0028, 0x9446,
      EVR_CS, "ImageProcessingApplied", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9454, 0x0028, 0x9454,
      EVR_CS, "MaskSelectionMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9474, 0x0028, 0x9474,
      EVR_CS, "LUTFunction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9478, 0x0028, 0x9478,
      EVR_FL, "MaskVisibilityPercentage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9501, 0x0028, 0x9501,
      EVR_SQ, "PixelShiftSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9502, 0x0028, 0x9502,
      EVR_SQ, "RegionPixelShiftSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9503, 0x0028, 0x9503,
      EVR_SS, "VerticesOfTheRegion", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9505, 0x0028, 0x9505,
      EVR_SQ, "MultiFramePresentationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9506, 0x0028, 0x9506,
      EVR_US, "PixelShiftFrameRange", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9507, 0x0028, 0x9507,
      EVR_US, "LUTFrameRange", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9520, 0x0028, 0x9520,
      EVR_DS, "ImageToEquipmentMappingMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0028, 0x9537, 0x0028, 0x9537,
      EVR_CS, "EquipmentCoordinateSystemIdentification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_LT, "ImageEnhancementID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_LT, "SubtractionMaskID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;2" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_DS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_SQ, "EdgeEnhancementSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Graphics Export V1.0" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_CDS Release 1" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_LT, "ZoomID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_FD, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
  , { 0x0029, 0x0000, 0x0029, 0x0000,
      EVR_DS, "Unknown", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0001, 0x0029, 0x0001,
      EVR_LT, "ImageEnhancement", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0001, 0x0029, 0x0001,
      EVR_US, "ConvolutionKernelSize", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0029, 0x0001, 0x0029, 0x0001,
      EVR_DS, "ZoomRectangle", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0001, 0x0029, 0x0001,
      EVR_FD, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0001, 0x0029, 0x0001,
      EVR_FD, "ImageCompressionFraction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0002, 0x0029, 0x0002,
      EVR_LT, "ConvolutionID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0002, 0x0029, 0x0002,
      EVR_DS, "ConvolutionKernelCoefficients", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0029, 0x0002, 0x0029, 0x0002,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;4" }
  , { 0x0029, 0x0002, 0x0029, 0x0002,
      EVR_FD, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0002, 0x0029, 0x0002,
      EVR_FD, "ImageQuality", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0003, 0x0029, 0x0003,
      EVR_LT, "ConvolutionType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0003, 0x0029, 0x0003,
      EVR_DS, "EdgeEnhancementGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CARDIO-D.R. 1.0" }
  , { 0x0029, 0x0003, 0x0029, 0x0003,
      EVR_DS, "ZoomFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0003, 0x0029, 0x0003,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0003, 0x0029, 0x0003,
      EVR_FD, "ImageBytesTransferred", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_LT, "ConvolutionKernelSizeID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_UN, "MaskingFunction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;2" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_LT, "PhotometricInterpretation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_US, "ZoomFunction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0004, 0x0029, 0x0004,
      EVR_SL, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0005, 0x0029, 0x0005,
      EVR_US, "ConvolutionKernelSize", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0005, 0x0029, 0x0005,
      EVR_FL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x0005, 0x0029, 0x0005,
      EVR_SL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;2" }
  , { 0x0029, 0x0005, 0x0029, 0x0005,
      EVR_DS, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0006, 0x0029, 0x0006,
      EVR_US, "ConvolutionKernel", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x0006, 0x0029, 0x0006,
      EVR_FL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x0006, 0x0029, 0x0006,
      EVR_DS, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0007, 0x0029, 0x0007,
      EVR_SL, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_CS, "CSAImageHeaderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_CS, "CSADataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA NON-IMAGE" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_CS, "MEDCOMOOGType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM OOG" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_SH, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_IS, "NumberOfExposureResults", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0029, 0x0008, 0x0029, 0x0008,
      EVR_CS, "MedComHeaderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0009, 0x0029, 0x0009,
      EVR_LO, "CSAImageHeaderVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0009, 0x0029, 0x0009,
      EVR_LO, "CSADataVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA NON-IMAGE" }
  , { 0x0029, 0x0009, 0x0029, 0x0009,
      EVR_LO, "MEDCOMOOGVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM OOG" }
  , { 0x0029, 0x0009, 0x0029, 0x0009,
      EVR_SH, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0009, 0x0029, 0x0009,
      EVR_LO, "MedComHeaderVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x000a, 0x0029, 0x000a,
      EVR_SS, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x000c, 0x0029, 0x000c,
      EVR_DS, "EnhancementGain", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x000c, 0x0029, 0x000c,
      EVR_UN, "ProprietaryMaskingParameters", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;2" }
  , { 0x0029, 0x000e, 0x0029, 0x000e,
      EVR_CS, "ZoomEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x000f, 0x0029, 0x000f,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x000f, 0x0029, 0x000f,
      EVR_CS, "ZoomSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_LT, "Commentline", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS IP" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_US, "ListOfGroupNumbers", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_LT, "WindowStyle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_OB, "CSAImageHeaderInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_US, "RowsOfSubmatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_OB, "CSADataInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA NON-IMAGE" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_OB, "MEDCOMOOGInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM OOG" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_OB, "MedComHeaderInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_DS, "FPMin", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_SH, "J2cParameterType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0010, 0x0029, 0x0010,
      EVR_US, "ListOfGroupNumbers", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_IS, "LineName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_US, "ColumnsOfSubmatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_IS, "ROIName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_IS, "AnnotationName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0011, 0x0029, 0x0011,
      EVR_US, "J2cPixelRepresentation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0012, 0x0029, 0x0012,
      EVR_LT, "LineNameFont", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0012, 0x0029, 0x0012,
      EVR_LT, "ROINameFont", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0012, 0x0029, 0x0012,
      EVR_LT, "AnnotationFont", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0012, 0x0029, 0x0012,
      EVR_US, "J2cBitsAllocated", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_UL, "LineNameDisplay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_LT, "ROINormalColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_LT, "AnnotationTextForegroundColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0013, 0x0029, 0x0013,
      EVR_US, "J2cPixelShiftValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0014, 0x0029, 0x0014,
      EVR_LT, "LineNormalColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0014, 0x0029, 0x0014,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0014, 0x0029, 0x0014,
      EVR_UL, "ROIFillPattern", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0014, 0x0029, 0x0014,
      EVR_LT, "AnnotationTextBackgroundColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0014, 0x0029, 0x0014,
      EVR_US, "J2cPlanarConfiguration", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_UL, "LineType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_LT, "ListOfShadowOwnerCodes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_UL, "ROIBpSeg", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_UL, "AnnotationTextBackingMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_SL, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_DS, "J2cRescaleIntercept", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0015, 0x0029, 0x0015,
      EVR_LT, "ListOfShadowOwnerCodes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0016, 0x0029, 0x0016,
      EVR_UL, "LineThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0016, 0x0029, 0x0016,
      EVR_UN, "ROIBpSegPairs", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0016, 0x0029, 0x0016,
      EVR_UL, "AnnotationTextJustification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0016, 0x0029, 0x0016,
      EVR_SL, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0017, 0x0029, 0x0017,
      EVR_UL, "LineStyle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0017, 0x0029, 0x0017,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0017, 0x0029, 0x0017,
      EVR_UL, "ROISeedSpace", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0017, 0x0029, 0x0017,
      EVR_UL, "AnnotationTextLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0017, 0x0029, 0x0017,
      EVR_SL, "LowerRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_UL, "LineDashLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_CS, "CSASeriesHeaderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_UN, "ROISeeds", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_LT, "AnnotationTextString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0018, 0x0029, 0x0018,
      EVR_SL, "UpperRangeOfPixels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0019, 0x0029, 0x0019,
      EVR_UL, "LineInteractivity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0019, 0x0029, 0x0019,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0019, 0x0029, 0x0019,
      EVR_LO, "CSASeriesHeaderVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0019, 0x0029, 0x0019,
      EVR_UL, "ROILineThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0019, 0x0029, 0x0019,
      EVR_UL, "AnnotationTextAttachMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x001a, 0x0029, 0x001a,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x001a, 0x0029, 0x001a,
      EVR_SL, "LengthOfTotalHeaderInBytes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x001b, 0x0029, 0x001b,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x001b, 0x0029, 0x001b,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x001c, 0x0029, 0x001c,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x001c, 0x0029, 0x001c,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x001d, 0x0029, 0x001d,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x001e, 0x0029, 0x001e,
      EVR_CS, "ImageEnhancementEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x001e, 0x0029, 0x001e,
      EVR_CS, "SubtractionMaskEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;2" }
  , { 0x0029, 0x001e, 0x0029, 0x001e,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x001f, 0x0029, 0x001f,
      EVR_CS, "ImageEnhancementSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;3" }
  , { 0x0029, 0x001f, 0x0029, 0x001f,
      EVR_CS, "SubtractionMaskSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;2" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_DS, "PixelAspectRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_DS, "EdgeEnhancementCoefficient", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS IP" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_LT, "LineMeasurementColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_FL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_CS, "ImageScanningDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_US, "ListOfElementNumbers", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_LT, "PixelQualityCode", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_OB, "CSASeriesHeaderInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA HEADER" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_UL, "ROILineStyle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_UL, "AnnotationTextCursorMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_LO, "PixelDataMD5SumPerFrame", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_OB, "MedComHistoryInformation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_DS, "FPMax", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0020, 0x0029, 0x0020,
      EVR_US, "ListOfElementNumbers", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_LT, "LineMeasurementFont", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_FL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_UL, "ROILineDashLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_UL, "AnnotationTextShadowOffsetX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0021, 0x0029, 0x0021,
      EVR_US, "HistogramPercentileLabels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_UL, "LineMeasurementDashLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_IS, "PixelQualityValue", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_UL, "ROIInteractivity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_UL, "AnnotationTextShadowOffsetY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0022, 0x0029, 0x0022,
      EVR_FD, "HistogramPercentileValues", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x0029, 0x0023, 0x0029, 0x0023,
      EVR_UL, "LinePointSpace", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0023, 0x0029, 0x0023,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0023, 0x0029, 0x0023,
      EVR_UL, "ROINamePosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0023, 0x0029, 0x0023,
      EVR_LT, "AnnotationLineColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0024, 0x0029, 0x0024,
      EVR_FD, "LinePoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0024, 0x0029, 0x0024,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0024, 0x0029, 0x0024,
      EVR_UL, "ROINameDisplay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0024, 0x0029, 0x0024,
      EVR_UL, "AnnotationLineThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0025, 0x0029, 0x0025,
      EVR_UL, "LineControlPointSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0025, 0x0029, 0x0025,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0025, 0x0029, 0x0025,
      EVR_LT, "ROILabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0025, 0x0029, 0x0025,
      EVR_UL, "AnnotationLineType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0025, 0x0029, 0x0025,
      EVR_LO, "ProcessedPixelDataQuality", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0026, 0x0029, 0x0026,
      EVR_UL, "LineControlPointSpace", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0026, 0x0029, 0x0026,
      EVR_UL, "ROIShape", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0026, 0x0029, 0x0026,
      EVR_UL, "AnnotationLineStyle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0026, 0x0029, 0x0026,
      EVR_SS, "VersionOfHeaderStructure", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0027, 0x0029, 0x0027,
      EVR_FD, "LineControlPoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0027, 0x0029, 0x0027,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0027, 0x0029, 0x0027,
      EVR_FD, "ROIShapeTilt", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0027, 0x0029, 0x0027,
      EVR_UL, "AnnotationLineDashLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0028, 0x0029, 0x0028,
      EVR_LT, "LineLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0028, 0x0029, 0x0028,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0028, 0x0029, 0x0028,
      EVR_UL, "ROIShapePointsCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0028, 0x0029, 0x0028,
      EVR_UL, "AnnotationLineAttachMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0029, 0x0029, 0x0029,
      EVR_UL, "LineDontSave", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Line V1.0" }
  , { 0x0029, 0x0029, 0x0029, 0x0029,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0029, 0x0029, 0x0029,
      EVR_UL, "ROIShapePointsSpace", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0029, 0x0029, 0x0029,
      EVR_UL, "AnnotationLinePointCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS IP" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_CS, "ExtendedReadingSizeValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_US, "ListOfTotalDisplayLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_FD, "ROIShapePoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_FD, "AnnotationLinePoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_DS, "ScaledMinimum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0030, 0x0029, 0x0030,
      EVR_US, "ListOfTotalDisplayLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_DS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_UL, "ROIShapeControlPointsCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIGISCAN IMAGE" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_UL, "AnnotationLineControlSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0031, 0x0029, 0x0031,
      EVR_LO, "PMTFInformation1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_DS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_UL, "ROIShapeControlPointsSpace", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIGISCAN IMAGE" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_LT, "AnnotationMarkerColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0032, 0x0029, 0x0032,
      EVR_UL, "PMTFInformation2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0033, 0x0029, 0x0033,
      EVR_FD, "ROIShapeControlPoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0033, 0x0029, 0x0033,
      EVR_UL, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0033, 0x0029, 0x0033,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIGISCAN IMAGE" }
  , { 0x0029, 0x0033, 0x0029, 0x0033,
      EVR_UL, "AnnotationMarkerType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0033, 0x0029, 0x0033,
      EVR_UL, "PMTFInformation3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_US, "MagnificationReductionRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_UL, "ROIDontSave", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette ROI V1.0" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIGISCAN IMAGE" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_UL, "AnnotationMarkerSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_SL, "AdvantageCompOverflow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0034, 0x0029, 0x0034,
      EVR_CS, "PMTFInformation4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0035, 0x0029, 0x0035,
      EVR_FD, "AnnotationMarkerLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0035, 0x0029, 0x0035,
      EVR_SL, "AdvantageCompUnderflow", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_IMPS_01" }
  , { 0x0029, 0x0035, 0x0029, 0x0035,
      EVR_UL, "PMTFInformation5", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0036, 0x0029, 0x0036,
      EVR_UL, "AnnotationMarkerAttachMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0037, 0x0029, 0x0037,
      EVR_LT, "AnnotationGeomColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0038, 0x0029, 0x0038,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0038, 0x0029, 0x0038,
      EVR_UL, "AnnotationGeomThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0039, 0x0029, 0x0039,
      EVR_UL, "AnnotationGeomLineStyle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS IP" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_LT, "ListOfDisplayPrefix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_LT, "MagnifyingGlassID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_UL, "AnnotationGeomDashLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_SQ, "ApplicationHeaderSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_DS, "ScaledMaximum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0040, 0x0029, 0x0040,
      EVR_LT, "ListOfDisplayPrefix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0041, 0x0029, 0x0041,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Sequence Ids V1.0" }
  , { 0x0029, 0x0041, 0x0029, 0x0041,
      EVR_DS, "MagnifyingGlassRectangle", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0041, 0x0029, 0x0041,
      EVR_UL, "AnnotationGeomFillPattern", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0041, 0x0029, 0x0041,
      EVR_CS, "ApplicationHeaderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0042, 0x0029, 0x0042,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Sequence Ids V1.0" }
  , { 0x0029, 0x0042, 0x0029, 0x0042,
      EVR_UL, "AnnotationInteractivity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0042, 0x0029, 0x0042,
      EVR_LO, "ApplicationHeaderID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0043, 0x0029, 0x0043,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0043, 0x0029, 0x0043,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Sequence Ids V1.0" }
  , { 0x0029, 0x0043, 0x0029, 0x0043,
      EVR_DS, "MagnifyingGlassFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0043, 0x0029, 0x0043,
      EVR_FD, "AnnotationArrowLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0043, 0x0029, 0x0043,
      EVR_LO, "ApplicationHeaderVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0044, 0x0029, 0x0044,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0044, 0x0029, 0x0044,
      EVR_CS, "LineDensityCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0029, 0x0044, 0x0029, 0x0044,
      EVR_US, "MagnifyingGlassFunction", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0044, 0x0029, 0x0044,
      EVR_FD, "AnnotationArrowAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x0044, 0x0029, 0x0044,
      EVR_OB, "ApplicationHeaderInfo", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0045, 0x0029, 0x0045,
      EVR_UL, "AnnotationDontSave", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette Annot V1.0" }
  , { 0x0029, 0x004c, 0x0029, 0x004c,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x004e, 0x0029, 0x004e,
      EVR_CS, "MagnifyingGlassEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x004f, 0x0029, 0x004f,
      EVR_CS, "MagnifyingGlassSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1;1" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_LT, "SceneText", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_CS, "DataCompressionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_LT, "ArchiveCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_LT, "ListOfDisplayPostfix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_US, "OriginOfSubmatrix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;5" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_LO, "WorkflowControlFlags", 8, 8, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_DS, "WindowMinimum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0050, 0x0029, 0x0050,
      EVR_LT, "ListOfDisplayPostfix", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0051, 0x0029, 0x0051,
      EVR_LT, "ExposureCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0051, 0x0029, 0x0051,
      EVR_CS, "ArchiveManagementFlagKeepOnline", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0052, 0x0029, 0x0052,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0052, 0x0029, 0x0052,
      EVR_LT, "SortCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0052, 0x0029, 0x0052,
      EVR_CS, "ArchiveManagementFlagDoNotArchive", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0053, 0x0029, 0x0053,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0053, 0x0029, 0x0053,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0053, 0x0029, 0x0053,
      EVR_CS, "ImageLocationStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0054, 0x0029, 0x0054,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0054, 0x0029, 0x0054,
      EVR_DS, "EstimatedRetrieveTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0055, 0x0029, 0x0055,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0055, 0x0029, 0x0055,
      EVR_CS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;5" }
  , { 0x0029, 0x0055, 0x0029, 0x0055,
      EVR_DS, "DataSizeOfRetrievedImages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0056, 0x0029, 0x0056,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0057, 0x0029, 0x0057,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Silhouette V1.0" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "ImageText", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "CompressionAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI Release 1" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "Splash", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_US, "ListOfTextPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "CompressionAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LO, "SeriesWorkflowStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER2" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_DS, "WindowMaximum", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_LT, "CompressionAlgorithm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI RELEASE 1" }
  , { 0x0029, 0x0060, 0x0029, 0x0060,
      EVR_US, "ListOfTextPosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0061, 0x0029, 0x0061,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0061, 0x0029, 0x0061,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0067, 0x0029, 0x0067,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0067, 0x0029, 0x0067,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0068, 0x0029, 0x0068,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x006a, 0x0029, 0x006a,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x006b, 0x0029, 0x006b,
      EVR_US, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_IS, "PixelShiftHorizontal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ISG shadow" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_LT, "ListOfTextConcatenation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED HG" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_SQ, "SiemensLinkSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_LT, "ListOfTextConcatenation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED MG" }
  , { 0x0029, 0x0070, 0x0029, 0x0070,
      EVR_LT, "WindowID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0071, 0x0029, 0x0071,
      EVR_AT, "ReferencedTag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0071, 0x0029, 0x0071,
      EVR_DS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0071, 0x0029, 0x0071,
      EVR_CS, "VideoInvertSubtracted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0072, 0x0029, 0x0072,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0072, 0x0029, 0x0072,
      EVR_CS, "ReferencedTagType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0072, 0x0029, 0x0072,
      EVR_CS, "VideoInvertNonsubtracted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0073, 0x0029, 0x0073,
      EVR_UL, "ReferencedValueLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0074, 0x0029, 0x0074,
      EVR_CS, "ReferencedObjectDeviceType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0075, 0x0029, 0x0075,
      EVR_OB, "ReferencedObjectDeviceLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0076, 0x0029, 0x0076,
      EVR_OB, "ReferencedObjectDeviceID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MEDCOM HEADER" }
  , { 0x0029, 0x0077, 0x0029, 0x0077,
      EVR_CS, "WindowSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0078, 0x0029, 0x0078,
      EVR_LT, "ECGDisplayPrintingID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0079, 0x0029, 0x0079,
      EVR_CS, "ECGDisplayPrinting", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x007e, 0x0029, 0x007e,
      EVR_CS, "ECGDisplayPrintingEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x007f, 0x0029, 0x007f,
      EVR_CS, "ECGDisplayPrintingSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0080, 0x0029, 0x0080,
      EVR_IS, "PixelShiftVertical", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0080, 0x0029, 0x0080,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ISG shadow" }
  , { 0x0029, 0x0080, 0x0029, 0x0080,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0080, 0x0029, 0x0080,
      EVR_IS, "ViewCenter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0080, 0x0029, 0x0080,
      EVR_LT, "PhysiologicalDisplayID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0081, 0x0029, 0x0081,
      EVR_IS, "ViewSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0081, 0x0029, 0x0081,
      EVR_US, "PreferredPhysiologicalChannelDisplay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0082, 0x0029, 0x0082,
      EVR_IS, "ViewZoom", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x0083, 0x0029, 0x0083,
      EVR_IS, "ViewTransform", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x0029, 0x008e, 0x0029, 0x008e,
      EVR_CS, "PhysiologicalDisplayEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x008f, 0x0029, 0x008f,
      EVR_CS, "PhysiologicalDisplaySelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x0090, 0x0029, 0x0090,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "CAMTRONICS" }
  , { 0x0029, 0x0090, 0x0029, 0x0090,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ISG shadow" }
  , { 0x0029, 0x0090, 0x0029, 0x0090,
      EVR_LO, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x0099, 0x0029, 0x0099,
      EVR_LT, "ShutterType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x009a, 0x0029, 0x009a,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;4" }
  , { 0x0029, 0x00a0, 0x0029, 0x00a0,
      EVR_US, "RowsOfRectangularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00a1, 0x0029, 0x00a1,
      EVR_US, "ColumnsOfRectangularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00a2, 0x0029, 0x00a2,
      EVR_US, "OriginOfRectangularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00b0, 0x0029, 0x00b0,
      EVR_US, "RadiusOfCircularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00b2, 0x0029, 0x00b2,
      EVR_US, "OriginOfCircularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00c0, 0x0029, 0x00c0,
      EVR_LT, "FunctionalShutterID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x00c0, 0x0029, 0x00c0,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00c1, 0x0029, 0x00c1,
      EVR_US, "ContourOfIrregularShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED DISPLAY" }
  , { 0x0029, 0x00c1, 0x0029, 0x00c1,
      EVR_US, "FieldOfShutter", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x00c1, 0x0029, 0x00c1,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00c2, 0x0029, 0x00c2,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00c3, 0x0029, 0x00c3,
      EVR_IS, "ScanResolution", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x00c3, 0x0029, 0x00c3,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00c4, 0x0029, 0x00c4,
      EVR_IS, "FieldOfView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x00c4, 0x0029, 0x00c4,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00c5, 0x0029, 0x00c5,
      EVR_LT, "FieldOfShutterRectangle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x00c5, 0x0029, 0x00c5,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;3" }
  , { 0x0029, 0x00cd, 0x0029, 0x00cd,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;1" }
  , { 0x0029, 0x00ce, 0x0029, 0x00ce,
      EVR_CS, "ShutterEnableStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x00cf, 0x0029, 0x00cf,
      EVR_CS, "ShutterSelectStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x0029, 0x00d0, 0x0029, 0x00d0,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x00d1, 0x0029, 0x00d1,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-GV-CT Release 1" }
  , { 0x0029, 0x00d5, 0x0029, 0x00d5,
      EVR_LT, "SliceThickness", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/PART" }
  , { 0x0029, 0x00e0, 0x0029, 0x00e0,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P-Private_ICS Release 1;4" }
  , { 0x0029, 0x1010, 0x0029, 0x1010,
      EVR_OB, "OriginalMeasuringData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "OCULUS Optikgeraete GmbH" }
  , { 0x0029, 0x1012, 0x0029, 0x1012,
      EVR_UL, "OriginalMeasuringDataLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "OCULUS Optikgeraete GmbH" }
  , { 0x0029, 0x1020, 0x0029, 0x1020,
      EVR_OB, "OriginalMeasuringRawData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "OCULUS Optikgeraete GmbH" }
  , { 0x0029, 0x1022, 0x0029, 0x1022,
      EVR_UL, "OriginalMeasuringRawDataLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "OCULUS Optikgeraete GmbH" }
  , { 0x0031, 0x0000, 0x0031, 0x0000,
      EVR_CS, "StudyStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA PACS Archive Mirroring 1.0" }
  , { 0x0031, 0x0001, 0x0031, 0x0001,
      EVR_UL, "DateTimeVerified", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AGFA PACS Archive Mirroring 1.0" }
  , { 0x0031, 0x0010, 0x0031, 0x0010,
      EVR_LT, "RequestUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0031, 0x0012, 0x0031, 0x0012,
      EVR_LT, "ExaminationReason", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0031, 0x0030, 0x0031, 0x0030,
      EVR_DA, "RequestedDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0031, 0x0032, 0x0031, 0x0032,
      EVR_TM, "WorklistRequestStartTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0031, 0x0033, 0x0031, 0x0033,
      EVR_TM, "WorklistRequestEndTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0031, 0x0045, 0x0031, 0x0045,
      EVR_LT, "RequestingPhysician", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0031, 0x004a, 0x0031, 0x004a,
      EVR_TM, "RequestedTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0031, 0x0050, 0x0031, 0x0050,
      EVR_LT, "RequestedPhysician", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
  , { 0x0031, 0x0080, 0x0031, 0x0080,
      EVR_LT, "RequestedLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x0032, 0x000a, 0x0032, 0x000a,
      EVR_CS, "RETIRED_StudyStatusID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x000c, 0x0032, 0x000c,
      EVR_CS, "RETIRED_StudyPriorityID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x0012, 0x0032, 0x0012,
      EVR_LO, "RETIRED_StudyIDIssuer", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x0032, 0x0032, 0x0032,
      EVR_DA, "RETIRED_StudyVerifiedDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x0033, 0x0032, 0x0033,
      EVR_TM, "RETIRED_StudyVerifiedTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x0034, 0x0032, 0x0034,
      EVR_DA, "RETIRED_StudyReadDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x0035, 0x0032, 0x0035,
      EVR_TM, "RETIRED_StudyReadTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1000, 0x0032, 0x1000,
      EVR_DA, "RETIRED_ScheduledStudyStartDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1001, 0x0032, 0x1001,
      EVR_TM, "RETIRED_ScheduledStudyStartTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1010, 0x0032, 0x1010,
      EVR_DA, "RETIRED_ScheduledStudyStopDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1011, 0x0032, 0x1011,
      EVR_TM, "RETIRED_ScheduledStudyStopTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1020, 0x0032, 0x1020,
      EVR_LO, "RETIRED_ScheduledStudyLocation", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1021, 0x0032, 0x1021,
      EVR_AE, "RETIRED_ScheduledStudyLocationAETitle", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1030, 0x0032, 0x1030,
      EVR_LO, "RETIRED_ReasonForStudy", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1031, 0x0032, 0x1031,
      EVR_SQ, "RequestingPhysicianIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1032, 0x0032, 0x1032,
      EVR_PN, "RequestingPhysician", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1033, 0x0032, 0x1033,
      EVR_LO, "RequestingService", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1034, 0x0032, 0x1034,
      EVR_SQ, "RequestingServiceCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1040, 0x0032, 0x1040,
      EVR_DA, "RETIRED_StudyArrivalDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1041, 0x0032, 0x1041,
      EVR_TM, "RETIRED_StudyArrivalTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1050, 0x0032, 0x1050,
      EVR_DA, "RETIRED_StudyCompletionDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1051, 0x0032, 0x1051,
      EVR_TM, "RETIRED_StudyCompletionTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1055, 0x0032, 0x1055,
      EVR_CS, "RETIRED_StudyComponentStatusID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1060, 0x0032, 0x1060,
      EVR_LO, "RequestedProcedureDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1064, 0x0032, 0x1064,
      EVR_SQ, "RequestedProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1065, 0x0032, 0x1065,
      EVR_SQ, "RequestedLateralityCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1066, 0x0032, 0x1066,
      EVR_UT, "ReasonForVisit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1067, 0x0032, 0x1067,
      EVR_SQ, "ReasonForVisitCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x1070, 0x0032, 0x1070,
      EVR_LO, "RequestedContrastAgent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0032, 0x4000, 0x0032, 0x4000,
      EVR_LT, "RETIRED_StudyComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0033, 0x0001, 0x0033, 0x0001,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GNHD_01" }
  , { 0x0033, 0x0002, 0x0033, 0x0002,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_CTHD_01" }
  , { 0x0033, 0x0002, 0x0033, 0x0002,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_GNHD_01" }
  , { 0x0033, 0x0005, 0x0033, 0x0005,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_YMHD_01" }
  , { 0x0033, 0x0006, 0x0033, 0x0006,
      EVR_UN, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_YMHD_01" }
  , { 0x0033, 0x0010, 0x0033, 0x0010,
      EVR_LT, "PatientStudyUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS RIS" }
#endif
  , { 0x0034, 0x0001, 0x0034, 0x0001,
      EVR_SQ, "FlowIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0002, 0x0034, 0x0002,
      EVR_OB, "FlowIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0003, 0x0034, 0x0003,
      EVR_UI, "FlowTransferSyntaxUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0004, 0x0034, 0x0004,
      EVR_UL, "FlowRTPSamplingRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0005, 0x0034, 0x0005,
      EVR_OB, "SourceIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0007, 0x0034, 0x0007,
      EVR_OB, "FrameOriginTimestamp", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0008, 0x0034, 0x0008,
      EVR_CS, "IncludesImagingSubject", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x0009, 0x0034, 0x0009,
      EVR_SQ, "FrameUsefulnessGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x000a, 0x0034, 0x000a,
      EVR_SQ, "RealTimeBulkDataFlowSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x000b, 0x0034, 0x000b,
      EVR_SQ, "CameraPositionGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x000c, 0x0034, 0x000c,
      EVR_CS, "IncludesInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0034, 0x000d, 0x0034, 0x000d,
      EVR_SQ, "TimeOfFrameGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0037, 0x0010, 0x0037, 0x0010,
      EVR_LO, "ReferringDepartment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
  , { 0x0037, 0x0020, 0x0037, 0x0020,
      EVR_US, "ScreenNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
  , { 0x0037, 0x0040, 0x0037, 0x0040,
      EVR_SH, "LeftOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
  , { 0x0037, 0x0042, 0x0037, 0x0042,
      EVR_SH, "RightOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
  , { 0x0037, 0x0050, 0x0037, 0x0050,
      EVR_CS, "Inversion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
  , { 0x0037, 0x0060, 0x0037, 0x0060,
      EVR_US, "DSA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_DRS_1" }
#endif
  , { 0x0038, 0x0004, 0x0038, 0x0004,
      EVR_SQ, "RETIRED_ReferencedPatientAliasSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0008, 0x0038, 0x0008,
      EVR_CS, "VisitStatusID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0010, 0x0038, 0x0010,
      EVR_LO, "AdmissionID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0011, 0x0038, 0x0011,
      EVR_LO, "RETIRED_IssuerOfAdmissionID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0014, 0x0038, 0x0014,
      EVR_SQ, "IssuerOfAdmissionIDSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0016, 0x0038, 0x0016,
      EVR_LO, "RouteOfAdmissions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x001a, 0x0038, 0x001a,
      EVR_DA, "RETIRED_ScheduledAdmissionDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x001b, 0x0038, 0x001b,
      EVR_TM, "RETIRED_ScheduledAdmissionTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x001c, 0x0038, 0x001c,
      EVR_DA, "RETIRED_ScheduledDischargeDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x001d, 0x0038, 0x001d,
      EVR_TM, "RETIRED_ScheduledDischargeTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x001e, 0x0038, 0x001e,
      EVR_LO, "RETIRED_ScheduledPatientInstitutionResidence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0020, 0x0038, 0x0020,
      EVR_DA, "AdmittingDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0021, 0x0038, 0x0021,
      EVR_TM, "AdmittingTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0030, 0x0038, 0x0030,
      EVR_DA, "RETIRED_DischargeDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0032, 0x0038, 0x0032,
      EVR_TM, "RETIRED_DischargeTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0040, 0x0038, 0x0040,
      EVR_LO, "RETIRED_DischargeDiagnosisDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0044, 0x0038, 0x0044,
      EVR_SQ, "RETIRED_DischargeDiagnosisCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0050, 0x0038, 0x0050,
      EVR_LO, "SpecialNeeds", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0060, 0x0038, 0x0060,
      EVR_LO, "ServiceEpisodeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0061, 0x0038, 0x0061,
      EVR_LO, "RETIRED_IssuerOfServiceEpisodeID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0062, 0x0038, 0x0062,
      EVR_LO, "ServiceEpisodeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0064, 0x0038, 0x0064,
      EVR_SQ, "IssuerOfServiceEpisodeIDSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0100, 0x0038, 0x0100,
      EVR_SQ, "PertinentDocumentsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0101, 0x0038, 0x0101,
      EVR_SQ, "PertinentResourcesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0102, 0x0038, 0x0102,
      EVR_LO, "ResourceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0300, 0x0038, 0x0300,
      EVR_LO, "CurrentPatientLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0400, 0x0038, 0x0400,
      EVR_LO, "PatientInstitutionResidence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0500, 0x0038, 0x0500,
      EVR_LO, "PatientState", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x0502, 0x0038, 0x0502,
      EVR_SQ, "PatientClinicalTrialParticipationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0038, 0x4000, 0x0038, 0x4000,
      EVR_LT, "VisitComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0039, 0x0080, 0x0039, 0x0080,
      EVR_IS, "PrivateEntityNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_DPO" }
  , { 0x0039, 0x0085, 0x0039, 0x0085,
      EVR_DA, "PrivateEntityDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_DPO" }
  , { 0x0039, 0x0090, 0x0039, 0x0090,
      EVR_TM, "PrivateEntityTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_DPO" }
  , { 0x0039, 0x0095, 0x0039, 0x0095,
      EVR_LO, "PrivateEntityLaunchCommand", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_DPO" }
  , { 0x0039, 0x00aa, 0x0039, 0x00aa,
      EVR_CS, "PrivateEntityType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_DPO" }
#endif
  , { 0x003a, 0x0004, 0x003a, 0x0004,
      EVR_CS, "WaveformOriginality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0005, 0x003a, 0x0005,
      EVR_US, "NumberOfWaveformChannels", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0010, 0x003a, 0x0010,
      EVR_UL, "NumberOfWaveformSamples", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x001a, 0x003a, 0x001a,
      EVR_DS, "SamplingFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0020, 0x003a, 0x0020,
      EVR_SH, "MultiplexGroupLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0200, 0x003a, 0x0200,
      EVR_SQ, "ChannelDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0202, 0x003a, 0x0202,
      EVR_IS, "WaveformChannelNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0203, 0x003a, 0x0203,
      EVR_SH, "ChannelLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0205, 0x003a, 0x0205,
      EVR_CS, "ChannelStatus", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0208, 0x003a, 0x0208,
      EVR_SQ, "ChannelSourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0209, 0x003a, 0x0209,
      EVR_SQ, "ChannelSourceModifiersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x020a, 0x003a, 0x020a,
      EVR_SQ, "SourceWaveformSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x020c, 0x003a, 0x020c,
      EVR_LO, "ChannelDerivationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0210, 0x003a, 0x0210,
      EVR_DS, "ChannelSensitivity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0211, 0x003a, 0x0211,
      EVR_SQ, "ChannelSensitivityUnitsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0212, 0x003a, 0x0212,
      EVR_DS, "ChannelSensitivityCorrectionFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0213, 0x003a, 0x0213,
      EVR_DS, "ChannelBaseline", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0214, 0x003a, 0x0214,
      EVR_DS, "ChannelTimeSkew", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0215, 0x003a, 0x0215,
      EVR_DS, "ChannelSampleSkew", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0218, 0x003a, 0x0218,
      EVR_DS, "ChannelOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x021a, 0x003a, 0x021a,
      EVR_US, "WaveformBitsStored", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0220, 0x003a, 0x0220,
      EVR_DS, "FilterLowFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0221, 0x003a, 0x0221,
      EVR_DS, "FilterHighFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0222, 0x003a, 0x0222,
      EVR_DS, "NotchFilterFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0223, 0x003a, 0x0223,
      EVR_DS, "NotchFilterBandwidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0230, 0x003a, 0x0230,
      EVR_FL, "WaveformDataDisplayScale", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0231, 0x003a, 0x0231,
      EVR_US, "WaveformDisplayBackgroundCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0240, 0x003a, 0x0240,
      EVR_SQ, "WaveformPresentationGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0241, 0x003a, 0x0241,
      EVR_US, "PresentationGroupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0242, 0x003a, 0x0242,
      EVR_SQ, "ChannelDisplaySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0244, 0x003a, 0x0244,
      EVR_US, "ChannelRecommendedDisplayCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0245, 0x003a, 0x0245,
      EVR_FL, "ChannelPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0246, 0x003a, 0x0246,
      EVR_CS, "DisplayShadingFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0247, 0x003a, 0x0247,
      EVR_FL, "FractionalChannelDisplayScale", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0248, 0x003a, 0x0248,
      EVR_FL, "AbsoluteChannelDisplayScale", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0300, 0x003a, 0x0300,
      EVR_SQ, "MultiplexedAudioChannelsDescriptionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0301, 0x003a, 0x0301,
      EVR_IS, "ChannelIdentificationCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0302, 0x003a, 0x0302,
      EVR_CS, "ChannelMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0310, 0x003a, 0x0310,
      EVR_UI, "MultiplexGroupUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0311, 0x003a, 0x0311,
      EVR_DS, "PowerlineFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0312, 0x003a, 0x0312,
      EVR_SQ, "ChannelImpedanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0313, 0x003a, 0x0313,
      EVR_DS, "ImpedanceValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0314, 0x003a, 0x0314,
      EVR_DT, "ImpedanceMeasurementDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0315, 0x003a, 0x0315,
      EVR_DS, "ImpedanceMeasurementFrequency", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0316, 0x003a, 0x0316,
      EVR_CS, "ImpedanceMeasurementCurrentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0317, 0x003a, 0x0317,
      EVR_CS, "WaveformAmplifierType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0318, 0x003a, 0x0318,
      EVR_SQ, "FilterLowFrequencyCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0319, 0x003a, 0x0319,
      EVR_SQ, "FilterHighFrequencyCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0320, 0x003a, 0x0320,
      EVR_SQ, "SummarizedFilterLookupTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0321, 0x003a, 0x0321,
      EVR_SQ, "NotchFilterCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0322, 0x003a, 0x0322,
      EVR_CS, "WaveformFilterType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0323, 0x003a, 0x0323,
      EVR_SQ, "AnalogFilterCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0324, 0x003a, 0x0324,
      EVR_DS, "AnalogFilterRollOff", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0325, 0x003a, 0x0325,
      EVR_SQ, "AnalogFilterTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0326, 0x003a, 0x0326,
      EVR_SQ, "DigitalFilterCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0327, 0x003a, 0x0327,
      EVR_IS, "DigitalFilterOrder", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0328, 0x003a, 0x0328,
      EVR_SQ, "DigitalFilterTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x0329, 0x003a, 0x0329,
      EVR_ST, "WaveformFilterDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x032a, 0x003a, 0x032a,
      EVR_SQ, "FilterLookupTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x032b, 0x003a, 0x032b,
      EVR_ST, "FilterLookupTableDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x032c, 0x003a, 0x032c,
      EVR_SQ, "FrequencyEncodingCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x032d, 0x003a, 0x032d,
      EVR_SQ, "MagnitudeEncodingCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x003a, 0x032e, 0x003a, 0x032e,
      EVR_OD, "FilterLookupTableData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0001, 0x0040, 0x0001,
      EVR_AE, "ScheduledStationAETitle", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0002, 0x0040, 0x0002,
      EVR_DA, "ScheduledProcedureStepStartDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0003, 0x0040, 0x0003,
      EVR_TM, "ScheduledProcedureStepStartTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0004, 0x0040, 0x0004,
      EVR_DA, "ScheduledProcedureStepEndDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0005, 0x0040, 0x0005,
      EVR_TM, "ScheduledProcedureStepEndTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0006, 0x0040, 0x0006,
      EVR_PN, "ScheduledPerformingPhysicianName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0007, 0x0040, 0x0007,
      EVR_LO, "ScheduledProcedureStepDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0008, 0x0040, 0x0008,
      EVR_SQ, "ScheduledProtocolCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0009, 0x0040, 0x0009,
      EVR_SH, "ScheduledProcedureStepID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x000a, 0x0040, 0x000a,
      EVR_SQ, "StageCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x000b, 0x0040, 0x000b,
      EVR_SQ, "ScheduledPerformingPhysicianIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0010, 0x0040, 0x0010,
      EVR_SH, "ScheduledStationName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0011, 0x0040, 0x0011,
      EVR_SH, "ScheduledProcedureStepLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0012, 0x0040, 0x0012,
      EVR_LO, "PreMedication", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0020, 0x0040, 0x0020,
      EVR_CS, "ScheduledProcedureStepStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0026, 0x0040, 0x0026,
      EVR_SQ, "OrderPlacerIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0027, 0x0040, 0x0027,
      EVR_SQ, "OrderFillerIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0031, 0x0040, 0x0031,
      EVR_UT, "LocalNamespaceEntityID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0032, 0x0040, 0x0032,
      EVR_UT, "UniversalEntityID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0033, 0x0040, 0x0033,
      EVR_CS, "UniversalEntityIDType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0035, 0x0040, 0x0035,
      EVR_CS, "IdentifierTypeCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0036, 0x0040, 0x0036,
      EVR_SQ, "AssigningFacilitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0039, 0x0040, 0x0039,
      EVR_SQ, "AssigningJurisdictionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x003a, 0x0040, 0x003a,
      EVR_SQ, "AssigningAgencyOrDepartmentCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0100, 0x0040, 0x0100,
      EVR_SQ, "ScheduledProcedureStepSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0220, 0x0040, 0x0220,
      EVR_SQ, "ReferencedNonImageCompositeSOPInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0241, 0x0040, 0x0241,
      EVR_AE, "PerformedStationAETitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0242, 0x0040, 0x0242,
      EVR_SH, "PerformedStationName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0243, 0x0040, 0x0243,
      EVR_SH, "PerformedLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0244, 0x0040, 0x0244,
      EVR_DA, "PerformedProcedureStepStartDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0245, 0x0040, 0x0245,
      EVR_TM, "PerformedProcedureStepStartTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0250, 0x0040, 0x0250,
      EVR_DA, "PerformedProcedureStepEndDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0251, 0x0040, 0x0251,
      EVR_TM, "PerformedProcedureStepEndTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0252, 0x0040, 0x0252,
      EVR_CS, "PerformedProcedureStepStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0253, 0x0040, 0x0253,
      EVR_SH, "PerformedProcedureStepID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0254, 0x0040, 0x0254,
      EVR_LO, "PerformedProcedureStepDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0255, 0x0040, 0x0255,
      EVR_LO, "PerformedProcedureTypeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0260, 0x0040, 0x0260,
      EVR_SQ, "PerformedProtocolCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0261, 0x0040, 0x0261,
      EVR_CS, "PerformedProtocolType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0270, 0x0040, 0x0270,
      EVR_SQ, "ScheduledStepAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0275, 0x0040, 0x0275,
      EVR_SQ, "RequestAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0280, 0x0040, 0x0280,
      EVR_ST, "CommentsOnThePerformedProcedureStep", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0281, 0x0040, 0x0281,
      EVR_SQ, "PerformedProcedureStepDiscontinuationReasonCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0293, 0x0040, 0x0293,
      EVR_SQ, "QuantitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0294, 0x0040, 0x0294,
      EVR_DS, "Quantity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0295, 0x0040, 0x0295,
      EVR_SQ, "MeasuringUnitsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0296, 0x0040, 0x0296,
      EVR_SQ, "BillingItemSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0300, 0x0040, 0x0300,
      EVR_US, "RETIRED_TotalTimeOfFluoroscopy", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0301, 0x0040, 0x0301,
      EVR_US, "RETIRED_TotalNumberOfExposures", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0302, 0x0040, 0x0302,
      EVR_US, "EntranceDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0303, 0x0040, 0x0303,
      EVR_US, "ExposedArea", 1, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0306, 0x0040, 0x0306,
      EVR_DS, "DistanceSourceToEntrance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0307, 0x0040, 0x0307,
      EVR_DS, "RETIRED_DistanceSourceToSupport", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x030e, 0x0040, 0x030e,
      EVR_SQ, "RETIRED_ExposureDoseSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0310, 0x0040, 0x0310,
      EVR_ST, "CommentsOnRadiationDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0312, 0x0040, 0x0312,
      EVR_DS, "XRayOutput", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0314, 0x0040, 0x0314,
      EVR_DS, "HalfValueLayer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0316, 0x0040, 0x0316,
      EVR_DS, "OrganDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0318, 0x0040, 0x0318,
      EVR_CS, "OrganExposed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0320, 0x0040, 0x0320,
      EVR_SQ, "BillingProcedureStepSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0321, 0x0040, 0x0321,
      EVR_SQ, "FilmConsumptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0324, 0x0040, 0x0324,
      EVR_SQ, "BillingSuppliesAndDevicesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0330, 0x0040, 0x0330,
      EVR_SQ, "RETIRED_ReferencedProcedureStepSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0340, 0x0040, 0x0340,
      EVR_SQ, "PerformedSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0400, 0x0040, 0x0400,
      EVR_LT, "CommentsOnTheScheduledProcedureStep", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0440, 0x0040, 0x0440,
      EVR_SQ, "ProtocolContextSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0441, 0x0040, 0x0441,
      EVR_SQ, "ContentItemModifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0500, 0x0040, 0x0500,
      EVR_SQ, "ScheduledSpecimenSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x050a, 0x0040, 0x050a,
      EVR_LO, "RETIRED_SpecimenAccessionNumber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0512, 0x0040, 0x0512,
      EVR_LO, "ContainerIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0513, 0x0040, 0x0513,
      EVR_SQ, "IssuerOfTheContainerIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0515, 0x0040, 0x0515,
      EVR_SQ, "AlternateContainerIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0518, 0x0040, 0x0518,
      EVR_SQ, "ContainerTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x051a, 0x0040, 0x051a,
      EVR_LO, "ContainerDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0520, 0x0040, 0x0520,
      EVR_SQ, "ContainerComponentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0550, 0x0040, 0x0550,
      EVR_SQ, "RETIRED_SpecimenSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0551, 0x0040, 0x0551,
      EVR_LO, "SpecimenIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0552, 0x0040, 0x0552,
      EVR_SQ, "RETIRED_SpecimenDescriptionSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0553, 0x0040, 0x0553,
      EVR_ST, "RETIRED_SpecimenDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0554, 0x0040, 0x0554,
      EVR_UI, "SpecimenUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0555, 0x0040, 0x0555,
      EVR_SQ, "AcquisitionContextSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0556, 0x0040, 0x0556,
      EVR_ST, "AcquisitionContextDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0560, 0x0040, 0x0560,
      EVR_SQ, "SpecimenDescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0562, 0x0040, 0x0562,
      EVR_SQ, "IssuerOfTheSpecimenIdentifierSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x059a, 0x0040, 0x059a,
      EVR_SQ, "SpecimenTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0600, 0x0040, 0x0600,
      EVR_LO, "SpecimenShortDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0602, 0x0040, 0x0602,
      EVR_UT, "SpecimenDetailedDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0610, 0x0040, 0x0610,
      EVR_SQ, "SpecimenPreparationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0612, 0x0040, 0x0612,
      EVR_SQ, "SpecimenPreparationStepContentItemSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0620, 0x0040, 0x0620,
      EVR_SQ, "SpecimenLocalizationContentItemSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x06fa, 0x0040, 0x06fa,
      EVR_LO, "RETIRED_SlideIdentifier", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x0710, 0x0040, 0x0710,
      EVR_SQ, "WholeSlideMicroscopyImageFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x071a, 0x0040, 0x071a,
      EVR_SQ, "ImageCenterPointCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x072a, 0x0040, 0x072a,
      EVR_DS, "XOffsetInSlideCoordinateSystem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x073a, 0x0040, 0x073a,
      EVR_DS, "YOffsetInSlideCoordinateSystem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x074a, 0x0040, 0x074a,
      EVR_DS, "ZOffsetInSlideCoordinateSystem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x08d8, 0x0040, 0x08d8,
      EVR_SQ, "RETIRED_PixelSpacingSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x08da, 0x0040, 0x08da,
      EVR_SQ, "RETIRED_CoordinateSystemAxisCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x08ea, 0x0040, 0x08ea,
      EVR_SQ, "MeasurementUnitsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x09f8, 0x0040, 0x09f8,
      EVR_SQ, "RETIRED_VitalStainCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1001, 0x0040, 0x1001,
      EVR_SH, "RequestedProcedureID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1002, 0x0040, 0x1002,
      EVR_LO, "ReasonForTheRequestedProcedure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1003, 0x0040, 0x1003,
      EVR_SH, "RequestedProcedurePriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1004, 0x0040, 0x1004,
      EVR_LO, "PatientTransportArrangements", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1005, 0x0040, 0x1005,
      EVR_LO, "RequestedProcedureLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1006, 0x0040, 0x1006,
      EVR_SH, "RETIRED_PlacerOrderNumberProcedure", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1007, 0x0040, 0x1007,
      EVR_SH, "RETIRED_FillerOrderNumberProcedure", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1008, 0x0040, 0x1008,
      EVR_LO, "ConfidentialityCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1009, 0x0040, 0x1009,
      EVR_SH, "ReportingPriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x100a, 0x0040, 0x100a,
      EVR_SQ, "ReasonForRequestedProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1010, 0x0040, 0x1010,
      EVR_PN, "NamesOfIntendedRecipientsOfResults", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1011, 0x0040, 0x1011,
      EVR_SQ, "IntendedRecipientsOfResultsIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1012, 0x0040, 0x1012,
      EVR_SQ, "ReasonForPerformedProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1060, 0x0040, 0x1060,
      EVR_LO, "RETIRED_RequestedProcedureDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1101, 0x0040, 0x1101,
      EVR_SQ, "PersonIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1102, 0x0040, 0x1102,
      EVR_ST, "PersonAddress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1103, 0x0040, 0x1103,
      EVR_LO, "PersonTelephoneNumbers", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1104, 0x0040, 0x1104,
      EVR_LT, "PersonTelecomInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x1400, 0x0040, 0x1400,
      EVR_LT, "RequestedProcedureComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2001, 0x0040, 0x2001,
      EVR_LO, "RETIRED_ReasonForTheImagingServiceRequest", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2004, 0x0040, 0x2004,
      EVR_DA, "IssueDateOfImagingServiceRequest", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2005, 0x0040, 0x2005,
      EVR_TM, "IssueTimeOfImagingServiceRequest", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2006, 0x0040, 0x2006,
      EVR_SH, "RETIRED_PlacerOrderNumberImagingServiceRequestRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2007, 0x0040, 0x2007,
      EVR_SH, "RETIRED_FillerOrderNumberImagingServiceRequestRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2008, 0x0040, 0x2008,
      EVR_PN, "OrderEnteredBy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2009, 0x0040, 0x2009,
      EVR_SH, "OrderEntererLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2010, 0x0040, 0x2010,
      EVR_SH, "OrderCallbackPhoneNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2011, 0x0040, 0x2011,
      EVR_LT, "OrderCallbackTelecomInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2016, 0x0040, 0x2016,
      EVR_LO, "PlacerOrderNumberImagingServiceRequest", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2017, 0x0040, 0x2017,
      EVR_LO, "FillerOrderNumberImagingServiceRequest", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x2400, 0x0040, 0x2400,
      EVR_LT, "ImagingServiceRequestComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x3001, 0x0040, 0x3001,
      EVR_LO, "ConfidentialityConstraintOnPatientDataDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4001, 0x0040, 0x4001,
      EVR_CS, "RETIRED_GeneralPurposeScheduledProcedureStepStatus", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4002, 0x0040, 0x4002,
      EVR_CS, "RETIRED_GeneralPurposePerformedProcedureStepStatus", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4003, 0x0040, 0x4003,
      EVR_CS, "RETIRED_GeneralPurposeScheduledProcedureStepPriority", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4004, 0x0040, 0x4004,
      EVR_SQ, "RETIRED_ScheduledProcessingApplicationsCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4005, 0x0040, 0x4005,
      EVR_DT, "ScheduledProcedureStepStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4006, 0x0040, 0x4006,
      EVR_CS, "RETIRED_MultipleCopiesFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4007, 0x0040, 0x4007,
      EVR_SQ, "RETIRED_PerformedProcessingApplicationsCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4008, 0x0040, 0x4008,
      EVR_DT, "ScheduledProcedureStepExpirationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4009, 0x0040, 0x4009,
      EVR_SQ, "HumanPerformerCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4010, 0x0040, 0x4010,
      EVR_DT, "ScheduledProcedureStepModificationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4011, 0x0040, 0x4011,
      EVR_DT, "ExpectedCompletionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4015, 0x0040, 0x4015,
      EVR_SQ, "RETIRED_ResultingGeneralPurposePerformedProcedureStepsSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4016, 0x0040, 0x4016,
      EVR_SQ, "RETIRED_ReferencedGeneralPurposeScheduledProcedureStepSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4018, 0x0040, 0x4018,
      EVR_SQ, "ScheduledWorkitemCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4019, 0x0040, 0x4019,
      EVR_SQ, "PerformedWorkitemCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4020, 0x0040, 0x4020,
      EVR_CS, "RETIRED_InputAvailabilityFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4021, 0x0040, 0x4021,
      EVR_SQ, "InputInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4022, 0x0040, 0x4022,
      EVR_SQ, "RETIRED_RelevantInformationSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4023, 0x0040, 0x4023,
      EVR_UI, "RETIRED_ReferencedGeneralPurposeScheduledProcedureStepTransactionUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4025, 0x0040, 0x4025,
      EVR_SQ, "ScheduledStationNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4026, 0x0040, 0x4026,
      EVR_SQ, "ScheduledStationClassCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4027, 0x0040, 0x4027,
      EVR_SQ, "ScheduledStationGeographicLocationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4028, 0x0040, 0x4028,
      EVR_SQ, "PerformedStationNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4029, 0x0040, 0x4029,
      EVR_SQ, "PerformedStationClassCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4030, 0x0040, 0x4030,
      EVR_SQ, "PerformedStationGeographicLocationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4031, 0x0040, 0x4031,
      EVR_SQ, "RETIRED_RequestedSubsequentWorkitemCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4032, 0x0040, 0x4032,
      EVR_SQ, "RETIRED_NonDICOMOutputCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4033, 0x0040, 0x4033,
      EVR_SQ, "OutputInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4034, 0x0040, 0x4034,
      EVR_SQ, "ScheduledHumanPerformersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4035, 0x0040, 0x4035,
      EVR_SQ, "ActualHumanPerformersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4036, 0x0040, 0x4036,
      EVR_LO, "HumanPerformerOrganization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4037, 0x0040, 0x4037,
      EVR_PN, "HumanPerformerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4040, 0x0040, 0x4040,
      EVR_CS, "RawDataHandling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4041, 0x0040, 0x4041,
      EVR_CS, "InputReadinessState", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4050, 0x0040, 0x4050,
      EVR_DT, "PerformedProcedureStepStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4051, 0x0040, 0x4051,
      EVR_DT, "PerformedProcedureStepEndDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4052, 0x0040, 0x4052,
      EVR_DT, "ProcedureStepCancellationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4070, 0x0040, 0x4070,
      EVR_SQ, "OutputDestinationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4071, 0x0040, 0x4071,
      EVR_SQ, "DICOMStorageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4072, 0x0040, 0x4072,
      EVR_SQ, "STOWRSStorageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4073, 0x0040, 0x4073,
      EVR_UR, "StorageURL", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x4074, 0x0040, 0x4074,
      EVR_SQ, "XDSStorageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x8302, 0x0040, 0x8302,
      EVR_DS, "EntranceDoseInmGy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x8303, 0x0040, 0x8303,
      EVR_CS, "EntranceDoseDerivation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9092, 0x0040, 0x9092,
      EVR_SQ, "ParametricMapFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9094, 0x0040, 0x9094,
      EVR_SQ, "ReferencedImageRealWorldValueMappingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9096, 0x0040, 0x9096,
      EVR_SQ, "RealWorldValueMappingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9098, 0x0040, 0x9098,
      EVR_SQ, "PixelValueMappingCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9210, 0x0040, 0x9210,
      EVR_SH, "LUTLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9211, 0x0040, 0x9211,
      EVR_xs, "RealWorldValueLastValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9212, 0x0040, 0x9212,
      EVR_FD, "RealWorldValueLUTData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9213, 0x0040, 0x9213,
      EVR_FD, "DoubleFloatRealWorldValueLastValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9214, 0x0040, 0x9214,
      EVR_FD, "DoubleFloatRealWorldValueFirstValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9216, 0x0040, 0x9216,
      EVR_xs, "RealWorldValueFirstValueMapped", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9220, 0x0040, 0x9220,
      EVR_SQ, "QuantityDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9224, 0x0040, 0x9224,
      EVR_FD, "RealWorldValueIntercept", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0x9225, 0x0040, 0x9225,
      EVR_FD, "RealWorldValueSlope", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa007, 0x0040, 0xa007,
      EVR_CS, "RETIRED_FindingsFlagTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa010, 0x0040, 0xa010,
      EVR_CS, "RelationshipType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa020, 0x0040, 0xa020,
      EVR_SQ, "RETIRED_FindingsSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa021, 0x0040, 0xa021,
      EVR_UI, "RETIRED_FindingsGroupUIDTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa022, 0x0040, 0xa022,
      EVR_UI, "RETIRED_ReferencedFindingsGroupUIDTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa023, 0x0040, 0xa023,
      EVR_DA, "RETIRED_FindingsGroupRecordingDateTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa024, 0x0040, 0xa024,
      EVR_TM, "RETIRED_FindingsGroupRecordingTimeTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa026, 0x0040, 0xa026,
      EVR_SQ, "RETIRED_FindingsSourceCategoryCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa027, 0x0040, 0xa027,
      EVR_LO, "VerifyingOrganization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa028, 0x0040, 0xa028,
      EVR_SQ, "RETIRED_DocumentingOrganizationIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa030, 0x0040, 0xa030,
      EVR_DT, "VerificationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa032, 0x0040, 0xa032,
      EVR_DT, "ObservationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa033, 0x0040, 0xa033,
      EVR_DT, "ObservationStartDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa040, 0x0040, 0xa040,
      EVR_CS, "ValueType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa043, 0x0040, 0xa043,
      EVR_SQ, "ConceptNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa047, 0x0040, 0xa047,
      EVR_LO, "RETIRED_MeasurementPrecisionDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa050, 0x0040, 0xa050,
      EVR_CS, "ContinuityOfContent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa057, 0x0040, 0xa057,
      EVR_CS, "RETIRED_UrgencyOrPriorityAlertsTrial", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa060, 0x0040, 0xa060,
      EVR_LO, "RETIRED_SequencingIndicatorTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa066, 0x0040, 0xa066,
      EVR_SQ, "RETIRED_DocumentIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa067, 0x0040, 0xa067,
      EVR_PN, "RETIRED_DocumentAuthorTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa068, 0x0040, 0xa068,
      EVR_SQ, "RETIRED_DocumentAuthorIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa070, 0x0040, 0xa070,
      EVR_SQ, "RETIRED_IdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa073, 0x0040, 0xa073,
      EVR_SQ, "VerifyingObserverSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa074, 0x0040, 0xa074,
      EVR_OB, "RETIRED_ObjectBinaryIdentifierTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa075, 0x0040, 0xa075,
      EVR_PN, "VerifyingObserverName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa076, 0x0040, 0xa076,
      EVR_SQ, "RETIRED_DocumentingObserverIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa078, 0x0040, 0xa078,
      EVR_SQ, "AuthorObserverSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa07a, 0x0040, 0xa07a,
      EVR_SQ, "ParticipantSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa07c, 0x0040, 0xa07c,
      EVR_SQ, "CustodialOrganizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa080, 0x0040, 0xa080,
      EVR_CS, "ParticipationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa082, 0x0040, 0xa082,
      EVR_DT, "ParticipationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa084, 0x0040, 0xa084,
      EVR_CS, "ObserverType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa085, 0x0040, 0xa085,
      EVR_SQ, "RETIRED_ProcedureIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa088, 0x0040, 0xa088,
      EVR_SQ, "VerifyingObserverIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa089, 0x0040, 0xa089,
      EVR_OB, "RETIRED_ObjectDirectoryBinaryIdentifierTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa090, 0x0040, 0xa090,
      EVR_SQ, "RETIRED_EquivalentCDADocumentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa0b0, 0x0040, 0xa0b0,
      EVR_US, "ReferencedWaveformChannels", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa110, 0x0040, 0xa110,
      EVR_DA, "RETIRED_DateOfDocumentOrVerbalTransactionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa112, 0x0040, 0xa112,
      EVR_TM, "RETIRED_TimeOfDocumentCreationOrVerbalTransactionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa120, 0x0040, 0xa120,
      EVR_DT, "DateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa121, 0x0040, 0xa121,
      EVR_DA, "Date", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa122, 0x0040, 0xa122,
      EVR_TM, "Time", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa123, 0x0040, 0xa123,
      EVR_PN, "PersonName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa124, 0x0040, 0xa124,
      EVR_UI, "UID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa125, 0x0040, 0xa125,
      EVR_CS, "RETIRED_ReportStatusIDTrial", 2, 2, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa130, 0x0040, 0xa130,
      EVR_CS, "TemporalRangeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa132, 0x0040, 0xa132,
      EVR_UL, "ReferencedSamplePositions", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa136, 0x0040, 0xa136,
      EVR_US, "RETIRED_ReferencedFrameNumbers", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa138, 0x0040, 0xa138,
      EVR_DS, "ReferencedTimeOffsets", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa13a, 0x0040, 0xa13a,
      EVR_DT, "ReferencedDateTime", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa160, 0x0040, 0xa160,
      EVR_UT, "TextValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa161, 0x0040, 0xa161,
      EVR_FD, "FloatingPointValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa162, 0x0040, 0xa162,
      EVR_SL, "RationalNumeratorValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa163, 0x0040, 0xa163,
      EVR_UL, "RationalDenominatorValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa167, 0x0040, 0xa167,
      EVR_SQ, "RETIRED_ObservationCategoryCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa168, 0x0040, 0xa168,
      EVR_SQ, "ConceptCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa16a, 0x0040, 0xa16a,
      EVR_ST, "RETIRED_BibliographicCitationTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa170, 0x0040, 0xa170,
      EVR_SQ, "PurposeOfReferenceCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa171, 0x0040, 0xa171,
      EVR_UI, "ObservationUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa172, 0x0040, 0xa172,
      EVR_UI, "RETIRED_ReferencedObservationUIDTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa173, 0x0040, 0xa173,
      EVR_CS, "RETIRED_ReferencedObservationClassTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa174, 0x0040, 0xa174,
      EVR_CS, "RETIRED_ReferencedObjectObservationClassTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa180, 0x0040, 0xa180,
      EVR_US, "AnnotationGroupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa192, 0x0040, 0xa192,
      EVR_DA, "RETIRED_ObservationDateTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa193, 0x0040, 0xa193,
      EVR_TM, "RETIRED_ObservationTimeTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa194, 0x0040, 0xa194,
      EVR_CS, "RETIRED_MeasurementAutomationTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa195, 0x0040, 0xa195,
      EVR_SQ, "ModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa224, 0x0040, 0xa224,
      EVR_ST, "RETIRED_IdentificationDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa290, 0x0040, 0xa290,
      EVR_CS, "RETIRED_CoordinatesSetGeometricTypeTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa296, 0x0040, 0xa296,
      EVR_SQ, "RETIRED_AlgorithmCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa297, 0x0040, 0xa297,
      EVR_ST, "RETIRED_AlgorithmDescriptionTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa29a, 0x0040, 0xa29a,
      EVR_SL, "RETIRED_PixelCoordinatesSetTrial", 2, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa300, 0x0040, 0xa300,
      EVR_SQ, "MeasuredValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa301, 0x0040, 0xa301,
      EVR_SQ, "NumericValueQualifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa307, 0x0040, 0xa307,
      EVR_PN, "RETIRED_CurrentObserverTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa30a, 0x0040, 0xa30a,
      EVR_DS, "NumericValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa313, 0x0040, 0xa313,
      EVR_SQ, "RETIRED_ReferencedAccessionSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa33a, 0x0040, 0xa33a,
      EVR_ST, "RETIRED_ReportStatusCommentTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa340, 0x0040, 0xa340,
      EVR_SQ, "RETIRED_ProcedureContextSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa352, 0x0040, 0xa352,
      EVR_PN, "RETIRED_VerbalSourceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa353, 0x0040, 0xa353,
      EVR_ST, "RETIRED_AddressTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa354, 0x0040, 0xa354,
      EVR_LO, "RETIRED_TelephoneNumberTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa358, 0x0040, 0xa358,
      EVR_SQ, "RETIRED_VerbalSourceIdentifierCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa360, 0x0040, 0xa360,
      EVR_SQ, "PredecessorDocumentsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa370, 0x0040, 0xa370,
      EVR_SQ, "ReferencedRequestSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa372, 0x0040, 0xa372,
      EVR_SQ, "PerformedProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa375, 0x0040, 0xa375,
      EVR_SQ, "CurrentRequestedProcedureEvidenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa380, 0x0040, 0xa380,
      EVR_SQ, "RETIRED_ReportDetailSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa385, 0x0040, 0xa385,
      EVR_SQ, "PertinentOtherEvidenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa390, 0x0040, 0xa390,
      EVR_SQ, "HL7StructuredDocumentReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa402, 0x0040, 0xa402,
      EVR_UI, "RETIRED_ObservationSubjectUIDTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa403, 0x0040, 0xa403,
      EVR_CS, "RETIRED_ObservationSubjectClassTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa404, 0x0040, 0xa404,
      EVR_SQ, "RETIRED_ObservationSubjectTypeCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa491, 0x0040, 0xa491,
      EVR_CS, "CompletionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa492, 0x0040, 0xa492,
      EVR_LO, "CompletionFlagDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa493, 0x0040, 0xa493,
      EVR_CS, "VerificationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa494, 0x0040, 0xa494,
      EVR_CS, "ArchiveRequested", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa496, 0x0040, 0xa496,
      EVR_CS, "PreliminaryFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa504, 0x0040, 0xa504,
      EVR_SQ, "ContentTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa525, 0x0040, 0xa525,
      EVR_SQ, "IdenticalDocumentsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa600, 0x0040, 0xa600,
      EVR_CS, "RETIRED_ObservationSubjectContextFlagTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa601, 0x0040, 0xa601,
      EVR_CS, "RETIRED_ObserverContextFlagTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa603, 0x0040, 0xa603,
      EVR_CS, "RETIRED_ProcedureContextFlagTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa730, 0x0040, 0xa730,
      EVR_SQ, "ContentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa731, 0x0040, 0xa731,
      EVR_SQ, "RETIRED_RelationshipSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa732, 0x0040, 0xa732,
      EVR_SQ, "RETIRED_RelationshipTypeCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa744, 0x0040, 0xa744,
      EVR_SQ, "RETIRED_LanguageCodeSequenceTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa801, 0x0040, 0xa801,
      EVR_SQ, "TabulatedValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa802, 0x0040, 0xa802,
      EVR_UL, "NumberOfTableRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa803, 0x0040, 0xa803,
      EVR_UL, "NumberOfTableColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa804, 0x0040, 0xa804,
      EVR_UL, "TableRowNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa805, 0x0040, 0xa805,
      EVR_UL, "TableColumnNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa806, 0x0040, 0xa806,
      EVR_SQ, "TableRowDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa807, 0x0040, 0xa807,
      EVR_SQ, "TableColumnDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa808, 0x0040, 0xa808,
      EVR_SQ, "CellValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xa992, 0x0040, 0xa992,
      EVR_ST, "RETIRED_UniformResourceLocatorTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xb020, 0x0040, 0xb020,
      EVR_SQ, "WaveformAnnotationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb00, 0x0040, 0xdb00,
      EVR_CS, "TemplateIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb06, 0x0040, 0xdb06,
      EVR_DT, "RETIRED_TemplateVersion", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb07, 0x0040, 0xdb07,
      EVR_DT, "RETIRED_TemplateLocalVersion", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb0b, 0x0040, 0xdb0b,
      EVR_CS, "RETIRED_TemplateExtensionFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb0c, 0x0040, 0xdb0c,
      EVR_UI, "RETIRED_TemplateExtensionOrganizationUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb0d, 0x0040, 0xdb0d,
      EVR_UI, "RETIRED_TemplateExtensionCreatorUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xdb73, 0x0040, 0xdb73,
      EVR_UL, "ReferencedContentItemIdentifier", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe001, 0x0040, 0xe001,
      EVR_ST, "HL7InstanceIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe004, 0x0040, 0xe004,
      EVR_DT, "HL7DocumentEffectiveTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe006, 0x0040, 0xe006,
      EVR_SQ, "HL7DocumentTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe008, 0x0040, 0xe008,
      EVR_SQ, "DocumentClassCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe010, 0x0040, 0xe010,
      EVR_UR, "RetrieveURI", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe011, 0x0040, 0xe011,
      EVR_UI, "RetrieveLocationUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe020, 0x0040, 0xe020,
      EVR_CS, "TypeOfInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe021, 0x0040, 0xe021,
      EVR_SQ, "DICOMRetrievalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe022, 0x0040, 0xe022,
      EVR_SQ, "DICOMMediaRetrievalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe023, 0x0040, 0xe023,
      EVR_SQ, "WADORetrievalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe024, 0x0040, 0xe024,
      EVR_SQ, "XDSRetrievalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe025, 0x0040, 0xe025,
      EVR_SQ, "WADORSRetrievalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe030, 0x0040, 0xe030,
      EVR_UI, "RepositoryUniqueID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0040, 0xe031, 0x0040, 0xe031,
      EVR_UI, "HomeCommunityID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0041, 0x0000, 0x0041, 0x0000,
      EVR_LT, "PapyrusComments", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0000, 0x0041, 0x0000,
      EVR_LT, "PapyrusComments", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0010, 0x0041, 0x0010,
      EVR_SQ, "PointerSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0010, 0x0041, 0x0010,
      EVR_US, "FolderType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0010, 0x0041, 0x0010,
      EVR_US, "NumberOfHardcopies", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0041, 0x0011, 0x0041, 0x0011,
      EVR_UL, "ImagePointer", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0011, 0x0041, 0x0011,
      EVR_LT, "PatientFolderDataSetID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0012, 0x0041, 0x0012,
      EVR_UL, "PixelOffset", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0013, 0x0041, 0x0013,
      EVR_SQ, "ImageIdentifierSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0014, 0x0041, 0x0014,
      EVR_SQ, "ExternalFileReferenceSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0015, 0x0041, 0x0015,
      EVR_US, "NumberOfImages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0020, 0x0041, 0x0020,
      EVR_LT, "FolderName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0020, 0x0041, 0x0020,
      EVR_LT, "FilmFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0041, 0x0020, 0x0041, 0x0020,
      EVR_FL, "AccumulatedFluoroscopyDose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0041, 0x0021, 0x0041, 0x0021,
      EVR_UI, "ReferencedSOPClassUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0022, 0x0041, 0x0022,
      EVR_UI, "ReferencedSOPInstanceUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0030, 0x0041, 0x0030,
      EVR_DA, "CreationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0030, 0x0041, 0x0030,
      EVR_LT, "FilmSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0041, 0x0030, 0x0041, 0x0030,
      EVR_FL, "AccumulatedExposureDose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0041, 0x0031, 0x0041, 0x0031,
      EVR_LT, "ReferencedFileName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0031, 0x0041, 0x0031,
      EVR_LT, "FullFilmFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS DLR.01" }
  , { 0x0041, 0x0032, 0x0041, 0x0032,
      EVR_LT, "ReferencedFilePath", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0032, 0x0041, 0x0032,
      EVR_TM, "CreationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0034, 0x0041, 0x0034,
      EVR_DA, "ModifiedDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0036, 0x0041, 0x0036,
      EVR_TM, "ModifiedTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0040, 0x0041, 0x0040,
      EVR_LT, "OwnerName", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0040, 0x0041, 0x0040,
      EVR_FL, "TotalDose", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0041, 0x0041, 0x0041, 0x0041,
      EVR_UI, "ReferencedImageSOPClassUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0041, 0x0041, 0x0041,
      EVR_FL, "TotalNumberOfFrames", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0041, 0x0042, 0x0041, 0x0042,
      EVR_UI, "ReferencedImageSOPInstanceUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0050, 0x0041, 0x0050,
      EVR_SQ, "ImageSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x0041, 0x0050, 0x0041, 0x0050,
      EVR_LT, "FolderStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0050, 0x0041, 0x0050,
      EVR_SQ, "ExposureInformationSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTEGRIS 1.0" }
  , { 0x0041, 0x0060, 0x0041, 0x0060,
      EVR_UL, "NumberOfImages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x0062, 0x0041, 0x0062,
      EVR_UL, "NumberOfOther", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00a0, 0x0041, 0x00a0,
      EVR_LT, "ExternalFolderElementDSID", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00a1, 0x0041, 0x00a1,
      EVR_US, "ExternalFolderElementDataSetType", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00a2, 0x0041, 0x00a2,
      EVR_LT, "ExternalFolderElementFileLocation", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00a3, 0x0041, 0x00a3,
      EVR_UL, "ExternalFolderElementLength", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00b0, 0x0041, 0x00b0,
      EVR_LT, "InternalFolderElementDSID", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00b1, 0x0041, 0x00b1,
      EVR_US, "InternalFolderElementDataSetType", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00b2, 0x0041, 0x00b2,
      EVR_UL, "InternalOffsetToDataSet", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
  , { 0x0041, 0x00b3, 0x0041, 0x00b3,
      EVR_UL, "InternalOffsetToImage", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PAPYRUS" }
#endif
  , { 0x0042, 0x0010, 0x0042, 0x0010,
      EVR_ST, "DocumentTitle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0042, 0x0011, 0x0042, 0x0011,
      EVR_OB, "EncapsulatedDocument", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0042, 0x0012, 0x0042, 0x0012,
      EVR_LO, "MIMETypeOfEncapsulatedDocument", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0042, 0x0013, 0x0042, 0x0013,
      EVR_SQ, "SourceInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0042, 0x0014, 0x0042, 0x0014,
      EVR_LO, "ListOfMIMETypes", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0042, 0x0015, 0x0042, 0x0015,
      EVR_UL, "EncapsulatedDocumentLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0043, 0x0001, 0x0043, 0x0001,
      EVR_SS, "BitmapOfPrescanOptions", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0002, 0x0043, 0x0002,
      EVR_SS, "GradientOffsetInX", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0003, 0x0043, 0x0003,
      EVR_SS, "GradientOffsetInY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0004, 0x0043, 0x0004,
      EVR_SS, "GradientOffsetInZ", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0005, 0x0043, 0x0005,
      EVR_SS, "ImageIsOriginalOrUnoriginal", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0006, 0x0043, 0x0006,
      EVR_SS, "NumberOfEPIShots", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0007, 0x0043, 0x0007,
      EVR_SS, "ViewsPerSegment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0008, 0x0043, 0x0008,
      EVR_SS, "RespiratoryRateInBPM", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0009, 0x0043, 0x0009,
      EVR_SS, "RespiratoryTriggerPoint", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000a, 0x0043, 0x000a,
      EVR_SS, "TypeOfReceiverUsed", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000b, 0x0043, 0x000b,
      EVR_DS, "PeakRateOfChangeOfGradientField", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000c, 0x0043, 0x000c,
      EVR_DS, "LimitsInUnitsOfPercent", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000d, 0x0043, 0x000d,
      EVR_DS, "PSDEstimatedLimit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000e, 0x0043, 0x000e,
      EVR_DS, "PSDEstimatedLimitInTeslaPerSecond", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x000f, 0x0043, 0x000f,
      EVR_DS, "SARAvgHead", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0010, 0x0043, 0x0010,
      EVR_US, "WindowValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0011, 0x0043, 0x0011,
      EVR_US, "TotalInputViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0012, 0x0043, 0x0012,
      EVR_SS, "XrayChain", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0013, 0x0043, 0x0013,
      EVR_SS, "ReconKernelParameters", 5, 5, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0014, 0x0043, 0x0014,
      EVR_SS, "CalibrationParameters", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0015, 0x0043, 0x0015,
      EVR_SS, "TotalOutputViews", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0016, 0x0043, 0x0016,
      EVR_SS, "NumberOfOverranges", 5, 5, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0017, 0x0043, 0x0017,
      EVR_DS, "IBHImageScaleFactors", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0018, 0x0043, 0x0018,
      EVR_DS, "BBHCoefficients", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0019, 0x0043, 0x0019,
      EVR_SS, "NumberOfBBHChainsToBlend", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001a, 0x0043, 0x001a,
      EVR_SL, "StartingChannelNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001b, 0x0043, 0x001b,
      EVR_SS, "PPScanParameters", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001c, 0x0043, 0x001c,
      EVR_SS, "GEImageIntegrity", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001d, 0x0043, 0x001d,
      EVR_SS, "LevelValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001e, 0x0043, 0x001e,
      EVR_DS, "DeltaStartTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0043, 0x001e, 0x0043, 0x001e,
      EVR_DS, "DeltaStartTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x001f, 0x0043, 0x001f,
      EVR_SL, "MaxOverrangesInAView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0020, 0x0043, 0x0020,
      EVR_DS, "AvgOverrangesAllViews", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0021, 0x0043, 0x0021,
      EVR_SS, "CorrectedAfterglowTerms", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0025, 0x0043, 0x0025,
      EVR_SS, "ReferenceChannels", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0026, 0x0043, 0x0026,
      EVR_US, "NoViewsRefChannelsBlocked", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0027, 0x0043, 0x0027,
      EVR_SH, "ScanPitchRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GE_GENESIS_REV3.0" }
  , { 0x0043, 0x0027, 0x0043, 0x0027,
      EVR_SH, "ScanPitchRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0028, 0x0043, 0x0028,
      EVR_OB, "UniqueImageIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0029, 0x0043, 0x0029,
      EVR_OB, "HistogramTables", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002a, 0x0043, 0x002a,
      EVR_OB, "UserDefinedData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002b, 0x0043, 0x002b,
      EVR_SS, "PrivateScanOptions", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002c, 0x0043, 0x002c,
      EVR_SS, "EffectiveEchoSpacing", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002d, 0x0043, 0x002d,
      EVR_SH, "StringSlopField1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002e, 0x0043, 0x002e,
      EVR_SH, "StringSlopField2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x002f, 0x0043, 0x002f,
      EVR_SS, "RawDataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0030, 0x0043, 0x0030,
      EVR_SS, "RawDataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0031, 0x0043, 0x0031,
      EVR_DS, "RACoordOfTargetReconCentre", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0032, 0x0043, 0x0032,
      EVR_SS, "RawDataType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0033, 0x0043, 0x0033,
      EVR_FL, "NegScanSpacing", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0034, 0x0043, 0x0034,
      EVR_IS, "OffsetFrequency", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0035, 0x0043, 0x0035,
      EVR_UL, "UserUsageTag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0036, 0x0043, 0x0036,
      EVR_UL, "UserFillMapMSW", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0037, 0x0043, 0x0037,
      EVR_UL, "UserFillMapLSW", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0038, 0x0043, 0x0038,
      EVR_FL, "User25ToUser48", 24, 24, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0039, 0x0043, 0x0039,
      EVR_IS, "SlopInteger6ToSlopInteger9", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0040, 0x0043, 0x0040,
      EVR_FL, "TriggerOnPosition", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0041, 0x0043, 0x0041,
      EVR_FL, "DegreeOfRotation", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0042, 0x0043, 0x0042,
      EVR_SL, "DASTriggerSource", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0043, 0x0043, 0x0043,
      EVR_SL, "DASFpaGain", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0044, 0x0043, 0x0044,
      EVR_SL, "DASOutputSource", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0045, 0x0043, 0x0045,
      EVR_SL, "DASAdInput", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0046, 0x0043, 0x0046,
      EVR_SL, "DASCalMode", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0047, 0x0043, 0x0047,
      EVR_SL, "DASCalFrequency", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0048, 0x0043, 0x0048,
      EVR_SL, "DASRegXm", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0049, 0x0043, 0x0049,
      EVR_SL, "DASAutoZero", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x004a, 0x0043, 0x004a,
      EVR_SS, "StartingChannelOfView", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x004b, 0x0043, 0x004b,
      EVR_SL, "DASXmPattern", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x004c, 0x0043, 0x004c,
      EVR_SS, "TGGCTriggerMode", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x004d, 0x0043, 0x004d,
      EVR_FL, "StartScanToXrayOnDelay", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x004e, 0x0043, 0x004e,
      EVR_FL, "DurationOfXrayOn", 4, 4, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0060, 0x0043, 0x0060,
      EVR_IS, "SlopInteger10ToSlopInteger17", 8, 8, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0061, 0x0043, 0x0061,
      EVR_UI, "ScannerStudyEntityUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0062, 0x0043, 0x0062,
      EVR_SH, "ScannerStudyID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x006f, 0x0043, 0x006f,
      EVR_DS, "ScannerTableEntry", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0070, 0x0043, 0x0070,
      EVR_LO, "ParadigmName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0071, 0x0043, 0x0071,
      EVR_ST, "ParadigmDescription", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0072, 0x0043, 0x0072,
      EVR_UI, "ParadigmUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0073, 0x0043, 0x0073,
      EVR_US, "ExperimentType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0074, 0x0043, 0x0074,
      EVR_US, "NumberOfRestVolumes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0075, 0x0043, 0x0075,
      EVR_US, "NumberOfActiveVolumes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0076, 0x0043, 0x0076,
      EVR_US, "NumberOfDummyScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0077, 0x0043, 0x0077,
      EVR_SH, "ApplicationName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0078, 0x0043, 0x0078,
      EVR_SH, "ApplicationVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0079, 0x0043, 0x0079,
      EVR_US, "SlicesPerVolume", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007a, 0x0043, 0x007a,
      EVR_US, "ExpectedTimePoints", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007b, 0x0043, 0x007b,
      EVR_FL, "RegressorValues", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007c, 0x0043, 0x007c,
      EVR_FL, "DelayAfterSliceGroup", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007d, 0x0043, 0x007d,
      EVR_US, "ReconModeFlagWord", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007e, 0x0043, 0x007e,
      EVR_LO, "PACCSpecificInformation", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x007f, 0x0043, 0x007f,
      EVR_DS, "EDWIScaleFactor", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0080, 0x0043, 0x0080,
      EVR_LO, "CoilIDData", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0081, 0x0043, 0x0081,
      EVR_LO, "GECoilName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0082, 0x0043, 0x0082,
      EVR_LO, "SystemConfigurationInformation", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0083, 0x0043, 0x0083,
      EVR_DS, "AssetRFactors", 1, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0084, 0x0043, 0x0084,
      EVR_LO, "AdditionalAssetData", 5, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0085, 0x0043, 0x0085,
      EVR_UT, "DebugDataTextFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0086, 0x0043, 0x0086,
      EVR_OB, "DebugDataBinaryFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0087, 0x0043, 0x0087,
      EVR_UT, "ScannerSoftwareVersionLongForm", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0088, 0x0043, 0x0088,
      EVR_UI, "PUREAcquisitionCalibrationSeriesUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0089, 0x0043, 0x0089,
      EVR_LO, "GoverningBodydBdtAndSARDefinition", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008a, 0x0043, 0x008a,
      EVR_CS, "PrivateInPlanePhaseEncodingDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008b, 0x0043, 0x008b,
      EVR_OB, "FMRIBinaryDataBlock", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008c, 0x0043, 0x008c,
      EVR_DS, "VoxelLocation", 6, 6, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008d, 0x0043, 0x008d,
      EVR_DS, "SATBandLocations", 7, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008e, 0x0043, 0x008e,
      EVR_DS, "SpectroPrescanValues", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x008f, 0x0043, 0x008f,
      EVR_DS, "SpectroParameters", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0090, 0x0043, 0x0090,
      EVR_LO, "SARDefinition", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0091, 0x0043, 0x0091,
      EVR_DS, "SARValue", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0092, 0x0043, 0x0092,
      EVR_LO, "ImageErrorText", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0093, 0x0043, 0x0093,
      EVR_DS, "SpectroQuantitationValues", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0094, 0x0043, 0x0094,
      EVR_DS, "SpectroRatioValues", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0095, 0x0043, 0x0095,
      EVR_LO, "PrescanReuseString", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0096, 0x0043, 0x0096,
      EVR_CS, "ContentQualification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0097, 0x0043, 0x0097,
      EVR_LO, "ImageFilteringParameters", 9, 9, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0098, 0x0043, 0x0098,
      EVR_UI, "ASSETAcquisitionCalibrationSeriesUID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x0099, 0x0043, 0x0099,
      EVR_LO, "ExtendedOptions", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009a, 0x0043, 0x009a,
      EVR_IS, "RxStackIdentification", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009b, 0x0043, 0x009b,
      EVR_DS, "NPWFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009c, 0x0043, 0x009c,
      EVR_OB, "ResearchTag1", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009d, 0x0043, 0x009d,
      EVR_OB, "ResearchTag2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009e, 0x0043, 0x009e,
      EVR_OB, "ResearchTag3", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
  , { 0x0043, 0x009f, 0x0043, 0x009f,
      EVR_OB, "ResearchTag4", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_PARM_01" }
#endif
  , { 0x0044, 0x0001, 0x0044, 0x0001,
      EVR_ST, "ProductPackageIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0002, 0x0044, 0x0002,
      EVR_CS, "SubstanceAdministrationApproval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0003, 0x0044, 0x0003,
      EVR_LT, "ApprovalStatusFurtherDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0004, 0x0044, 0x0004,
      EVR_DT, "ApprovalStatusDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0007, 0x0044, 0x0007,
      EVR_SQ, "ProductTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0008, 0x0044, 0x0008,
      EVR_LO, "ProductName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0009, 0x0044, 0x0009,
      EVR_LT, "ProductDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x000a, 0x0044, 0x000a,
      EVR_LO, "ProductLotIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x000b, 0x0044, 0x000b,
      EVR_DT, "ProductExpirationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0010, 0x0044, 0x0010,
      EVR_DT, "SubstanceAdministrationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0011, 0x0044, 0x0011,
      EVR_LO, "SubstanceAdministrationNotes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0012, 0x0044, 0x0012,
      EVR_LO, "SubstanceAdministrationDeviceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0013, 0x0044, 0x0013,
      EVR_SQ, "ProductParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0019, 0x0044, 0x0019,
      EVR_SQ, "SubstanceAdministrationParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0100, 0x0044, 0x0100,
      EVR_SQ, "ApprovalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0101, 0x0044, 0x0101,
      EVR_SQ, "AssertionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0102, 0x0044, 0x0102,
      EVR_UI, "AssertionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0103, 0x0044, 0x0103,
      EVR_SQ, "AsserterIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0104, 0x0044, 0x0104,
      EVR_DT, "AssertionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0105, 0x0044, 0x0105,
      EVR_DT, "AssertionExpirationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0106, 0x0044, 0x0106,
      EVR_UT, "AssertionComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0107, 0x0044, 0x0107,
      EVR_SQ, "RelatedAssertionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0108, 0x0044, 0x0108,
      EVR_UI, "ReferencedAssertionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x0109, 0x0044, 0x0109,
      EVR_SQ, "ApprovalSubjectSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0044, 0x010a, 0x0044, 0x010a,
      EVR_SQ, "OrganizationalRoleCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0045, 0x0004, 0x0045, 0x0004,
      EVR_CS, "AES", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0006, 0x0045, 0x0006,
      EVR_DS, "Angulation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0009, 0x0045, 0x0009,
      EVR_DS, "RealMagnificationFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x000b, 0x0045, 0x000b,
      EVR_CS, "SenographType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x000c, 0x0045, 0x000c,
      EVR_DS, "IntegrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x000d, 0x0045, 0x000d,
      EVR_DS, "ROIOriginXY", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0011, 0x0045, 0x0011,
      EVR_DS, "ReceptorSizeCmXY", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0012, 0x0045, 0x0012,
      EVR_IS, "ReceptorSizePixelsXY", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0013, 0x0045, 0x0013,
      EVR_ST, "Screen", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0014, 0x0045, 0x0014,
      EVR_DS, "PixelPitchMicrons", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0015, 0x0045, 0x0015,
      EVR_IS, "PixelDepthBits", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0016, 0x0045, 0x0016,
      EVR_IS, "BinningFactorXY", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x001b, 0x0045, 0x001b,
      EVR_CS, "ClinicalView", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x001d, 0x0045, 0x001d,
      EVR_DS, "MeanOfRawGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x001e, 0x0045, 0x001e,
      EVR_DS, "MeanOfOffsetGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x001f, 0x0045, 0x001f,
      EVR_DS, "MeanOfCorrectedGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0020, 0x0045, 0x0020,
      EVR_DS, "MeanOfRegionGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0021, 0x0045, 0x0021,
      EVR_DS, "MeanOfLogRegionGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0022, 0x0045, 0x0022,
      EVR_DS, "StandardDeviationOfRawGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0023, 0x0045, 0x0023,
      EVR_DS, "StandardDeviationOfCorrectedGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0024, 0x0045, 0x0024,
      EVR_DS, "StandardDeviationOfRegionGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0025, 0x0045, 0x0025,
      EVR_DS, "StandardDeviationOfLogRegionGrayLevels", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0026, 0x0045, 0x0026,
      EVR_OB, "MAOBuffer", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0027, 0x0045, 0x0027,
      EVR_IS, "SetNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0028, 0x0045, 0x0028,
      EVR_CS, "WindowingType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0029, 0x0045, 0x0029,
      EVR_DS, "WindowingParameters", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x002a, 0x0045, 0x002a,
      EVR_IS, "CrosshairCursorXCoordinates", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x002b, 0x0045, 0x002b,
      EVR_IS, "CrosshairCursorYCoordinates", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x0039, 0x0045, 0x0039,
      EVR_US, "VignetteRows", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003a, 0x0045, 0x003a,
      EVR_US, "VignetteColumns", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003b, 0x0045, 0x003b,
      EVR_US, "VignetteBitsAllocated", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003c, 0x0045, 0x003c,
      EVR_US, "VignetteBitsStored", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003d, 0x0045, 0x003d,
      EVR_US, "VignetteHighBit", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003e, 0x0045, 0x003e,
      EVR_US, "VignettePixelRepresentation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
  , { 0x0045, 0x003f, 0x0045, 0x003f,
      EVR_OB, "VignettePixelData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_SENO_02" }
#endif
  , { 0x0046, 0x0012, 0x0046, 0x0012,
      EVR_LO, "LensDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0014, 0x0046, 0x0014,
      EVR_SQ, "RightLensSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0015, 0x0046, 0x0015,
      EVR_SQ, "LeftLensSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0016, 0x0046, 0x0016,
      EVR_SQ, "UnspecifiedLateralityLensSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0018, 0x0046, 0x0018,
      EVR_SQ, "CylinderSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0028, 0x0046, 0x0028,
      EVR_SQ, "PrismSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0030, 0x0046, 0x0030,
      EVR_FD, "HorizontalPrismPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0032, 0x0046, 0x0032,
      EVR_CS, "HorizontalPrismBase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0034, 0x0046, 0x0034,
      EVR_FD, "VerticalPrismPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0036, 0x0046, 0x0036,
      EVR_CS, "VerticalPrismBase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0038, 0x0046, 0x0038,
      EVR_CS, "LensSegmentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0040, 0x0046, 0x0040,
      EVR_FD, "OpticalTransmittance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0042, 0x0046, 0x0042,
      EVR_FD, "ChannelWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0044, 0x0046, 0x0044,
      EVR_FD, "PupilSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0046, 0x0046, 0x0046,
      EVR_FD, "CornealSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0047, 0x0046, 0x0047,
      EVR_SQ, "CornealSizeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0050, 0x0046, 0x0050,
      EVR_SQ, "AutorefractionRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0052, 0x0046, 0x0052,
      EVR_SQ, "AutorefractionLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0060, 0x0046, 0x0060,
      EVR_FD, "DistancePupillaryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0062, 0x0046, 0x0062,
      EVR_FD, "NearPupillaryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0063, 0x0046, 0x0063,
      EVR_FD, "IntermediatePupillaryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0064, 0x0046, 0x0064,
      EVR_FD, "OtherPupillaryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0070, 0x0046, 0x0070,
      EVR_SQ, "KeratometryRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0071, 0x0046, 0x0071,
      EVR_SQ, "KeratometryLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0074, 0x0046, 0x0074,
      EVR_SQ, "SteepKeratometricAxisSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0075, 0x0046, 0x0075,
      EVR_FD, "RadiusOfCurvature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0076, 0x0046, 0x0076,
      EVR_FD, "KeratometricPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0077, 0x0046, 0x0077,
      EVR_FD, "KeratometricAxis", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0080, 0x0046, 0x0080,
      EVR_SQ, "FlatKeratometricAxisSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0092, 0x0046, 0x0092,
      EVR_CS, "BackgroundColor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0094, 0x0046, 0x0094,
      EVR_CS, "Optotype", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0095, 0x0046, 0x0095,
      EVR_CS, "OptotypePresentation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0097, 0x0046, 0x0097,
      EVR_SQ, "SubjectiveRefractionRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0098, 0x0046, 0x0098,
      EVR_SQ, "SubjectiveRefractionLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0100, 0x0046, 0x0100,
      EVR_SQ, "AddNearSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0101, 0x0046, 0x0101,
      EVR_SQ, "AddIntermediateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0102, 0x0046, 0x0102,
      EVR_SQ, "AddOtherSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0104, 0x0046, 0x0104,
      EVR_FD, "AddPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0106, 0x0046, 0x0106,
      EVR_FD, "ViewingDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0110, 0x0046, 0x0110,
      EVR_SQ, "CorneaMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0111, 0x0046, 0x0111,
      EVR_SQ, "SourceOfCorneaMeasurementDataCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0112, 0x0046, 0x0112,
      EVR_SQ, "SteepCornealAxisSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0113, 0x0046, 0x0113,
      EVR_SQ, "FlatCornealAxisSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0114, 0x0046, 0x0114,
      EVR_FD, "CornealPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0115, 0x0046, 0x0115,
      EVR_FD, "CornealAxis", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0116, 0x0046, 0x0116,
      EVR_SQ, "CorneaMeasurementMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0117, 0x0046, 0x0117,
      EVR_FL, "RefractiveIndexOfCornea", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0118, 0x0046, 0x0118,
      EVR_FL, "RefractiveIndexOfAqueousHumor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0121, 0x0046, 0x0121,
      EVR_SQ, "VisualAcuityTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0122, 0x0046, 0x0122,
      EVR_SQ, "VisualAcuityRightEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0123, 0x0046, 0x0123,
      EVR_SQ, "VisualAcuityLeftEyeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0124, 0x0046, 0x0124,
      EVR_SQ, "VisualAcuityBothEyesOpenSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0125, 0x0046, 0x0125,
      EVR_CS, "ViewingDistanceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0135, 0x0046, 0x0135,
      EVR_SS, "VisualAcuityModifiers", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0137, 0x0046, 0x0137,
      EVR_FD, "DecimalVisualAcuity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0139, 0x0046, 0x0139,
      EVR_LO, "OptotypeDetailedDefinition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0145, 0x0046, 0x0145,
      EVR_SQ, "ReferencedRefractiveMeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0146, 0x0046, 0x0146,
      EVR_FD, "SpherePower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0147, 0x0046, 0x0147,
      EVR_FD, "CylinderPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0201, 0x0046, 0x0201,
      EVR_CS, "CornealTopographySurface", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0202, 0x0046, 0x0202,
      EVR_FL, "CornealVertexLocation", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0203, 0x0046, 0x0203,
      EVR_FL, "PupilCentroidXCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0204, 0x0046, 0x0204,
      EVR_FL, "PupilCentroidYCoordinate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0205, 0x0046, 0x0205,
      EVR_FL, "EquivalentPupilRadius", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0207, 0x0046, 0x0207,
      EVR_SQ, "CornealTopographyMapTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0208, 0x0046, 0x0208,
      EVR_IS, "VerticesOfTheOutlineOfPupil", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0210, 0x0046, 0x0210,
      EVR_SQ, "CornealTopographyMappingNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0211, 0x0046, 0x0211,
      EVR_SQ, "MaximumCornealCurvatureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0212, 0x0046, 0x0212,
      EVR_FL, "MaximumCornealCurvature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0213, 0x0046, 0x0213,
      EVR_FL, "MaximumCornealCurvatureLocation", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0215, 0x0046, 0x0215,
      EVR_SQ, "MinimumKeratometricSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0218, 0x0046, 0x0218,
      EVR_SQ, "SimulatedKeratometricCylinderSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0220, 0x0046, 0x0220,
      EVR_FL, "AverageCornealPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0224, 0x0046, 0x0224,
      EVR_FL, "CornealISValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0227, 0x0046, 0x0227,
      EVR_FL, "AnalyzedArea", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0230, 0x0046, 0x0230,
      EVR_FL, "SurfaceRegularityIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0232, 0x0046, 0x0232,
      EVR_FL, "SurfaceAsymmetryIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0234, 0x0046, 0x0234,
      EVR_FL, "CornealEccentricityIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0236, 0x0046, 0x0236,
      EVR_FL, "KeratoconusPredictionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0238, 0x0046, 0x0238,
      EVR_FL, "DecimalPotentialVisualAcuity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0242, 0x0046, 0x0242,
      EVR_CS, "CornealTopographyMapQualityEvaluation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0244, 0x0046, 0x0244,
      EVR_SQ, "SourceImageCornealProcessedDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0247, 0x0046, 0x0247,
      EVR_FL, "CornealPointLocation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0248, 0x0046, 0x0248,
      EVR_CS, "CornealPointEstimated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0249, 0x0046, 0x0249,
      EVR_FL, "AxialPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0250, 0x0046, 0x0250,
      EVR_FL, "TangentialPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0251, 0x0046, 0x0251,
      EVR_FL, "RefractivePower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0252, 0x0046, 0x0252,
      EVR_FL, "RelativeElevation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0046, 0x0253, 0x0046, 0x0253,
      EVR_FL, "CornealWavefront", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0047, 0x0001, 0x0047, 0x0001,
      EVR_SQ, "ReconstructionParametersSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0050, 0x0047, 0x0050,
      EVR_UL, "VolumeVoxelCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0051, 0x0047, 0x0051,
      EVR_UL, "VolumeSegmentCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0053, 0x0047, 0x0053,
      EVR_US, "VolumeSliceSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0054, 0x0047, 0x0054,
      EVR_US, "VolumeSliceCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0055, 0x0047, 0x0055,
      EVR_SL, "VolumeThresholdValue", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0057, 0x0047, 0x0057,
      EVR_DS, "VolumeVoxelRatio", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0058, 0x0047, 0x0058,
      EVR_DS, "VolumeVoxelSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0059, 0x0047, 0x0059,
      EVR_US, "VolumeZPositionSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0060, 0x0047, 0x0060,
      EVR_DS, "VolumeBaseLine", 9, 9, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0061, 0x0047, 0x0061,
      EVR_DS, "VolumeCenterPoint", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0063, 0x0047, 0x0063,
      EVR_SL, "VolumeSkewBase", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0064, 0x0047, 0x0064,
      EVR_DS, "VolumeRegistrationTransformRotationMatrix", 9, 9, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0065, 0x0047, 0x0065,
      EVR_DS, "VolumeRegistrationTransformTranslationVector", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0070, 0x0047, 0x0070,
      EVR_DS, "KVPList", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0071, 0x0047, 0x0071,
      EVR_IS, "XRayTubeCurrentList", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0072, 0x0047, 0x0072,
      EVR_IS, "ExposureList", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0080, 0x0047, 0x0080,
      EVR_LO, "AcquisitionDLXIdentifier", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0085, 0x0047, 0x0085,
      EVR_SQ, "AcquisitionDLX2DSeriesSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0089, 0x0047, 0x0089,
      EVR_DS, "ContrastAgentVolumeList", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x008a, 0x0047, 0x008a,
      EVR_US, "NumberOfInjections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x008b, 0x0047, 0x008b,
      EVR_US, "FrameCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0091, 0x0047, 0x0091,
      EVR_LO, "XA3DReconstructionAlgorithmName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0092, 0x0047, 0x0092,
      EVR_CS, "XA3DReconstructionAlgorithmVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0093, 0x0047, 0x0093,
      EVR_DA, "DLXCalibrationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0094, 0x0047, 0x0094,
      EVR_TM, "DLXCalibrationTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0095, 0x0047, 0x0095,
      EVR_CS, "DLXCalibrationStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0096, 0x0047, 0x0096,
      EVR_IS, "UsedFrames", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0098, 0x0047, 0x0098,
      EVR_US, "TransformCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x0099, 0x0047, 0x0099,
      EVR_SQ, "TransformSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x009a, 0x0047, 0x009a,
      EVR_DS, "TransformRotationMatrix", 9, 9, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x009b, 0x0047, 0x009b,
      EVR_DS, "TransformTranslationVector", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x009c, 0x0047, 0x009c,
      EVR_LO, "TransformLabel", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b0, 0x0047, 0x00b0,
      EVR_SQ, "WireframeList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b1, 0x0047, 0x00b1,
      EVR_US, "WireframeCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b2, 0x0047, 0x00b2,
      EVR_US, "LocationSystem", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b5, 0x0047, 0x00b5,
      EVR_LO, "WireframeName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b6, 0x0047, 0x00b6,
      EVR_LO, "WireframeGroupName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b7, 0x0047, 0x00b7,
      EVR_LO, "WireframeColor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b8, 0x0047, 0x00b8,
      EVR_SL, "WireframeAttributes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00b9, 0x0047, 0x00b9,
      EVR_SL, "WireframePointCount", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00ba, 0x0047, 0x00ba,
      EVR_SL, "WireframeTimestamp", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00bb, 0x0047, 0x00bb,
      EVR_SQ, "WireframePointList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00bc, 0x0047, 0x00bc,
      EVR_DS, "WireframePointsCoordinates", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00c0, 0x0047, 0x00c0,
      EVR_DS, "VolumeUpperLeftHighCornerRAS", 3, 3, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00c1, 0x0047, 0x00c1,
      EVR_DS, "VolumeSliceToRASRotationMatrix", 9, 9, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00c2, 0x0047, 0x00c2,
      EVR_DS, "VolumeUpperLeftHighCornerTLOC", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00d1, 0x0047, 0x00d1,
      EVR_OB, "VolumeSegmentList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00d2, 0x0047, 0x00d2,
      EVR_OB, "VolumeGradientList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00d3, 0x0047, 0x00d3,
      EVR_OB, "VolumeDensityList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00d4, 0x0047, 0x00d4,
      EVR_OB, "VolumeZPositionList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
  , { 0x0047, 0x00d5, 0x0047, 0x00d5,
      EVR_OB, "VolumeOriginalIndexList", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_ADWSoft_3D1" }
#endif
  , { 0x0048, 0x0001, 0x0048, 0x0001,
      EVR_FL, "ImagedVolumeWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0002, 0x0048, 0x0002,
      EVR_FL, "ImagedVolumeHeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0003, 0x0048, 0x0003,
      EVR_FL, "ImagedVolumeDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0006, 0x0048, 0x0006,
      EVR_UL, "TotalPixelMatrixColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0007, 0x0048, 0x0007,
      EVR_UL, "TotalPixelMatrixRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0008, 0x0048, 0x0008,
      EVR_SQ, "TotalPixelMatrixOriginSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0010, 0x0048, 0x0010,
      EVR_CS, "SpecimenLabelInImage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0011, 0x0048, 0x0011,
      EVR_CS, "FocusMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0012, 0x0048, 0x0012,
      EVR_CS, "ExtendedDepthOfField", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0013, 0x0048, 0x0013,
      EVR_US, "NumberOfFocalPlanes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0014, 0x0048, 0x0014,
      EVR_FL, "DistanceBetweenFocalPlanes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0015, 0x0048, 0x0015,
      EVR_US, "RecommendedAbsentPixelCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0100, 0x0048, 0x0100,
      EVR_SQ, "IlluminatorTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0102, 0x0048, 0x0102,
      EVR_DS, "ImageOrientationSlide", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0105, 0x0048, 0x0105,
      EVR_SQ, "OpticalPathSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0106, 0x0048, 0x0106,
      EVR_SH, "OpticalPathIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0107, 0x0048, 0x0107,
      EVR_ST, "OpticalPathDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0108, 0x0048, 0x0108,
      EVR_SQ, "IlluminationColorCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0110, 0x0048, 0x0110,
      EVR_SQ, "SpecimenReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0111, 0x0048, 0x0111,
      EVR_DS, "CondenserLensPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0112, 0x0048, 0x0112,
      EVR_DS, "ObjectiveLensPower", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0113, 0x0048, 0x0113,
      EVR_DS, "ObjectiveLensNumericalAperture", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0114, 0x0048, 0x0114,
      EVR_CS, "ConfocalMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0115, 0x0048, 0x0115,
      EVR_CS, "TissueLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0116, 0x0048, 0x0116,
      EVR_SQ, "ConfocalMicroscopyImageFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0117, 0x0048, 0x0117,
      EVR_FD, "ImageAcquisitionDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0120, 0x0048, 0x0120,
      EVR_SQ, "PaletteColorLookupTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0200, 0x0048, 0x0200,
      EVR_SQ, "RETIRED_ReferencedImageNavigationSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0201, 0x0048, 0x0201,
      EVR_US, "RETIRED_TopLeftHandCornerOfLocalizerArea", 2, 2, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0202, 0x0048, 0x0202,
      EVR_US, "RETIRED_BottomRightHandCornerOfLocalizerArea", 2, 2, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0207, 0x0048, 0x0207,
      EVR_SQ, "OpticalPathIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x021a, 0x0048, 0x021a,
      EVR_SQ, "PlanePositionSlideSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x021e, 0x0048, 0x021e,
      EVR_SL, "ColumnPositionInTotalImagePixelMatrix", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x021f, 0x0048, 0x021f,
      EVR_SL, "RowPositionInTotalImagePixelMatrix", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0301, 0x0048, 0x0301,
      EVR_CS, "PixelOriginInterpretation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0302, 0x0048, 0x0302,
      EVR_UL, "NumberOfOpticalPaths", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0048, 0x0303, 0x0048, 0x0303,
      EVR_UL, "TotalPixelMatrixFocalPlanes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0004, 0x0050, 0x0004,
      EVR_CS, "CalibrationImage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0010, 0x0050, 0x0010,
      EVR_SQ, "DeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0012, 0x0050, 0x0012,
      EVR_SQ, "ContainerComponentTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0013, 0x0050, 0x0013,
      EVR_FD, "ContainerComponentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0014, 0x0050, 0x0014,
      EVR_DS, "DeviceLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0015, 0x0050, 0x0015,
      EVR_FD, "ContainerComponentWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0016, 0x0050, 0x0016,
      EVR_DS, "DeviceDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0017, 0x0050, 0x0017,
      EVR_CS, "DeviceDiameterUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0018, 0x0050, 0x0018,
      EVR_DS, "DeviceVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0019, 0x0050, 0x0019,
      EVR_DS, "InterMarkerDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x001a, 0x0050, 0x001a,
      EVR_CS, "ContainerComponentMaterial", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x001b, 0x0050, 0x001b,
      EVR_LO, "ContainerComponentID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x001c, 0x0050, 0x001c,
      EVR_FD, "ContainerComponentLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x001d, 0x0050, 0x001d,
      EVR_FD, "ContainerComponentDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x001e, 0x0050, 0x001e,
      EVR_LO, "ContainerComponentDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0020, 0x0050, 0x0020,
      EVR_LO, "DeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0050, 0x0021, 0x0050, 0x0021,
      EVR_ST, "LongDeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0051, 0x0010, 0x0051, 0x0010,
      EVR_LT, "ImageText", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
#endif
  , { 0x0052, 0x0001, 0x0052, 0x0001,
      EVR_FL, "ContrastBolusIngredientPercentByVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0002, 0x0052, 0x0002,
      EVR_FD, "OCTFocalDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0003, 0x0052, 0x0003,
      EVR_FD, "BeamSpotSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0004, 0x0052, 0x0004,
      EVR_FD, "EffectiveRefractiveIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0006, 0x0052, 0x0006,
      EVR_CS, "OCTAcquisitionDomain", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0007, 0x0052, 0x0007,
      EVR_FD, "OCTOpticalCenterWavelength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0008, 0x0052, 0x0008,
      EVR_FD, "AxialResolution", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0009, 0x0052, 0x0009,
      EVR_FD, "RangingDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0011, 0x0052, 0x0011,
      EVR_FD, "ALineRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0012, 0x0052, 0x0012,
      EVR_US, "ALinesPerFrame", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0013, 0x0052, 0x0013,
      EVR_FD, "CatheterRotationalRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0014, 0x0052, 0x0014,
      EVR_FD, "ALinePixelSpacing", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0016, 0x0052, 0x0016,
      EVR_SQ, "ModeOfPercutaneousAccessSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0025, 0x0052, 0x0025,
      EVR_SQ, "IntravascularOCTFrameTypeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0026, 0x0052, 0x0026,
      EVR_CS, "OCTZOffsetApplied", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0027, 0x0052, 0x0027,
      EVR_SQ, "IntravascularFrameContentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0028, 0x0052, 0x0028,
      EVR_FD, "IntravascularLongitudinalDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0029, 0x0052, 0x0029,
      EVR_SQ, "IntravascularOCTFrameContentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0030, 0x0052, 0x0030,
      EVR_SS, "OCTZOffsetCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0031, 0x0052, 0x0031,
      EVR_CS, "CatheterDirectionOfRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0033, 0x0052, 0x0033,
      EVR_FD, "SeamLineLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0034, 0x0052, 0x0034,
      EVR_FD, "FirstALineLocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0036, 0x0052, 0x0036,
      EVR_US, "SeamLineIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0038, 0x0052, 0x0038,
      EVR_US, "NumberOfPaddedALines", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x0039, 0x0052, 0x0039,
      EVR_CS, "InterpolationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0052, 0x003a, 0x0052, 0x003a,
      EVR_CS, "RefractiveIndexApplied", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0010, 0x0054, 0x0010,
      EVR_US, "EnergyWindowVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0011, 0x0054, 0x0011,
      EVR_US, "NumberOfEnergyWindows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0012, 0x0054, 0x0012,
      EVR_SQ, "EnergyWindowInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0013, 0x0054, 0x0013,
      EVR_SQ, "EnergyWindowRangeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0014, 0x0054, 0x0014,
      EVR_DS, "EnergyWindowLowerLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0015, 0x0054, 0x0015,
      EVR_DS, "EnergyWindowUpperLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0016, 0x0054, 0x0016,
      EVR_SQ, "RadiopharmaceuticalInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0017, 0x0054, 0x0017,
      EVR_IS, "ResidualSyringeCounts", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0018, 0x0054, 0x0018,
      EVR_SH, "EnergyWindowName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0020, 0x0054, 0x0020,
      EVR_US, "DetectorVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0021, 0x0054, 0x0021,
      EVR_US, "NumberOfDetectors", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0022, 0x0054, 0x0022,
      EVR_SQ, "DetectorInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0030, 0x0054, 0x0030,
      EVR_US, "PhaseVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0031, 0x0054, 0x0031,
      EVR_US, "NumberOfPhases", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0032, 0x0054, 0x0032,
      EVR_SQ, "PhaseInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0033, 0x0054, 0x0033,
      EVR_US, "NumberOfFramesInPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0036, 0x0054, 0x0036,
      EVR_IS, "PhaseDelay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0038, 0x0054, 0x0038,
      EVR_IS, "PauseBetweenFrames", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0039, 0x0054, 0x0039,
      EVR_CS, "PhaseDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0050, 0x0054, 0x0050,
      EVR_US, "RotationVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0051, 0x0054, 0x0051,
      EVR_US, "NumberOfRotations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0052, 0x0054, 0x0052,
      EVR_SQ, "RotationInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0053, 0x0054, 0x0053,
      EVR_US, "NumberOfFramesInRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0060, 0x0054, 0x0060,
      EVR_US, "RRIntervalVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0061, 0x0054, 0x0061,
      EVR_US, "NumberOfRRIntervals", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0062, 0x0054, 0x0062,
      EVR_SQ, "GatedInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0063, 0x0054, 0x0063,
      EVR_SQ, "DataInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0070, 0x0054, 0x0070,
      EVR_US, "TimeSlotVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0071, 0x0054, 0x0071,
      EVR_US, "NumberOfTimeSlots", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0072, 0x0054, 0x0072,
      EVR_SQ, "TimeSlotInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0073, 0x0054, 0x0073,
      EVR_DS, "TimeSlotTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0080, 0x0054, 0x0080,
      EVR_US, "SliceVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0081, 0x0054, 0x0081,
      EVR_US, "NumberOfSlices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0090, 0x0054, 0x0090,
      EVR_US, "AngularViewVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0100, 0x0054, 0x0100,
      EVR_US, "TimeSliceVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0101, 0x0054, 0x0101,
      EVR_US, "NumberOfTimeSlices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0200, 0x0054, 0x0200,
      EVR_DS, "StartAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0202, 0x0054, 0x0202,
      EVR_CS, "TypeOfDetectorMotion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0210, 0x0054, 0x0210,
      EVR_IS, "TriggerVector", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0211, 0x0054, 0x0211,
      EVR_US, "NumberOfTriggersInPhase", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0220, 0x0054, 0x0220,
      EVR_SQ, "ViewCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0222, 0x0054, 0x0222,
      EVR_SQ, "ViewModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0300, 0x0054, 0x0300,
      EVR_SQ, "RadionuclideCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0302, 0x0054, 0x0302,
      EVR_SQ, "AdministrationRouteCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0304, 0x0054, 0x0304,
      EVR_SQ, "RadiopharmaceuticalCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0306, 0x0054, 0x0306,
      EVR_SQ, "CalibrationDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0308, 0x0054, 0x0308,
      EVR_US, "EnergyWindowNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0400, 0x0054, 0x0400,
      EVR_SH, "ImageID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0410, 0x0054, 0x0410,
      EVR_SQ, "PatientOrientationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0412, 0x0054, 0x0412,
      EVR_SQ, "PatientOrientationModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0414, 0x0054, 0x0414,
      EVR_SQ, "PatientGantryRelationshipCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0500, 0x0054, 0x0500,
      EVR_CS, "SliceProgressionDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x0501, 0x0054, 0x0501,
      EVR_CS, "ScanProgressionDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1000, 0x0054, 0x1000,
      EVR_CS, "SeriesType", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1001, 0x0054, 0x1001,
      EVR_CS, "Units", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1002, 0x0054, 0x1002,
      EVR_CS, "CountsSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1004, 0x0054, 0x1004,
      EVR_CS, "ReprojectionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1006, 0x0054, 0x1006,
      EVR_CS, "SUVType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1100, 0x0054, 0x1100,
      EVR_CS, "RandomsCorrectionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1101, 0x0054, 0x1101,
      EVR_LO, "AttenuationCorrectionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1102, 0x0054, 0x1102,
      EVR_CS, "DecayCorrection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1103, 0x0054, 0x1103,
      EVR_LO, "ReconstructionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1104, 0x0054, 0x1104,
      EVR_LO, "DetectorLinesOfResponseUsed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1105, 0x0054, 0x1105,
      EVR_LO, "ScatterCorrectionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1200, 0x0054, 0x1200,
      EVR_DS, "AxialAcceptance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1201, 0x0054, 0x1201,
      EVR_IS, "AxialMash", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1202, 0x0054, 0x1202,
      EVR_IS, "TransverseMash", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1203, 0x0054, 0x1203,
      EVR_DS, "DetectorElementSize", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1210, 0x0054, 0x1210,
      EVR_DS, "CoincidenceWindowWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1220, 0x0054, 0x1220,
      EVR_CS, "SecondaryCountsType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1300, 0x0054, 0x1300,
      EVR_DS, "FrameReferenceTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1310, 0x0054, 0x1310,
      EVR_IS, "PrimaryPromptsCountsAccumulated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1311, 0x0054, 0x1311,
      EVR_IS, "SecondaryCountsAccumulated", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1320, 0x0054, 0x1320,
      EVR_DS, "SliceSensitivityFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1321, 0x0054, 0x1321,
      EVR_DS, "DecayFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1322, 0x0054, 0x1322,
      EVR_DS, "DoseCalibrationFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1323, 0x0054, 0x1323,
      EVR_DS, "ScatterFractionFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1324, 0x0054, 0x1324,
      EVR_DS, "DeadTimeFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1330, 0x0054, 0x1330,
      EVR_US, "ImageIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1400, 0x0054, 0x1400,
      EVR_CS, "RETIRED_CountsIncluded", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0054, 0x1401, 0x0054, 0x1401,
      EVR_CS, "RETIRED_DeadTimeCorrectionFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0055, 0x0046, 0x0055, 0x0046,
      EVR_LT, "CurrentWard", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x0060, 0x3000, 0x0060, 0x3000,
      EVR_SQ, "HistogramSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3002, 0x0060, 0x3002,
      EVR_US, "HistogramNumberOfBins", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3004, 0x0060, 0x3004,
      EVR_xs, "HistogramFirstBinValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3006, 0x0060, 0x3006,
      EVR_xs, "HistogramLastBinValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3008, 0x0060, 0x3008,
      EVR_US, "HistogramBinWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3010, 0x0060, 0x3010,
      EVR_LO, "HistogramExplanation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0060, 0x3020, 0x0060, 0x3020,
      EVR_UL, "HistogramData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0001, 0x0062, 0x0001,
      EVR_CS, "SegmentationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0002, 0x0062, 0x0002,
      EVR_SQ, "SegmentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0003, 0x0062, 0x0003,
      EVR_SQ, "SegmentedPropertyCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0004, 0x0062, 0x0004,
      EVR_US, "SegmentNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0005, 0x0062, 0x0005,
      EVR_LO, "SegmentLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0006, 0x0062, 0x0006,
      EVR_ST, "SegmentDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0007, 0x0062, 0x0007,
      EVR_SQ, "SegmentationAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0008, 0x0062, 0x0008,
      EVR_CS, "SegmentAlgorithmType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0009, 0x0062, 0x0009,
      EVR_LO, "SegmentAlgorithmName", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000a, 0x0062, 0x000a,
      EVR_SQ, "SegmentIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000b, 0x0062, 0x000b,
      EVR_US, "ReferencedSegmentNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000c, 0x0062, 0x000c,
      EVR_US, "RecommendedDisplayGrayscaleValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000d, 0x0062, 0x000d,
      EVR_US, "RecommendedDisplayCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000e, 0x0062, 0x000e,
      EVR_US, "MaximumFractionalValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x000f, 0x0062, 0x000f,
      EVR_SQ, "SegmentedPropertyTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0010, 0x0062, 0x0010,
      EVR_CS, "SegmentationFractionalType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0011, 0x0062, 0x0011,
      EVR_SQ, "SegmentedPropertyTypeModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0012, 0x0062, 0x0012,
      EVR_SQ, "UsedSegmentsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0013, 0x0062, 0x0013,
      EVR_CS, "SegmentsOverlap", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0020, 0x0062, 0x0020,
      EVR_UT, "TrackingID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0062, 0x0021, 0x0062, 0x0021,
      EVR_UI, "TrackingUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0002, 0x0064, 0x0002,
      EVR_SQ, "DeformableRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0003, 0x0064, 0x0003,
      EVR_UI, "SourceFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0005, 0x0064, 0x0005,
      EVR_SQ, "DeformableRegistrationGridSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0007, 0x0064, 0x0007,
      EVR_UL, "GridDimensions", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0008, 0x0064, 0x0008,
      EVR_FD, "GridResolution", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0009, 0x0064, 0x0009,
      EVR_OF, "VectorGridData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x000f, 0x0064, 0x000f,
      EVR_SQ, "PreDeformationMatrixRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0064, 0x0010, 0x0064, 0x0010,
      EVR_SQ, "PostDeformationMatrixRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0001, 0x0066, 0x0001,
      EVR_UL, "NumberOfSurfaces", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0002, 0x0066, 0x0002,
      EVR_SQ, "SurfaceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0003, 0x0066, 0x0003,
      EVR_UL, "SurfaceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0004, 0x0066, 0x0004,
      EVR_LT, "SurfaceComments", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0005, 0x0066, 0x0005,
      EVR_FL, "SurfaceOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0009, 0x0066, 0x0009,
      EVR_CS, "SurfaceProcessing", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x000a, 0x0066, 0x000a,
      EVR_FL, "SurfaceProcessingRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x000b, 0x0066, 0x000b,
      EVR_LO, "SurfaceProcessingDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x000c, 0x0066, 0x000c,
      EVR_FL, "RecommendedPresentationOpacity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x000d, 0x0066, 0x000d,
      EVR_CS, "RecommendedPresentationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x000e, 0x0066, 0x000e,
      EVR_CS, "FiniteVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0010, 0x0066, 0x0010,
      EVR_CS, "Manifold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0011, 0x0066, 0x0011,
      EVR_SQ, "SurfacePointsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0012, 0x0066, 0x0012,
      EVR_SQ, "SurfacePointsNormalsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0013, 0x0066, 0x0013,
      EVR_SQ, "SurfaceMeshPrimitivesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0015, 0x0066, 0x0015,
      EVR_UL, "NumberOfSurfacePoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0016, 0x0066, 0x0016,
      EVR_OF, "PointCoordinatesData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0017, 0x0066, 0x0017,
      EVR_FL, "PointPositionAccuracy", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0018, 0x0066, 0x0018,
      EVR_FL, "MeanPointDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0019, 0x0066, 0x0019,
      EVR_FL, "MaximumPointDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x001a, 0x0066, 0x001a,
      EVR_FL, "PointsBoundingBoxCoordinates", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x001b, 0x0066, 0x001b,
      EVR_FL, "AxisOfRotation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x001c, 0x0066, 0x001c,
      EVR_FL, "CenterOfRotation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x001e, 0x0066, 0x001e,
      EVR_UL, "NumberOfVectors", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x001f, 0x0066, 0x001f,
      EVR_US, "VectorDimensionality", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0020, 0x0066, 0x0020,
      EVR_FL, "VectorAccuracy", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0021, 0x0066, 0x0021,
      EVR_OF, "VectorCoordinateData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0022, 0x0066, 0x0022,
      EVR_OD, "DoublePointCoordinatesData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0023, 0x0066, 0x0023,
      EVR_OW, "RETIRED_TrianglePointIndexList", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0024, 0x0066, 0x0024,
      EVR_OW, "RETIRED_EdgePointIndexList", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0025, 0x0066, 0x0025,
      EVR_OW, "RETIRED_VertexPointIndexList", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0026, 0x0066, 0x0026,
      EVR_SQ, "TriangleStripSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0027, 0x0066, 0x0027,
      EVR_SQ, "TriangleFanSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0028, 0x0066, 0x0028,
      EVR_SQ, "LineSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0029, 0x0066, 0x0029,
      EVR_OW, "RETIRED_PrimitivePointIndexList", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002a, 0x0066, 0x002a,
      EVR_UL, "SurfaceCount", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002b, 0x0066, 0x002b,
      EVR_SQ, "ReferencedSurfaceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002c, 0x0066, 0x002c,
      EVR_UL, "ReferencedSurfaceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002d, 0x0066, 0x002d,
      EVR_SQ, "SegmentSurfaceGenerationAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002e, 0x0066, 0x002e,
      EVR_SQ, "SegmentSurfaceSourceInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x002f, 0x0066, 0x002f,
      EVR_SQ, "AlgorithmFamilyCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0030, 0x0066, 0x0030,
      EVR_SQ, "AlgorithmNameCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0031, 0x0066, 0x0031,
      EVR_LO, "AlgorithmVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0032, 0x0066, 0x0032,
      EVR_LT, "AlgorithmParameters", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0034, 0x0066, 0x0034,
      EVR_SQ, "FacetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0035, 0x0066, 0x0035,
      EVR_SQ, "SurfaceProcessingAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0036, 0x0066, 0x0036,
      EVR_LO, "AlgorithmName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0037, 0x0066, 0x0037,
      EVR_FL, "RecommendedPointRadius", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0038, 0x0066, 0x0038,
      EVR_FL, "RecommendedLineThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0040, 0x0066, 0x0040,
      EVR_OL, "LongPrimitivePointIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0041, 0x0066, 0x0041,
      EVR_OL, "LongTrianglePointIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0042, 0x0066, 0x0042,
      EVR_OL, "LongEdgePointIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0043, 0x0066, 0x0043,
      EVR_OL, "LongVertexPointIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0101, 0x0066, 0x0101,
      EVR_SQ, "TrackSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0102, 0x0066, 0x0102,
      EVR_SQ, "TrackSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0103, 0x0066, 0x0103,
      EVR_OW, "RecommendedDisplayCIELabValueList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0104, 0x0066, 0x0104,
      EVR_SQ, "TrackingAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0105, 0x0066, 0x0105,
      EVR_UL, "TrackSetNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0106, 0x0066, 0x0106,
      EVR_LO, "TrackSetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0107, 0x0066, 0x0107,
      EVR_UT, "TrackSetDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0108, 0x0066, 0x0108,
      EVR_SQ, "TrackSetAnatomicalTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0121, 0x0066, 0x0121,
      EVR_SQ, "MeasurementsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0124, 0x0066, 0x0124,
      EVR_SQ, "TrackSetStatisticsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0125, 0x0066, 0x0125,
      EVR_OF, "FloatingPointValues", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0129, 0x0066, 0x0129,
      EVR_OL, "TrackPointIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0130, 0x0066, 0x0130,
      EVR_SQ, "TrackStatisticsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0132, 0x0066, 0x0132,
      EVR_SQ, "MeasurementValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0133, 0x0066, 0x0133,
      EVR_SQ, "DiffusionAcquisitionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0066, 0x0134, 0x0066, 0x0134,
      EVR_SQ, "DiffusionModelCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6210, 0x0068, 0x6210,
      EVR_LO, "ImplantSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6221, 0x0068, 0x6221,
      EVR_LO, "ImplantTemplateVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6222, 0x0068, 0x6222,
      EVR_SQ, "ReplacedImplantTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6223, 0x0068, 0x6223,
      EVR_CS, "ImplantType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6224, 0x0068, 0x6224,
      EVR_SQ, "DerivationImplantTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6225, 0x0068, 0x6225,
      EVR_SQ, "OriginalImplantTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6226, 0x0068, 0x6226,
      EVR_DT, "EffectiveDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6230, 0x0068, 0x6230,
      EVR_SQ, "ImplantTargetAnatomySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6260, 0x0068, 0x6260,
      EVR_SQ, "InformationFromManufacturerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6265, 0x0068, 0x6265,
      EVR_SQ, "NotificationFromManufacturerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6270, 0x0068, 0x6270,
      EVR_DT, "InformationIssueDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6280, 0x0068, 0x6280,
      EVR_ST, "InformationSummary", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62a0, 0x0068, 0x62a0,
      EVR_SQ, "ImplantRegulatoryDisapprovalCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62a5, 0x0068, 0x62a5,
      EVR_FD, "OverallTemplateSpatialTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62c0, 0x0068, 0x62c0,
      EVR_SQ, "HPGLDocumentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62d0, 0x0068, 0x62d0,
      EVR_US, "HPGLDocumentID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62d5, 0x0068, 0x62d5,
      EVR_LO, "HPGLDocumentLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62e0, 0x0068, 0x62e0,
      EVR_SQ, "ViewOrientationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62f0, 0x0068, 0x62f0,
      EVR_SQ, "ViewOrientationModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x62f2, 0x0068, 0x62f2,
      EVR_FD, "HPGLDocumentScaling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6300, 0x0068, 0x6300,
      EVR_OB, "HPGLDocument", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6310, 0x0068, 0x6310,
      EVR_US, "HPGLContourPenNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6320, 0x0068, 0x6320,
      EVR_SQ, "HPGLPenSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6330, 0x0068, 0x6330,
      EVR_US, "HPGLPenNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6340, 0x0068, 0x6340,
      EVR_LO, "HPGLPenLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6345, 0x0068, 0x6345,
      EVR_ST, "HPGLPenDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6346, 0x0068, 0x6346,
      EVR_FD, "RecommendedRotationPoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6347, 0x0068, 0x6347,
      EVR_FD, "BoundingRectangle", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6350, 0x0068, 0x6350,
      EVR_US, "ImplantTemplate3DModelSurfaceNumber", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6360, 0x0068, 0x6360,
      EVR_SQ, "SurfaceModelDescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6380, 0x0068, 0x6380,
      EVR_LO, "SurfaceModelLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6390, 0x0068, 0x6390,
      EVR_FD, "SurfaceModelScalingFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63a0, 0x0068, 0x63a0,
      EVR_SQ, "MaterialsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63a4, 0x0068, 0x63a4,
      EVR_SQ, "CoatingMaterialsCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63a8, 0x0068, 0x63a8,
      EVR_SQ, "ImplantTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63ac, 0x0068, 0x63ac,
      EVR_SQ, "FixationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63b0, 0x0068, 0x63b0,
      EVR_SQ, "MatingFeatureSetsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63c0, 0x0068, 0x63c0,
      EVR_US, "MatingFeatureSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63d0, 0x0068, 0x63d0,
      EVR_LO, "MatingFeatureSetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63e0, 0x0068, 0x63e0,
      EVR_SQ, "MatingFeatureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x63f0, 0x0068, 0x63f0,
      EVR_US, "MatingFeatureID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6400, 0x0068, 0x6400,
      EVR_SQ, "MatingFeatureDegreeOfFreedomSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6410, 0x0068, 0x6410,
      EVR_US, "DegreeOfFreedomID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6420, 0x0068, 0x6420,
      EVR_CS, "DegreeOfFreedomType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6430, 0x0068, 0x6430,
      EVR_SQ, "TwoDMatingFeatureCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6440, 0x0068, 0x6440,
      EVR_US, "ReferencedHPGLDocumentID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6450, 0x0068, 0x6450,
      EVR_FD, "TwoDMatingPoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6460, 0x0068, 0x6460,
      EVR_FD, "TwoDMatingAxes", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6470, 0x0068, 0x6470,
      EVR_SQ, "TwoDDegreeOfFreedomSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6490, 0x0068, 0x6490,
      EVR_FD, "ThreeDDegreeOfFreedomAxis", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x64a0, 0x0068, 0x64a0,
      EVR_FD, "RangeOfFreedom", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x64c0, 0x0068, 0x64c0,
      EVR_FD, "ThreeDMatingPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x64d0, 0x0068, 0x64d0,
      EVR_FD, "ThreeDMatingAxes", 9, 9, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x64f0, 0x0068, 0x64f0,
      EVR_FD, "TwoDDegreeOfFreedomAxis", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6500, 0x0068, 0x6500,
      EVR_SQ, "PlanningLandmarkPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6510, 0x0068, 0x6510,
      EVR_SQ, "PlanningLandmarkLineSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6520, 0x0068, 0x6520,
      EVR_SQ, "PlanningLandmarkPlaneSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6530, 0x0068, 0x6530,
      EVR_US, "PlanningLandmarkID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6540, 0x0068, 0x6540,
      EVR_LO, "PlanningLandmarkDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6545, 0x0068, 0x6545,
      EVR_SQ, "PlanningLandmarkIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6550, 0x0068, 0x6550,
      EVR_SQ, "TwoDPointCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6560, 0x0068, 0x6560,
      EVR_FD, "TwoDPointCoordinates", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6590, 0x0068, 0x6590,
      EVR_FD, "ThreeDPointCoordinates", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x65a0, 0x0068, 0x65a0,
      EVR_SQ, "TwoDLineCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x65b0, 0x0068, 0x65b0,
      EVR_FD, "TwoDLineCoordinates", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x65d0, 0x0068, 0x65d0,
      EVR_FD, "ThreeDLineCoordinates", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x65e0, 0x0068, 0x65e0,
      EVR_SQ, "TwoDPlaneCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x65f0, 0x0068, 0x65f0,
      EVR_FD, "TwoDPlaneIntersection", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6610, 0x0068, 0x6610,
      EVR_FD, "ThreeDPlaneOrigin", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x6620, 0x0068, 0x6620,
      EVR_FD, "ThreeDPlaneNormal", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x7001, 0x0068, 0x7001,
      EVR_CS, "ModelModification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x7002, 0x0068, 0x7002,
      EVR_CS, "ModelMirroring", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x7003, 0x0068, 0x7003,
      EVR_SQ, "ModelUsageCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x7004, 0x0068, 0x7004,
      EVR_UI, "ModelGroupUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0068, 0x7005, 0x0068, 0x7005,
      EVR_UR, "RelativeURIReferenceWithinEncapsulatedDocument", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0001, 0x006a, 0x0001,
      EVR_CS, "AnnotationCoordinateType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0002, 0x006a, 0x0002,
      EVR_SQ, "AnnotationGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0003, 0x006a, 0x0003,
      EVR_UI, "AnnotationGroupUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0005, 0x006a, 0x0005,
      EVR_LO, "AnnotationGroupLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0006, 0x006a, 0x0006,
      EVR_UT, "AnnotationGroupDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0007, 0x006a, 0x0007,
      EVR_CS, "AnnotationGroupGenerationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0008, 0x006a, 0x0008,
      EVR_SQ, "AnnotationGroupAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0009, 0x006a, 0x0009,
      EVR_SQ, "AnnotationPropertyCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000a, 0x006a, 0x000a,
      EVR_SQ, "AnnotationPropertyTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000b, 0x006a, 0x000b,
      EVR_SQ, "AnnotationPropertyTypeModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000c, 0x006a, 0x000c,
      EVR_UL, "NumberOfAnnotations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000d, 0x006a, 0x000d,
      EVR_CS, "AnnotationAppliesToAllOpticalPaths", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000e, 0x006a, 0x000e,
      EVR_SH, "ReferencedOpticalPathIdentifier", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x000f, 0x006a, 0x000f,
      EVR_CS, "AnnotationAppliesToAllZPlanes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0010, 0x006a, 0x0010,
      EVR_FD, "CommonZCoordinateValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x006a, 0x0011, 0x006a, 0x0011,
      EVR_OL, "AnnotationIndexList", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0001, 0x0070, 0x0001,
      EVR_SQ, "GraphicAnnotationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0002, 0x0070, 0x0002,
      EVR_CS, "GraphicLayer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0003, 0x0070, 0x0003,
      EVR_CS, "BoundingBoxAnnotationUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0004, 0x0070, 0x0004,
      EVR_CS, "AnchorPointAnnotationUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0005, 0x0070, 0x0005,
      EVR_CS, "GraphicAnnotationUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0006, 0x0070, 0x0006,
      EVR_ST, "UnformattedTextValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0008, 0x0070, 0x0008,
      EVR_SQ, "TextObjectSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0009, 0x0070, 0x0009,
      EVR_SQ, "GraphicObjectSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0010, 0x0070, 0x0010,
      EVR_FL, "BoundingBoxTopLeftHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0011, 0x0070, 0x0011,
      EVR_FL, "BoundingBoxBottomRightHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0012, 0x0070, 0x0012,
      EVR_CS, "BoundingBoxTextHorizontalJustification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0014, 0x0070, 0x0014,
      EVR_FL, "AnchorPoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0015, 0x0070, 0x0015,
      EVR_CS, "AnchorPointVisibility", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0020, 0x0070, 0x0020,
      EVR_US, "GraphicDimensions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0021, 0x0070, 0x0021,
      EVR_US, "NumberOfGraphicPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0022, 0x0070, 0x0022,
      EVR_FL, "GraphicData", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0023, 0x0070, 0x0023,
      EVR_CS, "GraphicType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0024, 0x0070, 0x0024,
      EVR_CS, "GraphicFilled", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0040, 0x0070, 0x0040,
      EVR_IS, "RETIRED_ImageRotationRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0041, 0x0070, 0x0041,
      EVR_CS, "ImageHorizontalFlip", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0042, 0x0070, 0x0042,
      EVR_US, "ImageRotation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0050, 0x0070, 0x0050,
      EVR_US, "RETIRED_DisplayedAreaTopLeftHandCornerTrial", 2, 2, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0051, 0x0070, 0x0051,
      EVR_US, "RETIRED_DisplayedAreaBottomRightHandCornerTrial", 2, 2, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0052, 0x0070, 0x0052,
      EVR_SL, "DisplayedAreaTopLeftHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0053, 0x0070, 0x0053,
      EVR_SL, "DisplayedAreaBottomRightHandCorner", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x005a, 0x0070, 0x005a,
      EVR_SQ, "DisplayedAreaSelectionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0060, 0x0070, 0x0060,
      EVR_SQ, "GraphicLayerSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0062, 0x0070, 0x0062,
      EVR_IS, "GraphicLayerOrder", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0066, 0x0070, 0x0066,
      EVR_US, "GraphicLayerRecommendedDisplayGrayscaleValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0067, 0x0070, 0x0067,
      EVR_US, "RETIRED_GraphicLayerRecommendedDisplayRGBValue", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0068, 0x0070, 0x0068,
      EVR_LO, "GraphicLayerDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0080, 0x0070, 0x0080,
      EVR_CS, "ContentLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0081, 0x0070, 0x0081,
      EVR_LO, "ContentDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0082, 0x0070, 0x0082,
      EVR_DA, "PresentationCreationDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0083, 0x0070, 0x0083,
      EVR_TM, "PresentationCreationTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0084, 0x0070, 0x0084,
      EVR_PN, "ContentCreatorName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0086, 0x0070, 0x0086,
      EVR_SQ, "ContentCreatorIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0087, 0x0070, 0x0087,
      EVR_SQ, "AlternateContentDescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0100, 0x0070, 0x0100,
      EVR_CS, "PresentationSizeMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0101, 0x0070, 0x0101,
      EVR_DS, "PresentationPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0102, 0x0070, 0x0102,
      EVR_IS, "PresentationPixelAspectRatio", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0103, 0x0070, 0x0103,
      EVR_FL, "PresentationPixelMagnificationRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0207, 0x0070, 0x0207,
      EVR_LO, "GraphicGroupLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0208, 0x0070, 0x0208,
      EVR_ST, "GraphicGroupDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0209, 0x0070, 0x0209,
      EVR_SQ, "CompoundGraphicSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0226, 0x0070, 0x0226,
      EVR_UL, "CompoundGraphicInstanceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0227, 0x0070, 0x0227,
      EVR_LO, "FontName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0228, 0x0070, 0x0228,
      EVR_CS, "FontNameType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0229, 0x0070, 0x0229,
      EVR_LO, "CSSFontName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0230, 0x0070, 0x0230,
      EVR_FD, "RotationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0231, 0x0070, 0x0231,
      EVR_SQ, "TextStyleSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0232, 0x0070, 0x0232,
      EVR_SQ, "LineStyleSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0233, 0x0070, 0x0233,
      EVR_SQ, "FillStyleSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0234, 0x0070, 0x0234,
      EVR_SQ, "GraphicGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0241, 0x0070, 0x0241,
      EVR_US, "TextColorCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0242, 0x0070, 0x0242,
      EVR_CS, "HorizontalAlignment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0243, 0x0070, 0x0243,
      EVR_CS, "VerticalAlignment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0244, 0x0070, 0x0244,
      EVR_CS, "ShadowStyle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0245, 0x0070, 0x0245,
      EVR_FL, "ShadowOffsetX", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0246, 0x0070, 0x0246,
      EVR_FL, "ShadowOffsetY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0247, 0x0070, 0x0247,
      EVR_US, "ShadowColorCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0248, 0x0070, 0x0248,
      EVR_CS, "Underlined", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0249, 0x0070, 0x0249,
      EVR_CS, "Bold", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0250, 0x0070, 0x0250,
      EVR_CS, "Italic", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0251, 0x0070, 0x0251,
      EVR_US, "PatternOnColorCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0252, 0x0070, 0x0252,
      EVR_US, "PatternOffColorCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0253, 0x0070, 0x0253,
      EVR_FL, "LineThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0254, 0x0070, 0x0254,
      EVR_CS, "LineDashingStyle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0255, 0x0070, 0x0255,
      EVR_UL, "LinePattern", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0256, 0x0070, 0x0256,
      EVR_OB, "FillPattern", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0257, 0x0070, 0x0257,
      EVR_CS, "FillMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0258, 0x0070, 0x0258,
      EVR_FL, "ShadowOpacity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0261, 0x0070, 0x0261,
      EVR_FL, "GapLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0262, 0x0070, 0x0262,
      EVR_FL, "DiameterOfVisibility", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0273, 0x0070, 0x0273,
      EVR_FL, "RotationPoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0274, 0x0070, 0x0274,
      EVR_CS, "TickAlignment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0278, 0x0070, 0x0278,
      EVR_CS, "ShowTickLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0279, 0x0070, 0x0279,
      EVR_CS, "TickLabelAlignment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0282, 0x0070, 0x0282,
      EVR_CS, "CompoundGraphicUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0284, 0x0070, 0x0284,
      EVR_FL, "PatternOnOpacity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0285, 0x0070, 0x0285,
      EVR_FL, "PatternOffOpacity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0287, 0x0070, 0x0287,
      EVR_SQ, "MajorTicksSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0288, 0x0070, 0x0288,
      EVR_FL, "TickPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0289, 0x0070, 0x0289,
      EVR_SH, "TickLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0294, 0x0070, 0x0294,
      EVR_CS, "CompoundGraphicType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0295, 0x0070, 0x0295,
      EVR_UL, "GraphicGroupID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0306, 0x0070, 0x0306,
      EVR_CS, "ShapeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0308, 0x0070, 0x0308,
      EVR_SQ, "RegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0309, 0x0070, 0x0309,
      EVR_SQ, "MatrixRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x030a, 0x0070, 0x030a,
      EVR_SQ, "MatrixSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x030b, 0x0070, 0x030b,
      EVR_FD, "FrameOfReferenceToDisplayedCoordinateSystemTransformationMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x030c, 0x0070, 0x030c,
      EVR_CS, "FrameOfReferenceTransformationMatrixType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x030d, 0x0070, 0x030d,
      EVR_SQ, "RegistrationTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x030f, 0x0070, 0x030f,
      EVR_ST, "FiducialDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0310, 0x0070, 0x0310,
      EVR_SH, "FiducialIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0311, 0x0070, 0x0311,
      EVR_SQ, "FiducialIdentifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0312, 0x0070, 0x0312,
      EVR_FD, "ContourUncertaintyRadius", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0314, 0x0070, 0x0314,
      EVR_SQ, "UsedFiducialsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0315, 0x0070, 0x0315,
      EVR_SQ, "UsedRTStructureSetROISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0318, 0x0070, 0x0318,
      EVR_SQ, "GraphicCoordinatesDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x031a, 0x0070, 0x031a,
      EVR_UI, "FiducialUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x031b, 0x0070, 0x031b,
      EVR_UI, "ReferencedFiducialUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x031c, 0x0070, 0x031c,
      EVR_SQ, "FiducialSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x031e, 0x0070, 0x031e,
      EVR_SQ, "FiducialSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x031f, 0x0070, 0x031f,
      EVR_SQ, "FiducialsPropertyCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0401, 0x0070, 0x0401,
      EVR_US, "GraphicLayerRecommendedDisplayCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0402, 0x0070, 0x0402,
      EVR_SQ, "BlendingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0403, 0x0070, 0x0403,
      EVR_FL, "RelativeOpacity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0404, 0x0070, 0x0404,
      EVR_SQ, "ReferencedSpatialRegistrationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x0405, 0x0070, 0x0405,
      EVR_CS, "BlendingPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1101, 0x0070, 0x1101,
      EVR_UI, "PresentationDisplayCollectionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1102, 0x0070, 0x1102,
      EVR_UI, "PresentationSequenceCollectionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1103, 0x0070, 0x1103,
      EVR_US, "PresentationSequencePositionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1104, 0x0070, 0x1104,
      EVR_SQ, "RenderedImageReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1201, 0x0070, 0x1201,
      EVR_SQ, "VolumetricPresentationStateInputSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1202, 0x0070, 0x1202,
      EVR_CS, "PresentationInputType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1203, 0x0070, 0x1203,
      EVR_US, "InputSequencePositionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1204, 0x0070, 0x1204,
      EVR_CS, "Crop", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1205, 0x0070, 0x1205,
      EVR_US, "CroppingSpecificationIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1206, 0x0070, 0x1206,
      EVR_CS, "RETIRED_CompositingMethod", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1207, 0x0070, 0x1207,
      EVR_US, "VolumetricPresentationInputNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1208, 0x0070, 0x1208,
      EVR_CS, "ImageVolumeGeometry", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1209, 0x0070, 0x1209,
      EVR_UI, "VolumetricPresentationInputSetUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x120a, 0x0070, 0x120a,
      EVR_SQ, "VolumetricPresentationInputSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x120b, 0x0070, 0x120b,
      EVR_CS, "GlobalCrop", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x120c, 0x0070, 0x120c,
      EVR_US, "GlobalCroppingSpecificationIndex", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x120d, 0x0070, 0x120d,
      EVR_CS, "RenderingMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1301, 0x0070, 0x1301,
      EVR_SQ, "VolumeCroppingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1302, 0x0070, 0x1302,
      EVR_CS, "VolumeCroppingMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1303, 0x0070, 0x1303,
      EVR_FD, "BoundingBoxCrop", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1304, 0x0070, 0x1304,
      EVR_SQ, "ObliqueCroppingPlaneSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1305, 0x0070, 0x1305,
      EVR_FD, "Plane", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1306, 0x0070, 0x1306,
      EVR_FD, "PlaneNormal", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1309, 0x0070, 0x1309,
      EVR_US, "CroppingSpecificationNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1501, 0x0070, 0x1501,
      EVR_CS, "MultiPlanarReconstructionStyle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1502, 0x0070, 0x1502,
      EVR_CS, "MPRThicknessType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1503, 0x0070, 0x1503,
      EVR_FD, "MPRSlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1505, 0x0070, 0x1505,
      EVR_FD, "MPRTopLeftHandCorner", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1507, 0x0070, 0x1507,
      EVR_FD, "MPRViewWidthDirection", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1508, 0x0070, 0x1508,
      EVR_FD, "MPRViewWidth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x150c, 0x0070, 0x150c,
      EVR_UL, "NumberOfVolumetricCurvePoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x150d, 0x0070, 0x150d,
      EVR_OD, "VolumetricCurvePoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1511, 0x0070, 0x1511,
      EVR_FD, "MPRViewHeightDirection", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1512, 0x0070, 0x1512,
      EVR_FD, "MPRViewHeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1602, 0x0070, 0x1602,
      EVR_CS, "RenderProjection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1603, 0x0070, 0x1603,
      EVR_FD, "ViewpointPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1604, 0x0070, 0x1604,
      EVR_FD, "ViewpointLookAtPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1605, 0x0070, 0x1605,
      EVR_FD, "ViewpointUpDirection", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1606, 0x0070, 0x1606,
      EVR_FD, "RenderFieldOfView", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1607, 0x0070, 0x1607,
      EVR_FD, "SamplingStepSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1701, 0x0070, 0x1701,
      EVR_CS, "ShadingStyle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1702, 0x0070, 0x1702,
      EVR_FD, "AmbientReflectionIntensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1703, 0x0070, 0x1703,
      EVR_FD, "LightDirection", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1704, 0x0070, 0x1704,
      EVR_FD, "DiffuseReflectionIntensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1705, 0x0070, 0x1705,
      EVR_FD, "SpecularReflectionIntensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1706, 0x0070, 0x1706,
      EVR_FD, "Shininess", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1801, 0x0070, 0x1801,
      EVR_SQ, "PresentationStateClassificationComponentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1802, 0x0070, 0x1802,
      EVR_CS, "ComponentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1803, 0x0070, 0x1803,
      EVR_SQ, "ComponentInputSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1804, 0x0070, 0x1804,
      EVR_US, "VolumetricPresentationInputIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1805, 0x0070, 0x1805,
      EVR_SQ, "PresentationStateCompositorComponentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1806, 0x0070, 0x1806,
      EVR_SQ, "WeightingTransferFunctionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1807, 0x0070, 0x1807,
      EVR_US, "RETIRED_WeightingLookupTableDescriptor", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1808, 0x0070, 0x1808,
      EVR_OB, "RETIRED_WeightingLookupTableData", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1901, 0x0070, 0x1901,
      EVR_SQ, "VolumetricAnnotationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1903, 0x0070, 0x1903,
      EVR_SQ, "ReferencedStructuredContextSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1904, 0x0070, 0x1904,
      EVR_UI, "ReferencedContentItem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1905, 0x0070, 0x1905,
      EVR_SQ, "VolumetricPresentationInputAnnotationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1907, 0x0070, 0x1907,
      EVR_CS, "AnnotationClipping", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a01, 0x0070, 0x1a01,
      EVR_CS, "PresentationAnimationStyle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a03, 0x0070, 0x1a03,
      EVR_FD, "RecommendedAnimationRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a04, 0x0070, 0x1a04,
      EVR_SQ, "AnimationCurveSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a05, 0x0070, 0x1a05,
      EVR_FD, "AnimationStepSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a06, 0x0070, 0x1a06,
      EVR_FD, "SwivelRange", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a07, 0x0070, 0x1a07,
      EVR_OD, "VolumetricCurveUpDirections", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a08, 0x0070, 0x1a08,
      EVR_SQ, "VolumeStreamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1a09, 0x0070, 0x1a09,
      EVR_LO, "RGBATransferFunctionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b01, 0x0070, 0x1b01,
      EVR_SQ, "AdvancedBlendingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b02, 0x0070, 0x1b02,
      EVR_US, "BlendingInputNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b03, 0x0070, 0x1b03,
      EVR_SQ, "BlendingDisplayInputSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b04, 0x0070, 0x1b04,
      EVR_SQ, "BlendingDisplaySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b06, 0x0070, 0x1b06,
      EVR_CS, "BlendingMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b07, 0x0070, 0x1b07,
      EVR_CS, "TimeSeriesBlending", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b08, 0x0070, 0x1b08,
      EVR_CS, "GeometryForDisplay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b11, 0x0070, 0x1b11,
      EVR_SQ, "ThresholdSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b12, 0x0070, 0x1b12,
      EVR_SQ, "ThresholdValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b13, 0x0070, 0x1b13,
      EVR_CS, "ThresholdType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0070, 0x1b14, 0x0070, 0x1b14,
      EVR_FD, "ThresholdValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0002, 0x0072, 0x0002,
      EVR_SH, "HangingProtocolName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0004, 0x0072, 0x0004,
      EVR_LO, "HangingProtocolDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0006, 0x0072, 0x0006,
      EVR_CS, "HangingProtocolLevel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0008, 0x0072, 0x0008,
      EVR_LO, "HangingProtocolCreator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x000a, 0x0072, 0x000a,
      EVR_DT, "HangingProtocolCreationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x000c, 0x0072, 0x000c,
      EVR_SQ, "HangingProtocolDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x000e, 0x0072, 0x000e,
      EVR_SQ, "HangingProtocolUserIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0010, 0x0072, 0x0010,
      EVR_LO, "HangingProtocolUserGroupName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0012, 0x0072, 0x0012,
      EVR_SQ, "SourceHangingProtocolSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0014, 0x0072, 0x0014,
      EVR_US, "NumberOfPriorsReferenced", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0020, 0x0072, 0x0020,
      EVR_SQ, "ImageSetsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0022, 0x0072, 0x0022,
      EVR_SQ, "ImageSetSelectorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0024, 0x0072, 0x0024,
      EVR_CS, "ImageSetSelectorUsageFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0026, 0x0072, 0x0026,
      EVR_AT, "SelectorAttribute", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0028, 0x0072, 0x0028,
      EVR_US, "SelectorValueNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0030, 0x0072, 0x0030,
      EVR_SQ, "TimeBasedImageSetsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0032, 0x0072, 0x0032,
      EVR_US, "ImageSetNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0034, 0x0072, 0x0034,
      EVR_CS, "ImageSetSelectorCategory", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0038, 0x0072, 0x0038,
      EVR_US, "RelativeTime", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x003a, 0x0072, 0x003a,
      EVR_CS, "RelativeTimeUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x003c, 0x0072, 0x003c,
      EVR_SS, "AbstractPriorValue", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x003e, 0x0072, 0x003e,
      EVR_SQ, "AbstractPriorCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0040, 0x0072, 0x0040,
      EVR_LO, "ImageSetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0050, 0x0072, 0x0050,
      EVR_CS, "SelectorAttributeVR", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0052, 0x0072, 0x0052,
      EVR_AT, "SelectorSequencePointer", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0054, 0x0072, 0x0054,
      EVR_LO, "SelectorSequencePointerPrivateCreator", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0056, 0x0072, 0x0056,
      EVR_LO, "SelectorAttributePrivateCreator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x005e, 0x0072, 0x005e,
      EVR_AE, "SelectorAEValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x005f, 0x0072, 0x005f,
      EVR_AS, "SelectorASValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0060, 0x0072, 0x0060,
      EVR_AT, "SelectorATValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0061, 0x0072, 0x0061,
      EVR_DA, "SelectorDAValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0062, 0x0072, 0x0062,
      EVR_CS, "SelectorCSValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0063, 0x0072, 0x0063,
      EVR_DT, "SelectorDTValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0064, 0x0072, 0x0064,
      EVR_IS, "SelectorISValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0065, 0x0072, 0x0065,
      EVR_OB, "SelectorOBValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0066, 0x0072, 0x0066,
      EVR_LO, "SelectorLOValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0067, 0x0072, 0x0067,
      EVR_OF, "SelectorOFValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0068, 0x0072, 0x0068,
      EVR_LT, "SelectorLTValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0069, 0x0072, 0x0069,
      EVR_OW, "SelectorOWValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006a, 0x0072, 0x006a,
      EVR_PN, "SelectorPNValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006b, 0x0072, 0x006b,
      EVR_TM, "SelectorTMValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006c, 0x0072, 0x006c,
      EVR_SH, "SelectorSHValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006d, 0x0072, 0x006d,
      EVR_UN, "SelectorUNValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006e, 0x0072, 0x006e,
      EVR_ST, "SelectorSTValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x006f, 0x0072, 0x006f,
      EVR_UC, "SelectorUCValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0070, 0x0072, 0x0070,
      EVR_UT, "SelectorUTValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0071, 0x0072, 0x0071,
      EVR_UR, "SelectorURValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0072, 0x0072, 0x0072,
      EVR_DS, "SelectorDSValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0073, 0x0072, 0x0073,
      EVR_OD, "SelectorODValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0074, 0x0072, 0x0074,
      EVR_FD, "SelectorFDValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0075, 0x0072, 0x0075,
      EVR_OL, "SelectorOLValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0076, 0x0072, 0x0076,
      EVR_FL, "SelectorFLValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0078, 0x0072, 0x0078,
      EVR_UL, "SelectorULValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x007a, 0x0072, 0x007a,
      EVR_US, "SelectorUSValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x007c, 0x0072, 0x007c,
      EVR_SL, "SelectorSLValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x007e, 0x0072, 0x007e,
      EVR_SS, "SelectorSSValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x007f, 0x0072, 0x007f,
      EVR_UI, "SelectorUIValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0080, 0x0072, 0x0080,
      EVR_SQ, "SelectorCodeSequenceValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0081, 0x0072, 0x0081,
      EVR_OV, "SelectorOVValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0082, 0x0072, 0x0082,
      EVR_SV, "SelectorSVValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0083, 0x0072, 0x0083,
      EVR_UV, "SelectorUVValue", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0100, 0x0072, 0x0100,
      EVR_US, "NumberOfScreens", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0102, 0x0072, 0x0102,
      EVR_SQ, "NominalScreenDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0104, 0x0072, 0x0104,
      EVR_US, "NumberOfVerticalPixels", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0106, 0x0072, 0x0106,
      EVR_US, "NumberOfHorizontalPixels", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0108, 0x0072, 0x0108,
      EVR_FD, "DisplayEnvironmentSpatialPosition", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x010a, 0x0072, 0x010a,
      EVR_US, "ScreenMinimumGrayscaleBitDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x010c, 0x0072, 0x010c,
      EVR_US, "ScreenMinimumColorBitDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x010e, 0x0072, 0x010e,
      EVR_US, "ApplicationMaximumRepaintTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0200, 0x0072, 0x0200,
      EVR_SQ, "DisplaySetsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0202, 0x0072, 0x0202,
      EVR_US, "DisplaySetNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0203, 0x0072, 0x0203,
      EVR_LO, "DisplaySetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0204, 0x0072, 0x0204,
      EVR_US, "DisplaySetPresentationGroup", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0206, 0x0072, 0x0206,
      EVR_LO, "DisplaySetPresentationGroupDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0208, 0x0072, 0x0208,
      EVR_CS, "PartialDataDisplayHandling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0210, 0x0072, 0x0210,
      EVR_SQ, "SynchronizedScrollingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0212, 0x0072, 0x0212,
      EVR_US, "DisplaySetScrollingGroup", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0214, 0x0072, 0x0214,
      EVR_SQ, "NavigationIndicatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0216, 0x0072, 0x0216,
      EVR_US, "NavigationDisplaySet", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0218, 0x0072, 0x0218,
      EVR_US, "ReferenceDisplaySets", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0300, 0x0072, 0x0300,
      EVR_SQ, "ImageBoxesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0302, 0x0072, 0x0302,
      EVR_US, "ImageBoxNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0304, 0x0072, 0x0304,
      EVR_CS, "ImageBoxLayoutType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0306, 0x0072, 0x0306,
      EVR_US, "ImageBoxTileHorizontalDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0308, 0x0072, 0x0308,
      EVR_US, "ImageBoxTileVerticalDimension", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0310, 0x0072, 0x0310,
      EVR_CS, "ImageBoxScrollDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0312, 0x0072, 0x0312,
      EVR_CS, "ImageBoxSmallScrollType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0314, 0x0072, 0x0314,
      EVR_US, "ImageBoxSmallScrollAmount", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0316, 0x0072, 0x0316,
      EVR_CS, "ImageBoxLargeScrollType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0318, 0x0072, 0x0318,
      EVR_US, "ImageBoxLargeScrollAmount", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0320, 0x0072, 0x0320,
      EVR_US, "ImageBoxOverlapPriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0330, 0x0072, 0x0330,
      EVR_FD, "CineRelativeToRealTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0400, 0x0072, 0x0400,
      EVR_SQ, "FilterOperationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0402, 0x0072, 0x0402,
      EVR_CS, "FilterByCategory", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0404, 0x0072, 0x0404,
      EVR_CS, "FilterByAttributePresence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0406, 0x0072, 0x0406,
      EVR_CS, "FilterByOperator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0420, 0x0072, 0x0420,
      EVR_US, "StructuredDisplayBackgroundCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0421, 0x0072, 0x0421,
      EVR_US, "EmptyImageBoxCIELabValue", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0422, 0x0072, 0x0422,
      EVR_SQ, "StructuredDisplayImageBoxSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0424, 0x0072, 0x0424,
      EVR_SQ, "StructuredDisplayTextBoxSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0427, 0x0072, 0x0427,
      EVR_SQ, "ReferencedFirstFrameSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0430, 0x0072, 0x0430,
      EVR_SQ, "ImageBoxSynchronizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0432, 0x0072, 0x0432,
      EVR_US, "SynchronizedImageBoxList", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0434, 0x0072, 0x0434,
      EVR_CS, "TypeOfSynchronization", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0500, 0x0072, 0x0500,
      EVR_CS, "BlendingOperationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0510, 0x0072, 0x0510,
      EVR_CS, "ReformattingOperationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0512, 0x0072, 0x0512,
      EVR_FD, "ReformattingThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0514, 0x0072, 0x0514,
      EVR_FD, "ReformattingInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0516, 0x0072, 0x0516,
      EVR_CS, "ReformattingOperationInitialViewDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0520, 0x0072, 0x0520,
      EVR_CS, "ThreeDRenderingType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0600, 0x0072, 0x0600,
      EVR_SQ, "SortingOperationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0602, 0x0072, 0x0602,
      EVR_CS, "SortByCategory", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0604, 0x0072, 0x0604,
      EVR_CS, "SortingDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0700, 0x0072, 0x0700,
      EVR_CS, "DisplaySetPatientOrientation", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0702, 0x0072, 0x0702,
      EVR_CS, "VOIType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0704, 0x0072, 0x0704,
      EVR_CS, "PseudoColorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0705, 0x0072, 0x0705,
      EVR_SQ, "PseudoColorPaletteInstanceReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0706, 0x0072, 0x0706,
      EVR_CS, "ShowGrayscaleInverted", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0710, 0x0072, 0x0710,
      EVR_CS, "ShowImageTrueSizeFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0712, 0x0072, 0x0712,
      EVR_CS, "ShowGraphicAnnotationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0714, 0x0072, 0x0714,
      EVR_CS, "ShowPatientDemographicsFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0716, 0x0072, 0x0716,
      EVR_CS, "ShowAcquisitionTechniquesFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0717, 0x0072, 0x0717,
      EVR_CS, "DisplaySetHorizontalJustification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0072, 0x0718, 0x0072, 0x0718,
      EVR_CS, "DisplaySetVerticalJustification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x0120, 0x0074, 0x0120,
      EVR_FD, "ContinuationStartMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x0121, 0x0074, 0x0121,
      EVR_FD, "ContinuationEndMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1000, 0x0074, 0x1000,
      EVR_CS, "ProcedureStepState", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1002, 0x0074, 0x1002,
      EVR_SQ, "ProcedureStepProgressInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1004, 0x0074, 0x1004,
      EVR_DS, "ProcedureStepProgress", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1006, 0x0074, 0x1006,
      EVR_ST, "ProcedureStepProgressDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1007, 0x0074, 0x1007,
      EVR_SQ, "ProcedureStepProgressParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1008, 0x0074, 0x1008,
      EVR_SQ, "ProcedureStepCommunicationsURISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x100a, 0x0074, 0x100a,
      EVR_UR, "ContactURI", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x100c, 0x0074, 0x100c,
      EVR_LO, "ContactDisplayName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x100e, 0x0074, 0x100e,
      EVR_SQ, "ProcedureStepDiscontinuationReasonCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1020, 0x0074, 0x1020,
      EVR_SQ, "BeamTaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1022, 0x0074, 0x1022,
      EVR_CS, "BeamTaskType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1024, 0x0074, 0x1024,
      EVR_IS, "RETIRED_BeamOrderIndexTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1025, 0x0074, 0x1025,
      EVR_CS, "AutosequenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1026, 0x0074, 0x1026,
      EVR_FD, "TableTopVerticalAdjustedPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1027, 0x0074, 0x1027,
      EVR_FD, "TableTopLongitudinalAdjustedPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1028, 0x0074, 0x1028,
      EVR_FD, "TableTopLateralAdjustedPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x102a, 0x0074, 0x102a,
      EVR_FD, "PatientSupportAdjustedAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x102b, 0x0074, 0x102b,
      EVR_FD, "TableTopEccentricAdjustedAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x102c, 0x0074, 0x102c,
      EVR_FD, "TableTopPitchAdjustedAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x102d, 0x0074, 0x102d,
      EVR_FD, "TableTopRollAdjustedAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1030, 0x0074, 0x1030,
      EVR_SQ, "DeliveryVerificationImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1032, 0x0074, 0x1032,
      EVR_CS, "VerificationImageTiming", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1034, 0x0074, 0x1034,
      EVR_CS, "DoubleExposureFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1036, 0x0074, 0x1036,
      EVR_CS, "DoubleExposureOrdering", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1038, 0x0074, 0x1038,
      EVR_DS, "RETIRED_DoubleExposureMetersetTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x103a, 0x0074, 0x103a,
      EVR_DS, "RETIRED_DoubleExposureFieldDeltaTrial", 4, 4, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1040, 0x0074, 0x1040,
      EVR_SQ, "RelatedReferenceRTImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1042, 0x0074, 0x1042,
      EVR_SQ, "GeneralMachineVerificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1044, 0x0074, 0x1044,
      EVR_SQ, "ConventionalMachineVerificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1046, 0x0074, 0x1046,
      EVR_SQ, "IonMachineVerificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1048, 0x0074, 0x1048,
      EVR_SQ, "FailedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x104a, 0x0074, 0x104a,
      EVR_SQ, "OverriddenAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x104c, 0x0074, 0x104c,
      EVR_SQ, "ConventionalControlPointVerificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x104e, 0x0074, 0x104e,
      EVR_SQ, "IonControlPointVerificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1050, 0x0074, 0x1050,
      EVR_SQ, "AttributeOccurrenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1052, 0x0074, 0x1052,
      EVR_AT, "AttributeOccurrencePointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1054, 0x0074, 0x1054,
      EVR_UL, "AttributeItemSelector", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1056, 0x0074, 0x1056,
      EVR_LO, "AttributeOccurrencePrivateCreator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1057, 0x0074, 0x1057,
      EVR_IS, "SelectorSequencePointerItems", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1200, 0x0074, 0x1200,
      EVR_CS, "ScheduledProcedureStepPriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1202, 0x0074, 0x1202,
      EVR_LO, "WorklistLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1204, 0x0074, 0x1204,
      EVR_LO, "ProcedureStepLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1210, 0x0074, 0x1210,
      EVR_SQ, "ScheduledProcessingParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1212, 0x0074, 0x1212,
      EVR_SQ, "PerformedProcessingParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1216, 0x0074, 0x1216,
      EVR_SQ, "UnifiedProcedureStepPerformedProcedureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1220, 0x0074, 0x1220,
      EVR_SQ, "RETIRED_RelatedProcedureStepSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1222, 0x0074, 0x1222,
      EVR_LO, "RETIRED_ProcedureStepRelationshipType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1224, 0x0074, 0x1224,
      EVR_SQ, "ReplacedProcedureStepSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1230, 0x0074, 0x1230,
      EVR_LO, "DeletionLock", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1234, 0x0074, 0x1234,
      EVR_AE, "ReceivingAE", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1236, 0x0074, 0x1236,
      EVR_AE, "RequestingAE", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1238, 0x0074, 0x1238,
      EVR_LT, "ReasonForCancellation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1242, 0x0074, 0x1242,
      EVR_CS, "SCPStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1244, 0x0074, 0x1244,
      EVR_CS, "SubscriptionListStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1246, 0x0074, 0x1246,
      EVR_CS, "UnifiedProcedureStepListStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1324, 0x0074, 0x1324,
      EVR_UL, "BeamOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1338, 0x0074, 0x1338,
      EVR_FD, "DoubleExposureMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x133a, 0x0074, 0x133a,
      EVR_FD, "DoubleExposureFieldDelta", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1401, 0x0074, 0x1401,
      EVR_SQ, "BrachyTaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1402, 0x0074, 0x1402,
      EVR_DS, "ContinuationStartTotalReferenceAirKerma", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1403, 0x0074, 0x1403,
      EVR_DS, "ContinuationEndTotalReferenceAirKerma", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1404, 0x0074, 0x1404,
      EVR_IS, "ContinuationPulseNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1405, 0x0074, 0x1405,
      EVR_SQ, "ChannelDeliveryOrderSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1406, 0x0074, 0x1406,
      EVR_IS, "ReferencedChannelNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1407, 0x0074, 0x1407,
      EVR_DS, "StartCumulativeTimeWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1408, 0x0074, 0x1408,
      EVR_DS, "EndCumulativeTimeWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x1409, 0x0074, 0x1409,
      EVR_SQ, "OmittedChannelSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x140a, 0x0074, 0x140a,
      EVR_CS, "ReasonForChannelOmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x140b, 0x0074, 0x140b,
      EVR_LO, "ReasonForChannelOmissionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x140c, 0x0074, 0x140c,
      EVR_IS, "ChannelDeliveryOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x140d, 0x0074, 0x140d,
      EVR_SQ, "ChannelDeliveryContinuationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0074, 0x140e, 0x0074, 0x140e,
      EVR_SQ, "OmittedApplicationSetupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0001, 0x0076, 0x0001,
      EVR_LO, "ImplantAssemblyTemplateName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0003, 0x0076, 0x0003,
      EVR_LO, "ImplantAssemblyTemplateIssuer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0006, 0x0076, 0x0006,
      EVR_LO, "ImplantAssemblyTemplateVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0008, 0x0076, 0x0008,
      EVR_SQ, "ReplacedImplantAssemblyTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x000a, 0x0076, 0x000a,
      EVR_CS, "ImplantAssemblyTemplateType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x000c, 0x0076, 0x000c,
      EVR_SQ, "OriginalImplantAssemblyTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x000e, 0x0076, 0x000e,
      EVR_SQ, "DerivationImplantAssemblyTemplateSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0010, 0x0076, 0x0010,
      EVR_SQ, "ImplantAssemblyTemplateTargetAnatomySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0020, 0x0076, 0x0020,
      EVR_SQ, "ProcedureTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0030, 0x0076, 0x0030,
      EVR_LO, "SurgicalTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0032, 0x0076, 0x0032,
      EVR_SQ, "ComponentTypesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0034, 0x0076, 0x0034,
      EVR_SQ, "ComponentTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0036, 0x0076, 0x0036,
      EVR_CS, "ExclusiveComponentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0038, 0x0076, 0x0038,
      EVR_CS, "MandatoryComponentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0040, 0x0076, 0x0040,
      EVR_SQ, "ComponentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0055, 0x0076, 0x0055,
      EVR_US, "ComponentID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0060, 0x0076, 0x0060,
      EVR_SQ, "ComponentAssemblySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0070, 0x0076, 0x0070,
      EVR_US, "Component1ReferencedID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0080, 0x0076, 0x0080,
      EVR_US, "Component1ReferencedMatingFeatureSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x0090, 0x0076, 0x0090,
      EVR_US, "Component1ReferencedMatingFeatureID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x00a0, 0x0076, 0x00a0,
      EVR_US, "Component2ReferencedID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x00b0, 0x0076, 0x00b0,
      EVR_US, "Component2ReferencedMatingFeatureSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0076, 0x00c0, 0x0076, 0x00c0,
      EVR_US, "Component2ReferencedMatingFeatureID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0001, 0x0078, 0x0001,
      EVR_LO, "ImplantTemplateGroupName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0010, 0x0078, 0x0010,
      EVR_ST, "ImplantTemplateGroupDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0020, 0x0078, 0x0020,
      EVR_LO, "ImplantTemplateGroupIssuer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0024, 0x0078, 0x0024,
      EVR_LO, "ImplantTemplateGroupVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0026, 0x0078, 0x0026,
      EVR_SQ, "ReplacedImplantTemplateGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0028, 0x0078, 0x0028,
      EVR_SQ, "ImplantTemplateGroupTargetAnatomySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x002a, 0x0078, 0x002a,
      EVR_SQ, "ImplantTemplateGroupMembersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x002e, 0x0078, 0x002e,
      EVR_US, "ImplantTemplateGroupMemberID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0050, 0x0078, 0x0050,
      EVR_FD, "ThreeDImplantTemplateGroupMemberMatchingPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0060, 0x0078, 0x0060,
      EVR_FD, "ThreeDImplantTemplateGroupMemberMatchingAxes", 9, 9, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0070, 0x0078, 0x0070,
      EVR_SQ, "ImplantTemplateGroupMemberMatching2DCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x0090, 0x0078, 0x0090,
      EVR_FD, "TwoDImplantTemplateGroupMemberMatchingPoint", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00a0, 0x0078, 0x00a0,
      EVR_FD, "TwoDImplantTemplateGroupMemberMatchingAxes", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00b0, 0x0078, 0x00b0,
      EVR_SQ, "ImplantTemplateGroupVariationDimensionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00b2, 0x0078, 0x00b2,
      EVR_LO, "ImplantTemplateGroupVariationDimensionName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00b4, 0x0078, 0x00b4,
      EVR_SQ, "ImplantTemplateGroupVariationDimensionRankSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00b6, 0x0078, 0x00b6,
      EVR_US, "ReferencedImplantTemplateGroupMemberID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0078, 0x00b8, 0x0078, 0x00b8,
      EVR_US, "ImplantTemplateGroupVariationDimensionRank", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0001, 0x0080, 0x0001,
      EVR_SQ, "SurfaceScanAcquisitionTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0002, 0x0080, 0x0002,
      EVR_SQ, "SurfaceScanModeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0003, 0x0080, 0x0003,
      EVR_SQ, "RegistrationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0004, 0x0080, 0x0004,
      EVR_FD, "ShotDurationTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0005, 0x0080, 0x0005,
      EVR_FD, "ShotOffsetTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0006, 0x0080, 0x0006,
      EVR_US, "SurfacePointPresentationValueData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0007, 0x0080, 0x0007,
      EVR_US, "SurfacePointColorCIELabValueData", 3, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0008, 0x0080, 0x0008,
      EVR_SQ, "UVMappingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0009, 0x0080, 0x0009,
      EVR_SH, "TextureLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0010, 0x0080, 0x0010,
      EVR_OF, "UValueData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0011, 0x0080, 0x0011,
      EVR_OF, "VValueData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0012, 0x0080, 0x0012,
      EVR_SQ, "ReferencedTextureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0080, 0x0013, 0x0080, 0x0013,
      EVR_SQ, "ReferencedSurfaceDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0001, 0x0082, 0x0001,
      EVR_CS, "AssessmentSummary", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0003, 0x0082, 0x0003,
      EVR_UT, "AssessmentSummaryDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0004, 0x0082, 0x0004,
      EVR_SQ, "AssessedSOPInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0005, 0x0082, 0x0005,
      EVR_SQ, "ReferencedComparisonSOPInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0006, 0x0082, 0x0006,
      EVR_UL, "NumberOfAssessmentObservations", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0007, 0x0082, 0x0007,
      EVR_SQ, "AssessmentObservationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0008, 0x0082, 0x0008,
      EVR_CS, "ObservationSignificance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x000a, 0x0082, 0x000a,
      EVR_UT, "ObservationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x000c, 0x0082, 0x000c,
      EVR_SQ, "StructuredConstraintObservationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0010, 0x0082, 0x0010,
      EVR_SQ, "AssessedAttributeValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0016, 0x0082, 0x0016,
      EVR_LO, "AssessmentSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0017, 0x0082, 0x0017,
      EVR_SQ, "AssessmentRequesterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0018, 0x0082, 0x0018,
      EVR_LO, "SelectorAttributeName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0019, 0x0082, 0x0019,
      EVR_LO, "SelectorAttributeKeyword", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0021, 0x0082, 0x0021,
      EVR_SQ, "AssessmentTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0022, 0x0082, 0x0022,
      EVR_SQ, "ObservationBasisCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0023, 0x0082, 0x0023,
      EVR_LO, "AssessmentLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0032, 0x0082, 0x0032,
      EVR_CS, "ConstraintType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0033, 0x0082, 0x0033,
      EVR_UT, "SpecificationSelectionGuidance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0034, 0x0082, 0x0034,
      EVR_SQ, "ConstraintValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0035, 0x0082, 0x0035,
      EVR_SQ, "RecommendedDefaultValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0036, 0x0082, 0x0036,
      EVR_CS, "ConstraintViolationSignificance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0037, 0x0082, 0x0037,
      EVR_UT, "ConstraintViolationCondition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0082, 0x0038, 0x0082, 0x0038,
      EVR_CS, "ModifiableConstraintFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0087, 0x0010, 0x0087, 0x0010,
      EVR_CS, "MediaType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113708.794.1.1.2.0" }
  , { 0x0087, 0x0020, 0x0087, 0x0020,
      EVR_CS, "MediaLocation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113708.794.1.1.2.0" }
  , { 0x0087, 0x0050, 0x0087, 0x0050,
      EVR_IS, "EstimatedRetrieveTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "1.2.840.113708.794.1.1.2.0" }
#endif
  , { 0x0088, 0x0130, 0x0088, 0x0130,
      EVR_SH, "StorageMediaFileSetID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0140, 0x0088, 0x0140,
      EVR_UI, "StorageMediaFileSetUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0200, 0x0088, 0x0200,
      EVR_SQ, "IconImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0904, 0x0088, 0x0904,
      EVR_LO, "RETIRED_TopicTitle", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0906, 0x0088, 0x0906,
      EVR_ST, "RETIRED_TopicSubject", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0910, 0x0088, 0x0910,
      EVR_LO, "RETIRED_TopicAuthor", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0088, 0x0912, 0x0088, 0x0912,
      EVR_LO, "RETIRED_TopicKeywords", 1, 32, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0089, 0x0010, 0x0089, 0x0010,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "DIDI TO PCR 1.1" }
  , { 0x0095, 0x0001, 0x0095, 0x0001,
      EVR_LT, "ExaminationFolderID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0095, 0x0004, 0x0095, 0x0004,
      EVR_UL, "FolderReportedStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0095, 0x0005, 0x0095, 0x0005,
      EVR_LT, "FolderReportingRadiologist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0095, 0x0007, 0x0095, 0x0007,
      EVR_LT, "SIENETISAPLA", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x0099, 0x0002, 0x0099, 0x0002,
      EVR_UL, "DataObjectAttributes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIENET" }
  , { 0x00e1, 0x0001, 0x00e1, 0x0001,
      EVR_US, "DataDictionaryVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0014, 0x00e1, 0x0014,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0022, 0x00e1, 0x0022,
      EVR_DS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0023, 0x00e1, 0x0023,
      EVR_DS, "Unknown", 2, 2, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0024, 0x00e1, 0x0024,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0025, 0x00e1, 0x0025,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x00e1, 0x0040, 0x00e1, 0x0040,
      EVR_SH, "OffsetFromCTMRImages", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
#endif
  , { 0x0100, 0x0410, 0x0100, 0x0410,
      EVR_CS, "SOPInstanceStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0100, 0x0420, 0x0100, 0x0420,
      EVR_DT, "SOPAuthorizationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0100, 0x0424, 0x0100, 0x0424,
      EVR_LT, "SOPAuthorizationComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0100, 0x0426, 0x0100, 0x0426,
      EVR_LO, "AuthorizationEquipmentCertificationNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0135, 0x0010, 0x0135, 0x0010,
      EVR_LO, "UltrasoundScannerName", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SONOWAND AS" }
  , { 0x0135, 0x0011, 0x0135, 0x0011,
      EVR_LO, "TransducerSerial", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SONOWAND AS" }
  , { 0x0135, 0x0012, 0x0135, 0x0012,
      EVR_LO, "ProbeApplication", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SONOWAND AS" }
  , { 0x0193, 0x0002, 0x0193, 0x0002,
      EVR_DS, "RISKey", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0307, 0x0001, 0x0307, 0x0001,
      EVR_UN, "RISWorklistIMGEF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x0309, 0x0001, 0x0309, 0x0001,
      EVR_UN, "RISReportIMGEF", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x0400, 0x0005, 0x0400, 0x0005,
      EVR_US, "MACIDNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0010, 0x0400, 0x0010,
      EVR_UI, "MACCalculationTransferSyntaxUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0015, 0x0400, 0x0015,
      EVR_CS, "MACAlgorithm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0020, 0x0400, 0x0020,
      EVR_AT, "DataElementsSigned", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0100, 0x0400, 0x0100,
      EVR_UI, "DigitalSignatureUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0105, 0x0400, 0x0105,
      EVR_DT, "DigitalSignatureDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0110, 0x0400, 0x0110,
      EVR_CS, "CertificateType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0115, 0x0400, 0x0115,
      EVR_OB, "CertificateOfSigner", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0120, 0x0400, 0x0120,
      EVR_OB, "Signature", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0305, 0x0400, 0x0305,
      EVR_CS, "CertifiedTimestampType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0310, 0x0400, 0x0310,
      EVR_OB, "CertifiedTimestamp", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0401, 0x0400, 0x0401,
      EVR_SQ, "DigitalSignaturePurposeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0402, 0x0400, 0x0402,
      EVR_SQ, "ReferencedDigitalSignatureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0403, 0x0400, 0x0403,
      EVR_SQ, "ReferencedSOPInstanceMACSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0404, 0x0400, 0x0404,
      EVR_OB, "MAC", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0500, 0x0400, 0x0500,
      EVR_SQ, "EncryptedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0510, 0x0400, 0x0510,
      EVR_UI, "EncryptedContentTransferSyntaxUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0520, 0x0400, 0x0520,
      EVR_OB, "EncryptedContent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0550, 0x0400, 0x0550,
      EVR_SQ, "ModifiedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0551, 0x0400, 0x0551,
      EVR_SQ, "NonconformingModifiedAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0552, 0x0400, 0x0552,
      EVR_OB, "NonconformingDataElementValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0561, 0x0400, 0x0561,
      EVR_SQ, "OriginalAttributesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0562, 0x0400, 0x0562,
      EVR_DT, "AttributeModificationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0563, 0x0400, 0x0563,
      EVR_LO, "ModifyingSystem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0564, 0x0400, 0x0564,
      EVR_LO, "SourceOfPreviousValues", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0565, 0x0400, 0x0565,
      EVR_CS, "ReasonForTheAttributeModification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x0400, 0x0600, 0x0400, 0x0600,
      EVR_CS, "InstanceOriginStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x0601, 0x0000, 0x0601, 0x0000,
      EVR_SH, "ImplementationVersion", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0020, 0x0601, 0x0020,
      EVR_DS, "RelativeTablePosition", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0021, 0x0601, 0x0021,
      EVR_DS, "RelativeTableHeight", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0030, 0x0601, 0x0030,
      EVR_SH, "SurviewDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0031, 0x0601, 0x0031,
      EVR_DS, "SurviewLength", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0050, 0x0601, 0x0050,
      EVR_SH, "ImageViewType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0070, 0x0601, 0x0070,
      EVR_DS, "BatchNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0071, 0x0601, 0x0071,
      EVR_DS, "BatchSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
  , { 0x0601, 0x0072, 0x0601, 0x0072,
      EVR_DS, "BatchSliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "ELSCINT1" }
#endif
  , { 0x1000, 0x0010, 0x1000, 0x0010,
      EVR_US, "RETIRED_EscapeTriplet", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1000, 0x0011, 0x1000, 0x0011,
      EVR_US, "RETIRED_RunLengthTriplet", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1000, 0x0012, 0x1000, 0x0012,
      EVR_US, "RETIRED_HuffmanTableSize", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1000, 0x0013, 0x1000, 0x0013,
      EVR_US, "RETIRED_HuffmanTableTriplet", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1000, 0x0014, 0x1000, 0x0014,
      EVR_US, "RETIRED_ShiftTableSize", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1000, 0x0015, 0x1000, 0x0015,
      EVR_US, "RETIRED_ShiftTableTriplet", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x1010, 0x0004, 0x1010, 0x0004,
      EVR_US, "RETIRED_ZonalMap", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x1369, 0x0000, 0x1369, 0x0000,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "AEGIS_DICOM_2.00" }
#endif
  , { 0x2000, 0x0010, 0x2000, 0x0010,
      EVR_IS, "NumberOfCopies", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x001e, 0x2000, 0x001e,
      EVR_SQ, "PrinterConfigurationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0020, 0x2000, 0x0020,
      EVR_CS, "PrintPriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0030, 0x2000, 0x0030,
      EVR_CS, "MediumType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0040, 0x2000, 0x0040,
      EVR_CS, "FilmDestination", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0050, 0x2000, 0x0050,
      EVR_LO, "FilmSessionLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0060, 0x2000, 0x0060,
      EVR_IS, "MemoryAllocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0061, 0x2000, 0x0061,
      EVR_IS, "MaximumMemoryAllocation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0062, 0x2000, 0x0062,
      EVR_CS, "RETIRED_ColorImagePrintingFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0063, 0x2000, 0x0063,
      EVR_CS, "RETIRED_CollationFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0065, 0x2000, 0x0065,
      EVR_CS, "RETIRED_AnnotationFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0067, 0x2000, 0x0067,
      EVR_CS, "RETIRED_ImageOverlayFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0069, 0x2000, 0x0069,
      EVR_CS, "RETIRED_PresentationLUTFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x006a, 0x2000, 0x006a,
      EVR_CS, "RETIRED_ImageBoxPresentationLUTFlag", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x00a0, 0x2000, 0x00a0,
      EVR_US, "MemoryBitDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x00a1, 0x2000, 0x00a1,
      EVR_US, "PrintingBitDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x00a2, 0x2000, 0x00a2,
      EVR_SQ, "MediaInstalledSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x00a4, 0x2000, 0x00a4,
      EVR_SQ, "OtherMediaAvailableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x00a8, 0x2000, 0x00a8,
      EVR_SQ, "SupportedImageDisplayFormatsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0500, 0x2000, 0x0500,
      EVR_SQ, "ReferencedFilmBoxSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2000, 0x0510, 0x2000, 0x0510,
      EVR_SQ, "RETIRED_ReferencedStoredPrintSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x2001, 0x0001, 0x2001, 0x0001,
      EVR_FL, "ChemicalShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0001, 0x2001, 0x0001,
      EVR_FL, "ChemicalShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0002, 0x2001, 0x0002,
      EVR_IS, "ChemicalShiftNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0002, 0x2001, 0x0002,
      EVR_IS, "ChemicalShiftNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0003, 0x2001, 0x0003,
      EVR_FL, "DiffusionBFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0003, 0x2001, 0x0003,
      EVR_FL, "DiffusionBFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0004, 0x2001, 0x0004,
      EVR_CS, "DiffusionDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0004, 0x2001, 0x0004,
      EVR_CS, "DiffusionDirection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0006, 0x2001, 0x0006,
      EVR_CS, "ImageEnhanced", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0006, 0x2001, 0x0006,
      EVR_CS, "ImageEnhanced", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0007, 0x2001, 0x0007,
      EVR_CS, "ImageTypeEDES", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0007, 0x2001, 0x0007,
      EVR_CS, "ImageTypeEDES", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0008, 0x2001, 0x0008,
      EVR_IS, "PhaseNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0008, 0x2001, 0x0008,
      EVR_IS, "PhaseNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0009, 0x2001, 0x0009,
      EVR_FL, "ImagePrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0009, 0x2001, 0x0009,
      EVR_FL, "ImagePrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x000a, 0x2001, 0x000a,
      EVR_IS, "SliceNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x000a, 0x2001, 0x000a,
      EVR_IS, "SliceNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x000b, 0x2001, 0x000b,
      EVR_CS, "SliceOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x000b, 0x2001, 0x000b,
      EVR_CS, "SliceOrientation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x000c, 0x2001, 0x000c,
      EVR_CS, "ArrhythmiaRejection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x000c, 0x2001, 0x000c,
      EVR_CS, "ArrhythmiaRejection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x000e, 0x2001, 0x000e,
      EVR_CS, "CardiacCycled", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x000e, 0x2001, 0x000e,
      EVR_CS, "CardiacCycled", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x000f, 0x2001, 0x000f,
      EVR_SS, "CardiacGateWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x000f, 0x2001, 0x000f,
      EVR_SS, "CardiacGateWidth", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0010, 0x2001, 0x0010,
      EVR_CS, "CardiacSync", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0010, 0x2001, 0x0010,
      EVR_CS, "CardiacSync", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0011, 0x2001, 0x0011,
      EVR_FL, "DiffusionEchoTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0011, 0x2001, 0x0011,
      EVR_FL, "DiffusionEchoTime", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0012, 0x2001, 0x0012,
      EVR_CS, "DynamicSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0012, 0x2001, 0x0012,
      EVR_CS, "DynamicSeries", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0013, 0x2001, 0x0013,
      EVR_SL, "EPIFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0013, 0x2001, 0x0013,
      EVR_SL, "EPIFactor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0014, 0x2001, 0x0014,
      EVR_SL, "NumberOfEchoes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0014, 0x2001, 0x0014,
      EVR_SL, "NumberOfEchoes", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0015, 0x2001, 0x0015,
      EVR_SS, "NumberOfLocations", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0015, 0x2001, 0x0015,
      EVR_SS, "NumberOfLocations", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0016, 0x2001, 0x0016,
      EVR_SS, "NumberOfPCDirections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0016, 0x2001, 0x0016,
      EVR_SS, "NumberOfPCDirections", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0017, 0x2001, 0x0017,
      EVR_SL, "NumberOfPhasesMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0017, 0x2001, 0x0017,
      EVR_SL, "NumberOfPhasesMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0018, 0x2001, 0x0018,
      EVR_SL, "NumberOfSlicesMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0018, 0x2001, 0x0018,
      EVR_SL, "NumberOfSlicesMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0019, 0x2001, 0x0019,
      EVR_CS, "PartialMatrixScanned", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0019, 0x2001, 0x0019,
      EVR_CS, "PartialMatrixScanned", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x001a, 0x2001, 0x001a,
      EVR_FL, "PCVelocity", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x001a, 0x2001, 0x001a,
      EVR_FL, "PCVelocity", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x001b, 0x2001, 0x001b,
      EVR_FL, "PrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x001b, 0x2001, 0x001b,
      EVR_FL, "PrepulseDelay", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x001c, 0x2001, 0x001c,
      EVR_CS, "PrepulseType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x001c, 0x2001, 0x001c,
      EVR_CS, "PrepulseType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x001d, 0x2001, 0x001d,
      EVR_IS, "ReconstructionNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x001d, 0x2001, 0x001d,
      EVR_IS, "ReconstructionNumberMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x001f, 0x2001, 0x001f,
      EVR_CS, "RespirationSync", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x001f, 0x2001, 0x001f,
      EVR_CS, "RespirationSync", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0020, 0x2001, 0x0020,
      EVR_LO, "ScanningTechnique", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0020, 0x2001, 0x0020,
      EVR_LO, "ScanningTechnique", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0021, 0x2001, 0x0021,
      EVR_CS, "SPIR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0021, 0x2001, 0x0021,
      EVR_CS, "SPIR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0022, 0x2001, 0x0022,
      EVR_FL, "WaterFatShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0022, 0x2001, 0x0022,
      EVR_FL, "WaterFatShift", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0023, 0x2001, 0x0023,
      EVR_DS, "FlipAnglePhilips", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0023, 0x2001, 0x0023,
      EVR_DS, "FlipAnglePhilips", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0024, 0x2001, 0x0024,
      EVR_CS, "SeriesIsInteractive", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0024, 0x2001, 0x0024,
      EVR_CS, "SeriesIsInteractive", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0025, 0x2001, 0x0025,
      EVR_SH, "EchoTimeDisplayMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0025, 0x2001, 0x0025,
      EVR_SH, "EchoTimeDisplayMR", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0026, 0x2001, 0x0026,
      EVR_CS, "PresentationStateSubtractionActive", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0026, 0x2001, 0x0026,
      EVR_CS, "PresentationStateSubtractionActive", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x002d, 0x2001, 0x002d,
      EVR_SS, "StackNumberOfSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x002d, 0x2001, 0x002d,
      EVR_SS, "StackNumberOfSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0032, 0x2001, 0x0032,
      EVR_FL, "StackRadialAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0032, 0x2001, 0x0032,
      EVR_FL, "StackRadialAngle", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0033, 0x2001, 0x0033,
      EVR_CS, "StackRadialAxis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0033, 0x2001, 0x0033,
      EVR_CS, "StackRadialAxis", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0035, 0x2001, 0x0035,
      EVR_SS, "StackSliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0035, 0x2001, 0x0035,
      EVR_SS, "StackSliceNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0036, 0x2001, 0x0036,
      EVR_CS, "StackType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0036, 0x2001, 0x0036,
      EVR_CS, "StackType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x003f, 0x2001, 0x003f,
      EVR_CS, "ZoomMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x003f, 0x2001, 0x003f,
      EVR_CS, "ZoomMode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0058, 0x2001, 0x0058,
      EVR_UL, "ContrastTransferTaste", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0058, 0x2001, 0x0058,
      EVR_UL, "ContrastTransferTaste", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x005f, 0x2001, 0x005f,
      EVR_SQ, "StackSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x005f, 0x2001, 0x005f,
      EVR_SQ, "StackSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0060, 0x2001, 0x0060,
      EVR_SL, "NumberOfStacks", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0060, 0x2001, 0x0060,
      EVR_SL, "NumberOfStacks", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0061, 0x2001, 0x0061,
      EVR_CS, "SeriesTransmitted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0061, 0x2001, 0x0061,
      EVR_CS, "SeriesTransmitted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0062, 0x2001, 0x0062,
      EVR_CS, "SeriesCommitted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0062, 0x2001, 0x0062,
      EVR_CS, "SeriesCommitted", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0063, 0x2001, 0x0063,
      EVR_CS, "ExaminationSource", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0063, 0x2001, 0x0063,
      EVR_CS, "ExaminationSource", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0067, 0x2001, 0x0067,
      EVR_CS, "LinearPresentationGLTrafoShapeSub", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0067, 0x2001, 0x0067,
      EVR_CS, "LinearPresentationGLTrafoShapeSub", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0077, 0x2001, 0x0077,
      EVR_CS, "GLTrafoType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0077, 0x2001, 0x0077,
      EVR_CS, "GLTrafoType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x007b, 0x2001, 0x007b,
      EVR_IS, "AcquisitionNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x007b, 0x2001, 0x007b,
      EVR_IS, "AcquisitionNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x0081, 0x2001, 0x0081,
      EVR_IS, "NumberOfDynamicScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x0081, 0x2001, 0x0081,
      EVR_IS, "NumberOfDynamicScans", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x009f, 0x2001, 0x009f,
      EVR_US, "PixelProcessingKernelSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x009f, 0x2001, 0x009f,
      EVR_US, "PixelProcessingKernelSize", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x00a1, 0x2001, 0x00a1,
      EVR_CS, "IsRawImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x00a1, 0x2001, 0x00a1,
      EVR_CS, "IsRawImage", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x00f1, 0x2001, 0x00f1,
      EVR_FL, "ProspectiveMotionCorrection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x00f1, 0x2001, 0x00f1,
      EVR_FL, "ProspectiveMotionCorrection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2001, 0x00f2, 0x2001, 0x00f2,
      EVR_FL, "RetrospectiveMotionCorrection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips Imaging DD 001" }
  , { 0x2001, 0x00f2, 0x2001, 0x00f2,
      EVR_FL, "RetrospectiveMotionCorrection", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS IMAGING DD 001" }
  , { 0x2005, 0x0002, 0x2005, 0x0002,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 005" }
  , { 0x2005, 0x0005, 0x2005, 0x0005,
      EVR_CS, "SynergyReconstructionType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x0005, 0x2005, 0x0005,
      EVR_CS, "SynergyReconstructionType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x001e, 0x2005, 0x001e,
      EVR_SH, "MIPProtocol", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x001e, 0x2005, 0x001e,
      EVR_SH, "MIPProtocol", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x001f, 0x2005, 0x001f,
      EVR_SH, "MPRProtocol", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x001f, 0x2005, 0x001f,
      EVR_SH, "MPRProtocol", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x0020, 0x2005, 0x0020,
      EVR_SL, "NumberOfChemicalShifts", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x0020, 0x2005, 0x0020,
      EVR_SL, "NumberOfChemicalShifts", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x002d, 0x2005, 0x002d,
      EVR_SS, "NumberOfStackSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x002d, 0x2005, 0x002d,
      EVR_SS, "NumberOfStackSlices", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x0083, 0x2005, 0x0083,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x0083, 0x2005, 0x0083,
      EVR_SQ, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x00a1, 0x2005, 0x00a1,
      EVR_CS, "SyncraScanType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x00a1, 0x2005, 0x00a1,
      EVR_CS, "SyncraScanType", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x00b0, 0x2005, 0x00b0,
      EVR_FL, "DiffusionDirectionRL", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x00b0, 0x2005, 0x00b0,
      EVR_FL, "DiffusionDirectionRL", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x00b1, 0x2005, 0x00b1,
      EVR_FL, "DiffusionDirectionAP", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x00b1, 0x2005, 0x00b1,
      EVR_FL, "DiffusionDirectionAP", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
  , { 0x2005, 0x00b2, 0x2005, 0x00b2,
      EVR_FL, "DiffusionDirectionFH", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Philips MR Imaging DD 001" }
  , { 0x2005, 0x00b2, 0x2005, 0x00b2,
      EVR_FL, "DiffusionDirectionFH", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR IMAGING DD 001" }
#endif
  , { 0x2010, 0x0010, 0x2010, 0x0010,
      EVR_ST, "ImageDisplayFormat", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0030, 0x2010, 0x0030,
      EVR_CS, "AnnotationDisplayFormatID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0040, 0x2010, 0x0040,
      EVR_CS, "FilmOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0050, 0x2010, 0x0050,
      EVR_CS, "FilmSizeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0052, 0x2010, 0x0052,
      EVR_CS, "PrinterResolutionID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0054, 0x2010, 0x0054,
      EVR_CS, "DefaultPrinterResolutionID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0060, 0x2010, 0x0060,
      EVR_CS, "MagnificationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0080, 0x2010, 0x0080,
      EVR_CS, "SmoothingType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x00a6, 0x2010, 0x00a6,
      EVR_CS, "DefaultMagnificationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x00a7, 0x2010, 0x00a7,
      EVR_CS, "OtherMagnificationTypesAvailable", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x00a8, 0x2010, 0x00a8,
      EVR_CS, "DefaultSmoothingType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x00a9, 0x2010, 0x00a9,
      EVR_CS, "OtherSmoothingTypesAvailable", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0100, 0x2010, 0x0100,
      EVR_CS, "BorderDensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0110, 0x2010, 0x0110,
      EVR_CS, "EmptyImageDensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0120, 0x2010, 0x0120,
      EVR_US, "MinDensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0130, 0x2010, 0x0130,
      EVR_US, "MaxDensity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0140, 0x2010, 0x0140,
      EVR_CS, "Trim", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0150, 0x2010, 0x0150,
      EVR_ST, "ConfigurationInformation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0152, 0x2010, 0x0152,
      EVR_LT, "ConfigurationInformationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0154, 0x2010, 0x0154,
      EVR_IS, "MaximumCollatedFilms", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x015e, 0x2010, 0x015e,
      EVR_US, "Illumination", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0160, 0x2010, 0x0160,
      EVR_US, "ReflectedAmbientLight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0376, 0x2010, 0x0376,
      EVR_DS, "PrinterPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0500, 0x2010, 0x0500,
      EVR_SQ, "ReferencedFilmSessionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0510, 0x2010, 0x0510,
      EVR_SQ, "ReferencedImageBoxSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2010, 0x0520, 0x2010, 0x0520,
      EVR_SQ, "ReferencedBasicAnnotationBoxSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x2011, 0x0011, 0x2011, 0x0011,
      EVR_CS, "ImagePositionSpecifyingFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
#endif
  , { 0x2020, 0x0010, 0x2020, 0x0010,
      EVR_US, "ImageBoxPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0020, 0x2020, 0x0020,
      EVR_CS, "Polarity", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0030, 0x2020, 0x0030,
      EVR_DS, "RequestedImageSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0040, 0x2020, 0x0040,
      EVR_CS, "RequestedDecimateCropBehavior", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0050, 0x2020, 0x0050,
      EVR_CS, "RequestedResolutionID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x00a0, 0x2020, 0x00a0,
      EVR_CS, "RequestedImageSizeFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x00a2, 0x2020, 0x00a2,
      EVR_CS, "DecimateCropResult", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0110, 0x2020, 0x0110,
      EVR_SQ, "BasicGrayscaleImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0111, 0x2020, 0x0111,
      EVR_SQ, "BasicColorImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0130, 0x2020, 0x0130,
      EVR_SQ, "RETIRED_ReferencedImageOverlayBoxSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2020, 0x0140, 0x2020, 0x0140,
      EVR_SQ, "RETIRED_ReferencedVOILUTBoxSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2030, 0x0010, 0x2030, 0x0010,
      EVR_US, "AnnotationPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2030, 0x0020, 0x2030, 0x0020,
      EVR_LO, "TextString", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0010, 0x2040, 0x0010,
      EVR_SQ, "RETIRED_ReferencedOverlayPlaneSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0011, 0x2040, 0x0011,
      EVR_US, "RETIRED_ReferencedOverlayPlaneGroups", 1, 99, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0020, 0x2040, 0x0020,
      EVR_SQ, "RETIRED_OverlayPixelDataSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0060, 0x2040, 0x0060,
      EVR_CS, "RETIRED_OverlayMagnificationType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0070, 0x2040, 0x0070,
      EVR_CS, "RETIRED_OverlaySmoothingType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0072, 0x2040, 0x0072,
      EVR_CS, "RETIRED_OverlayOrImageMagnification", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0074, 0x2040, 0x0074,
      EVR_US, "RETIRED_MagnifyToNumberOfColumns", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0080, 0x2040, 0x0080,
      EVR_CS, "RETIRED_OverlayForegroundDensity", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0082, 0x2040, 0x0082,
      EVR_CS, "RETIRED_OverlayBackgroundDensity", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0090, 0x2040, 0x0090,
      EVR_CS, "RETIRED_OverlayMode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0100, 0x2040, 0x0100,
      EVR_CS, "RETIRED_ThresholdDensity", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2040, 0x0500, 0x2040, 0x0500,
      EVR_SQ, "RETIRED_ReferencedImageBoxSequenceRetired", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2050, 0x0010, 0x2050, 0x0010,
      EVR_SQ, "PresentationLUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2050, 0x0020, 0x2050, 0x0020,
      EVR_CS, "PresentationLUTShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2050, 0x0500, 0x2050, 0x0500,
      EVR_SQ, "ReferencedPresentationLUTSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0010, 0x2100, 0x0010,
      EVR_SH, "RETIRED_PrintJobID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0020, 0x2100, 0x0020,
      EVR_CS, "ExecutionStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0030, 0x2100, 0x0030,
      EVR_CS, "ExecutionStatusInfo", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0040, 0x2100, 0x0040,
      EVR_DA, "CreationDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0050, 0x2100, 0x0050,
      EVR_TM, "CreationTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0070, 0x2100, 0x0070,
      EVR_AE, "Originator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0140, 0x2100, 0x0140,
      EVR_AE, "DestinationAE", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0160, 0x2100, 0x0160,
      EVR_SH, "OwnerID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0170, 0x2100, 0x0170,
      EVR_IS, "NumberOfFilms", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2100, 0x0500, 0x2100, 0x0500,
      EVR_SQ, "RETIRED_ReferencedPrintJobSequencePullStoredPrint", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2110, 0x0010, 0x2110, 0x0010,
      EVR_CS, "PrinterStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2110, 0x0020, 0x2110, 0x0020,
      EVR_CS, "PrinterStatusInfo", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2110, 0x0030, 0x2110, 0x0030,
      EVR_LO, "PrinterName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2110, 0x0099, 0x2110, 0x0099,
      EVR_SH, "RETIRED_PrintQueueID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2120, 0x0010, 0x2120, 0x0010,
      EVR_CS, "RETIRED_QueueStatus", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2120, 0x0050, 0x2120, 0x0050,
      EVR_SQ, "RETIRED_PrintJobDescriptionSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2120, 0x0070, 0x2120, 0x0070,
      EVR_SQ, "RETIRED_ReferencedPrintJobSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0010, 0x2130, 0x0010,
      EVR_SQ, "RETIRED_PrintManagementCapabilitiesSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0015, 0x2130, 0x0015,
      EVR_SQ, "RETIRED_PrinterCharacteristicsSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0030, 0x2130, 0x0030,
      EVR_SQ, "RETIRED_FilmBoxContentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0040, 0x2130, 0x0040,
      EVR_SQ, "RETIRED_ImageBoxContentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0050, 0x2130, 0x0050,
      EVR_SQ, "RETIRED_AnnotationContentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0060, 0x2130, 0x0060,
      EVR_SQ, "RETIRED_ImageOverlayBoxContentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x0080, 0x2130, 0x0080,
      EVR_SQ, "RETIRED_PresentationLUTContentSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x00a0, 0x2130, 0x00a0,
      EVR_SQ, "ProposedStudySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2130, 0x00c0, 0x2130, 0x00c0,
      EVR_SQ, "OriginalImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0001, 0x2200, 0x0001,
      EVR_CS, "LabelUsingInformationExtractedFromInstances", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0002, 0x2200, 0x0002,
      EVR_UT, "LabelText", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0003, 0x2200, 0x0003,
      EVR_CS, "LabelStyleSelection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0004, 0x2200, 0x0004,
      EVR_LT, "MediaDisposition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0005, 0x2200, 0x0005,
      EVR_LT, "BarcodeValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0006, 0x2200, 0x0006,
      EVR_CS, "BarcodeSymbology", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0007, 0x2200, 0x0007,
      EVR_CS, "AllowMediaSplitting", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0008, 0x2200, 0x0008,
      EVR_CS, "IncludeNonDICOMObjects", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0009, 0x2200, 0x0009,
      EVR_CS, "IncludeDisplayApplication", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000a, 0x2200, 0x000a,
      EVR_CS, "PreserveCompositeInstancesAfterMediaCreation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000b, 0x2200, 0x000b,
      EVR_US, "TotalNumberOfPiecesOfMediaCreated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000c, 0x2200, 0x000c,
      EVR_LO, "RequestedMediaApplicationProfile", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000d, 0x2200, 0x000d,
      EVR_SQ, "ReferencedStorageMediaSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000e, 0x2200, 0x000e,
      EVR_AT, "FailureAttributes", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x000f, 0x2200, 0x000f,
      EVR_CS, "AllowLossyCompression", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x2200, 0x0020, 0x2200, 0x0020,
      EVR_CS, "RequestPriority", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0002, 0x3002, 0x0002,
      EVR_SH, "RTImageLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0003, 0x3002, 0x0003,
      EVR_LO, "RTImageName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0004, 0x3002, 0x0004,
      EVR_ST, "RTImageDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x000a, 0x3002, 0x000a,
      EVR_CS, "ReportedValuesOrigin", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x000c, 0x3002, 0x000c,
      EVR_CS, "RTImagePlane", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x000d, 0x3002, 0x000d,
      EVR_DS, "XRayImageReceptorTranslation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x000e, 0x3002, 0x000e,
      EVR_DS, "XRayImageReceptorAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0010, 0x3002, 0x0010,
      EVR_DS, "RTImageOrientation", 6, 6, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0011, 0x3002, 0x0011,
      EVR_DS, "ImagePlanePixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0012, 0x3002, 0x0012,
      EVR_DS, "RTImagePosition", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0020, 0x3002, 0x0020,
      EVR_SH, "RadiationMachineName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0022, 0x3002, 0x0022,
      EVR_DS, "RadiationMachineSAD", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0024, 0x3002, 0x0024,
      EVR_DS, "RadiationMachineSSD", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0026, 0x3002, 0x0026,
      EVR_DS, "RTImageSID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0028, 0x3002, 0x0028,
      EVR_DS, "SourceToReferenceObjectDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0029, 0x3002, 0x0029,
      EVR_IS, "FractionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0030, 0x3002, 0x0030,
      EVR_SQ, "ExposureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0032, 0x3002, 0x0032,
      EVR_DS, "MetersetExposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0034, 0x3002, 0x0034,
      EVR_DS, "DiaphragmPosition", 4, 4, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0040, 0x3002, 0x0040,
      EVR_SQ, "FluenceMapSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0041, 0x3002, 0x0041,
      EVR_CS, "FluenceDataSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0042, 0x3002, 0x0042,
      EVR_DS, "FluenceDataScale", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0050, 0x3002, 0x0050,
      EVR_SQ, "PrimaryFluenceModeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0051, 0x3002, 0x0051,
      EVR_CS, "FluenceMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0052, 0x3002, 0x0052,
      EVR_SH, "FluenceModeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0100, 0x3002, 0x0100,
      EVR_IS, "SelectedFrameNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0101, 0x3002, 0x0101,
      EVR_SQ, "SelectedFrameFunctionalGroupsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0102, 0x3002, 0x0102,
      EVR_SQ, "RTImageFrameGeneralContentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0103, 0x3002, 0x0103,
      EVR_SQ, "RTImageFrameContextSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0104, 0x3002, 0x0104,
      EVR_SQ, "RTImageScopeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0105, 0x3002, 0x0105,
      EVR_CS, "BeamModifierCoordinatesPresenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0106, 0x3002, 0x0106,
      EVR_FD, "StartCumulativeMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0107, 0x3002, 0x0107,
      EVR_FD, "StopCumulativeMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0108, 0x3002, 0x0108,
      EVR_SQ, "RTAcquisitionPatientPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0109, 0x3002, 0x0109,
      EVR_SQ, "RTImageFrameImagingDevicePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010a, 0x3002, 0x010a,
      EVR_SQ, "RTImageFramekVRadiationAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010b, 0x3002, 0x010b,
      EVR_SQ, "RTImageFrameMVRadiationAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010c, 0x3002, 0x010c,
      EVR_SQ, "RTImageFrameRadiationAcquisitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010d, 0x3002, 0x010d,
      EVR_SQ, "ImagingSourcePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010e, 0x3002, 0x010e,
      EVR_SQ, "ImageReceptorPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x010f, 0x3002, 0x010f,
      EVR_FD, "DevicePositionToEquipmentMappingMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0110, 0x3002, 0x0110,
      EVR_SQ, "DevicePositionParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0111, 0x3002, 0x0111,
      EVR_CS, "ImagingSourceLocationSpecificationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0112, 0x3002, 0x0112,
      EVR_SQ, "ImagingDeviceLocationMatrixSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0113, 0x3002, 0x0113,
      EVR_SQ, "ImagingDeviceLocationParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0114, 0x3002, 0x0114,
      EVR_SQ, "ImagingApertureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0115, 0x3002, 0x0115,
      EVR_CS, "ImagingApertureSpecificationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0116, 0x3002, 0x0116,
      EVR_US, "NumberOfAcquisitionDevices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0117, 0x3002, 0x0117,
      EVR_SQ, "AcquisitionDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0118, 0x3002, 0x0118,
      EVR_SQ, "AcquisitionTaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0119, 0x3002, 0x0119,
      EVR_SQ, "AcquisitionTaskWorkitemCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011a, 0x3002, 0x011a,
      EVR_SQ, "AcquisitionSubtaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011b, 0x3002, 0x011b,
      EVR_SQ, "SubtaskWorkitemCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011c, 0x3002, 0x011c,
      EVR_US, "AcquisitionTaskIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011d, 0x3002, 0x011d,
      EVR_US, "AcquisitionSubtaskIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011e, 0x3002, 0x011e,
      EVR_SQ, "ReferencedBaselineParametersRTRadiationInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x011f, 0x3002, 0x011f,
      EVR_SQ, "PositionAcquisitionTemplateIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0120, 0x3002, 0x0120,
      EVR_ST, "PositionAcquisitionTemplateID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0121, 0x3002, 0x0121,
      EVR_LO, "PositionAcquisitionTemplateName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0122, 0x3002, 0x0122,
      EVR_SQ, "PositionAcquisitionTemplateCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0123, 0x3002, 0x0123,
      EVR_LT, "PositionAcquisitionTemplateDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0124, 0x3002, 0x0124,
      EVR_SQ, "AcquisitionTaskApplicabilitySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0125, 0x3002, 0x0125,
      EVR_SQ, "ProjectionImagingAcquisitionParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0126, 0x3002, 0x0126,
      EVR_SQ, "CTImagingAcquisitionParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0127, 0x3002, 0x0127,
      EVR_SQ, "KVImagingGenerationParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0128, 0x3002, 0x0128,
      EVR_SQ, "MVImagingGenerationParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0129, 0x3002, 0x0129,
      EVR_CS, "AcquisitionSignalType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012a, 0x3002, 0x012a,
      EVR_CS, "AcquisitionMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012b, 0x3002, 0x012b,
      EVR_SQ, "ScanStartPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012c, 0x3002, 0x012c,
      EVR_SQ, "ScanStopPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012d, 0x3002, 0x012d,
      EVR_FD, "ImagingSourceToBeamModifierDefinitionPlaneDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012e, 0x3002, 0x012e,
      EVR_CS, "ScanArcType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x012f, 0x3002, 0x012f,
      EVR_CS, "DetectorPositioningType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0130, 0x3002, 0x0130,
      EVR_SQ, "AdditionalRTAccessoryDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0131, 0x3002, 0x0131,
      EVR_SQ, "DeviceSpecificAcquisitionParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0132, 0x3002, 0x0132,
      EVR_SQ, "ReferencedPositionReferenceInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0133, 0x3002, 0x0133,
      EVR_SQ, "EnergyDerivationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0134, 0x3002, 0x0134,
      EVR_FD, "MaximumCumulativeMetersetExposure", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0135, 0x3002, 0x0135,
      EVR_SQ, "AcquisitionInitiationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3002, 0x0136, 0x3002, 0x0136,
      EVR_SQ, "RTConeBeamImagingGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0001, 0x3004, 0x0001,
      EVR_CS, "DVHType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0002, 0x3004, 0x0002,
      EVR_CS, "DoseUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0004, 0x3004, 0x0004,
      EVR_CS, "DoseType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0005, 0x3004, 0x0005,
      EVR_CS, "SpatialTransformOfDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0006, 0x3004, 0x0006,
      EVR_LO, "DoseComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0008, 0x3004, 0x0008,
      EVR_DS, "NormalizationPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x000a, 0x3004, 0x000a,
      EVR_CS, "DoseSummationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x000c, 0x3004, 0x000c,
      EVR_DS, "GridFrameOffsetVector", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x000e, 0x3004, 0x000e,
      EVR_DS, "DoseGridScaling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0010, 0x3004, 0x0010,
      EVR_SQ, "RETIRED_RTDoseROISequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0012, 0x3004, 0x0012,
      EVR_DS, "DoseValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0014, 0x3004, 0x0014,
      EVR_CS, "TissueHeterogeneityCorrection", 1, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0016, 0x3004, 0x0016,
      EVR_SQ, "RecommendedIsodoseLevelSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0040, 0x3004, 0x0040,
      EVR_DS, "DVHNormalizationPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0042, 0x3004, 0x0042,
      EVR_DS, "DVHNormalizationDoseValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0050, 0x3004, 0x0050,
      EVR_SQ, "DVHSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0052, 0x3004, 0x0052,
      EVR_DS, "DVHDoseScaling", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0054, 0x3004, 0x0054,
      EVR_CS, "DVHVolumeUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0056, 0x3004, 0x0056,
      EVR_IS, "DVHNumberOfBins", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0058, 0x3004, 0x0058,
      EVR_DS, "DVHData", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0060, 0x3004, 0x0060,
      EVR_SQ, "DVHReferencedROISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0062, 0x3004, 0x0062,
      EVR_CS, "DVHROIContributionType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0070, 0x3004, 0x0070,
      EVR_DS, "DVHMinimumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0072, 0x3004, 0x0072,
      EVR_DS, "DVHMaximumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3004, 0x0074, 0x3004, 0x0074,
      EVR_DS, "DVHMeanDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0002, 0x3006, 0x0002,
      EVR_SH, "StructureSetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0004, 0x3006, 0x0004,
      EVR_LO, "StructureSetName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0006, 0x3006, 0x0006,
      EVR_ST, "StructureSetDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0008, 0x3006, 0x0008,
      EVR_DA, "StructureSetDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0009, 0x3006, 0x0009,
      EVR_TM, "StructureSetTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0010, 0x3006, 0x0010,
      EVR_SQ, "ReferencedFrameOfReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0012, 0x3006, 0x0012,
      EVR_SQ, "RTReferencedStudySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0014, 0x3006, 0x0014,
      EVR_SQ, "RTReferencedSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0016, 0x3006, 0x0016,
      EVR_SQ, "ContourImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0018, 0x3006, 0x0018,
      EVR_SQ, "PredecessorStructureSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0020, 0x3006, 0x0020,
      EVR_SQ, "StructureSetROISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0022, 0x3006, 0x0022,
      EVR_IS, "ROINumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0024, 0x3006, 0x0024,
      EVR_UI, "ReferencedFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0026, 0x3006, 0x0026,
      EVR_LO, "ROIName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0028, 0x3006, 0x0028,
      EVR_ST, "ROIDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x002a, 0x3006, 0x002a,
      EVR_IS, "ROIDisplayColor", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x002c, 0x3006, 0x002c,
      EVR_DS, "ROIVolume", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x002d, 0x3006, 0x002d,
      EVR_DT, "ROIDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x002e, 0x3006, 0x002e,
      EVR_DT, "ROIObservationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0030, 0x3006, 0x0030,
      EVR_SQ, "RTRelatedROISequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0033, 0x3006, 0x0033,
      EVR_CS, "RTROIRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0036, 0x3006, 0x0036,
      EVR_CS, "ROIGenerationAlgorithm", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0037, 0x3006, 0x0037,
      EVR_SQ, "ROIDerivationAlgorithmIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0038, 0x3006, 0x0038,
      EVR_LO, "ROIGenerationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0039, 0x3006, 0x0039,
      EVR_SQ, "ROIContourSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0040, 0x3006, 0x0040,
      EVR_SQ, "ContourSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0042, 0x3006, 0x0042,
      EVR_CS, "ContourGeometricType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0044, 0x3006, 0x0044,
      EVR_DS, "RETIRED_ContourSlabThickness", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0045, 0x3006, 0x0045,
      EVR_DS, "RETIRED_ContourOffsetVector", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0046, 0x3006, 0x0046,
      EVR_IS, "NumberOfContourPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0048, 0x3006, 0x0048,
      EVR_IS, "ContourNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0049, 0x3006, 0x0049,
      EVR_IS, "RETIRED_AttachedContours", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004a, 0x3006, 0x004a,
      EVR_SQ, "SourcePixelPlanesCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004b, 0x3006, 0x004b,
      EVR_SQ, "SourceSeriesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004c, 0x3006, 0x004c,
      EVR_SQ, "SourceSeriesInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004d, 0x3006, 0x004d,
      EVR_SQ, "ROICreatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004e, 0x3006, 0x004e,
      EVR_SQ, "ROIInterpreterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x004f, 0x3006, 0x004f,
      EVR_SQ, "ROIObservationContextCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0050, 0x3006, 0x0050,
      EVR_DS, "ContourData", 3, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0080, 0x3006, 0x0080,
      EVR_SQ, "RTROIObservationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0082, 0x3006, 0x0082,
      EVR_IS, "ObservationNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0084, 0x3006, 0x0084,
      EVR_IS, "ReferencedROINumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0085, 0x3006, 0x0085,
      EVR_SH, "RETIRED_ROIObservationLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0086, 0x3006, 0x0086,
      EVR_SQ, "RTROIIdentificationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x0088, 0x3006, 0x0088,
      EVR_ST, "RETIRED_ROIObservationDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00a0, 0x3006, 0x00a0,
      EVR_SQ, "RelatedRTROIObservationsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00a4, 0x3006, 0x00a4,
      EVR_CS, "RTROIInterpretedType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00a6, 0x3006, 0x00a6,
      EVR_PN, "ROIInterpreter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b0, 0x3006, 0x00b0,
      EVR_SQ, "ROIPhysicalPropertiesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b2, 0x3006, 0x00b2,
      EVR_CS, "ROIPhysicalProperty", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b4, 0x3006, 0x00b4,
      EVR_DS, "ROIPhysicalPropertyValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b6, 0x3006, 0x00b6,
      EVR_SQ, "ROIElementalCompositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b7, 0x3006, 0x00b7,
      EVR_US, "ROIElementalCompositionAtomicNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b8, 0x3006, 0x00b8,
      EVR_FL, "ROIElementalCompositionAtomicMassFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00b9, 0x3006, 0x00b9,
      EVR_SQ, "RETIRED_AdditionalRTROIIdentificationCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c0, 0x3006, 0x00c0,
      EVR_SQ, "RETIRED_FrameOfReferenceRelationshipSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c2, 0x3006, 0x00c2,
      EVR_UI, "RETIRED_RelatedFrameOfReferenceUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c4, 0x3006, 0x00c4,
      EVR_CS, "RETIRED_FrameOfReferenceTransformationType", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c6, 0x3006, 0x00c6,
      EVR_DS, "FrameOfReferenceTransformationMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c8, 0x3006, 0x00c8,
      EVR_LO, "FrameOfReferenceTransformationComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00c9, 0x3006, 0x00c9,
      EVR_SQ, "PatientLocationCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00ca, 0x3006, 0x00ca,
      EVR_SQ, "PatientLocationCoordinatesCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3006, 0x00cb, 0x3006, 0x00cb,
      EVR_SQ, "PatientSupportPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0010, 0x3008, 0x0010,
      EVR_SQ, "MeasuredDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0012, 0x3008, 0x0012,
      EVR_ST, "MeasuredDoseDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0014, 0x3008, 0x0014,
      EVR_CS, "MeasuredDoseType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0016, 0x3008, 0x0016,
      EVR_DS, "MeasuredDoseValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0020, 0x3008, 0x0020,
      EVR_SQ, "TreatmentSessionBeamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0021, 0x3008, 0x0021,
      EVR_SQ, "TreatmentSessionIonBeamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0022, 0x3008, 0x0022,
      EVR_IS, "CurrentFractionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0024, 0x3008, 0x0024,
      EVR_DA, "TreatmentControlPointDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0025, 0x3008, 0x0025,
      EVR_TM, "TreatmentControlPointTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x002a, 0x3008, 0x002a,
      EVR_CS, "TreatmentTerminationStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x002b, 0x3008, 0x002b,
      EVR_SH, "RETIRED_TreatmentTerminationCode", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x002c, 0x3008, 0x002c,
      EVR_CS, "TreatmentVerificationStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0030, 0x3008, 0x0030,
      EVR_SQ, "ReferencedTreatmentRecordSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0032, 0x3008, 0x0032,
      EVR_DS, "SpecifiedPrimaryMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0033, 0x3008, 0x0033,
      EVR_DS, "SpecifiedSecondaryMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0036, 0x3008, 0x0036,
      EVR_DS, "DeliveredPrimaryMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0037, 0x3008, 0x0037,
      EVR_DS, "DeliveredSecondaryMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x003a, 0x3008, 0x003a,
      EVR_DS, "SpecifiedTreatmentTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x003b, 0x3008, 0x003b,
      EVR_DS, "DeliveredTreatmentTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0040, 0x3008, 0x0040,
      EVR_SQ, "ControlPointDeliverySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0041, 0x3008, 0x0041,
      EVR_SQ, "IonControlPointDeliverySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0042, 0x3008, 0x0042,
      EVR_DS, "SpecifiedMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0044, 0x3008, 0x0044,
      EVR_DS, "DeliveredMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0045, 0x3008, 0x0045,
      EVR_FL, "MetersetRateSet", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0046, 0x3008, 0x0046,
      EVR_FL, "MetersetRateDelivered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0047, 0x3008, 0x0047,
      EVR_FL, "ScanSpotMetersetsDelivered", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0048, 0x3008, 0x0048,
      EVR_DS, "DoseRateDelivered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0050, 0x3008, 0x0050,
      EVR_SQ, "TreatmentSummaryCalculatedDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0052, 0x3008, 0x0052,
      EVR_DS, "CumulativeDoseToDoseReference", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0054, 0x3008, 0x0054,
      EVR_DA, "FirstTreatmentDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0056, 0x3008, 0x0056,
      EVR_DA, "MostRecentTreatmentDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x005a, 0x3008, 0x005a,
      EVR_IS, "NumberOfFractionsDelivered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0060, 0x3008, 0x0060,
      EVR_SQ, "OverrideSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0061, 0x3008, 0x0061,
      EVR_AT, "ParameterSequencePointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0062, 0x3008, 0x0062,
      EVR_AT, "OverrideParameterPointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0063, 0x3008, 0x0063,
      EVR_IS, "ParameterItemIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0064, 0x3008, 0x0064,
      EVR_IS, "MeasuredDoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0065, 0x3008, 0x0065,
      EVR_AT, "ParameterPointer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0066, 0x3008, 0x0066,
      EVR_ST, "OverrideReason", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0067, 0x3008, 0x0067,
      EVR_US, "ParameterValueNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0068, 0x3008, 0x0068,
      EVR_SQ, "CorrectedParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x006a, 0x3008, 0x006a,
      EVR_FL, "CorrectionValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0070, 0x3008, 0x0070,
      EVR_SQ, "CalculatedDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0072, 0x3008, 0x0072,
      EVR_IS, "CalculatedDoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0074, 0x3008, 0x0074,
      EVR_ST, "CalculatedDoseReferenceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0076, 0x3008, 0x0076,
      EVR_DS, "CalculatedDoseReferenceDoseValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0078, 0x3008, 0x0078,
      EVR_DS, "StartMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x007a, 0x3008, 0x007a,
      EVR_DS, "EndMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0080, 0x3008, 0x0080,
      EVR_SQ, "ReferencedMeasuredDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0082, 0x3008, 0x0082,
      EVR_IS, "ReferencedMeasuredDoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0090, 0x3008, 0x0090,
      EVR_SQ, "ReferencedCalculatedDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0092, 0x3008, 0x0092,
      EVR_IS, "ReferencedCalculatedDoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00a0, 0x3008, 0x00a0,
      EVR_SQ, "BeamLimitingDeviceLeafPairsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00a1, 0x3008, 0x00a1,
      EVR_SQ, "EnhancedRTBeamLimitingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00a2, 0x3008, 0x00a2,
      EVR_SQ, "EnhancedRTBeamLimitingOpeningSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00a3, 0x3008, 0x00a3,
      EVR_CS, "EnhancedRTBeamLimitingDeviceDefinitionFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00a4, 0x3008, 0x00a4,
      EVR_FD, "ParallelRTBeamDelimiterOpeningExtents", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00b0, 0x3008, 0x00b0,
      EVR_SQ, "RecordedWedgeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00c0, 0x3008, 0x00c0,
      EVR_SQ, "RecordedCompensatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00d0, 0x3008, 0x00d0,
      EVR_SQ, "RecordedBlockSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00d1, 0x3008, 0x00d1,
      EVR_SQ, "RecordedBlockSlabSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00e0, 0x3008, 0x00e0,
      EVR_SQ, "TreatmentSummaryMeasuredDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00f0, 0x3008, 0x00f0,
      EVR_SQ, "RecordedSnoutSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00f2, 0x3008, 0x00f2,
      EVR_SQ, "RecordedRangeShifterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00f4, 0x3008, 0x00f4,
      EVR_SQ, "RecordedLateralSpreadingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x00f6, 0x3008, 0x00f6,
      EVR_SQ, "RecordedRangeModulatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0100, 0x3008, 0x0100,
      EVR_SQ, "RecordedSourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0105, 0x3008, 0x0105,
      EVR_LO, "SourceSerialNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0110, 0x3008, 0x0110,
      EVR_SQ, "TreatmentSessionApplicationSetupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0116, 0x3008, 0x0116,
      EVR_CS, "ApplicationSetupCheck", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0120, 0x3008, 0x0120,
      EVR_SQ, "RecordedBrachyAccessoryDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0122, 0x3008, 0x0122,
      EVR_IS, "ReferencedBrachyAccessoryDeviceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0130, 0x3008, 0x0130,
      EVR_SQ, "RecordedChannelSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0132, 0x3008, 0x0132,
      EVR_DS, "SpecifiedChannelTotalTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0134, 0x3008, 0x0134,
      EVR_DS, "DeliveredChannelTotalTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0136, 0x3008, 0x0136,
      EVR_IS, "SpecifiedNumberOfPulses", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0138, 0x3008, 0x0138,
      EVR_IS, "DeliveredNumberOfPulses", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x013a, 0x3008, 0x013a,
      EVR_DS, "SpecifiedPulseRepetitionInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x013c, 0x3008, 0x013c,
      EVR_DS, "DeliveredPulseRepetitionInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0140, 0x3008, 0x0140,
      EVR_SQ, "RecordedSourceApplicatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0142, 0x3008, 0x0142,
      EVR_IS, "ReferencedSourceApplicatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0150, 0x3008, 0x0150,
      EVR_SQ, "RecordedChannelShieldSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0152, 0x3008, 0x0152,
      EVR_IS, "ReferencedChannelShieldNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0160, 0x3008, 0x0160,
      EVR_SQ, "BrachyControlPointDeliveredSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0162, 0x3008, 0x0162,
      EVR_DA, "SafePositionExitDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0164, 0x3008, 0x0164,
      EVR_TM, "SafePositionExitTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0166, 0x3008, 0x0166,
      EVR_DA, "SafePositionReturnDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0168, 0x3008, 0x0168,
      EVR_TM, "SafePositionReturnTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0171, 0x3008, 0x0171,
      EVR_SQ, "PulseSpecificBrachyControlPointDeliveredSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0172, 0x3008, 0x0172,
      EVR_US, "PulseNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0173, 0x3008, 0x0173,
      EVR_SQ, "BrachyPulseControlPointDeliveredSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0200, 0x3008, 0x0200,
      EVR_CS, "CurrentTreatmentStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0202, 0x3008, 0x0202,
      EVR_ST, "TreatmentStatusComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0220, 0x3008, 0x0220,
      EVR_SQ, "FractionGroupSummarySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0223, 0x3008, 0x0223,
      EVR_IS, "ReferencedFractionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0224, 0x3008, 0x0224,
      EVR_CS, "FractionGroupType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0230, 0x3008, 0x0230,
      EVR_CS, "BeamStopperPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0240, 0x3008, 0x0240,
      EVR_SQ, "FractionStatusSummarySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0250, 0x3008, 0x0250,
      EVR_DA, "TreatmentDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3008, 0x0251, 0x3008, 0x0251,
      EVR_TM, "TreatmentTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0002, 0x300a, 0x0002,
      EVR_SH, "RTPlanLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0003, 0x300a, 0x0003,
      EVR_LO, "RTPlanName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0004, 0x300a, 0x0004,
      EVR_ST, "RTPlanDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0006, 0x300a, 0x0006,
      EVR_DA, "RTPlanDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0007, 0x300a, 0x0007,
      EVR_TM, "RTPlanTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0009, 0x300a, 0x0009,
      EVR_LO, "TreatmentProtocols", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x000a, 0x300a, 0x000a,
      EVR_CS, "PlanIntent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x000b, 0x300a, 0x000b,
      EVR_LO, "RETIRED_TreatmentSites", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x000c, 0x300a, 0x000c,
      EVR_CS, "RTPlanGeometry", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x000e, 0x300a, 0x000e,
      EVR_ST, "PrescriptionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0010, 0x300a, 0x0010,
      EVR_SQ, "DoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0012, 0x300a, 0x0012,
      EVR_IS, "DoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0013, 0x300a, 0x0013,
      EVR_UI, "DoseReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0014, 0x300a, 0x0014,
      EVR_CS, "DoseReferenceStructureType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0015, 0x300a, 0x0015,
      EVR_CS, "NominalBeamEnergyUnit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0016, 0x300a, 0x0016,
      EVR_LO, "DoseReferenceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0018, 0x300a, 0x0018,
      EVR_DS, "DoseReferencePointCoordinates", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x001a, 0x300a, 0x001a,
      EVR_DS, "NominalPriorDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0020, 0x300a, 0x0020,
      EVR_CS, "DoseReferenceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0021, 0x300a, 0x0021,
      EVR_DS, "ConstraintWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0022, 0x300a, 0x0022,
      EVR_DS, "DeliveryWarningDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0023, 0x300a, 0x0023,
      EVR_DS, "DeliveryMaximumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0025, 0x300a, 0x0025,
      EVR_DS, "TargetMinimumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0026, 0x300a, 0x0026,
      EVR_DS, "TargetPrescriptionDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0027, 0x300a, 0x0027,
      EVR_DS, "TargetMaximumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0028, 0x300a, 0x0028,
      EVR_DS, "TargetUnderdoseVolumeFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x002a, 0x300a, 0x002a,
      EVR_DS, "OrganAtRiskFullVolumeDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x002b, 0x300a, 0x002b,
      EVR_DS, "OrganAtRiskLimitDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x002c, 0x300a, 0x002c,
      EVR_DS, "OrganAtRiskMaximumDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x002d, 0x300a, 0x002d,
      EVR_DS, "OrganAtRiskOverdoseVolumeFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0040, 0x300a, 0x0040,
      EVR_SQ, "ToleranceTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0042, 0x300a, 0x0042,
      EVR_IS, "ToleranceTableNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0043, 0x300a, 0x0043,
      EVR_SH, "ToleranceTableLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0044, 0x300a, 0x0044,
      EVR_DS, "GantryAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0046, 0x300a, 0x0046,
      EVR_DS, "BeamLimitingDeviceAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0048, 0x300a, 0x0048,
      EVR_SQ, "BeamLimitingDeviceToleranceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x004a, 0x300a, 0x004a,
      EVR_DS, "BeamLimitingDevicePositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x004b, 0x300a, 0x004b,
      EVR_FL, "SnoutPositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x004c, 0x300a, 0x004c,
      EVR_DS, "PatientSupportAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x004e, 0x300a, 0x004e,
      EVR_DS, "TableTopEccentricAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x004f, 0x300a, 0x004f,
      EVR_FL, "TableTopPitchAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0050, 0x300a, 0x0050,
      EVR_FL, "TableTopRollAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0051, 0x300a, 0x0051,
      EVR_DS, "TableTopVerticalPositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0052, 0x300a, 0x0052,
      EVR_DS, "TableTopLongitudinalPositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0053, 0x300a, 0x0053,
      EVR_DS, "TableTopLateralPositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0054, 0x300a, 0x0054,
      EVR_UI, "TableTopPositionAlignmentUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0055, 0x300a, 0x0055,
      EVR_CS, "RTPlanRelationship", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0070, 0x300a, 0x0070,
      EVR_SQ, "FractionGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0071, 0x300a, 0x0071,
      EVR_IS, "FractionGroupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0072, 0x300a, 0x0072,
      EVR_LO, "FractionGroupDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0078, 0x300a, 0x0078,
      EVR_IS, "NumberOfFractionsPlanned", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0079, 0x300a, 0x0079,
      EVR_IS, "NumberOfFractionPatternDigitsPerDay", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x007a, 0x300a, 0x007a,
      EVR_IS, "RepeatFractionCycleLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x007b, 0x300a, 0x007b,
      EVR_LT, "FractionPattern", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0080, 0x300a, 0x0080,
      EVR_IS, "NumberOfBeams", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0082, 0x300a, 0x0082,
      EVR_DS, "RETIRED_BeamDoseSpecificationPoint", 3, 3, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0083, 0x300a, 0x0083,
      EVR_UI, "ReferencedDoseReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0084, 0x300a, 0x0084,
      EVR_DS, "BeamDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0086, 0x300a, 0x0086,
      EVR_DS, "BeamMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0088, 0x300a, 0x0088,
      EVR_FL, "BeamDosePointDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0089, 0x300a, 0x0089,
      EVR_FL, "BeamDosePointEquivalentDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008a, 0x300a, 0x008a,
      EVR_FL, "BeamDosePointSSD", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008b, 0x300a, 0x008b,
      EVR_CS, "BeamDoseMeaning", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008c, 0x300a, 0x008c,
      EVR_SQ, "BeamDoseVerificationControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008d, 0x300a, 0x008d,
      EVR_FL, "RETIRED_AverageBeamDosePointDepth", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008e, 0x300a, 0x008e,
      EVR_FL, "RETIRED_AverageBeamDosePointEquivalentDepth", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x008f, 0x300a, 0x008f,
      EVR_FL, "RETIRED_AverageBeamDosePointSSD", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0090, 0x300a, 0x0090,
      EVR_CS, "BeamDoseType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0091, 0x300a, 0x0091,
      EVR_DS, "AlternateBeamDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0092, 0x300a, 0x0092,
      EVR_CS, "AlternateBeamDoseType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0093, 0x300a, 0x0093,
      EVR_CS, "DepthValueAveragingFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0094, 0x300a, 0x0094,
      EVR_DS, "BeamDosePointSourceToExternalContourDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00a0, 0x300a, 0x00a0,
      EVR_IS, "NumberOfBrachyApplicationSetups", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00a2, 0x300a, 0x00a2,
      EVR_DS, "BrachyApplicationSetupDoseSpecificationPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00a4, 0x300a, 0x00a4,
      EVR_DS, "BrachyApplicationSetupDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b0, 0x300a, 0x00b0,
      EVR_SQ, "BeamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b2, 0x300a, 0x00b2,
      EVR_SH, "TreatmentMachineName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b3, 0x300a, 0x00b3,
      EVR_CS, "PrimaryDosimeterUnit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b4, 0x300a, 0x00b4,
      EVR_DS, "SourceAxisDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b6, 0x300a, 0x00b6,
      EVR_SQ, "BeamLimitingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00b8, 0x300a, 0x00b8,
      EVR_CS, "RTBeamLimitingDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ba, 0x300a, 0x00ba,
      EVR_DS, "SourceToBeamLimitingDeviceDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00bb, 0x300a, 0x00bb,
      EVR_FL, "IsocenterToBeamLimitingDeviceDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00bc, 0x300a, 0x00bc,
      EVR_IS, "NumberOfLeafJawPairs", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00be, 0x300a, 0x00be,
      EVR_DS, "LeafPositionBoundaries", 3, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c0, 0x300a, 0x00c0,
      EVR_IS, "BeamNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c2, 0x300a, 0x00c2,
      EVR_LO, "BeamName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c3, 0x300a, 0x00c3,
      EVR_ST, "BeamDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c4, 0x300a, 0x00c4,
      EVR_CS, "BeamType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c5, 0x300a, 0x00c5,
      EVR_FD, "BeamDeliveryDurationLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c6, 0x300a, 0x00c6,
      EVR_CS, "RadiationType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c7, 0x300a, 0x00c7,
      EVR_CS, "HighDoseTechniqueType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00c8, 0x300a, 0x00c8,
      EVR_IS, "ReferenceImageNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ca, 0x300a, 0x00ca,
      EVR_SQ, "PlannedVerificationImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00cc, 0x300a, 0x00cc,
      EVR_LO, "ImagingDeviceSpecificAcquisitionParameters", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ce, 0x300a, 0x00ce,
      EVR_CS, "TreatmentDeliveryType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d0, 0x300a, 0x00d0,
      EVR_IS, "NumberOfWedges", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d1, 0x300a, 0x00d1,
      EVR_SQ, "WedgeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d2, 0x300a, 0x00d2,
      EVR_IS, "WedgeNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d3, 0x300a, 0x00d3,
      EVR_CS, "WedgeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d4, 0x300a, 0x00d4,
      EVR_SH, "WedgeID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d5, 0x300a, 0x00d5,
      EVR_IS, "WedgeAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d6, 0x300a, 0x00d6,
      EVR_DS, "WedgeFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d7, 0x300a, 0x00d7,
      EVR_FL, "TotalWedgeTrayWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d8, 0x300a, 0x00d8,
      EVR_DS, "WedgeOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00d9, 0x300a, 0x00d9,
      EVR_FL, "IsocenterToWedgeTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00da, 0x300a, 0x00da,
      EVR_DS, "SourceToWedgeTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00db, 0x300a, 0x00db,
      EVR_FL, "WedgeThinEdgePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00dc, 0x300a, 0x00dc,
      EVR_SH, "BolusID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00dd, 0x300a, 0x00dd,
      EVR_ST, "BolusDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00de, 0x300a, 0x00de,
      EVR_DS, "EffectiveWedgeAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e0, 0x300a, 0x00e0,
      EVR_IS, "NumberOfCompensators", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e1, 0x300a, 0x00e1,
      EVR_SH, "MaterialID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e2, 0x300a, 0x00e2,
      EVR_DS, "TotalCompensatorTrayFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e3, 0x300a, 0x00e3,
      EVR_SQ, "CompensatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e4, 0x300a, 0x00e4,
      EVR_IS, "CompensatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e5, 0x300a, 0x00e5,
      EVR_SH, "CompensatorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e6, 0x300a, 0x00e6,
      EVR_DS, "SourceToCompensatorTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e7, 0x300a, 0x00e7,
      EVR_IS, "CompensatorRows", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e8, 0x300a, 0x00e8,
      EVR_IS, "CompensatorColumns", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00e9, 0x300a, 0x00e9,
      EVR_DS, "CompensatorPixelSpacing", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ea, 0x300a, 0x00ea,
      EVR_DS, "CompensatorPosition", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00eb, 0x300a, 0x00eb,
      EVR_DS, "CompensatorTransmissionData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ec, 0x300a, 0x00ec,
      EVR_DS, "CompensatorThicknessData", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ed, 0x300a, 0x00ed,
      EVR_IS, "NumberOfBoli", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ee, 0x300a, 0x00ee,
      EVR_CS, "CompensatorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00ef, 0x300a, 0x00ef,
      EVR_SH, "CompensatorTrayID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f0, 0x300a, 0x00f0,
      EVR_IS, "NumberOfBlocks", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f2, 0x300a, 0x00f2,
      EVR_DS, "TotalBlockTrayFactor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f3, 0x300a, 0x00f3,
      EVR_FL, "TotalBlockTrayWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f4, 0x300a, 0x00f4,
      EVR_SQ, "BlockSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f5, 0x300a, 0x00f5,
      EVR_SH, "BlockTrayID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f6, 0x300a, 0x00f6,
      EVR_DS, "SourceToBlockTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f7, 0x300a, 0x00f7,
      EVR_FL, "IsocenterToBlockTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f8, 0x300a, 0x00f8,
      EVR_CS, "BlockType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00f9, 0x300a, 0x00f9,
      EVR_LO, "AccessoryCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00fa, 0x300a, 0x00fa,
      EVR_CS, "BlockDivergence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00fb, 0x300a, 0x00fb,
      EVR_CS, "BlockMountingPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00fc, 0x300a, 0x00fc,
      EVR_IS, "BlockNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x00fe, 0x300a, 0x00fe,
      EVR_LO, "BlockName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0100, 0x300a, 0x0100,
      EVR_DS, "BlockThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0102, 0x300a, 0x0102,
      EVR_DS, "BlockTransmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0104, 0x300a, 0x0104,
      EVR_IS, "BlockNumberOfPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0106, 0x300a, 0x0106,
      EVR_DS, "BlockData", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0107, 0x300a, 0x0107,
      EVR_SQ, "ApplicatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0108, 0x300a, 0x0108,
      EVR_SH, "ApplicatorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0109, 0x300a, 0x0109,
      EVR_CS, "ApplicatorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x010a, 0x300a, 0x010a,
      EVR_LO, "ApplicatorDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x010c, 0x300a, 0x010c,
      EVR_DS, "CumulativeDoseReferenceCoefficient", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x010e, 0x300a, 0x010e,
      EVR_DS, "FinalCumulativeMetersetWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0110, 0x300a, 0x0110,
      EVR_IS, "NumberOfControlPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0111, 0x300a, 0x0111,
      EVR_SQ, "ControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0112, 0x300a, 0x0112,
      EVR_IS, "ControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0114, 0x300a, 0x0114,
      EVR_DS, "NominalBeamEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0115, 0x300a, 0x0115,
      EVR_DS, "DoseRateSet", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0116, 0x300a, 0x0116,
      EVR_SQ, "WedgePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0118, 0x300a, 0x0118,
      EVR_CS, "WedgePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x011a, 0x300a, 0x011a,
      EVR_SQ, "BeamLimitingDevicePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x011c, 0x300a, 0x011c,
      EVR_DS, "LeafJawPositions", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x011e, 0x300a, 0x011e,
      EVR_DS, "GantryAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x011f, 0x300a, 0x011f,
      EVR_CS, "GantryRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0120, 0x300a, 0x0120,
      EVR_DS, "BeamLimitingDeviceAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0121, 0x300a, 0x0121,
      EVR_CS, "BeamLimitingDeviceRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0122, 0x300a, 0x0122,
      EVR_DS, "PatientSupportAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0123, 0x300a, 0x0123,
      EVR_CS, "PatientSupportRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0124, 0x300a, 0x0124,
      EVR_DS, "TableTopEccentricAxisDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0125, 0x300a, 0x0125,
      EVR_DS, "TableTopEccentricAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0126, 0x300a, 0x0126,
      EVR_CS, "TableTopEccentricRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0128, 0x300a, 0x0128,
      EVR_DS, "TableTopVerticalPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0129, 0x300a, 0x0129,
      EVR_DS, "TableTopLongitudinalPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x012a, 0x300a, 0x012a,
      EVR_DS, "TableTopLateralPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x012c, 0x300a, 0x012c,
      EVR_DS, "IsocenterPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x012e, 0x300a, 0x012e,
      EVR_DS, "SurfaceEntryPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0130, 0x300a, 0x0130,
      EVR_DS, "SourceToSurfaceDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0131, 0x300a, 0x0131,
      EVR_FL, "AverageBeamDosePointSourceToExternalContourDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0132, 0x300a, 0x0132,
      EVR_FL, "SourceToExternalContourDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0133, 0x300a, 0x0133,
      EVR_FL, "ExternalContourEntryPoint", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0134, 0x300a, 0x0134,
      EVR_DS, "CumulativeMetersetWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0140, 0x300a, 0x0140,
      EVR_FL, "TableTopPitchAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0142, 0x300a, 0x0142,
      EVR_CS, "TableTopPitchRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0144, 0x300a, 0x0144,
      EVR_FL, "TableTopRollAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0146, 0x300a, 0x0146,
      EVR_CS, "TableTopRollRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0148, 0x300a, 0x0148,
      EVR_FL, "HeadFixationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x014a, 0x300a, 0x014a,
      EVR_FL, "GantryPitchAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x014c, 0x300a, 0x014c,
      EVR_CS, "GantryPitchRotationDirection", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x014e, 0x300a, 0x014e,
      EVR_FL, "GantryPitchAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0150, 0x300a, 0x0150,
      EVR_CS, "FixationEye", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0151, 0x300a, 0x0151,
      EVR_DS, "ChairHeadFramePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0152, 0x300a, 0x0152,
      EVR_DS, "HeadFixationAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0153, 0x300a, 0x0153,
      EVR_DS, "ChairHeadFramePositionTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0154, 0x300a, 0x0154,
      EVR_DS, "FixationLightAzimuthalAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0155, 0x300a, 0x0155,
      EVR_DS, "FixationLightPolarAngleTolerance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0180, 0x300a, 0x0180,
      EVR_SQ, "PatientSetupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0182, 0x300a, 0x0182,
      EVR_IS, "PatientSetupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0183, 0x300a, 0x0183,
      EVR_LO, "PatientSetupLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0184, 0x300a, 0x0184,
      EVR_LO, "PatientAdditionalPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0190, 0x300a, 0x0190,
      EVR_SQ, "FixationDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0192, 0x300a, 0x0192,
      EVR_CS, "FixationDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0194, 0x300a, 0x0194,
      EVR_SH, "FixationDeviceLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0196, 0x300a, 0x0196,
      EVR_ST, "FixationDeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0198, 0x300a, 0x0198,
      EVR_SH, "FixationDevicePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0199, 0x300a, 0x0199,
      EVR_FL, "FixationDevicePitchAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x019a, 0x300a, 0x019a,
      EVR_FL, "FixationDeviceRollAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01a0, 0x300a, 0x01a0,
      EVR_SQ, "ShieldingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01a2, 0x300a, 0x01a2,
      EVR_CS, "ShieldingDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01a4, 0x300a, 0x01a4,
      EVR_SH, "ShieldingDeviceLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01a6, 0x300a, 0x01a6,
      EVR_ST, "ShieldingDeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01a8, 0x300a, 0x01a8,
      EVR_SH, "ShieldingDevicePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01b0, 0x300a, 0x01b0,
      EVR_CS, "SetupTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01b2, 0x300a, 0x01b2,
      EVR_ST, "SetupTechniqueDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01b4, 0x300a, 0x01b4,
      EVR_SQ, "SetupDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01b6, 0x300a, 0x01b6,
      EVR_CS, "SetupDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01b8, 0x300a, 0x01b8,
      EVR_SH, "SetupDeviceLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01ba, 0x300a, 0x01ba,
      EVR_ST, "SetupDeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01bc, 0x300a, 0x01bc,
      EVR_DS, "SetupDeviceParameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01d0, 0x300a, 0x01d0,
      EVR_ST, "SetupReferenceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01d2, 0x300a, 0x01d2,
      EVR_DS, "TableTopVerticalSetupDisplacement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01d4, 0x300a, 0x01d4,
      EVR_DS, "TableTopLongitudinalSetupDisplacement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x01d6, 0x300a, 0x01d6,
      EVR_DS, "TableTopLateralSetupDisplacement", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0200, 0x300a, 0x0200,
      EVR_CS, "BrachyTreatmentTechnique", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0202, 0x300a, 0x0202,
      EVR_CS, "BrachyTreatmentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0206, 0x300a, 0x0206,
      EVR_SQ, "TreatmentMachineSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0210, 0x300a, 0x0210,
      EVR_SQ, "SourceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0212, 0x300a, 0x0212,
      EVR_IS, "SourceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0214, 0x300a, 0x0214,
      EVR_CS, "SourceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0216, 0x300a, 0x0216,
      EVR_LO, "SourceManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0218, 0x300a, 0x0218,
      EVR_DS, "ActiveSourceDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x021a, 0x300a, 0x021a,
      EVR_DS, "ActiveSourceLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x021b, 0x300a, 0x021b,
      EVR_SH, "SourceModelID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x021c, 0x300a, 0x021c,
      EVR_LO, "SourceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0222, 0x300a, 0x0222,
      EVR_DS, "SourceEncapsulationNominalThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0224, 0x300a, 0x0224,
      EVR_DS, "SourceEncapsulationNominalTransmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0226, 0x300a, 0x0226,
      EVR_LO, "SourceIsotopeName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0228, 0x300a, 0x0228,
      EVR_DS, "SourceIsotopeHalfLife", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0229, 0x300a, 0x0229,
      EVR_CS, "SourceStrengthUnits", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x022a, 0x300a, 0x022a,
      EVR_DS, "ReferenceAirKermaRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x022b, 0x300a, 0x022b,
      EVR_DS, "SourceStrength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x022c, 0x300a, 0x022c,
      EVR_DA, "SourceStrengthReferenceDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x022e, 0x300a, 0x022e,
      EVR_TM, "SourceStrengthReferenceTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0230, 0x300a, 0x0230,
      EVR_SQ, "ApplicationSetupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0232, 0x300a, 0x0232,
      EVR_CS, "ApplicationSetupType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0234, 0x300a, 0x0234,
      EVR_IS, "ApplicationSetupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0236, 0x300a, 0x0236,
      EVR_LO, "ApplicationSetupName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0238, 0x300a, 0x0238,
      EVR_LO, "ApplicationSetupManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0240, 0x300a, 0x0240,
      EVR_IS, "TemplateNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0242, 0x300a, 0x0242,
      EVR_SH, "TemplateType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0244, 0x300a, 0x0244,
      EVR_LO, "TemplateName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0250, 0x300a, 0x0250,
      EVR_DS, "TotalReferenceAirKerma", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0260, 0x300a, 0x0260,
      EVR_SQ, "BrachyAccessoryDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0262, 0x300a, 0x0262,
      EVR_IS, "BrachyAccessoryDeviceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0263, 0x300a, 0x0263,
      EVR_SH, "BrachyAccessoryDeviceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0264, 0x300a, 0x0264,
      EVR_CS, "BrachyAccessoryDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0266, 0x300a, 0x0266,
      EVR_LO, "BrachyAccessoryDeviceName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x026a, 0x300a, 0x026a,
      EVR_DS, "BrachyAccessoryDeviceNominalThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x026c, 0x300a, 0x026c,
      EVR_DS, "BrachyAccessoryDeviceNominalTransmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0271, 0x300a, 0x0271,
      EVR_DS, "ChannelEffectiveLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0272, 0x300a, 0x0272,
      EVR_DS, "ChannelInnerLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0273, 0x300a, 0x0273,
      EVR_SH, "AfterloaderChannelID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0274, 0x300a, 0x0274,
      EVR_DS, "SourceApplicatorTipLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0280, 0x300a, 0x0280,
      EVR_SQ, "ChannelSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0282, 0x300a, 0x0282,
      EVR_IS, "ChannelNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0284, 0x300a, 0x0284,
      EVR_DS, "ChannelLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0286, 0x300a, 0x0286,
      EVR_DS, "ChannelTotalTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0288, 0x300a, 0x0288,
      EVR_CS, "SourceMovementType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x028a, 0x300a, 0x028a,
      EVR_IS, "NumberOfPulses", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x028c, 0x300a, 0x028c,
      EVR_DS, "PulseRepetitionInterval", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0290, 0x300a, 0x0290,
      EVR_IS, "SourceApplicatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0291, 0x300a, 0x0291,
      EVR_SH, "SourceApplicatorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0292, 0x300a, 0x0292,
      EVR_CS, "SourceApplicatorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0294, 0x300a, 0x0294,
      EVR_LO, "SourceApplicatorName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0296, 0x300a, 0x0296,
      EVR_DS, "SourceApplicatorLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0298, 0x300a, 0x0298,
      EVR_LO, "SourceApplicatorManufacturer", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x029c, 0x300a, 0x029c,
      EVR_DS, "SourceApplicatorWallNominalThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x029e, 0x300a, 0x029e,
      EVR_DS, "SourceApplicatorWallNominalTransmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02a0, 0x300a, 0x02a0,
      EVR_DS, "SourceApplicatorStepSize", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02a1, 0x300a, 0x02a1,
      EVR_IS, "ApplicatorShapeReferencedROINumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02a2, 0x300a, 0x02a2,
      EVR_IS, "TransferTubeNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02a4, 0x300a, 0x02a4,
      EVR_DS, "TransferTubeLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02b0, 0x300a, 0x02b0,
      EVR_SQ, "ChannelShieldSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02b2, 0x300a, 0x02b2,
      EVR_IS, "ChannelShieldNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02b3, 0x300a, 0x02b3,
      EVR_SH, "ChannelShieldID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02b4, 0x300a, 0x02b4,
      EVR_LO, "ChannelShieldName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02b8, 0x300a, 0x02b8,
      EVR_DS, "ChannelShieldNominalThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02ba, 0x300a, 0x02ba,
      EVR_DS, "ChannelShieldNominalTransmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02c8, 0x300a, 0x02c8,
      EVR_DS, "FinalCumulativeTimeWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02d0, 0x300a, 0x02d0,
      EVR_SQ, "BrachyControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02d2, 0x300a, 0x02d2,
      EVR_DS, "ControlPointRelativePosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02d4, 0x300a, 0x02d4,
      EVR_DS, "ControlPoint3DPosition", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02d6, 0x300a, 0x02d6,
      EVR_DS, "CumulativeTimeWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e0, 0x300a, 0x02e0,
      EVR_CS, "CompensatorDivergence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e1, 0x300a, 0x02e1,
      EVR_CS, "CompensatorMountingPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e2, 0x300a, 0x02e2,
      EVR_DS, "SourceToCompensatorDistance", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e3, 0x300a, 0x02e3,
      EVR_FL, "TotalCompensatorTrayWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e4, 0x300a, 0x02e4,
      EVR_FL, "IsocenterToCompensatorTrayDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e5, 0x300a, 0x02e5,
      EVR_FL, "CompensatorColumnOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e6, 0x300a, 0x02e6,
      EVR_FL, "IsocenterToCompensatorDistances", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e7, 0x300a, 0x02e7,
      EVR_FL, "CompensatorRelativeStoppingPowerRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02e8, 0x300a, 0x02e8,
      EVR_FL, "CompensatorMillingToolDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02ea, 0x300a, 0x02ea,
      EVR_SQ, "IonRangeCompensatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02eb, 0x300a, 0x02eb,
      EVR_LT, "CompensatorDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x02ec, 0x300a, 0x02ec,
      EVR_CS, "CompensatorSurfaceRepresentationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0302, 0x300a, 0x0302,
      EVR_IS, "RadiationMassNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0304, 0x300a, 0x0304,
      EVR_IS, "RadiationAtomicNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0306, 0x300a, 0x0306,
      EVR_SS, "RadiationChargeState", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0308, 0x300a, 0x0308,
      EVR_CS, "ScanMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0309, 0x300a, 0x0309,
      EVR_CS, "ModulatedScanModeType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x030a, 0x300a, 0x030a,
      EVR_FL, "VirtualSourceAxisDistances", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x030c, 0x300a, 0x030c,
      EVR_SQ, "SnoutSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x030d, 0x300a, 0x030d,
      EVR_FL, "SnoutPosition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x030f, 0x300a, 0x030f,
      EVR_SH, "SnoutID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0312, 0x300a, 0x0312,
      EVR_IS, "NumberOfRangeShifters", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0314, 0x300a, 0x0314,
      EVR_SQ, "RangeShifterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0316, 0x300a, 0x0316,
      EVR_IS, "RangeShifterNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0318, 0x300a, 0x0318,
      EVR_SH, "RangeShifterID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0320, 0x300a, 0x0320,
      EVR_CS, "RangeShifterType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0322, 0x300a, 0x0322,
      EVR_LO, "RangeShifterDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0330, 0x300a, 0x0330,
      EVR_IS, "NumberOfLateralSpreadingDevices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0332, 0x300a, 0x0332,
      EVR_SQ, "LateralSpreadingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0334, 0x300a, 0x0334,
      EVR_IS, "LateralSpreadingDeviceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0336, 0x300a, 0x0336,
      EVR_SH, "LateralSpreadingDeviceID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0338, 0x300a, 0x0338,
      EVR_CS, "LateralSpreadingDeviceType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x033a, 0x300a, 0x033a,
      EVR_LO, "LateralSpreadingDeviceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x033c, 0x300a, 0x033c,
      EVR_FL, "LateralSpreadingDeviceWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0340, 0x300a, 0x0340,
      EVR_IS, "NumberOfRangeModulators", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0342, 0x300a, 0x0342,
      EVR_SQ, "RangeModulatorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0344, 0x300a, 0x0344,
      EVR_IS, "RangeModulatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0346, 0x300a, 0x0346,
      EVR_SH, "RangeModulatorID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0348, 0x300a, 0x0348,
      EVR_CS, "RangeModulatorType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x034a, 0x300a, 0x034a,
      EVR_LO, "RangeModulatorDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x034c, 0x300a, 0x034c,
      EVR_SH, "BeamCurrentModulationID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0350, 0x300a, 0x0350,
      EVR_CS, "PatientSupportType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0352, 0x300a, 0x0352,
      EVR_SH, "PatientSupportID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0354, 0x300a, 0x0354,
      EVR_LO, "PatientSupportAccessoryCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0355, 0x300a, 0x0355,
      EVR_LO, "TrayAccessoryCode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0356, 0x300a, 0x0356,
      EVR_FL, "FixationLightAzimuthalAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0358, 0x300a, 0x0358,
      EVR_FL, "FixationLightPolarAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x035a, 0x300a, 0x035a,
      EVR_FL, "MetersetRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0360, 0x300a, 0x0360,
      EVR_SQ, "RangeShifterSettingsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0362, 0x300a, 0x0362,
      EVR_LO, "RangeShifterSetting", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0364, 0x300a, 0x0364,
      EVR_FL, "IsocenterToRangeShifterDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0366, 0x300a, 0x0366,
      EVR_FL, "RangeShifterWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0370, 0x300a, 0x0370,
      EVR_SQ, "LateralSpreadingDeviceSettingsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0372, 0x300a, 0x0372,
      EVR_LO, "LateralSpreadingDeviceSetting", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0374, 0x300a, 0x0374,
      EVR_FL, "IsocenterToLateralSpreadingDeviceDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0380, 0x300a, 0x0380,
      EVR_SQ, "RangeModulatorSettingsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0382, 0x300a, 0x0382,
      EVR_FL, "RangeModulatorGatingStartValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0384, 0x300a, 0x0384,
      EVR_FL, "RangeModulatorGatingStopValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0386, 0x300a, 0x0386,
      EVR_FL, "RangeModulatorGatingStartWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0388, 0x300a, 0x0388,
      EVR_FL, "RangeModulatorGatingStopWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x038a, 0x300a, 0x038a,
      EVR_FL, "IsocenterToRangeModulatorDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x038f, 0x300a, 0x038f,
      EVR_FL, "ScanSpotTimeOffset", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0390, 0x300a, 0x0390,
      EVR_SH, "ScanSpotTuneID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0391, 0x300a, 0x0391,
      EVR_IS, "ScanSpotPrescribedIndices", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0392, 0x300a, 0x0392,
      EVR_IS, "NumberOfScanSpotPositions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0393, 0x300a, 0x0393,
      EVR_CS, "ScanSpotReordered", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0394, 0x300a, 0x0394,
      EVR_FL, "ScanSpotPositionMap", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0395, 0x300a, 0x0395,
      EVR_CS, "ScanSpotReorderingAllowed", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0396, 0x300a, 0x0396,
      EVR_FL, "ScanSpotMetersetWeights", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0398, 0x300a, 0x0398,
      EVR_FL, "ScanningSpotSize", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0399, 0x300a, 0x0399,
      EVR_FL, "ScanSpotSizesDelivered", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x039a, 0x300a, 0x039a,
      EVR_IS, "NumberOfPaintings", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03a0, 0x300a, 0x03a0,
      EVR_SQ, "IonToleranceTableSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03a2, 0x300a, 0x03a2,
      EVR_SQ, "IonBeamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03a4, 0x300a, 0x03a4,
      EVR_SQ, "IonBeamLimitingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03a6, 0x300a, 0x03a6,
      EVR_SQ, "IonBlockSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03a8, 0x300a, 0x03a8,
      EVR_SQ, "IonControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03aa, 0x300a, 0x03aa,
      EVR_SQ, "IonWedgeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x03ac, 0x300a, 0x03ac,
      EVR_SQ, "IonWedgePositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0401, 0x300a, 0x0401,
      EVR_SQ, "ReferencedSetupImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0402, 0x300a, 0x0402,
      EVR_ST, "SetupImageComment", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0410, 0x300a, 0x0410,
      EVR_SQ, "MotionSynchronizationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0412, 0x300a, 0x0412,
      EVR_FL, "ControlPointOrientation", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0420, 0x300a, 0x0420,
      EVR_SQ, "GeneralAccessorySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0421, 0x300a, 0x0421,
      EVR_SH, "GeneralAccessoryID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0422, 0x300a, 0x0422,
      EVR_ST, "GeneralAccessoryDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0423, 0x300a, 0x0423,
      EVR_CS, "GeneralAccessoryType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0424, 0x300a, 0x0424,
      EVR_IS, "GeneralAccessoryNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0425, 0x300a, 0x0425,
      EVR_FL, "SourceToGeneralAccessoryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0426, 0x300a, 0x0426,
      EVR_DS, "IsocenterToGeneralAccessoryDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0431, 0x300a, 0x0431,
      EVR_SQ, "ApplicatorGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0432, 0x300a, 0x0432,
      EVR_CS, "ApplicatorApertureShape", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0433, 0x300a, 0x0433,
      EVR_FL, "ApplicatorOpening", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0434, 0x300a, 0x0434,
      EVR_FL, "ApplicatorOpeningX", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0435, 0x300a, 0x0435,
      EVR_FL, "ApplicatorOpeningY", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0436, 0x300a, 0x0436,
      EVR_FL, "SourceToApplicatorMountingPositionDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0440, 0x300a, 0x0440,
      EVR_IS, "NumberOfBlockSlabItems", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0441, 0x300a, 0x0441,
      EVR_SQ, "BlockSlabSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0442, 0x300a, 0x0442,
      EVR_DS, "BlockSlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0443, 0x300a, 0x0443,
      EVR_US, "BlockSlabNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0450, 0x300a, 0x0450,
      EVR_SQ, "DeviceMotionControlSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0451, 0x300a, 0x0451,
      EVR_CS, "DeviceMotionExecutionMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0452, 0x300a, 0x0452,
      EVR_CS, "DeviceMotionObservationMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0453, 0x300a, 0x0453,
      EVR_SQ, "DeviceMotionParameterCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0501, 0x300a, 0x0501,
      EVR_FL, "DistalDepthFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0502, 0x300a, 0x0502,
      EVR_FL, "DistalDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0503, 0x300a, 0x0503,
      EVR_FL, "NominalRangeModulationFractions", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0504, 0x300a, 0x0504,
      EVR_FL, "NominalRangeModulatedRegionDepths", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0505, 0x300a, 0x0505,
      EVR_SQ, "DepthDoseParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0506, 0x300a, 0x0506,
      EVR_SQ, "DeliveredDepthDoseParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0507, 0x300a, 0x0507,
      EVR_FL, "DeliveredDistalDepthFraction", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0508, 0x300a, 0x0508,
      EVR_FL, "DeliveredDistalDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0509, 0x300a, 0x0509,
      EVR_FL, "DeliveredNominalRangeModulationFractions", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0510, 0x300a, 0x0510,
      EVR_FL, "DeliveredNominalRangeModulatedRegionDepths", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0511, 0x300a, 0x0511,
      EVR_CS, "DeliveredReferenceDoseDefinition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0512, 0x300a, 0x0512,
      EVR_CS, "ReferenceDoseDefinition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0600, 0x300a, 0x0600,
      EVR_US, "RTControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0601, 0x300a, 0x0601,
      EVR_US, "RadiationGenerationModeIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0602, 0x300a, 0x0602,
      EVR_US, "ReferencedDefinedDeviceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0603, 0x300a, 0x0603,
      EVR_US, "RadiationDoseIdentificationIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0604, 0x300a, 0x0604,
      EVR_US, "NumberOfRTControlPoints", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0605, 0x300a, 0x0605,
      EVR_US, "ReferencedRadiationGenerationModeIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0606, 0x300a, 0x0606,
      EVR_US, "TreatmentPositionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0607, 0x300a, 0x0607,
      EVR_US, "ReferencedDeviceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0608, 0x300a, 0x0608,
      EVR_LO, "TreatmentPositionGroupLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0609, 0x300a, 0x0609,
      EVR_UI, "TreatmentPositionGroupUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060a, 0x300a, 0x060a,
      EVR_SQ, "TreatmentPositionGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060b, 0x300a, 0x060b,
      EVR_US, "ReferencedTreatmentPositionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060c, 0x300a, 0x060c,
      EVR_US, "ReferencedRadiationDoseIdentificationIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060d, 0x300a, 0x060d,
      EVR_FD, "RTAccessoryHolderWaterEquivalentThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060e, 0x300a, 0x060e,
      EVR_US, "ReferencedRTAccessoryHolderDeviceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x060f, 0x300a, 0x060f,
      EVR_CS, "RTAccessoryHolderSlotExistenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0610, 0x300a, 0x0610,
      EVR_SQ, "RTAccessoryHolderSlotSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0611, 0x300a, 0x0611,
      EVR_LO, "RTAccessoryHolderSlotID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0612, 0x300a, 0x0612,
      EVR_FD, "RTAccessoryHolderSlotDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0613, 0x300a, 0x0613,
      EVR_FD, "RTAccessorySlotDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0614, 0x300a, 0x0614,
      EVR_SQ, "RTAccessoryHolderDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0615, 0x300a, 0x0615,
      EVR_LO, "RTAccessoryDeviceSlotID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0616, 0x300a, 0x0616,
      EVR_SQ, "RTRadiationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0617, 0x300a, 0x0617,
      EVR_SQ, "RadiationDoseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0618, 0x300a, 0x0618,
      EVR_SQ, "RadiationDoseIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0619, 0x300a, 0x0619,
      EVR_LO, "RadiationDoseIdentificationLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061a, 0x300a, 0x061a,
      EVR_CS, "ReferenceDoseType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061b, 0x300a, 0x061b,
      EVR_CS, "PrimaryDoseValueIndicator", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061c, 0x300a, 0x061c,
      EVR_SQ, "DoseValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061d, 0x300a, 0x061d,
      EVR_CS, "DoseValuePurpose", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061e, 0x300a, 0x061e,
      EVR_FD, "ReferenceDosePointCoordinates", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x061f, 0x300a, 0x061f,
      EVR_SQ, "RadiationDoseValuesParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0620, 0x300a, 0x0620,
      EVR_SQ, "MetersetToDoseMappingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0621, 0x300a, 0x0621,
      EVR_SQ, "ExpectedInVivoMeasurementValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0622, 0x300a, 0x0622,
      EVR_US, "ExpectedInVivoMeasurementValueIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0623, 0x300a, 0x0623,
      EVR_LO, "RadiationDoseInVivoMeasurementLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0624, 0x300a, 0x0624,
      EVR_FD, "RadiationDoseCentralAxisDisplacement", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0625, 0x300a, 0x0625,
      EVR_FD, "RadiationDoseValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0626, 0x300a, 0x0626,
      EVR_FD, "RadiationDoseSourceToSkinDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0627, 0x300a, 0x0627,
      EVR_FD, "RadiationDoseMeasurementPointCoordinates", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0628, 0x300a, 0x0628,
      EVR_FD, "RadiationDoseSourceToExternalContourDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0629, 0x300a, 0x0629,
      EVR_SQ, "RTToleranceSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062a, 0x300a, 0x062a,
      EVR_LO, "RTToleranceSetLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062b, 0x300a, 0x062b,
      EVR_SQ, "AttributeToleranceValuesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062c, 0x300a, 0x062c,
      EVR_FD, "ToleranceValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062d, 0x300a, 0x062d,
      EVR_SQ, "PatientSupportPositionToleranceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062e, 0x300a, 0x062e,
      EVR_FD, "TreatmentTimeLimit", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x062f, 0x300a, 0x062f,
      EVR_SQ, "CArmPhotonElectronControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0630, 0x300a, 0x0630,
      EVR_SQ, "ReferencedRTRadiationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0631, 0x300a, 0x0631,
      EVR_SQ, "ReferencedRTInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0632, 0x300a, 0x0632,
      EVR_SQ, "RETIRED_ReferencedRTPatientSetupSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0634, 0x300a, 0x0634,
      EVR_FD, "SourceToPatientSurfaceDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0635, 0x300a, 0x0635,
      EVR_SQ, "TreatmentMachineSpecialModeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0636, 0x300a, 0x0636,
      EVR_US, "IntendedNumberOfFractions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0637, 0x300a, 0x0637,
      EVR_CS, "RTRadiationSetIntent", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0638, 0x300a, 0x0638,
      EVR_CS, "RTRadiationPhysicalAndGeometricContentDetailFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0639, 0x300a, 0x0639,
      EVR_CS, "RTRecordFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063a, 0x300a, 0x063a,
      EVR_SQ, "TreatmentDeviceIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063b, 0x300a, 0x063b,
      EVR_SQ, "ReferencedRTPhysicianIntentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063c, 0x300a, 0x063c,
      EVR_FD, "CumulativeMeterset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063d, 0x300a, 0x063d,
      EVR_FD, "DeliveryRate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063e, 0x300a, 0x063e,
      EVR_SQ, "DeliveryRateUnitSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x063f, 0x300a, 0x063f,
      EVR_SQ, "TreatmentPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0640, 0x300a, 0x0640,
      EVR_FD, "RadiationSourceAxisDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0641, 0x300a, 0x0641,
      EVR_US, "NumberOfRTBeamLimitingDevices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0642, 0x300a, 0x0642,
      EVR_FD, "RTBeamLimitingDeviceProximalDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0643, 0x300a, 0x0643,
      EVR_FD, "RTBeamLimitingDeviceDistalDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0644, 0x300a, 0x0644,
      EVR_SQ, "ParallelRTBeamDelimiterDeviceOrientationLabelCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0645, 0x300a, 0x0645,
      EVR_FD, "BeamModifierOrientationAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0646, 0x300a, 0x0646,
      EVR_SQ, "FixedRTBeamDelimiterDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0647, 0x300a, 0x0647,
      EVR_SQ, "ParallelRTBeamDelimiterDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0648, 0x300a, 0x0648,
      EVR_US, "NumberOfParallelRTBeamDelimiters", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0649, 0x300a, 0x0649,
      EVR_FD, "ParallelRTBeamDelimiterBoundaries", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064a, 0x300a, 0x064a,
      EVR_FD, "ParallelRTBeamDelimiterPositions", 2, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064b, 0x300a, 0x064b,
      EVR_FD, "RTBeamLimitingDeviceOffset", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064c, 0x300a, 0x064c,
      EVR_SQ, "RTBeamDelimiterGeometrySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064d, 0x300a, 0x064d,
      EVR_SQ, "RTBeamLimitingDeviceDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064e, 0x300a, 0x064e,
      EVR_CS, "ParallelRTBeamDelimiterOpeningMode", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x064f, 0x300a, 0x064f,
      EVR_CS, "ParallelRTBeamDelimiterLeafMountingSide", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0650, 0x300a, 0x0650,
      EVR_UI, "RETIRED_PatientSetupUID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0651, 0x300a, 0x0651,
      EVR_SQ, "WedgeDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0652, 0x300a, 0x0652,
      EVR_FD, "RadiationBeamWedgeAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0653, 0x300a, 0x0653,
      EVR_FD, "RadiationBeamWedgeThinEdgeDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0654, 0x300a, 0x0654,
      EVR_FD, "RadiationBeamEffectiveWedgeAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0655, 0x300a, 0x0655,
      EVR_US, "NumberOfWedgePositions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0656, 0x300a, 0x0656,
      EVR_SQ, "RTBeamLimitingDeviceOpeningSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0657, 0x300a, 0x0657,
      EVR_US, "NumberOfRTBeamLimitingDeviceOpenings", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0658, 0x300a, 0x0658,
      EVR_SQ, "RadiationDosimeterUnitSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0659, 0x300a, 0x0659,
      EVR_SQ, "RTDeviceDistanceReferenceLocationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065a, 0x300a, 0x065a,
      EVR_SQ, "RadiationDeviceConfigurationAndCommissioningKeySequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065b, 0x300a, 0x065b,
      EVR_SQ, "PatientSupportPositionParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065c, 0x300a, 0x065c,
      EVR_CS, "PatientSupportPositionSpecificationMethod", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065d, 0x300a, 0x065d,
      EVR_SQ, "PatientSupportPositionDeviceParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065e, 0x300a, 0x065e,
      EVR_US, "DeviceOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x065f, 0x300a, 0x065f,
      EVR_US, "PatientSupportPositionParameterOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0660, 0x300a, 0x0660,
      EVR_SQ, "PatientSupportPositionDeviceToleranceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0661, 0x300a, 0x0661,
      EVR_US, "PatientSupportPositionToleranceOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0662, 0x300a, 0x0662,
      EVR_SQ, "CompensatorDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0663, 0x300a, 0x0663,
      EVR_CS, "CompensatorMapOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0664, 0x300a, 0x0664,
      EVR_OF, "CompensatorProximalThicknessMap", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0665, 0x300a, 0x0665,
      EVR_OF, "CompensatorDistalThicknessMap", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0666, 0x300a, 0x0666,
      EVR_FD, "CompensatorBasePlaneOffset", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0667, 0x300a, 0x0667,
      EVR_SQ, "CompensatorShapeFabricationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0668, 0x300a, 0x0668,
      EVR_SQ, "CompensatorShapeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0669, 0x300a, 0x0669,
      EVR_FD, "RadiationBeamCompensatorMillingToolDiameter", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066a, 0x300a, 0x066a,
      EVR_SQ, "BlockDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066b, 0x300a, 0x066b,
      EVR_OF, "BlockEdgeData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066c, 0x300a, 0x066c,
      EVR_CS, "BlockOrientation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066d, 0x300a, 0x066d,
      EVR_FD, "RadiationBeamBlockThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066e, 0x300a, 0x066e,
      EVR_FD, "RadiationBeamBlockSlabThickness", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x066f, 0x300a, 0x066f,
      EVR_SQ, "BlockEdgeDataSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0670, 0x300a, 0x0670,
      EVR_US, "NumberOfRTAccessoryHolders", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0671, 0x300a, 0x0671,
      EVR_SQ, "GeneralAccessoryDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0672, 0x300a, 0x0672,
      EVR_US, "NumberOfGeneralAccessories", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0673, 0x300a, 0x0673,
      EVR_SQ, "BolusDefinitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0674, 0x300a, 0x0674,
      EVR_US, "NumberOfBoluses", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0675, 0x300a, 0x0675,
      EVR_UI, "EquipmentFrameOfReferenceUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0676, 0x300a, 0x0676,
      EVR_ST, "EquipmentFrameOfReferenceDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0677, 0x300a, 0x0677,
      EVR_SQ, "EquipmentReferencePointCoordinatesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0678, 0x300a, 0x0678,
      EVR_SQ, "EquipmentReferencePointCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0679, 0x300a, 0x0679,
      EVR_FD, "RTBeamLimitingDeviceAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067a, 0x300a, 0x067a,
      EVR_FD, "SourceRollAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067b, 0x300a, 0x067b,
      EVR_SQ, "RadiationGenerationModeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067c, 0x300a, 0x067c,
      EVR_SH, "RadiationGenerationModeLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067d, 0x300a, 0x067d,
      EVR_ST, "RadiationGenerationModeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067e, 0x300a, 0x067e,
      EVR_SQ, "RadiationGenerationModeMachineCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x067f, 0x300a, 0x067f,
      EVR_SQ, "RadiationTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0680, 0x300a, 0x0680,
      EVR_DS, "NominalEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0681, 0x300a, 0x0681,
      EVR_DS, "MinimumNominalEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0682, 0x300a, 0x0682,
      EVR_DS, "MaximumNominalEnergy", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0683, 0x300a, 0x0683,
      EVR_SQ, "RadiationFluenceModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0684, 0x300a, 0x0684,
      EVR_SQ, "EnergyUnitCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0685, 0x300a, 0x0685,
      EVR_US, "NumberOfRadiationGenerationModes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0686, 0x300a, 0x0686,
      EVR_SQ, "PatientSupportDevicesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0687, 0x300a, 0x0687,
      EVR_US, "NumberOfPatientSupportDevices", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0688, 0x300a, 0x0688,
      EVR_FD, "RTBeamModifierDefinitionDistance", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0689, 0x300a, 0x0689,
      EVR_SQ, "BeamAreaLimitSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x068a, 0x300a, 0x068a,
      EVR_SQ, "ReferencedRTPrescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x068b, 0x300a, 0x068b,
      EVR_CS, "DoseValueInterpretation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0700, 0x300a, 0x0700,
      EVR_UI, "TreatmentSessionUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0701, 0x300a, 0x0701,
      EVR_CS, "RTRadiationUsage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0702, 0x300a, 0x0702,
      EVR_SQ, "ReferencedRTRadiationSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0703, 0x300a, 0x0703,
      EVR_SQ, "ReferencedRTRadiationRecordSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0704, 0x300a, 0x0704,
      EVR_US, "RTRadiationSetDeliveryNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0705, 0x300a, 0x0705,
      EVR_US, "ClinicalFractionNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0706, 0x300a, 0x0706,
      EVR_CS, "RTTreatmentFractionCompletionStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0707, 0x300a, 0x0707,
      EVR_CS, "RTRadiationSetUsage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0708, 0x300a, 0x0708,
      EVR_CS, "TreatmentDeliveryContinuationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0709, 0x300a, 0x0709,
      EVR_CS, "TreatmentRecordContentOrigin", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0714, 0x300a, 0x0714,
      EVR_CS, "RTTreatmentTerminationStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0715, 0x300a, 0x0715,
      EVR_SQ, "RTTreatmentTerminationReasonCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0716, 0x300a, 0x0716,
      EVR_SQ, "MachineSpecificTreatmentTerminationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0722, 0x300a, 0x0722,
      EVR_SQ, "RTRadiationSalvageRecordControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0723, 0x300a, 0x0723,
      EVR_CS, "StartingMetersetValueKnownFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0730, 0x300a, 0x0730,
      EVR_ST, "TreatmentTerminationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0731, 0x300a, 0x0731,
      EVR_SQ, "TreatmentToleranceViolationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0732, 0x300a, 0x0732,
      EVR_CS, "TreatmentToleranceViolationCategory", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0733, 0x300a, 0x0733,
      EVR_SQ, "TreatmentToleranceViolationAttributeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0734, 0x300a, 0x0734,
      EVR_ST, "TreatmentToleranceViolationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0735, 0x300a, 0x0735,
      EVR_ST, "TreatmentToleranceViolationIdentification", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0736, 0x300a, 0x0736,
      EVR_DT, "TreatmentToleranceViolationDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x073a, 0x300a, 0x073a,
      EVR_DT, "RecordedRTControlPointDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x073b, 0x300a, 0x073b,
      EVR_US, "ReferencedRadiationRTControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x073e, 0x300a, 0x073e,
      EVR_SQ, "AlternateValueSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x073f, 0x300a, 0x073f,
      EVR_SQ, "ConfirmationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0740, 0x300a, 0x0740,
      EVR_SQ, "InterlockSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0741, 0x300a, 0x0741,
      EVR_DT, "InterlockDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0742, 0x300a, 0x0742,
      EVR_ST, "InterlockDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0743, 0x300a, 0x0743,
      EVR_SQ, "InterlockOriginatingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0744, 0x300a, 0x0744,
      EVR_SQ, "InterlockCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0745, 0x300a, 0x0745,
      EVR_SQ, "InterlockResolutionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0746, 0x300a, 0x0746,
      EVR_SQ, "InterlockResolutionUserSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0760, 0x300a, 0x0760,
      EVR_DT, "OverrideDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0761, 0x300a, 0x0761,
      EVR_SQ, "TreatmentToleranceViolationTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0762, 0x300a, 0x0762,
      EVR_SQ, "TreatmentToleranceViolationCauseCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0772, 0x300a, 0x0772,
      EVR_SQ, "MeasuredMetersetToDoseMappingSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0773, 0x300a, 0x0773,
      EVR_US, "ReferencedExpectedInVivoMeasurementValueIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0774, 0x300a, 0x0774,
      EVR_SQ, "DoseMeasurementDeviceCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0780, 0x300a, 0x0780,
      EVR_SQ, "AdditionalParameterRecordingInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0783, 0x300a, 0x0783,
      EVR_ST, "InterlockOriginDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0784, 0x300a, 0x0784,
      EVR_SQ, "RTPatientPositionScopeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0785, 0x300a, 0x0785,
      EVR_UI, "ReferencedTreatmentPositionGroupUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0786, 0x300a, 0x0786,
      EVR_US, "RadiationOrderIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0787, 0x300a, 0x0787,
      EVR_SQ, "OmittedRadiationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0788, 0x300a, 0x0788,
      EVR_SQ, "ReasonForOmissionCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0789, 0x300a, 0x0789,
      EVR_SQ, "RTDeliveryStartPatientPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078a, 0x300a, 0x078a,
      EVR_SQ, "RTTreatmentPreparationPatientPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078b, 0x300a, 0x078b,
      EVR_SQ, "ReferencedRTTreatmentPreparationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078c, 0x300a, 0x078c,
      EVR_SQ, "ReferencedPatientSetupPhotoSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078d, 0x300a, 0x078d,
      EVR_SQ, "PatientTreatmentPreparationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078e, 0x300a, 0x078e,
      EVR_LT, "PatientTreatmentPreparationProcedureParameterDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x078f, 0x300a, 0x078f,
      EVR_SQ, "PatientTreatmentPreparationDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0790, 0x300a, 0x0790,
      EVR_SQ, "PatientTreatmentPreparationProcedureSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0791, 0x300a, 0x0791,
      EVR_SQ, "PatientTreatmentPreparationProcedureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0792, 0x300a, 0x0792,
      EVR_LT, "PatientTreatmentPreparationMethodDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0793, 0x300a, 0x0793,
      EVR_SQ, "PatientTreatmentPreparationProcedureParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0794, 0x300a, 0x0794,
      EVR_LT, "PatientSetupPhotoDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0795, 0x300a, 0x0795,
      EVR_US, "PatientTreatmentPreparationProcedureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0796, 0x300a, 0x0796,
      EVR_US, "ReferencedPatientSetupProcedureIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0797, 0x300a, 0x0797,
      EVR_SQ, "RTRadiationTaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0798, 0x300a, 0x0798,
      EVR_SQ, "RTPatientPositionDisplacementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x0799, 0x300a, 0x0799,
      EVR_SQ, "RTPatientPositionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079a, 0x300a, 0x079a,
      EVR_LO, "DisplacementReferenceLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079b, 0x300a, 0x079b,
      EVR_FD, "DisplacementMatrix", 16, 16, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079c, 0x300a, 0x079c,
      EVR_SQ, "PatientSupportDisplacementSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079d, 0x300a, 0x079d,
      EVR_SQ, "DisplacementReferenceLocationCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079e, 0x300a, 0x079e,
      EVR_CS, "RTRadiationSetDeliveryUsage", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x079f, 0x300a, 0x079f,
      EVR_SQ, "PatientTreatmentPreparationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x07a0, 0x300a, 0x07a0,
      EVR_SQ, "PatientToEquipmentRelationshipSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300a, 0x07a1, 0x300a, 0x07a1,
      EVR_SQ, "ImagingEquipmentToTreatmentDeliveryDeviceRelationshipSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0002, 0x300c, 0x0002,
      EVR_SQ, "ReferencedRTPlanSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0004, 0x300c, 0x0004,
      EVR_SQ, "ReferencedBeamSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0006, 0x300c, 0x0006,
      EVR_IS, "ReferencedBeamNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0007, 0x300c, 0x0007,
      EVR_IS, "ReferencedReferenceImageNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0008, 0x300c, 0x0008,
      EVR_DS, "StartCumulativeMetersetWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0009, 0x300c, 0x0009,
      EVR_DS, "EndCumulativeMetersetWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x000a, 0x300c, 0x000a,
      EVR_SQ, "ReferencedBrachyApplicationSetupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x000c, 0x300c, 0x000c,
      EVR_IS, "ReferencedBrachyApplicationSetupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x000e, 0x300c, 0x000e,
      EVR_IS, "ReferencedSourceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0020, 0x300c, 0x0020,
      EVR_SQ, "ReferencedFractionGroupSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0022, 0x300c, 0x0022,
      EVR_IS, "ReferencedFractionGroupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0040, 0x300c, 0x0040,
      EVR_SQ, "ReferencedVerificationImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0042, 0x300c, 0x0042,
      EVR_SQ, "ReferencedReferenceImageSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0050, 0x300c, 0x0050,
      EVR_SQ, "ReferencedDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0051, 0x300c, 0x0051,
      EVR_IS, "ReferencedDoseReferenceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0055, 0x300c, 0x0055,
      EVR_SQ, "BrachyReferencedDoseReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0060, 0x300c, 0x0060,
      EVR_SQ, "ReferencedStructureSetSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x006a, 0x300c, 0x006a,
      EVR_IS, "ReferencedPatientSetupNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0080, 0x300c, 0x0080,
      EVR_SQ, "ReferencedDoseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00a0, 0x300c, 0x00a0,
      EVR_IS, "ReferencedToleranceTableNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00b0, 0x300c, 0x00b0,
      EVR_SQ, "ReferencedBolusSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00c0, 0x300c, 0x00c0,
      EVR_IS, "ReferencedWedgeNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00d0, 0x300c, 0x00d0,
      EVR_IS, "ReferencedCompensatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00e0, 0x300c, 0x00e0,
      EVR_IS, "ReferencedBlockNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00f0, 0x300c, 0x00f0,
      EVR_IS, "ReferencedControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00f2, 0x300c, 0x00f2,
      EVR_SQ, "ReferencedControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00f4, 0x300c, 0x00f4,
      EVR_IS, "ReferencedStartControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x00f6, 0x300c, 0x00f6,
      EVR_IS, "ReferencedStopControlPointIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0100, 0x300c, 0x0100,
      EVR_IS, "ReferencedRangeShifterNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0102, 0x300c, 0x0102,
      EVR_IS, "ReferencedLateralSpreadingDeviceNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0104, 0x300c, 0x0104,
      EVR_IS, "ReferencedRangeModulatorNumber", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0111, 0x300c, 0x0111,
      EVR_SQ, "OmittedBeamTaskSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0112, 0x300c, 0x0112,
      EVR_CS, "ReasonForOmission", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0113, 0x300c, 0x0113,
      EVR_LO, "ReasonForOmissionDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0114, 0x300c, 0x0114,
      EVR_SQ, "PrescriptionOverviewSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0115, 0x300c, 0x0115,
      EVR_FL, "TotalPrescriptionDose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0116, 0x300c, 0x0116,
      EVR_SQ, "PlanOverviewSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0117, 0x300c, 0x0117,
      EVR_US, "PlanOverviewIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0118, 0x300c, 0x0118,
      EVR_US, "ReferencedPlanOverviewIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0119, 0x300c, 0x0119,
      EVR_US, "NumberOfFractionsIncluded", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0120, 0x300c, 0x0120,
      EVR_SQ, "DoseCalibrationConditionsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0121, 0x300c, 0x0121,
      EVR_FD, "AbsorbedDoseToMetersetRatio", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0122, 0x300c, 0x0122,
      EVR_FD, "DelineatedRadiationFieldSize", 2, 2, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0123, 0x300c, 0x0123,
      EVR_CS, "DoseCalibrationConditionsVerifiedFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0124, 0x300c, 0x0124,
      EVR_FD, "CalibrationReferencePointDepth", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0125, 0x300c, 0x0125,
      EVR_SQ, "GatingBeamHoldTransitionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0126, 0x300c, 0x0126,
      EVR_CS, "BeamHoldTransition", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0127, 0x300c, 0x0127,
      EVR_DT, "BeamHoldTransitionDateTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0128, 0x300c, 0x0128,
      EVR_SQ, "BeamHoldOriginatingDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300c, 0x0129, 0x300c, 0x0129,
      EVR_CS, "BeamHoldTransitionTriggerSource", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300e, 0x0002, 0x300e, 0x0002,
      EVR_CS, "ApprovalStatus", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300e, 0x0004, 0x300e, 0x0004,
      EVR_DA, "ReviewDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300e, 0x0005, 0x300e, 0x0005,
      EVR_TM, "ReviewTime", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x300e, 0x0008, 0x300e, 0x0008,
      EVR_PN, "ReviewerName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0001, 0x3010, 0x0001,
      EVR_SQ, "RadiobiologicalDoseEffectSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0002, 0x3010, 0x0002,
      EVR_CS, "RadiobiologicalDoseEffectFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0003, 0x3010, 0x0003,
      EVR_SQ, "EffectiveDoseCalculationMethodCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0004, 0x3010, 0x0004,
      EVR_SQ, "EffectiveDoseCalculationMethodCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0005, 0x3010, 0x0005,
      EVR_LO, "EffectiveDoseCalculationMethodDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0006, 0x3010, 0x0006,
      EVR_UI, "ConceptualVolumeUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0007, 0x3010, 0x0007,
      EVR_SQ, "OriginatingSOPInstanceReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0008, 0x3010, 0x0008,
      EVR_SQ, "ConceptualVolumeConstituentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0009, 0x3010, 0x0009,
      EVR_SQ, "EquivalentConceptualVolumeInstanceReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000a, 0x3010, 0x000a,
      EVR_SQ, "EquivalentConceptualVolumesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000b, 0x3010, 0x000b,
      EVR_UI, "ReferencedConceptualVolumeUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000c, 0x3010, 0x000c,
      EVR_UT, "ConceptualVolumeCombinationExpression", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000d, 0x3010, 0x000d,
      EVR_US, "ConceptualVolumeConstituentIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000e, 0x3010, 0x000e,
      EVR_CS, "ConceptualVolumeCombinationFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x000f, 0x3010, 0x000f,
      EVR_ST, "ConceptualVolumeCombinationDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0010, 0x3010, 0x0010,
      EVR_CS, "ConceptualVolumeSegmentationDefinedFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0011, 0x3010, 0x0011,
      EVR_SQ, "ConceptualVolumeSegmentationReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0012, 0x3010, 0x0012,
      EVR_SQ, "ConceptualVolumeConstituentSegmentationReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0013, 0x3010, 0x0013,
      EVR_UI, "ConstituentConceptualVolumeUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0014, 0x3010, 0x0014,
      EVR_SQ, "DerivationConceptualVolumeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0015, 0x3010, 0x0015,
      EVR_UI, "SourceConceptualVolumeUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0016, 0x3010, 0x0016,
      EVR_SQ, "ConceptualVolumeDerivationAlgorithmSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0017, 0x3010, 0x0017,
      EVR_ST, "ConceptualVolumeDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0018, 0x3010, 0x0018,
      EVR_SQ, "SourceConceptualVolumeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0019, 0x3010, 0x0019,
      EVR_SQ, "AuthorIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001a, 0x3010, 0x001a,
      EVR_LO, "ManufacturerModelVersion", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001b, 0x3010, 0x001b,
      EVR_UC, "DeviceAlternateIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001c, 0x3010, 0x001c,
      EVR_CS, "DeviceAlternateIdentifierType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001d, 0x3010, 0x001d,
      EVR_LT, "DeviceAlternateIdentifierFormat", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001e, 0x3010, 0x001e,
      EVR_LO, "SegmentationCreationTemplateLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x001f, 0x3010, 0x001f,
      EVR_UI, "SegmentationTemplateUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0020, 0x3010, 0x0020,
      EVR_US, "ReferencedSegmentReferenceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0021, 0x3010, 0x0021,
      EVR_SQ, "SegmentReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0022, 0x3010, 0x0022,
      EVR_US, "SegmentReferenceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0023, 0x3010, 0x0023,
      EVR_SQ, "DirectSegmentReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0024, 0x3010, 0x0024,
      EVR_SQ, "CombinationSegmentReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0025, 0x3010, 0x0025,
      EVR_SQ, "ConceptualVolumeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0026, 0x3010, 0x0026,
      EVR_SQ, "SegmentedRTAccessoryDeviceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0027, 0x3010, 0x0027,
      EVR_SQ, "SegmentCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0028, 0x3010, 0x0028,
      EVR_SQ, "RelatedSegmentCharacteristicsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0029, 0x3010, 0x0029,
      EVR_US, "SegmentCharacteristicsPrecedence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002a, 0x3010, 0x002a,
      EVR_SQ, "RTSegmentAnnotationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002b, 0x3010, 0x002b,
      EVR_SQ, "SegmentAnnotationCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002c, 0x3010, 0x002c,
      EVR_SQ, "SegmentAnnotationTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002d, 0x3010, 0x002d,
      EVR_LO, "DeviceLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002e, 0x3010, 0x002e,
      EVR_SQ, "DeviceTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x002f, 0x3010, 0x002f,
      EVR_SQ, "SegmentAnnotationTypeModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0030, 0x3010, 0x0030,
      EVR_SQ, "PatientEquipmentRelationshipCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0031, 0x3010, 0x0031,
      EVR_UI, "ReferencedFiducialsUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0032, 0x3010, 0x0032,
      EVR_SQ, "PatientTreatmentOrientationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0033, 0x3010, 0x0033,
      EVR_SH, "UserContentLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0034, 0x3010, 0x0034,
      EVR_LO, "UserContentLongLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0035, 0x3010, 0x0035,
      EVR_SH, "EntityLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0036, 0x3010, 0x0036,
      EVR_LO, "EntityName", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0037, 0x3010, 0x0037,
      EVR_ST, "EntityDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0038, 0x3010, 0x0038,
      EVR_LO, "EntityLongLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0039, 0x3010, 0x0039,
      EVR_US, "DeviceIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003a, 0x3010, 0x003a,
      EVR_US, "RTTreatmentPhaseIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003b, 0x3010, 0x003b,
      EVR_UI, "RTTreatmentPhaseUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003c, 0x3010, 0x003c,
      EVR_US, "RTPrescriptionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003d, 0x3010, 0x003d,
      EVR_US, "RTSegmentAnnotationIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003e, 0x3010, 0x003e,
      EVR_US, "BasisRTTreatmentPhaseIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x003f, 0x3010, 0x003f,
      EVR_US, "RelatedRTTreatmentPhaseIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0040, 0x3010, 0x0040,
      EVR_US, "ReferencedRTTreatmentPhaseIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0041, 0x3010, 0x0041,
      EVR_US, "ReferencedRTPrescriptionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0042, 0x3010, 0x0042,
      EVR_US, "ReferencedParentRTPrescriptionIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0043, 0x3010, 0x0043,
      EVR_ST, "ManufacturerDeviceIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0044, 0x3010, 0x0044,
      EVR_SQ, "InstanceLevelReferencedPerformedProcedureStepSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0045, 0x3010, 0x0045,
      EVR_CS, "RTTreatmentPhaseIntentPresenceFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0046, 0x3010, 0x0046,
      EVR_CS, "RadiotherapyTreatmentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0047, 0x3010, 0x0047,
      EVR_CS, "TeletherapyRadiationType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0048, 0x3010, 0x0048,
      EVR_CS, "BrachytherapySourceType", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0049, 0x3010, 0x0049,
      EVR_SQ, "ReferencedRTTreatmentPhaseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004a, 0x3010, 0x004a,
      EVR_SQ, "ReferencedDirectSegmentInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004b, 0x3010, 0x004b,
      EVR_SQ, "IntendedRTTreatmentPhaseSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004c, 0x3010, 0x004c,
      EVR_DA, "IntendedPhaseStartDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004d, 0x3010, 0x004d,
      EVR_DA, "IntendedPhaseEndDate", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004e, 0x3010, 0x004e,
      EVR_SQ, "RTTreatmentPhaseIntervalSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x004f, 0x3010, 0x004f,
      EVR_CS, "TemporalRelationshipIntervalAnchor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0050, 0x3010, 0x0050,
      EVR_FD, "MinimumNumberOfIntervalDays", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0051, 0x3010, 0x0051,
      EVR_FD, "MaximumNumberOfIntervalDays", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0052, 0x3010, 0x0052,
      EVR_UI, "PertinentSOPClassesInStudy", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0053, 0x3010, 0x0053,
      EVR_UI, "PertinentSOPClassesInSeries", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0054, 0x3010, 0x0054,
      EVR_LO, "RTPrescriptionLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0055, 0x3010, 0x0055,
      EVR_SQ, "RTPhysicianIntentPredecessorSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0056, 0x3010, 0x0056,
      EVR_LO, "RTTreatmentApproachLabel", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0057, 0x3010, 0x0057,
      EVR_SQ, "RTPhysicianIntentSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0058, 0x3010, 0x0058,
      EVR_US, "RTPhysicianIntentIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0059, 0x3010, 0x0059,
      EVR_CS, "RTTreatmentIntentType", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005a, 0x3010, 0x005a,
      EVR_UT, "RTPhysicianIntentNarrative", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005b, 0x3010, 0x005b,
      EVR_SQ, "RTProtocolCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005c, 0x3010, 0x005c,
      EVR_ST, "ReasonForSuperseding", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005d, 0x3010, 0x005d,
      EVR_SQ, "RTDiagnosisCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005e, 0x3010, 0x005e,
      EVR_US, "ReferencedRTPhysicianIntentIndex", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x005f, 0x3010, 0x005f,
      EVR_SQ, "RTPhysicianIntentInputInstanceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0060, 0x3010, 0x0060,
      EVR_SQ, "RTAnatomicPrescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0061, 0x3010, 0x0061,
      EVR_UT, "PriorTreatmentDoseDescription", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0062, 0x3010, 0x0062,
      EVR_SQ, "PriorTreatmentReferenceSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0063, 0x3010, 0x0063,
      EVR_CS, "DosimetricObjectiveEvaluationScope", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0064, 0x3010, 0x0064,
      EVR_SQ, "TherapeuticRoleCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0065, 0x3010, 0x0065,
      EVR_SQ, "TherapeuticRoleTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0066, 0x3010, 0x0066,
      EVR_US, "ConceptualVolumeOptimizationPrecedence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0067, 0x3010, 0x0067,
      EVR_SQ, "ConceptualVolumeCategoryCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0068, 0x3010, 0x0068,
      EVR_CS, "ConceptualVolumeBlockingConstraint", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0069, 0x3010, 0x0069,
      EVR_SQ, "ConceptualVolumeTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006a, 0x3010, 0x006a,
      EVR_SQ, "ConceptualVolumeTypeModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006b, 0x3010, 0x006b,
      EVR_SQ, "RTPrescriptionSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006c, 0x3010, 0x006c,
      EVR_SQ, "DosimetricObjectiveSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006d, 0x3010, 0x006d,
      EVR_SQ, "DosimetricObjectiveTypeCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006e, 0x3010, 0x006e,
      EVR_UI, "DosimetricObjectiveUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x006f, 0x3010, 0x006f,
      EVR_UI, "ReferencedDosimetricObjectiveUID", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0070, 0x3010, 0x0070,
      EVR_SQ, "DosimetricObjectiveParameterSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0071, 0x3010, 0x0071,
      EVR_SQ, "ReferencedDosimetricObjectivesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0073, 0x3010, 0x0073,
      EVR_CS, "AbsoluteDosimetricObjectiveFlag", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0074, 0x3010, 0x0074,
      EVR_FD, "DosimetricObjectiveWeight", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0075, 0x3010, 0x0075,
      EVR_CS, "DosimetricObjectivePurpose", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0076, 0x3010, 0x0076,
      EVR_SQ, "PlanningInputInformationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0077, 0x3010, 0x0077,
      EVR_LO, "TreatmentSite", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0078, 0x3010, 0x0078,
      EVR_SQ, "TreatmentSiteCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0079, 0x3010, 0x0079,
      EVR_SQ, "FractionPatternSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007a, 0x3010, 0x007a,
      EVR_UT, "TreatmentTechniqueNotes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007b, 0x3010, 0x007b,
      EVR_UT, "PrescriptionNotes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007c, 0x3010, 0x007c,
      EVR_IS, "NumberOfIntervalFractions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007d, 0x3010, 0x007d,
      EVR_US, "NumberOfFractions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007e, 0x3010, 0x007e,
      EVR_US, "IntendedDeliveryDuration", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x007f, 0x3010, 0x007f,
      EVR_UT, "FractionationNotes", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0080, 0x3010, 0x0080,
      EVR_SQ, "RTTreatmentTechniqueCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0081, 0x3010, 0x0081,
      EVR_SQ, "PrescriptionNotesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0082, 0x3010, 0x0082,
      EVR_SQ, "FractionBasedRelationshipSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0083, 0x3010, 0x0083,
      EVR_CS, "FractionBasedRelationshipIntervalAnchor", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0084, 0x3010, 0x0084,
      EVR_FD, "MinimumHoursBetweenFractions", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0085, 0x3010, 0x0085,
      EVR_TM, "IntendedFractionStartTime", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0086, 0x3010, 0x0086,
      EVR_LT, "IntendedStartDayOfWeek", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0087, 0x3010, 0x0087,
      EVR_SQ, "WeekdayFractionPatternSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0088, 0x3010, 0x0088,
      EVR_SQ, "DeliveryTimeStructureCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0089, 0x3010, 0x0089,
      EVR_SQ, "TreatmentSiteModifierCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0090, 0x3010, 0x0090,
      EVR_CS, "RETIRED_RoboticBaseLocationIndicator", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0091, 0x3010, 0x0091,
      EVR_SQ, "RoboticPathNodeSetCodeSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0092, 0x3010, 0x0092,
      EVR_UL, "RoboticNodeIdentifier", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0093, 0x3010, 0x0093,
      EVR_FD, "RTTreatmentSourceCoordinates", 3, 3, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0094, 0x3010, 0x0094,
      EVR_FD, "RadiationSourceCoordinateSystemYawAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0095, 0x3010, 0x0095,
      EVR_FD, "RadiationSourceCoordinateSystemRollAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0096, 0x3010, 0x0096,
      EVR_FD, "RadiationSourceCoordinateSystemPitchAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0097, 0x3010, 0x0097,
      EVR_SQ, "RoboticPathControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0098, 0x3010, 0x0098,
      EVR_SQ, "TomotherapeuticControlPointSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x0099, 0x3010, 0x0099,
      EVR_FD, "TomotherapeuticLeafOpenDurations", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x009a, 0x3010, 0x009a,
      EVR_FD, "TomotherapeuticLeafInitialClosedDurations", 1, -1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x3010, 0x00a0, 0x3010, 0x00a0,
      EVR_SQ, "ConceptualVolumeIdentificationSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x3f01, 0x0001, 0x3f01, 0x0001,
      EVR_LO, "InstitutionCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0002, 0x3f01, 0x0002,
      EVR_LO, "RoutedTransferAE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0003, 0x3f01, 0x0003,
      EVR_LO, "SourceAE", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0004, 0x3f01, 0x0004,
      EVR_SH, "DeferredValidation", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0005, 0x3f01, 0x0005,
      EVR_LO, "SeriesOwner", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0006, 0x3f01, 0x0006,
      EVR_LO, "OrderGroupNumber", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0007, 0x3f01, 0x0007,
      EVR_SH, "StrippedPixelData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
  , { 0x3f01, 0x0008, 0x3f01, 0x0008,
      EVR_SH, "PendingMoveRequest", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "INTELERAD MEDICAL SYSTEMS" }
#endif
  , { 0x4000, 0x0010, 0x4000, 0x0010,
      EVR_LT, "RETIRED_Arbitrary", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4000, 0x4000, 0x4000, 0x4000,
      EVR_LT, "RETIRED_TextComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0040, 0x4008, 0x0040,
      EVR_SH, "RETIRED_ResultsID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0042, 0x4008, 0x0042,
      EVR_LO, "RETIRED_ResultsIDIssuer", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0050, 0x4008, 0x0050,
      EVR_SQ, "RETIRED_ReferencedInterpretationSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x00ff, 0x4008, 0x00ff,
      EVR_CS, "RETIRED_ReportProductionStatusTrial", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0100, 0x4008, 0x0100,
      EVR_DA, "RETIRED_InterpretationRecordedDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0101, 0x4008, 0x0101,
      EVR_TM, "RETIRED_InterpretationRecordedTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0102, 0x4008, 0x0102,
      EVR_PN, "RETIRED_InterpretationRecorder", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0103, 0x4008, 0x0103,
      EVR_LO, "RETIRED_ReferenceToRecordedSound", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0108, 0x4008, 0x0108,
      EVR_DA, "RETIRED_InterpretationTranscriptionDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0109, 0x4008, 0x0109,
      EVR_TM, "RETIRED_InterpretationTranscriptionTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x010a, 0x4008, 0x010a,
      EVR_PN, "RETIRED_InterpretationTranscriber", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x010b, 0x4008, 0x010b,
      EVR_ST, "RETIRED_InterpretationText", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x010c, 0x4008, 0x010c,
      EVR_PN, "RETIRED_InterpretationAuthor", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0111, 0x4008, 0x0111,
      EVR_SQ, "RETIRED_InterpretationApproverSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0112, 0x4008, 0x0112,
      EVR_DA, "RETIRED_InterpretationApprovalDate", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0113, 0x4008, 0x0113,
      EVR_TM, "RETIRED_InterpretationApprovalTime", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0114, 0x4008, 0x0114,
      EVR_PN, "RETIRED_PhysicianApprovingInterpretation", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0115, 0x4008, 0x0115,
      EVR_LT, "RETIRED_InterpretationDiagnosisDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0117, 0x4008, 0x0117,
      EVR_SQ, "RETIRED_InterpretationDiagnosisCodeSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0118, 0x4008, 0x0118,
      EVR_SQ, "RETIRED_ResultsDistributionListSequence", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0119, 0x4008, 0x0119,
      EVR_PN, "RETIRED_DistributionName", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x011a, 0x4008, 0x011a,
      EVR_LO, "RETIRED_DistributionAddress", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0200, 0x4008, 0x0200,
      EVR_SH, "RETIRED_InterpretationID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0202, 0x4008, 0x0202,
      EVR_LO, "RETIRED_InterpretationIDIssuer", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0210, 0x4008, 0x0210,
      EVR_CS, "RETIRED_InterpretationTypeID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0212, 0x4008, 0x0212,
      EVR_CS, "RETIRED_InterpretationStatusID", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x0300, 0x4008, 0x0300,
      EVR_ST, "RETIRED_Impressions", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4008, 0x4000, 0x4008, 0x4000,
      EVR_ST, "RETIRED_ResultsComments", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x4009, 0x0001, 0x4009, 0x0001,
      EVR_LT, "ReportID", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x0020, 0x4009, 0x0020,
      EVR_LT, "ReportStatus", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x0030, 0x4009, 0x0030,
      EVR_DA, "ReportCreationDate", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x0070, 0x4009, 0x0070,
      EVR_LT, "ReportApprovingPhysician", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x00e0, 0x4009, 0x00e0,
      EVR_LT, "ReportText", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x00e1, 0x4009, 0x00e1,
      EVR_LT, "ReportAuthor", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
  , { 0x4009, 0x00e3, 0x4009, 0x00e3,
      EVR_LT, "ReportingRadiologist", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS ISI" }
#endif
  , { 0x4010, 0x0001, 0x4010, 0x0001,
      EVR_CS, "LowEnergyDetectors", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x0002, 0x4010, 0x0002,
      EVR_CS, "HighEnergyDetectors", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x0004, 0x4010, 0x0004,
      EVR_SQ, "DetectorGeometrySequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1001, 0x4010, 0x1001,
      EVR_SQ, "ThreatROIVoxelSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1004, 0x4010, 0x1004,
      EVR_FL, "ThreatROIBase", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1005, 0x4010, 0x1005,
      EVR_FL, "ThreatROIExtents", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1006, 0x4010, 0x1006,
      EVR_OB, "ThreatROIBitmap", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1007, 0x4010, 0x1007,
      EVR_SH, "RouteSegmentID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1008, 0x4010, 0x1008,
      EVR_CS, "GantryType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1009, 0x4010, 0x1009,
      EVR_CS, "OOIOwnerType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x100a, 0x4010, 0x100a,
      EVR_SQ, "RouteSegmentSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1010, 0x4010, 0x1010,
      EVR_US, "PotentialThreatObjectID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1011, 0x4010, 0x1011,
      EVR_SQ, "ThreatSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1012, 0x4010, 0x1012,
      EVR_CS, "ThreatCategory", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1013, 0x4010, 0x1013,
      EVR_LT, "ThreatCategoryDescription", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1014, 0x4010, 0x1014,
      EVR_CS, "ATDAbilityAssessment", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1015, 0x4010, 0x1015,
      EVR_CS, "ATDAssessmentFlag", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1016, 0x4010, 0x1016,
      EVR_FL, "ATDAssessmentProbability", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1017, 0x4010, 0x1017,
      EVR_FL, "Mass", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1018, 0x4010, 0x1018,
      EVR_FL, "Density", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1019, 0x4010, 0x1019,
      EVR_FL, "ZEffective", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101a, 0x4010, 0x101a,
      EVR_SH, "BoardingPassID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101b, 0x4010, 0x101b,
      EVR_FL, "CenterOfMass", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101c, 0x4010, 0x101c,
      EVR_FL, "CenterOfPTO", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101d, 0x4010, 0x101d,
      EVR_FL, "BoundingPolygon", 6, -1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101e, 0x4010, 0x101e,
      EVR_SH, "RouteSegmentStartLocationID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x101f, 0x4010, 0x101f,
      EVR_SH, "RouteSegmentEndLocationID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1020, 0x4010, 0x1020,
      EVR_CS, "RouteSegmentLocationIDType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1021, 0x4010, 0x1021,
      EVR_CS, "AbortReason", 1, -1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1023, 0x4010, 0x1023,
      EVR_FL, "VolumeOfPTO", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1024, 0x4010, 0x1024,
      EVR_CS, "AbortFlag", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1025, 0x4010, 0x1025,
      EVR_DT, "RouteSegmentStartTime", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1026, 0x4010, 0x1026,
      EVR_DT, "RouteSegmentEndTime", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1027, 0x4010, 0x1027,
      EVR_CS, "TDRType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1028, 0x4010, 0x1028,
      EVR_CS, "InternationalRouteSegment", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1029, 0x4010, 0x1029,
      EVR_LO, "ThreatDetectionAlgorithmAndVersion", 1, -1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x102a, 0x4010, 0x102a,
      EVR_SH, "AssignedLocation", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x102b, 0x4010, 0x102b,
      EVR_DT, "AlarmDecisionTime", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1031, 0x4010, 0x1031,
      EVR_CS, "AlarmDecision", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1033, 0x4010, 0x1033,
      EVR_US, "NumberOfTotalObjects", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1034, 0x4010, 0x1034,
      EVR_US, "NumberOfAlarmObjects", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1037, 0x4010, 0x1037,
      EVR_SQ, "PTORepresentationSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1038, 0x4010, 0x1038,
      EVR_SQ, "ATDAssessmentSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1039, 0x4010, 0x1039,
      EVR_CS, "TIPType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x103a, 0x4010, 0x103a,
      EVR_CS, "DICOSVersion", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1041, 0x4010, 0x1041,
      EVR_DT, "OOIOwnerCreationTime", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1042, 0x4010, 0x1042,
      EVR_CS, "OOIType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1043, 0x4010, 0x1043,
      EVR_FL, "OOISize", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1044, 0x4010, 0x1044,
      EVR_CS, "AcquisitionStatus", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1045, 0x4010, 0x1045,
      EVR_SQ, "BasisMaterialsCodeSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1046, 0x4010, 0x1046,
      EVR_CS, "PhantomType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1047, 0x4010, 0x1047,
      EVR_SQ, "OOIOwnerSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1048, 0x4010, 0x1048,
      EVR_CS, "ScanType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1051, 0x4010, 0x1051,
      EVR_LO, "ItineraryID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1052, 0x4010, 0x1052,
      EVR_SH, "ItineraryIDType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1053, 0x4010, 0x1053,
      EVR_LO, "ItineraryIDAssigningAuthority", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1054, 0x4010, 0x1054,
      EVR_SH, "RouteID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1055, 0x4010, 0x1055,
      EVR_SH, "RouteIDAssigningAuthority", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1056, 0x4010, 0x1056,
      EVR_CS, "InboundArrivalType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1058, 0x4010, 0x1058,
      EVR_SH, "CarrierID", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1059, 0x4010, 0x1059,
      EVR_CS, "CarrierIDAssigningAuthority", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1060, 0x4010, 0x1060,
      EVR_FL, "SourceOrientation", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1061, 0x4010, 0x1061,
      EVR_FL, "SourcePosition", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1062, 0x4010, 0x1062,
      EVR_FL, "BeltHeight", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1064, 0x4010, 0x1064,
      EVR_SQ, "AlgorithmRoutingCodeSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1067, 0x4010, 0x1067,
      EVR_CS, "TransportClassification", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1068, 0x4010, 0x1068,
      EVR_LT, "OOITypeDescriptor", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1069, 0x4010, 0x1069,
      EVR_FL, "TotalProcessingTime", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x106c, 0x4010, 0x106c,
      EVR_OB, "DetectorCalibrationData", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x106d, 0x4010, 0x106d,
      EVR_CS, "AdditionalScreeningPerformed", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x106e, 0x4010, 0x106e,
      EVR_CS, "AdditionalInspectionSelectionCriteria", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x106f, 0x4010, 0x106f,
      EVR_SQ, "AdditionalInspectionMethodSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1070, 0x4010, 0x1070,
      EVR_CS, "AITDeviceType", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1071, 0x4010, 0x1071,
      EVR_SQ, "QRMeasurementsSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1072, 0x4010, 0x1072,
      EVR_SQ, "TargetMaterialSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1073, 0x4010, 0x1073,
      EVR_FD, "SNRThreshold", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1075, 0x4010, 0x1075,
      EVR_DS, "ImageScaleRepresentation", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1076, 0x4010, 0x1076,
      EVR_SQ, "ReferencedPTOSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1077, 0x4010, 0x1077,
      EVR_SQ, "ReferencedTDRInstanceSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1078, 0x4010, 0x1078,
      EVR_ST, "PTOLocationDescription", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x1079, 0x4010, 0x1079,
      EVR_SQ, "AnomalyLocatorIndicatorSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x107a, 0x4010, 0x107a,
      EVR_FL, "AnomalyLocatorIndicator", 3, 3, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x107b, 0x4010, 0x107b,
      EVR_SQ, "PTORegionSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x107c, 0x4010, 0x107c,
      EVR_CS, "InspectionSelectionCriteria", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x107d, 0x4010, 0x107d,
      EVR_SQ, "SecondaryInspectionMethodSequence", 1, 1, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4010, 0x107e, 0x4010, 0x107e,
      EVR_DS, "PRCSToRCSOrientation", 6, 6, "DICOM/DICOS",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x4ffe, 0x0001, 0x4ffe, 0x0001,
      EVR_SQ, "MACParametersSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x50f1, 0x0006, 0x50f1, 0x0006,
      EVR_CS, "EnergySubtractionParam", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x50f1, 0x0007, 0x50f1, 0x0007,
      EVR_CS, "SubtractionRegistrationResult", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x50f1, 0x0008, 0x50f1, 0x0008,
      EVR_CS, "EnergySubtractionParam2", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x50f1, 0x0009, 0x50f1, 0x0009,
      EVR_SL, "AfinConversionCoefficient", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x50f1, 0x0010, 0x50f1, 0x0010,
      EVR_CS, "FilmOutputFormat", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
  , { 0x50f1, 0x0020, 0x50f1, 0x0020,
      EVR_CS, "ImageProcessingModificationFlag", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "FDMS 1.0" }
#endif
  , { 0x5200, 0x9229, 0x5200, 0x9229,
      EVR_SQ, "SharedFunctionalGroupsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5200, 0x9230, 0x5200, 0x9230,
      EVR_SQ, "PerFrameFunctionalGroupsSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x0100, 0x5400, 0x0100,
      EVR_SQ, "WaveformSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x0110, 0x5400, 0x0110,
      EVR_ox, "ChannelMinimumValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x0112, 0x5400, 0x0112,
      EVR_ox, "ChannelMaximumValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x1004, 0x5400, 0x1004,
      EVR_US, "WaveformBitsAllocated", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x1006, 0x5400, 0x1006,
      EVR_CS, "WaveformSampleInterpretation", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x100a, 0x5400, 0x100a,
      EVR_ox, "WaveformPaddingValue", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5400, 0x1010, 0x5400, 0x1010,
      EVR_ox, "WaveformData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5600, 0x0010, 0x5600, 0x0010,
      EVR_OF, "FirstOrderPhaseCorrectionAngle", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x5600, 0x0020, 0x5600, 0x0020,
      EVR_OF, "SpectroscopyData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x6001, 0x0000, 0x6001, 0x0000,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "PHILIPS MR/LAST" }
  , { 0x6021, 0x0000, 0x6021, 0x0000,
      EVR_LT, "ImageGraphicsFormatCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x6021, 0x0000, 0x6021, 0x0000,
      EVR_LT, "OsteoContourComment", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  OST" }
  , { 0x6021, 0x0010, 0x6021, 0x0010,
      EVR_LT, "ImageGraphics", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x6021, 0x0010, 0x6021, 0x0010,
      EVR_US, "OsteoContourBuffer", 256, 256, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CT VA0  OST" }
  , { 0x7001, 0x0001, 0x7001, 0x0001,
      EVR_UI, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Picker NM Private Group" }
  , { 0x7001, 0x0002, 0x7001, 0x0002,
      EVR_OB, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "Picker NM Private Group" }
  , { 0x7001, 0x0010, 0x7001, 0x0010,
      EVR_LT, "Dummy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x7003, 0x0010, 0x7003, 0x0010,
      EVR_LT, "Header", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
  , { 0x7005, 0x0010, 0x7005, 0x0010,
      EVR_LT, "Dummy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS MED" }
#endif
  , { 0x7fe0, 0x0001, 0x7fe0, 0x0001,
      EVR_OV, "ExtendedOffsetTable", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0002, 0x7fe0, 0x0002,
      EVR_OV, "ExtendedOffsetTableLengths", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0003, 0x7fe0, 0x0003,
      EVR_UV, "EncapsulatedPixelDataValueTotalLength", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0008, 0x7fe0, 0x0008,
      EVR_OF, "FloatPixelData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0009, 0x7fe0, 0x0009,
      EVR_OD, "DoubleFloatPixelData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0010, 0x7fe0, 0x0010,
      EVR_px, "PixelData", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0020, 0x7fe0, 0x0020,
      EVR_OW, "RETIRED_CoefficientsSDVN", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0030, 0x7fe0, 0x0030,
      EVR_OW, "RETIRED_CoefficientsSDHN", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x7fe0, 0x0040, 0x7fe0, 0x0040,
      EVR_OW, "RETIRED_CoefficientsSDDN", 1, 1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x7fe1, 0x0000, 0x7fe1, 0x0000,
      EVR_OB, "BinaryData", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CM VA0  CMS" }
  , { 0x7fe1, 0x0010, 0x7fe1, 0x0010,
      EVR_OB, "CSAData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS CSA NON-IMAGE" }
  , { 0x7fe1, 0x0010, 0x7fe1, 0x0010,
      EVR_ox, "PixelData", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SPI-P Release 1" }
  , { 0x7fe1, 0x1060, 0x7fe1, 0x1060,
      EVR_px, "IllegalPrivatePixelSequence", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "GEMS_Ultrasound_MovieGroup_001" }
  , { 0x7fe3, 0x0000, 0x7fe3, 0x0000,
      EVR_LT, "ImageGraphicsFormatCode", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS NUMARIS II" }
  , { 0x7fe3, 0x0010, 0x7fe3, 0x0010,
      EVR_OB, "ImageGraphics", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS NUMARIS II" }
  , { 0x7fe3, 0x0020, 0x7fe3, 0x0020,
      EVR_OB, "ImageGraphicsDummy", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "SIEMENS NUMARIS II" }
  , { 0x7ff1, 0x0001, 0x7ff1, 0x0001,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x7ff1, 0x0001, 0x7ff1, 0x0001,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0002, 0x7ff1, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x7ff1, 0x0002, 0x7ff1, 0x0002,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0003, 0x7ff1, 0x0003,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
  , { 0x7ff1, 0x0003, 0x7ff1, 0x0003,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0004, 0x7ff1, 0x0004,
      EVR_IS, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0005, 0x7ff1, 0x0005,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0007, 0x7ff1, 0x0007,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0008, 0x7ff1, 0x0008,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0009, 0x7ff1, 0x0009,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x000a, 0x7ff1, 0x000a,
      EVR_LT, "Unknown", 1, 1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x000b, 0x7ff1, 0x000b,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x000c, 0x7ff1, 0x000c,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x000d, 0x7ff1, 0x000d,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_CT_1.0" }
  , { 0x7ff1, 0x0010, 0x7ff1, 0x0010,
      EVR_US, "Unknown", 1, -1, "PrivateTag",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      "TOSHIBA_MEC_1.0" }
#endif
  , { 0xfffa, 0xfffa, 0xfffa, 0xfffa,
      EVR_SQ, "DigitalSignaturesSequence", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0xfffc, 0xfffc, 0xfffc, 0xfffc,
      EVR_OB, "DataSetTrailingPadding", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0xfffe, 0xe000, 0xfffe, 0xe000,
      EVR_na, "Item", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0xfffe, 0xe00d, 0xfffe, 0xe00d,
      EVR_na, "ItemDelimitationItem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0xfffe, 0xe0dd, 0xfffe, 0xe0dd,
      EVR_na, "SequenceDelimitationItem", 1, 1, "DICOM",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0010, 0x60ff, 0x0010,
      EVR_US, "OverlayRows", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0011, 0x60ff, 0x0011,
      EVR_US, "OverlayColumns", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0015, 0x60ff, 0x0015,
      EVR_IS, "NumberOfFramesInOverlay", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0022, 0x60ff, 0x0022,
      EVR_LO, "OverlayDescription", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0040, 0x60ff, 0x0040,
      EVR_CS, "OverlayType", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0045, 0x60ff, 0x0045,
      EVR_LO, "OverlaySubtype", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0050, 0x60ff, 0x0050,
      EVR_SS, "OverlayOrigin", 2, 2, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0051, 0x60ff, 0x0051,
      EVR_US, "ImageFrameOrigin", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0100, 0x60ff, 0x0100,
      EVR_US, "OverlayBitsAllocated", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0102, 0x60ff, 0x0102,
      EVR_US, "OverlayBitPosition", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1001, 0x60ff, 0x1001,
      EVR_CS, "OverlayActivationLayer", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1301, 0x60ff, 0x1301,
      EVR_IS, "ROIArea", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1302, 0x60ff, 0x1302,
      EVR_DS, "ROIMean", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1303, 0x60ff, 0x1303,
      EVR_DS, "ROIStandardDeviation", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1500, 0x60ff, 0x1500,
      EVR_LO, "OverlayLabel", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x3000, 0x60ff, 0x3000,
      EVR_ox, "OverlayData", 1, 1, "DICOM",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x0020, 0x3100, 0x0020, 0x31ff,
      EVR_CS, "RETIRED_SourceImageIDs", 1, -1, "DICOM/retired",
      DcmDictRange_Unspecified, DcmDictRange_Even,
      NULL }
  , { 0x5000, 0x0005, 0x50ff, 0x0005,
      EVR_US, "RETIRED_CurveDimensions", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0010, 0x50ff, 0x0010,
      EVR_US, "RETIRED_NumberOfPoints", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0020, 0x50ff, 0x0020,
      EVR_CS, "RETIRED_TypeOfData", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0022, 0x50ff, 0x0022,
      EVR_LO, "RETIRED_CurveDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0030, 0x50ff, 0x0030,
      EVR_SH, "RETIRED_AxisUnits", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0040, 0x50ff, 0x0040,
      EVR_SH, "RETIRED_AxisLabels", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0103, 0x50ff, 0x0103,
      EVR_US, "RETIRED_DataValueRepresentation", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0104, 0x50ff, 0x0104,
      EVR_US, "RETIRED_MinimumCoordinateValue", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0105, 0x50ff, 0x0105,
      EVR_US, "RETIRED_MaximumCoordinateValue", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0106, 0x50ff, 0x0106,
      EVR_SH, "RETIRED_CurveRange", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0110, 0x50ff, 0x0110,
      EVR_US, "RETIRED_CurveDataDescriptor", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0112, 0x50ff, 0x0112,
      EVR_US, "RETIRED_CoordinateStartValue", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x0114, 0x50ff, 0x0114,
      EVR_US, "RETIRED_CoordinateStepValue", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x1001, 0x50ff, 0x1001,
      EVR_CS, "RETIRED_CurveActivationLayer", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2000, 0x50ff, 0x2000,
      EVR_US, "RETIRED_AudioType", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2002, 0x50ff, 0x2002,
      EVR_US, "RETIRED_AudioSampleFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2004, 0x50ff, 0x2004,
      EVR_US, "RETIRED_NumberOfChannels", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2006, 0x50ff, 0x2006,
      EVR_UL, "RETIRED_NumberOfSamples", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2008, 0x50ff, 0x2008,
      EVR_UL, "RETIRED_SampleRate", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x200a, 0x50ff, 0x200a,
      EVR_UL, "RETIRED_TotalTime", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x200c, 0x50ff, 0x200c,
      EVR_ox, "RETIRED_AudioSampleData", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x200e, 0x50ff, 0x200e,
      EVR_LT, "RETIRED_AudioComments", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2500, 0x50ff, 0x2500,
      EVR_LO, "RETIRED_CurveLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2600, 0x50ff, 0x2600,
      EVR_SQ, "RETIRED_CurveReferencedOverlaySequence", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x2610, 0x50ff, 0x2610,
      EVR_US, "RETIRED_CurveReferencedOverlayGroup", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x5000, 0x3000, 0x50ff, 0x3000,
      EVR_ox, "RETIRED_CurveData", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0012, 0x60ff, 0x0012,
      EVR_US, "RETIRED_OverlayPlanes", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0052, 0x60ff, 0x0052,
      EVR_US, "RETIRED_OverlayPlaneOrigin", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0060, 0x60ff, 0x0060,
      EVR_CS, "RETIRED_OverlayCompressionCode", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0061, 0x60ff, 0x0061,
      EVR_SH, "RETIRED_OverlayCompressionOriginator", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0062, 0x60ff, 0x0062,
      EVR_SH, "RETIRED_OverlayCompressionLabel", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0063, 0x60ff, 0x0063,
      EVR_CS, "RETIRED_OverlayCompressionDescription", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0066, 0x60ff, 0x0066,
      EVR_AT, "RETIRED_OverlayCompressionStepPointers", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0068, 0x60ff, 0x0068,
      EVR_US, "RETIRED_OverlayRepeatInterval", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0069, 0x60ff, 0x0069,
      EVR_US, "RETIRED_OverlayBitsGrouped", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0110, 0x60ff, 0x0110,
      EVR_CS, "RETIRED_OverlayFormat", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0200, 0x60ff, 0x0200,
      EVR_US, "RETIRED_OverlayLocation", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0800, 0x60ff, 0x0800,
      EVR_CS, "RETIRED_OverlayCodeLabel", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0802, 0x60ff, 0x0802,
      EVR_US, "RETIRED_OverlayNumberOfTables", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0803, 0x60ff, 0x0803,
      EVR_AT, "RETIRED_OverlayCodeTableLocation", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x0804, 0x60ff, 0x0804,
      EVR_US, "RETIRED_OverlayBitsForCodeWord", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1100, 0x60ff, 0x1100,
      EVR_US, "RETIRED_OverlayDescriptorGray", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1101, 0x60ff, 0x1101,
      EVR_US, "RETIRED_OverlayDescriptorRed", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1102, 0x60ff, 0x1102,
      EVR_US, "RETIRED_OverlayDescriptorGreen", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1103, 0x60ff, 0x1103,
      EVR_US, "RETIRED_OverlayDescriptorBlue", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1200, 0x60ff, 0x1200,
      EVR_US, "RETIRED_OverlaysGray", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1201, 0x60ff, 0x1201,
      EVR_US, "RETIRED_OverlaysRed", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1202, 0x60ff, 0x1202,
      EVR_US, "RETIRED_OverlaysGreen", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x1203, 0x60ff, 0x1203,
      EVR_US, "RETIRED_OverlaysBlue", 1, -1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x6000, 0x4000, 0x60ff, 0x4000,
      EVR_LT, "RETIRED_OverlayComments", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x7f00, 0x0010, 0x7fff, 0x0010,
      EVR_ox, "RETIRED_VariablePixelData", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x7f00, 0x0011, 0x7fff, 0x0011,
      EVR_US, "RETIRED_VariableNextDataGroup", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x7f00, 0x0020, 0x7fff, 0x0020,
      EVR_OW, "RETIRED_VariableCoefficientsSDVN", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x7f00, 0x0030, 0x7fff, 0x0030,
      EVR_OW, "RETIRED_VariableCoefficientsSDHN", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x7f00, 0x0040, 0x7fff, 0x0040,
      EVR_OW, "RETIRED_VariableCoefficientsSDDN", 1, 1, "DICOM/retired",
      DcmDictRange_Even, DcmDictRange_Unspecified,
      NULL }
  , { 0x0009, 0x0000, 0xffff, 0x0000,
      EVR_UL, "PrivateGroupLength", 1, 1, "PRIVATE",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      NULL }
  , { 0x0009, 0x0010, 0xffff, 0x00ff,
      EVR_LO, "PrivateCreator", 1, 1, "PRIVATE",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      NULL }
  , { 0x0001, 0x0000, 0x0007, 0x0000,
      EVR_UL, "IllegalGroupLength", 1, 1, "ILLEGAL",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      NULL }
  , { 0x0001, 0x0010, 0x0007, 0x00ff,
      EVR_LO, "IllegalPrivateCreator", 1, 1, "ILLEGAL",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      NULL }
  , { 0x0000, 0x0000, 0xffff, 0x0000,
      EVR_UL, "GenericGroupLength", 1, 1, "GENERIC",
      DcmDictRange_Unspecified, DcmDictRange_Unspecified,
      NULL }
#ifdef ENABLE_PRIVATE_TAGS
  , { 0x7001, 0x0004, 0x70ff, 0x0004,
      EVR_ST, "TextAnnotation", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "DLX_ANNOT_01" }
  , { 0x7001, 0x0005, 0x70ff, 0x0005,
      EVR_IS, "Box", 2, 2, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "DLX_ANNOT_01" }
  , { 0x7001, 0x0007, 0x70ff, 0x0007,
      EVR_IS, "ArrowEnd", 2, 2, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "DLX_ANNOT_01" }
  , { 0x6001, 0x0001, 0x60ff, 0x0001,
      EVR_US, "GrayPaletteColorLookupTableDescriptor", 3, 3, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "DLX_LKUP_01" }
  , { 0x6001, 0x0002, 0x60ff, 0x0002,
      EVR_US, "GrayPaletteColorLookupTableData", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "DLX_LKUP_01" }
  , { 0x6001, 0x0000, 0x60ff, 0x0000,
      EVR_IS, "OverlayID", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0001, 0x60ff, 0x0001,
      EVR_LT, "LinkedOverlays", 1, -1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0010, 0x60ff, 0x0010,
      EVR_US, "OverlayRows", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0011, 0x60ff, 0x0011,
      EVR_US, "OverlayColumns", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0040, 0x60ff, 0x0040,
      EVR_LO, "OverlayType", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0050, 0x60ff, 0x0050,
      EVR_US, "OverlayOrigin", 1, -1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0060, 0x60ff, 0x0060,
      EVR_LO, "Editable", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0070, 0x60ff, 0x0070,
      EVR_LO, "OverlayFont", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0072, 0x60ff, 0x0072,
      EVR_LO, "OverlayStyle", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0074, 0x60ff, 0x0074,
      EVR_US, "OverlayFontSize", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0076, 0x60ff, 0x0076,
      EVR_LO, "OverlayColor", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0078, 0x60ff, 0x0078,
      EVR_US, "ShadowSize", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0080, 0x60ff, 0x0080,
      EVR_LO, "FillPattern", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x0082, 0x60ff, 0x0082,
      EVR_US, "OverlayPenSize", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00a0, 0x60ff, 0x00a0,
      EVR_LO, "Label", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00a2, 0x60ff, 0x00a2,
      EVR_LT, "PostItText", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00a4, 0x60ff, 0x00a4,
      EVR_US, "AnchorPoint", 2, 2, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00b0, 0x60ff, 0x00b0,
      EVR_LO, "ROIType", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00b2, 0x60ff, 0x00b2,
      EVR_LT, "AttachedAnnotation", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00ba, 0x60ff, 0x00ba,
      EVR_US, "ContourPoints", 1, -1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00bc, 0x60ff, 0x00bc,
      EVR_US, "MaskData", 1, -1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
  , { 0x6001, 0x00c0, 0x60ff, 0x00c0,
      EVR_SQ, "UINOverlaySequence", 1, 1, "PrivateTag",
      DcmDictRange_Odd, DcmDictRange_Unspecified,
      "PAPYRUS 3.0" }
#endif

};

static const size_t simpleBuiltinDict_count =
    sizeof(simpleBuiltinDict)/sizeof(DBI_SimpleEntry);


void
DcmDataDictionary::loadBuiltinDictionary()
{
    DcmDictEntry* e = NULL;
    const DBI_SimpleEntry *b = simpleBuiltinDict;
    for (size_t i=0; i<simpleBuiltinDict_count; ++i) {
        b = simpleBuiltinDict + i;
        e = new DcmDictEntry(b->group, b->element,
            b->upperGroup, b->upperElement, b->evr,
            b->tagName, b->vmMin, b->vmMax,
            b->standardVersion, OFFalse, b->privateCreator);
        e->setGroupRangeRestriction(b->groupRestriction);
        e->setElementRangeRestriction(b->elementRestriction);
        addEntry(e);
    }
}


#else // DCM_DICT_DEFAULT == DCM_DICT_DEFAULT_USE_BUILTIN
void
DcmDataDictionary::loadBuiltinDictionary()
{
/*
 ** Empty Stub.
 **
 ** We don't want a populated built-in data dictionary. In order to enable it,
 ** define DCM_DICT_DEFAULT == DCM_DICT_DEFAULT_USE_BUILTIN (1).
 ** To re-create the builtin dictionary from a textfile like dicom.dic,
 ** use dcmdata/libsrc/mkdictbi.
*/
}

#endif // DCM_DICT_DEFAULT == DCM_DICT_DEFAULT_USE_BUILTIN

